#include "copyright.h"
#include "port.h"
#include "yylex.h"
#include "y.tab.h"
#include "smm.h"
#include "st.h"
#include "vov.h"

char *strsav();
int upper_strcmp();
int upper_strhash();

#define GETC()	    (*npushback != 0 ? getc_pushback() : \
			((*is_macro) ? getc_macro() : mygetc(lexin)))
#define UNGETC(c)   (input_stream[n_input_stream-1].pushback[(*npushback)++] = c);


/* 
 *  Define the stack of input streams (macros and files)
 */
#define MAXSTACK 24

struct input_stream_stack_struct {
    char *filename;		/* name of file (or name of macro) */
    int line;			/* current line number */
    int is_macro;		/* indicates entry is a macro */
    int npushback;		/* number of characters pushed back */
    char pushback[10];		/* characters pushed back on this stream */
    union input_type_union {
	FILE *fp;		/* file pointer */
	char *macro;		/* string defining the macro */
    } input_type;
} input_stream[MAXSTACK];

int n_input_stream;		/* number of input streams */

typedef struct resword_struct resword;		/* reserved word */

#define RESERVED 0
#define MACRO	 1
struct resword_struct {
    int type;			/* RESERVED or MACRO */
    union {
	int token;		/* token of reserved word */
	char *def;		/* def of macro */
    } u;
};


/* 
 *  Fast access to the current information on the input_stream stack
 *  yyline and yyfile are available for use by the parser
 */
static FILE *lexin;		/* current file pointer */
static int *npushback;		/* current pushback counter */
static int *is_macro;		/* current is_macro indicator */
int *yyline;			/* current line number in that file */
char *yyfile;			/* current file name */


/* We pass information in yytext/yylength for use by yyparse routines */
char yytext[1024];			/* the token read from the file */
int yylength;


/* Define the symbol tables */
st_table *reserved;

/* magic character list (initialized in yylex_init()) */
static int magic_char_lookup[256];

/* Reserved word list */
static struct reserved_words_struct {
    char *keyword;
    int token_value;
} reserved_words[] = {
    "ALIAS", T_ALIAS,
    "ARRAY", T_ARRAY,
    "ATERM", T_ATERM,
    "BAG", T_BAG,
    "BOX", T_BOX,
    "BY", T_BY,
    "CLOCK", T_CLOCK,
    "CONNECTOR", T_CONNECTOR,
    "CREATE", T_CREATE,
    "EDITSTYLE", T_EDITSTYLE,
    "ELSE", T_ELSE,
    "ENDIF", T_END,
    "ENDMACRO", T_ENDMACRO,
    "ENDMODEL", T_ENDMODEL,
    "FACET", T_FACET,
    "FROM", T_FROM,
    "FTERM", T_FTERM,
    "GROUND", T_GROUND,
    "IF", T_IF,
    "INOUT", T_INOUT,
    "INPUT", T_INPUT,
    "INSTANCE", T_INSTANCE,
    "IPROP", T_IPROP,
    "LAYER", T_LAYER,
    "MACRO", T_MACRO,
    "MODEL", T_MODEL,
    "NAME", T_INAME,
    "NET", T_NET,
    "OF", T_OF,
    "OUTCLOCK", T_OUTCLOCK,
    "OUTGROUND", T_OUTGROUND,
    "OUTPUT", T_OUTPUT,
    "OUTSUPPLY", T_OUTSUPPLY,
    "PATH", T_PATH,
    "PROMOTE", T_PROMOTE,
    "REQUIRE", T_REQUIRE,
    "RPROP", T_RPROP,
    "SUPPLY", T_SUPPLY,
    "SPROP", T_SPROP,
    "TECHNOLOGY", T_TECHNOLOGY,
    "THEN", T_THEN,
    "TRISTATE", T_TRISTATE,
    "TO", T_TO,
    "UNCONNECTED", T_UNCONNECTED,
    "VIEWTYPE", T_VIEWTYPE,
    "OCT_NO_TRANSFORM", T_OCT_NO_TRANSFORM,
    "OCT_MIRROR_X", T_OCT_MIRROR_X,
    "OCT_MIRROR_Y", T_OCT_MIRROR_Y,
    "OCT_ROT90", T_OCT_ROT90,
    "OCT_ROT180", T_OCT_ROT180,
    "OCT_ROT270", T_OCT_ROT270,
    "OCT_MX_ROT90", T_OCT_MX_ROT90,
    "OCT_MY_ROT90", T_OCT_MY_ROT90,
};
#define NUM_RESERVED_WORDS	\
	(sizeof(reserved_words) / sizeof(struct reserved_words_struct))


static struct magic_chars_struct {
    char c;
    int token_value;
} magic_chars[] = {
    ':', T_COLON,
    '=', T_EQ,
    ',', T_COMMA,
    ';', T_NEXT,
    '&', T_CONCAT,
    '(', T_OPEN_PAREN,
    ')', T_CLOSE_PAREN,
    '[', T_OPEN_SQUARE,
    ']', T_CLOSE_SQUARE,
    '<', T_OPEN_ANGLE,
    '>', T_CLOSE_ANGLE,
    '#', T_BASE,
    '+', T_PLUS,
    '-', T_MINUS,
    '*', T_TIMES,
    '|', T_DIV,
    '@', T_MOD,
};
#define NUM_MAGIC_CHARS		\
	sizeof(magic_chars) / sizeof(struct magic_chars_struct)


#define LINELENGTH 1025

static char line[LINELENGTH] = "";
static int linepos = 0, linenum = 0;
extern int errorcount;

where ()
{
	int i;

	if (*is_macro) {
		(void) fprintf(stderr, "error in macro %s at line %d\n", 
  			yyfile, *yyline);
	} else {
		(void) fprintf(stderr, "error in file %s at line %d\n", 
    		yyfile, *yyline);
	}
   	for(i = n_input_stream-2; i >= 0; i--) {
		(void) fprintf(stderr, "called from ");
		if (input_stream[i].is_macro) {
   			(void) fprintf(stderr, "macro %s at line %d\n", 
			input_stream[i].filename, input_stream[i].line);
		} else {
   			(void) fprintf(stderr, "file %s at line %d\n", 
			input_stream[i].filename, input_stream[i].line);
		}
   	}
}

yyerror (s)
register char *s;
{
#ifdef DUMB
	if (s) {
		(void) fputs(s, stderr);
		fprintf(stderr, " at line %d\n", linenum);
		exit(1);
	}
#endif
	yyshowerror (s, 1 /* exit */);
}

yysoft (s)
register char *s;
{
	extern int warning;

	errorcount++;
	if (! warning) return;

	if (s) {
		(void) fputs(s, stderr);
		putc('\n', stderr);
		return;
	}
	yyshowerror (s, 0 /* exit */);
}

yyshowerror (s, fatal)
register char *s;
int fatal;
{
	int lp;

	/* print position */
	if (linepos < 0) {
		(void) fprintf (stderr, "unexpected end of file at line %d\n", linenum);
	}
	else {
		(void) fputs(line, stderr);
		for (lp = 0; lp < linepos - 1; lp++)
			if (line[lp] == '\t')
				putc('\t', stderr);
			else	putc(' ', stderr);
		(void) fputs ("^\n", stderr);
	}
	/* print message */
	(void) fprintf(stderr, "[%d] ", errorcount);
	where ();
	if (s) {
		(void) fprintf(stderr, "%s\n", s);
	}
	if (fatal) exit(-1);
	else return;
}

int mygetc (inputfile)
FILE *inputfile;
{
    /* If EOF in a 'REQUIRE'd file or EOL (LL April 91) */
    if ( linepos < 0 || line[linepos] == '\0') {
	if (fgets (line, (LINELENGTH - 1), inputfile)) {
	    linenum++;
	    linepos = 0;
	    /* insert here listing generation */
	}
	else {
	    linepos = - 1;
	    return (EOF);
	}
    }
    return ((int) line[linepos++]);
}

/*
 *  yylex_init -- called first to initialize some tables to speed the
 *  processing of individual characters
 */
yylex_init()
{
    register int i;
    register resword *word;

    /* Define a lookup table for the magic characters */
    for(i = 0; i < 256; i++)
	magic_char_lookup[i] = 0;
    for(i = 0; i < NUM_MAGIC_CHARS; i++)
	magic_char_lookup[magic_chars[i].c] = magic_chars[i].token_value;

    /* Hash the reserved words into a reserved words symbol table */
    reserved = st_init_table(upper_strcmp, upper_strhash);
    for(i = 0; i < NUM_RESERVED_WORDS; i++) {
	word = MALLOC(resword);
	word->type = RESERVED;
	word->u.token = reserved_words[i].token_value;
	(void) st_insert( reserved, reserved_words[i].keyword, (char *) word);
    }
}


/*  getc_pushback -- fetch character from the pushback buffer */
int getc_pushback()
{
    return input_stream[n_input_stream-1].pushback[--(*npushback)];
}


/*  getc_macro -- fetch a character from an active macro  */
int getc_macro()
{
    register int c;
    char **macro = &(input_stream[n_input_stream-1].input_type.macro);

    if ((c = *(*macro)++) == '\0') {
	return EOF;
    } else {
	return c;
    }
}



/* read_string -- read text characters up to a specified terminator */
int read_string(word, terminator)
char *word;
int terminator;		/* character to terminate on */
{
    register int i = 0, c;

    while ((c = GETC()) != EOF && (c != terminator))
	word[i++] = c;
    word[i] = '\0';
    return i;
}

#ifdef FOO
/* read_number -- read digit characters up to the first nondigit */
int read_number(word, first_char)
char *word;
int first_char;
{
    register int i = 0, c;

    word[i++] = first_char;
    while ((c = GETC()) != EOF && isxdigit(c))
	word[i++] = c;
    word[i] = '\0';
    UNGETC(c);
    return i;
}
#endif

/* read_token -- read alphabetics until first nonalphabetic */
int read_token(word, first_char)
register char *word;
register int first_char;
{
    register int i = 0, c;

    word[i++] = first_char;
    while ((c = GETC()) != EOF && 
	(isalpha(c) || isdigit(c) || c == '_' || c == '%' 
		|| c == '$' || c == '.' || c == '/'))
	    word[i++] = c;
    word[i] = '\0';
    UNGETC(c);
    return i;
}


/* parse_require -- parse the REQUIRE mechanism for including other files */
parse_require()
{
    int c;
    char filename[256], *pfilename;

    if ((c = yylex()) != T_NAME && (c != T_STRING)) {
	yyerror("REQUIRE file name missing");
    }

    (void) strcpy(filename, yytext);
    if (strchr(filename, '.') == NIL(char)) {
	(void) strcat(filename, ".nrq");
    }
    if ((pfilename = strrchr(filename, ']')) == NIL(char)) {
	pfilename = filename;
    } else {
	pfilename += 1;		/* skip past the bracket */
    }
    if ((c = yylex()) != T_NEXT) {
	yyerror("REQUIRE syntax error (missing semicolon)");
    }
    if (yyopenfile(pfilename, /*use_stdin*/ 0, /*is_a_macro*/ 0) < 0) {
	yyerror("REQUIRE file not found");
    }
}


/* parse_macro -- parse macro definitions */
parse_macro()
{
    int i, c, macro_len, nbuf;
    char buffer[32], *macro_name, *temp;
    resword *word;

    /* read the macro name */
    c = yylex();
    macro_name = strsav(yytext);
    if ((c = yylex()) != T_EQ) {
	yyerror("MACRO Syntax (warning: parameters not supported)");
    }

    /* allow redefinition of a macro */
    if (st_lookup(reserved, macro_name, (char **) &word) == 1) {
	if (word->type = MACRO) {
	    temp = macro_name;
	    (void) st_delete(reserved, &temp, (char **) &word);
	    FREE(temp);
	    FREE(word->u.def);
	    FREE(word);
	} else {
	    yyerror("attempt to MACRO define reserved word");
	}
    }
    word = MALLOC(resword);
    (void) st_insert(reserved, macro_name, (char *) word);

    word->type = MACRO;

    /* read the macro text */
    word->u.def = ALLOC(char, 32);
    word->u.def[0] = '\0';
    macro_len = 0;
    nbuf = 0;

    while((c = GETC()) != EOF) {
	if (c == '\n') {
	    (*yyline)++;
	}
	/* Check for the end of the macro */
	if (c == '$') {
	    yylength = read_token(yytext, c);
	    if (upper_strcmp(yytext, "$ENDMACRO") == 0) {
		break;
	    } else {
		for(i = 0; i < yylength; i++) {
		    save_char(buffer, &nbuf, &(word->u.def), &macro_len, 
			yytext[i]);
		}
	    }
	} else {
	    save_char(buffer, &nbuf, &(word->u.def), &macro_len, c);
	}
    }

    save_char(buffer, &nbuf, &(word->u.def), &macro_len, EOF);

    if ((c == EOF) || (c = yylex()) != T_NEXT) {
	yyerror("MACRO Syntax (missing semicolon)");
    }
}


save_char(buffer, nbuf, macro_def, macro_len, c)
char *buffer;
int *nbuf;		/* will be changed */
char **macro_def;
int *macro_len;
int c;
{
    if (c != EOF) {
	buffer[(*nbuf)++] = c;
    }
    if (c == EOF || *nbuf >= 32) {
	*macro_len += *nbuf;
	*macro_def = REALLOC(char, *macro_def, *macro_len + *nbuf + 1);
	buffer[*nbuf] = '\0';
	(void) strcat(*macro_def, buffer);
	*nbuf = 0;
    }
}

int dump_tokens = 0;
yylex()
{
    int c = yylex_stub();
    if (dump_tokens) {
	(void) printf("YYLEX returns %d %s\n", c, yytext);
    }
    return c;
}


yylex_stub()
{
    register int c;
    resword *word;

start:
    c = GETC();

    /* Check for end of file, keep returning EOF until yacc satisfied */
    if (c == EOF) {
	yyclosefile();
	if (n_input_stream == 0) {
	    return EOF;
	} else {
	    goto start;
	}

    /* ignore whitespace for the most part */
    } else if (isspace(c)) {
	if (c == '\n') (*yyline)++;
	goto start;

    /* Check for a comment (comments are until end of line) */
    } else if (c == '!') {
	while ((c = GETC()) != EOF && (c != '\n'))
	    ;
	(*yyline)++;
	goto start;

    /* check for quoted name */
    } else if (c == '"') {
	yylength = read_string(yytext, '"');
	return T_NAME;

    /* check for string */
    } else if (c == '\'') {
	yylength = read_string(yytext, '\'');
	return T_STRING;

    /* check for special character */
    } else if (magic_char_lookup[c] != 0) {
	yytext[0] = c;
	yytext[1] = '\0';
	yylength = 1;
	return magic_char_lookup[c];

    /* check for number string */
#ifdef FOO
    } else if (isdigit(c)) {
	yylength = read_number(yytext, c);
	return T_NUMBER;
#endif

    /* check for variable */
    } else if (c == '%') {
	yylength = read_token(yytext, c);
	return T_VAR;

    /* otherwise must be a reserved word or identifier */
    } else if (isalpha(c) || (c == '$') || (c == '/') ||
		isdigit(c) || (c == '.')) {
	yylength = read_token(yytext, c);

	/* Check the reserved words symbol table for keywords */
	if (!st_lookup(reserved, yytext, (char **) &word)) {
	    return T_NAME;
	} else if (word->type == MACRO) {
	    (void) yyopenfile(yytext, /*use_stdin*/0, /* is_a_macro */ 1);
	    input_stream[n_input_stream-1].input_type.macro = word->u.def;
	    input_stream[n_input_stream-1].line = 1;
	    goto start;
	} else if (word->u.token == T_MACRO) {
	    parse_macro();
	    goto start;
	} else if (word->u.token == T_REQUIRE) {
	    parse_require();
	    goto start;
	} else {
	    return word->u.token;
	}

    /* else its something we don't quite understand: let YACC complain */
    } else {
	yytext[0] = c;
	yytext[1] = '\0';
	yylength = 1;
	return T_NAME;
    }
}



int yyopenfile(s, use_stdin, is_a_macro)
char *s;			/* name of the file */
int use_stdin;			/* if 1, use standard input */
int is_a_macro;			/* if 1, we are opening a macro */
{
    FILE *fp;

    if (n_input_stream >= MAXSTACK) {
	yyerror("Input file/Macro Nesting too deep");
    }

    input_stream[n_input_stream].filename = strsav(s);
    input_stream[n_input_stream].line = 0;
    input_stream[n_input_stream].is_macro = is_a_macro;
    input_stream[n_input_stream].npushback = 0;

    /* actually open the file if its a file */
    if (! input_stream[n_input_stream].is_macro) {
	if (use_stdin) {
	   fp = stdin;
	} else {
	    VOVinput( VOV_UNIX_FILE, s );
	    if ((fp = fopen(s, "r")) == NULL) {
		(void) fprintf(stderr, "bdnet: Unable to open \"%s\"\n", s);
		return -1;
	    }
	}
	input_stream[n_input_stream].input_type.fp = fp;
    }

    /* Setup the pointers to the current input stream */
    if (! input_stream[n_input_stream].is_macro) {
	lexin = input_stream[n_input_stream].input_type.fp;
    }
    yyfile = input_stream[n_input_stream].filename;
    yyline = &(input_stream[n_input_stream].line);
    is_macro = &(input_stream[n_input_stream].is_macro);
    npushback = &(input_stream[n_input_stream].npushback);

    n_input_stream++;
    *yyline = 1;
    return 0;
}


yyclosefile()
{
    /* pop one level on the stack */
    if (n_input_stream == 0)
	return;
    n_input_stream--;

    /* close the file it was a file, otherwise do nothing */
    if (! input_stream[n_input_stream].is_macro)
	if (input_stream[n_input_stream].input_type.fp != stdin)
	    (void) fclose(input_stream[n_input_stream].input_type.fp);

    /* Reinitialize the globals which point to the current entry */
    if (n_input_stream > 0) {
	lexin = input_stream[n_input_stream-1].input_type.fp;
	yyfile = input_stream[n_input_stream-1].filename;
	yyline = &(input_stream[n_input_stream-1].line);
	is_macro = &(input_stream[n_input_stream-1].is_macro);
	npushback = &(input_stream[n_input_stream-1].npushback);
    }

    /* note that yyfile, yyline, and is_macro point to the (still valid)
    last filename and line number if n_input_stream == 0
    (in case an error is detected past eof) */
}


yyreparse(string)
char *string;
{
    (void) yyopenfile("<re-parse>", /*use_stdin*/0, /* is_a_macro */ 1);
    input_stream[n_input_stream-1].input_type.macro = string;
    (void) yyparse();
}
