#include "copyright.h"
#include "port.h"
#include "utility.h"
#include <setjmp.h>
#include "options.h"
#include "oct.h"
#include "oh.h"
#include "st.h"
#include "list.h"
#include "octdiff.h"

#define EXACT_P		( ! (curMatchTablePtr->flags & INEXACT))

static int nullObjCmp(), termCmp(), netCmp(), instCmp();
static int polygonCmp(), boxCmp(), circleCmp(), pathCmp(), labelCmp();
static int pointCmp(), edgeCmp();
static int propCmp(), bagCmp(), layerCmp();
static int (*objCmpTable[])() = {
    nullObjCmp,		/* OCT_UNDEFINED_OBJECT */
    nullObjCmp,		/* OCT_FACET */
    termCmp,		/* OCT_TERM */
    netCmp,		/* OCT_NET */
    instCmp,		/* OCT_INSTANCE */
    polygonCmp,		/* OCT_POLYGON */
    boxCmp,		/* OCT_BOX */
    circleCmp,		/* OCT_CIRCLE */
    pathCmp,		/* OCT_PATH */
    labelCmp,		/* OCT_LABEL */
    propCmp,		/* OCT_PROP */
    bagCmp,		/* OCT_BAG */
    layerCmp,		/* OCT_LAYER */
    pointCmp,		/* OCT_POINT */
    edgeCmp,		/* OCT_EDGE */
    nullObjCmp,		/* OCT_FORMAL */
    nullObjCmp,		/* OCT_MASTER */
    nullObjCmp,		/* OCT_CHANGE_LIST */
    nullObjCmp,		/* OCT_CHANGE_RECORD */
    nullObjCmp,		/* OCT_??? */
    nullObjCmp,		/* OCT_??? */
    nullObjCmp,		/* OCT_??? */
    nullObjCmp,		/* OCT_??? */
    nullObjCmp,		/* OCT_??? */
};

/*
 *	compareObj
 *
 *  Takes two object pointers, and returns <0, 0, or >0.  Compare first
 *  using the fields of the object structures, then points (for polygons
 *  or paths), finally on octId of corresponding object in facet1 (this
 *  is oct_null_id if the obj is not part of a known pair).
 */
int compareObj(obj1Ptr, obj2Ptr)
octObject *obj1Ptr;
octObject *obj2Ptr;
{
    int ret;
    matchData *dataPtr;

    if ((ret = obj1Ptr->type - obj2Ptr->type) == 0) {
	if ((ret = (*objCmpTable[obj1Ptr->type]) (obj1Ptr, obj2Ptr)) == 0) {
	    dataPtr = getMatchData(obj1Ptr->objectId);
	    ret = (int) (octIdIsNull(dataPtr->matchId) ?
			oct_null_id : ((dataPtr->objFlags & FACET_1_P) ?
					obj1Ptr->objectId : dataPtr->matchId));
	    dataPtr = getMatchData(obj2Ptr->objectId);
	    ret -= (int) (octIdIsNull(dataPtr->matchId) ?
			oct_null_id : ((dataPtr->objFlags & FACET_1_P) ?
					obj2Ptr->objectId : dataPtr->matchId));
	}
    }
    return(ret);
}

static int tmp__cmp;

#define VALCMP(field)	(tmp__cmp = ((int) obj1Ptr->contents.field -	\
						(int) obj2Ptr->contents.field))
#define POINTCMP(field)	(VALCMP(field.x) ? tmp__cmp : VALCMP(field.y))
#define BOXCMP(field)	(POINTCMP(field.lowerLeft) ?	\
				    tmp__cmp : POINTCMP(field.upperRight))
#define FLTCMP(field)						\
    ((obj1Ptr->contents.field - obj2Ptr->contents.field) == 0 ? 0 :	\
	(obj1Ptr->contents.field - obj2Ptr->contents.field) < 0 ? -1 : 1)

#define STRCMP(field)	(tmp__cmp = str__cmp(obj1Ptr->contents.field,	\
						obj2Ptr->contents.field))
static int str__cmp(s1, s2)
char *s1, *s2;
{
    if (s1 == s2) {
	return(0);
    } else if (s1 == NIL(char)){
	return(1);
    } else if (s2 == NIL(char)){
	return(-1);
    } else {
	return(strcmp(s1, s2));
    }
}

/*ARGSUSED*/
static int nullObjCmp(obj1Ptr, obj2Ptr)
octObject *obj1Ptr, *obj2Ptr;
{
    return(0);
}

static int termCmp(obj1Ptr, obj2Ptr)
octObject *obj1Ptr, *obj2Ptr;
{
    return(STRCMP(term.name));
}

static int netCmp(obj1Ptr, obj2Ptr)
octObject *obj1Ptr, *obj2Ptr;
{
    if (EXACT_P) {
	return(STRCMP(net.name));
    } else {
	return(0);
    }
}

static int instCmp(obj1Ptr, obj2Ptr)
octObject *obj1Ptr, *obj2Ptr;
{
    if ( ! ((EXACT_P && STRCMP(instance.name)) ||
	    STRCMP(instance.master) ||
	    STRCMP(instance.view) ||
#if OCT_VERSION == 2
	    STRCMP(instance.facet) ||
#endif
	    STRCMP(instance.version) ||
	    POINTCMP(instance.transform.translation) ||
	    (EXACT_P && VALCMP(instance.transform.transformType)))) {
	if (EXACT_P && obj1Ptr->contents.instance.transform.transformType ==
							OCT_FULL_TRANSFORM) {
	    FLTCMP(instance.transform.generalTransform[0][0]) ||
		    FLTCMP(instance.transform.generalTransform[0][1]) ||
		    FLTCMP(instance.transform.generalTransform[1][0]) ||
		    FLTCMP(instance.transform.generalTransform[1][1]);
	}
    }
    return(tmp__cmp);
}

static int polygonCmp(obj1Ptr, obj2Ptr)
octObject *obj1Ptr, *obj2Ptr;
{
    static struct octPoint *p1List = NIL(struct octPoint);
    static struct octPoint *p2List = NIL(struct octPoint);
    static int size1 = 0;
    static int size2 = 0;
    octStatus stat;
    int32 n1, n2;
    int i;

    n1 = size1;
    stat = octGetPoints(obj1Ptr, &n1, p1List);
    if (stat == OCT_TOO_SMALL || n1 > size1) {
	size1 = n1;
	p1List = REALLOC(struct octPoint, p1List, n1);
	stat = octGetPoints(obj1Ptr, &n1, p1List);
    }
    OCT_ASSERT(stat);

    n2 = size2;
    stat = octGetPoints(obj2Ptr, &n2, p2List);
    if (stat == OCT_TOO_SMALL || n2 > size2) {
	size2 = n2;
	p2List = REALLOC(struct octPoint, p2List, n2);
	stat = octGetPoints(obj2Ptr, &n2, p2List);
    }
    OCT_ASSERT(stat);

    if (n1 != n2) return(n2 - n1);

    tmp__cmp = 0;
    for (i = 0; i < n1; i++) {
	if ((tmp__cmp = p2List[i].x - p1List[i].x) ||
		(tmp__cmp = p2List[i].y - p1List[i].y)) {
	    break;
	}
    }
    return(tmp__cmp);
}

static int boxCmp(obj1Ptr, obj2Ptr)
octObject *obj1Ptr, *obj2Ptr;
{
    return(BOXCMP(box));
}

static int circleCmp(obj1Ptr, obj2Ptr)
octObject *obj1Ptr, *obj2Ptr;
{
    POINTCMP(circle.center) ||
	    VALCMP(circle.startingAngle) ||
	    VALCMP(circle.endingAngle) ||
	    VALCMP(circle.innerRadius) ||
	    VALCMP(circle.outerRadius);
    return(tmp__cmp);
}

static int pathCmp(obj1Ptr, obj2Ptr)
octObject *obj1Ptr, *obj2Ptr;
{
    (EXACT_P && VALCMP(path.width)) ||
	    (tmp__cmp = polygonCmp(obj1Ptr, obj2Ptr));
    return(tmp__cmp);
}

static int labelCmp(obj1Ptr, obj2Ptr)
octObject *obj1Ptr, *obj2Ptr;
{
    (EXACT_P && STRCMP(label.label)) ||
#if OCT_VERSION == 2
	    BOXCMP(label.region) ||
	    (EXACT_P && VALCMP(label.textHeight)) ||
	    (EXACT_P && VALCMP(label.vertJust)) ||
	    (EXACT_P && VALCMP(label.horizJust)) ||
	    (EXACT_P && VALCMP(label.lineJust));
#else
	    POINTCMP(label.position);
#endif
    return(tmp__cmp);
}

static int pointCmp(obj1Ptr, obj2Ptr)
octObject *obj1Ptr, *obj2Ptr;
{
    return(POINTCMP(point));
}

static int edgeCmp(obj1Ptr, obj2Ptr)
octObject *obj1Ptr, *obj2Ptr;
{
    POINTCMP(edge.start) || POINTCMP(edge.end);
    return(tmp__cmp);
}

static int propCmp(obj1Ptr, obj2Ptr)
octObject *obj1Ptr, *obj2Ptr;
{
    int i;

    if (STRCMP(prop.name) || VALCMP(prop.type) || ! EXACT_P) {
	return(tmp__cmp);
    }

    switch (obj1Ptr->contents.prop.type) {
	case OCT_NULL:
	    break;
	case OCT_INTEGER:
	    VALCMP(prop.value.integer);
	    break;
	case OCT_REAL:
	    FLTCMP(prop.value.real);
	    break;
	case OCT_STRING:
	    STRCMP(prop.value.string);
	    break;
	case OCT_ID:
	    VALCMP(prop.value.id);
	    break;
	case OCT_STRANGER:
	    if ( ! VALCMP(prop.value.stranger.length)) {
		for (	i = 0;
			i < obj1Ptr->contents.prop.value.stranger.length &&
			    ! VALCMP(prop.value.stranger.contents[i]);
			i++) {
		}
	    }
	    break;
	case OCT_REAL_ARRAY:
	    if ( ! VALCMP(prop.value.real_array.length)){
		for (	i = 0;
			i < obj1Ptr->contents.prop.value.real_array.length &&
			    ! FLTCMP(prop.value.real_array.array[i]);
			i++) {
		}
	    }
	    break;
	case OCT_INTEGER_ARRAY:
	    if ( ! VALCMP(prop.value.integer_array.length)){
		for (	i = 0;
			i < obj1Ptr->contents.prop.value.integer_array.length &&
			    ! VALCMP(prop.value.integer_array.array[i]);
			i++ ) {
		}
	    }
	    break;
    }
    return(tmp__cmp);
}

static int bagCmp(obj1Ptr, obj2Ptr)
octObject *obj1Ptr, *obj2Ptr;
{
    return(STRCMP(bag.name));
}

static int layerCmp(obj1Ptr, obj2Ptr)
octObject *obj1Ptr, *obj2Ptr;
{
    return(STRCMP(layer.name));
}
