#include "copyright.h"
#include "port.h"
#include "utility.h"
#include <setjmp.h>
#include "options.h"
#include "oct.h"
#include "oh.h"
#include "st.h"
#include "list.h"
#include "octdiff.h"

optionStruct optionList[] = {
    { "s",	"",		"silent:  no diffs -- just exit status"	},
    { "x",	"",		"exact:  only identical objects match"	},
    { OPT_RARG,	"cell:view[:facet]",	"reference facet"		},
    { OPT_RARG,	"cell:view[:facet]",	"test facet"			},
    { 0,	0,		0					}
};

octObject facet1, facet2;
matchTable *curMatchTablePtr = NIL(matchTable);

main(argc, argv)
int argc;
char *argv[];
{
    int option;
    int flags = INEXACT;
    extern int optind;

    while ((option = optGetOption(argc, argv)) != EOF) {
	switch (option) {
	    case 's':		/* silent */
		flags |= QUICK_CHECK;
		break;
	    case 'x':		/* exact */
		flags &= ~INEXACT;
		break;
	}
    }

    if (optind != argc - 2) optUsage();

    exit(octdiff(flags, argv[optind], argv[optind+1]));
}

int octdiff(flags, facet1Name, facet2Name)
int flags;
char *facet1Name;
char *facet2Name;
{
    int match();
    void printStats();

    octBegin();

    ohUnpackDefaults(&facet1, "r", "::contents");
    if (ohUnpackFacetName(&facet1, facet1Name) != OCT_OK ||
		    (facet2 = facet1,
		     ohUnpackFacetName(&facet2, facet2Name) != OCT_OK)) {
	optUsage();
    }

    if (octOpenFacet(&facet1) < OCT_OK) {
	(void) fprintf(stderr, "%s:  couldn't open %s:%s:%s (%s)\n",
		    optProgName,
		    facet1.contents.facet.cell,
		    facet1.contents.facet.view,
		    facet1.contents.facet.facet,
		    octErrorString());
	exit(2);
    }

    if (octOpenFacet(&facet2) < OCT_OK) {
	(void) fprintf(stderr, "%s:  couldn't open %s:%s:%s (%s)\n",
		    optProgName,
		    facet2.contents.facet.cell,
		    facet2.contents.facet.view,
		    facet2.contents.facet.facet,
		    octErrorString());
	exit(2);
    }

    if (match(facet1.objectId, facet2.objectId, flags)) {
	if ( ! (flags & QUICK_CHECK)) {
	    printStats();
	}
    }

    octEnd();

    if (curMatchTablePtr && curMatchTablePtr->errorP == 0) {
	return(0);
    } else {
	return(1);
    }
}

int match(id1, id2, flags)
octId id1;
octId id2;
int flags;
{
    void matchObjects();

    pushTable(flags & ~INEXACT);
    if (curMatchTablePtr->flags & QUICK_CHECK) {
	if (setjmp(curMatchTablePtr->jmpBuf) != 0) {
	    popTable(0);
	    return(0);
	}
    }
    (void) markMatched(1, &id1, 1, &id2);

    if ( ! (curMatchTablePtr->flags & GUESS)) {
	/* run the simpler, non-guessing algorithm a few times at first */
	matchObjects();
	if (recycleSloppyObjs()) {
	    matchObjects();
	    (void) recycleSloppyObjs();
	}

	curMatchTablePtr->flags |= GUESS;
    }

    do {
	matchObjects();
    } while(recycleSloppyObjs());

    /* try again, seeing if the unmatchables can maybe be matched */
    /* also relax the exact-match constraint, if appropriate */
    (void) recycleErrorObjs();
    curMatchTablePtr->flags |= flags;
    do {
	matchObjects();
    } while(recycleSloppyObjs());

    popTable(1);
    return(1);
}

static void matchObjects()
{
    octObject obj1, obj2;
    octGenerator gen;
    lsList list1, list2;
    lsList sortedGenList();
    int matchGenObjects();

    while (!octIdIsNull(obj1.objectId = nextMatchedObj())) {
	OCT_ASSERT(octGetById(&obj1));
	obj2.objectId = getMatchData(obj1.objectId)->matchId;
	OCT_ASSERT(octGetById(&obj2));

	if (obj1.type != OCT_FACET) {
	    OCT_ASSERT(octInitGenContainers(&obj1, OCT_ALL_MASK, &gen));
	    list1 = sortedGenList(&gen);
	    OCT_ASSERT(octInitGenContainers(&obj2, OCT_ALL_MASK, &gen));
	    list2 = sortedGenList(&gen);
	    markContainers(obj1.objectId, matchGenObjects(list1, list2));
	}
	OCT_ASSERT(octInitGenContents(&obj1, OCT_ALL_MASK, &gen));
	list1 = sortedGenList(&gen);
	OCT_ASSERT(octInitGenContents(&obj2, OCT_ALL_MASK, &gen));
	list2 = sortedGenList(&gen);
	markContents(obj1.objectId, matchGenObjects(list1, list2));
    }
}

int matchGenObjects(list1, list2)
lsList list1, list2;
{
    octId *id1List, *id2List;
    int n1, n2, i;
    octObject obj1, obj2;
    int diff, newFlags, guessingP;
    int attFlags = 0;
    int getEqualIdList();

    guessingP = (curMatchTablePtr->flags & GUESS);

    id1List = ALLOC(octId, lsLength(list1));
    id2List = ALLOC(octId, lsLength(list2));

    n1 = getEqualIdList(list1, id1List, &obj1);
    n2 = getEqualIdList(list2, id2List, &obj2);

    while (n1 != 0 || n2 != 0) {
	if (n1 != 0 && n2 != 0) {
	    diff = compareObj(&obj1, &obj2);
	} else {
	    diff = (n1 == 0) ? 1 : -1;
	}
	if (diff < 0) {
	    /* peel off group 1 */
	    attFlags |= markMatched(n1, id1List, 0, NIL(octId));
	    n1 = getEqualIdList(list1, id1List, &obj1);
	} else if (diff > 0) {
	    /* peel off group 2 */
	    attFlags |= markMatched(0, NIL(octId), n2, id2List);
	    n2 = getEqualIdList(list2, id2List, &obj2);
	} else {
	    /* peel off both groups */
	    attFlags |= (newFlags = markMatched(n1, id1List, n2, id2List));
	    if (guessingP && (newFlags & SLOPPY_P)) {
		guessingP = 0;
		for (i = 0; i < n2; i++) {
		    if (match(id1List[0], id2List[i], QUICK_CHECK)) {
			break;
		    }
		}
		if (i == n2) {
		    markUnmatchable(id1List[0]);
		}
	    }
	    n1 = getEqualIdList(list1, id1List, &obj1);
	    n2 = getEqualIdList(list2, id2List, &obj2);
	}
    }

    FREE(id1List);
    FREE(id2List);
    lsDestroy(list1, (void (*)()) 0);
    lsDestroy(list2, (void (*)()) 0);

    return(attFlags);
}

lsList sortedGenList(genPtr)
octGenerator *genPtr;
{
    lsList list;
    static octObject *genObjPtr = NIL(octObject);
    octStatus stat;

    if (genObjPtr == NIL(octObject)) {
	genObjPtr = ALLOC(octObject, 1);
    }

    list = lsCreate();
    while ((stat = octGenerate(genPtr, genObjPtr)) == OCT_OK) {
	lsNewEnd(list, (lsGeneric) genObjPtr, LS_NH);
	genObjPtr = ALLOC(octObject, 1);
    }
    OCT_ASSERT(stat);

    lsSort(list, compareObj);

    return(list);
}

int getEqualIdList(list, idList, objPtr)
lsList list;
octId *idList;
octObject *objPtr;
{
    lsGeneric dummy;
    int n = 0;

    if (lsDelBegin(list, &dummy) == LS_NOMORE) return(0);

    *objPtr = *((octObject *) dummy);
    FREE(dummy);
    idList[n++] = objPtr->objectId;

    while (lsFirstItem(list, &dummy, LS_NH) == LS_OK &&
			compareObj(objPtr, (octObject *) dummy) == 0) {
	idList[n++] = ((octObject *) dummy)->objectId;
	FREE(dummy);
	(void) lsDelBegin(list, &dummy);
    }

    return(n);
}

void printStats()
{
    st_generator *gen;
    char *key, *val;
    matchData *dataPtr;
    octObject *objPtr;
    lsList unmatched1, unmatched2, modified, diffContents;
    void printUnmatchedObjs(), printModifiedObjs(), printAttachmentChanges();

    unmatched1 = lsCreate();
    unmatched2 = lsCreate();
    modified = lsCreate();
    diffContents = lsCreate();

    gen = st_init_gen(curMatchTablePtr->table);
    while (st_gen(gen, &key, &val)) {
	dataPtr = (matchData *) val;
	if (dataPtr->objFlags & UNMATCHED_P || octIdIsNull(dataPtr->matchId)) {
	    /* unmatched object */
	    objPtr = ALLOC(octObject, 1);
	    objPtr->objectId = (octId) key;
	    OCT_ASSERT(octGetById(objPtr));
	    lsNewEnd((dataPtr->objFlags & FACET_1_P) ?
						    unmatched1 : unmatched2,
			    (lsGeneric) objPtr, LS_NH);
	} else if (dataPtr->objFlags & FACET_1_P) {
	    if ( ! (dataPtr->objFlags & MATCHED_P)) {
		/* inexactly matched object */
		objPtr = ALLOC(octObject, 1);
		objPtr->objectId = (octId) key;
		OCT_ASSERT(octGetById(objPtr));
		lsNewEnd(modified, (lsGeneric) objPtr, LS_NH);
	    }
	    if (dataPtr->contentsFlags & DIFFERENT_P) {
		/* object with funny contents attachments */
		objPtr = ALLOC(octObject, 1);
		objPtr->objectId = (octId) key;
		OCT_ASSERT(octGetById(objPtr));
		lsNewEnd(diffContents, (lsGeneric) objPtr, LS_NH);
	    }
	}
    }
    st_free_gen(gen);

    lsSort(unmatched1, compareObj);
    lsSort(unmatched2, compareObj);
    lsSort(modified, compareObj);
    lsSort(diffContents, compareObj);

    printUnmatchedObjs("<<< ", &facet1, unmatched1);
    printUnmatchedObjs(">>> ", &facet2, unmatched2);
    printModifiedObjs(modified);
    printAttachmentChanges(diffContents);
}

static void printUnmatchedObjs(prefix, facetPtr, list)
char *prefix;
octObject *facetPtr;
lsList list;
{
    int n;
    lsGeneric dummy;
    octObject *objPtr;
    void printObject(), killObject();

    n = lsLength(list);
    if (n > 0) {
	(void) printf("%s %d object%s found only in %s:%s:%s\n",
		prefix, n, ((n == 1) ? "" : "s"),
		facetPtr->contents.facet.cell,
		facetPtr->contents.facet.view,
		facetPtr->contents.facet.facet);
	(void) printf("----------\n");
	while (lsDelBegin(list, &dummy) != LS_NOMORE) {
	    objPtr = (octObject *) dummy;
	    printObject(prefix, objPtr);
	    killObject(objPtr);
	    FREE(objPtr);
	}
	(void) printf("\n");
    }

    lsDestroy(list, (void (*)()) 0);
}

static void printModifiedObjs(list)
lsList list;
{
    int n;
    lsGeneric dummy;
    octObject *objPtr;
    void printObject();

    n = lsLength(list);
    if (n > 0) {
	(void) printf("<<< modified objects as found in %s:%s:%s\n",
		facet1.contents.facet.cell,
		facet1.contents.facet.view,
		facet1.contents.facet.facet);
	(void) printf(">>> modified objects as found in %s:%s:%s\n",
		facet2.contents.facet.cell,
		facet2.contents.facet.view,
		facet2.contents.facet.facet);
	(void) printf("----------\n");
	while (lsDelBegin(list, &dummy) != LS_NOMORE) {
	    objPtr = (octObject *) dummy;
	    printObject("<<<", objPtr);
	    objPtr->objectId = getMatchData(objPtr->objectId)->matchId;
	    OCT_ASSERT(octGetById(objPtr));
	    printObject(">>>", objPtr);
	    FREE(objPtr);
	}
	(void) printf("\n");
    }

    lsDestroy(list, (void (*)()) 0);
}

static void printAttachmentChanges(list)
lsList list;
{
    lsGeneric dummy;
    octObject *container1Ptr, container2, content1, content2;
    octGenerator gen;
    lsList list1, list2;
    octId *id1List, *id2List;
    int n1, n2;
    int diff;
    lsList sortedGenList();
    int getEqualIdList();
    void printAttachedObj();

    while (lsDelBegin(list, &dummy) != LS_NOMORE) {
	container1Ptr = (octObject *) dummy;
	container2.objectId = getMatchData(container1Ptr->objectId)->matchId;
	OCT_ASSERT(octGetById(&container2));

	OCT_ASSERT(octInitGenContents(container1Ptr, OCT_ALL_MASK, &gen));
	list1 = sortedGenList(&gen);
	OCT_ASSERT(octInitGenContents(&container2, OCT_ALL_MASK, &gen));
	list2 = sortedGenList(&gen);

	id1List = ALLOC(octId, lsLength(list1));
	id2List = ALLOC(octId, lsLength(list2));

	n1 = getEqualIdList(list1, id1List, &content1);
	n2 = getEqualIdList(list2, id2List, &content2);

	while (n1 != 0 || n2 != 0) {
	    if (n1 != 0 && n2 != 0) {
		diff = compareObj(&content1, &content2);
	    } else {
		diff = (n1 == 0) ? 1 : -1;
	    }
	    if (diff < 0) {
		/* peel off group 1 */
		printAttachedObj("<<<", container1Ptr, n1, &content1);
		n1 = getEqualIdList(list1, id1List, &content1);
	    } else if (diff > 0) {
		/* peel off group 2 */
		printAttachedObj(">>>", container1Ptr, n2, &content2);
		n2 = getEqualIdList(list2, id2List, &content2);
	    } else {
		/* peel off both groups */
		if (n1 != n2) {
		    printAttachedObj("<<<", container1Ptr, n1, &content1);
		    printAttachedObj(">>>", container1Ptr, n2, &content2);
		}
		n1 = getEqualIdList(list1, id1List, &content1);
		n2 = getEqualIdList(list2, id2List, &content2);
	    }
	}

	FREE(id1List);
	FREE(id2List);
	lsDestroy(list1, (void (*)()) 0);
	lsDestroy(list2, (void (*)()) 0);
    }
}

static void printAttachedObj(prefix, containerPtr, n, contentPtr)
char *prefix;
octObject *containerPtr;
int n;
octObject *contentPtr;
{
    static int headerPrintedP = 0;
    static octId lastContainerId = oct_null_id;
    void printObject();

    if ( ! headerPrintedP) {
	headerPrintedP = 1;
	(void) printf("<<< attachments found only in %s:%s:%s\n",
		facet1.contents.facet.cell,
		facet1.contents.facet.view,
		facet1.contents.facet.facet);
	(void) printf(">>> attachments found only in %s:%s:%s\n",
		facet2.contents.facet.cell,
		facet2.contents.facet.view,
		facet2.contents.facet.facet);
    }

    if (!octIdsEqual(containerPtr->objectId, lastContainerId)) {
	lastContainerId = containerPtr->objectId;
	(void) printf("----------\n");
	printObject("[container] ", containerPtr);
    }

    (void) printf("%s [attached %d time%s] ", prefix, n, ((n == 1) ? "" : "s"));

    printObject("", contentPtr);
}

void printObject(prefix, objPtr)
char *prefix;
octObject *objPtr;
{
    (void) printf(prefix);
    /*OCT_ASSERT(octPrintObject(stdout, objPtr), 0);*/
    (void) octPrintObject(stdout, objPtr, 0);
}

void killObject(objPtr)
octObject *objPtr;
{
    octGenerator gen;
    octStatus stat;
    octObject attObj;

    if (objPtr->type == OCT_TERM &&
			    !octIdIsNull(objPtr->contents.term.instanceId)) {
	/* can't delete actual term -- detach it from all but its instance */
	OCT_ASSERT(octInitGenContents(objPtr, OCT_ALL_MASK, &gen));
	while ((stat = octGenerate(&gen, &attObj)) == OCT_OK) {
	    OCT_ASSERT(octDetach(objPtr, &attObj));
	}
	OCT_ASSERT(stat);
	OCT_ASSERT(octInitGenContainers(objPtr, OCT_ALL_MASK, &gen));
	while ((stat = octGenerate(&gen, &attObj)) == OCT_OK) {
	    if (!octIdsEqual(attObj.objectId, objPtr->contents.term.instanceId)) {
		OCT_ASSERT(octDetach(objPtr, &attObj));
	    }
	}
	OCT_ASSERT(stat);
    } else {
	OCT_ASSERT(octDelete(objPtr));
    }
}
