#include "port.h"
#include "utility.h"
#include "options.h"
#include "oct.h"
#include "oh.h"

optionStruct optionList[] = {
    {"N",	"newpath",	"new path name"},
    {"O",	"oldpath",	"old path name"},
    {"r",       0,              "refresh instances"},
    {"i",       0,              "instantiate \"interface\" facet" },
    {"c",       0,              "instantiate \"contents\" facet" },
    {"o",	"[cell]:view",	"output cell:view"},
    {"v",       0,              "verbose"},
    { OPT_RARG,	"cell:view",	"facet defaults to `contents'" },
    {0, 0, 0}};


/* 
* ohReplaceInstance() is almost good, except that it does not allow
* us to force a change of instance when the terminals have changed.
* The output of this routine may be incorrect, but at least
* we have updated the instances.
*/
void
replaceInstance(newInst, oldInst, force)
    octObject *newInst, *oldInst;
    int force;
{
    octObject new_term, term, facet;
    octGenerator gen;
    octStatus status;


    octGetFacet(oldInst, &facet);
    if ( octCreate(&facet, newInst) != OCT_OK ) {
	errRaise( "octmvlib", 1, "New path %s is not valid",
		 newInst->contents.instance.master );
    }

    /*
     *  Copy attachements for objects attached to the instance
     */
    ohAttachContents(oldInst, newInst, OCT_ALL_MASK &~ OCT_TERM_MASK);
    ohAttachContainers(oldInst, newInst, OCT_ALL_MASK &~ OCT_FACET_MASK);
    /*
     *  Copy all objects attached to each terminal
     */
    OH_ASSERT(octInitGenContents(oldInst, OCT_TERM_MASK, &gen));
    while ((status = octGenerate(&gen, &term)) == OCT_OK) {
	/* Find corresponding terminal on the new cell */
	new_term = term;
	status = octGetByName(newInst, &new_term);
	if (status == OCT_OK) {
	    ohAttachContents(&term, &new_term, OCT_ALL_MASK);
	    ohAttachContainers(&term, &new_term, OCT_ALL_MASK&~OCT_INSTANCE_MASK);
	} else if (status == OCT_NOT_FOUND) {
	    if ( force ) {
		fprintf( stderr,  "Formal/Actual Terminal mismatch: %s",
			ohFormatName(&term) );
	    } else {
		errRaise(OH_PKG_NAME, OCT_ERROR, "Formal/Actual Terminal mismatch: %s",
			 ohFormatName(&term));
	    }
	} else {
	    OH_ASSERT(status);
	}
    }
    OH_ASSERT(status);
    OH_ASSERT(octDelete(oldInst));		/* delete the old instance */
    return;
}


main(argc, argv)
int argc;
char **argv;
{
    char *input_name = 0, *output_name = 0;
    char *newpath, *oldpath;
    octObject outfacet, facet, instance, newinstance;
    octGenerator gen;
    int oldlen;
    int option;
    int verbose = 0;
    int force = 0;
    char* forceFacet = 0;
    int count = 0;		/* Used to prevent infinite looping in replace loop. */
    int subst = 0;		/* Number of substitutions. */
    extern int optind;
    extern char *optarg;

    /* parse command line options*/
    output_name = NIL(char);
    newpath = oldpath = NIL(char);
    
    while ((option = optGetOption(argc, argv)) != EOF) {
	switch(option) {
	case 'i': forceFacet = "interface"; break;
	case 'c': forceFacet = "contents"; break;
	case 'r':
	    force = 1;
	    oldpath = "";
	    newpath = "";
	    break;
	case 'O': oldpath = util_strsav(optarg); break;
	case 'N': newpath = util_strsav(optarg); break;

	case 'o': output_name = util_strsav(optarg); break;
	case 'v': verbose = 1; break;
	default:
	    optUsage();
	}
    }

    if (newpath == NIL(char)) {
	(void) fprintf(stderr, "no newpath was specified\n");
	optUsage();
    }
    if (oldpath == NIL(char)) {
	(void) fprintf(stderr, "no oldpath was specified\n");
	optUsage();
    }

    /*
     *  Finish the command line argument parsing 
     */
    if (optind != argc-1) {
	optUsage();
    } else {
	input_name = argv[optind];
    }

    octBegin();
    
    /*
     *  Get the input and output facet names
     */
    ohWorkFacet( &facet, input_name, output_name );

    {
	octStatus   status = octOpenFacet(&facet);
	if ( status != OCT_OLD_FACET && status != OCT_INCONSISTENT && status != OCT_ALREADY_OPEN ) {
	    errRaise( "octmvlib", 1, "Cannot open %s\n%s\n",
		     ohFormatName( &facet ), octErrorString);
	}
    }

    OH_ASSERT(octInitGenContents(&facet, OCT_INSTANCE_MASK, &gen));

    oldlen = strlen(oldpath);

    /* Avoid infinite loops. */
    count = ohCountContents( &facet, OCT_INSTANCE_MASK );
    while (octGenerate(&gen, &instance) == OCT_OK) {
	if ( count-- == 0 ) break;
	if (strncmp(instance.contents.instance.master, oldpath, oldlen) == 0) {
	    char newname[2048];

	    (void) strcpy(newname, newpath);
	    (void) strcat(newname, &instance.contents.instance.master[oldlen]);
	    newinstance = instance;
	    newinstance.contents.instance.master = newname;
	    if ( forceFacet ) {
		newinstance.contents.instance.facet = forceFacet;
	    }
	    if ( verbose ) {
		printf( "Replacing \n\tOLD: %s\n\tNEW: %s\n", ohFormatName(&instance), ohFormatName(&newinstance) );
		fflush( stdout );
	    }

	    makeSureItIsNotABadInstance(&newinstance,&instance);

	    replaceInstance(&newinstance, &instance, force);
	    subst++;
	}
    }
    printf( "octmvlib: %d substitutions.\n", subst );
    octFreeGenerator( &gen );
    OH_ASSERT(octCloseFacet(&facet));
    
    exit(0);
}


#define STREQ(a,b) (!strcmp(a,b) )

int makeSureItIsNotABadInstance(newinst,inst)
    octObject *newinst, *inst;
{
    struct octInstance *ni = &newinst->contents.instance;
    struct octInstance *oi = &inst->contents.instance;
    
    if (STREQ(ni->master,oi->master )  &&
	STREQ(ni->view, oi->view )  &&
	STREQ(ni->facet, oi->facet ) ) 
      {
	  octObject facet, bag;
	  octGetFacet( inst, &facet );
	
	  if (ohGetByBagName(&facet,&bag,OCT_INCONSISTENT_BAG)==OCT_OK) {
	      if ( octIsAttached( &bag, inst ) == OCT_OK ) {
		  errRaise( "octmvlib", 1, 
			   "%s is already inconsistent and cannot be replaced with itself", ohFormatName(inst) );
	      }
	}
    }
}
