#include "port.h"
#include "utility.h"
#include "oct.h"
#include "oh.h"
#include "options.h"
#include "octprop.h"
#ifdef hpux
#include <regex.h>
#endif

optionStruct optionList[] = {
    { "p",	"{[type](container_name)}name=value",
	    "set property `name' to `value'"				},
    { "b",	"{[type](container_name)}name",
	    "create bag `name'"						},
    { "v",	"{[type](container_name)}name",
	    "print the value of property `name'"			},
    { OPT_ELLIP, 0,		0					},
    { OPT_RARG,	"cell:view[:facet]", "put properties on this facet"	},
    { OPT_ELLIP, 0,		0					},
    { OPT_DESC,	0,
      "One or more properties can be specified for one or more facets"	},
    { 0,	0,		0					}
};

main(argc, argv)
int argc;
char *argv[];
{
    int option;		/* option letter being processed */
    extern int optind;	/* index of current argument */
    extern char *optarg;/* argument string for options with arguments */
    octObject facet;	/* the facet to be modified */
    propDesc *propList	/* list of objects to be added */
	    = NIL(propDesc);
    int propCount = 0;	/* number of properties to be added */
    char *facetMode = "r";	/* mode for opening facet(s) */

    while ((option = optGetOption(argc, argv)) != EOF) {
	switch (option) {
	    case 'p':		/* property name/value */
		propList = REALLOC(propDesc, propList, propCount+1);
		collectProperty(optarg, &propList[propCount++]);
		facetMode = "a";
		break;
	    case 'b':		/* bag name */
		propList = REALLOC(propDesc, propList, propCount+1);
		collectBag(optarg, &propList[propCount++]);
		facetMode = "a";
		break;
	    case 'v':		/* property name (print value) */
		propList = REALLOC(propDesc, propList, propCount+1);
		collectPropVal(optarg, &propList[propCount++]);
		break;
	}
    }

    if (optind >= argc) {
	(void) fprintf(stderr, "%s: you must specify at least one facet",
			    optProgName);
	optUsage();
    }

    octBegin();

    ohUnpackDefaults(&facet, facetMode, "::contents");
    while (optind < argc) {
	if (ohUnpackFacetName(&facet, argv[optind++]) != OCT_OK) {
	    optUsage();
	}
	insertObjects(&facet, propList, propCount);
    }

    octEnd();

    exit(0);
}

collectProperty(propString, propPtr)
char *propString;	/* textual description of desired property */
propDesc *propPtr;	/* pointer to object to be filled in */
{
    char *valPtr;
    int count, i;
    long atol();
    double atof();
    char *handleContainers();

    valPtr = strchr(propString, '=');
    if (valPtr == NIL(char)) optUsage();
    *valPtr++ = '\0';

    propPtr->obj.type = OCT_PROP;
    propPtr->obj.contents.prop.name = handleContainers(propString, propPtr);
    if (isInteger(valPtr)) {
	propPtr->obj.contents.prop.type = OCT_INTEGER;
	propPtr->obj.contents.prop.value.integer = atol(valPtr);
    } else if (isReal(valPtr)) {
	propPtr->obj.contents.prop.type = OCT_REAL;
	propPtr->obj.contents.prop.value.real = atof(valPtr);
    } else if (count = isIntegerArray(valPtr)) {
	propPtr->obj.contents.prop.type = OCT_INTEGER_ARRAY;
	propPtr->obj.contents.prop.value.integer_array.length = count;
	propPtr->obj.contents.prop.value.integer_array.array =
							ALLOC(int32, count);
	for (i = 0; i < count; i++) {
	    propPtr->obj.contents.prop.value.integer_array.array[i] = atoi(valPtr);
	    valPtr = strchr(valPtr, ',') + 1;
	}
    } else if (count = isRealArray(valPtr)) {
	propPtr->obj.contents.prop.type = OCT_REAL_ARRAY;
	propPtr->obj.contents.prop.value.real_array.length = count;
	propPtr->obj.contents.prop.value.real_array.array =
							ALLOC(double, count);
	for (i = 0; i < count; i++) {
	    propPtr->obj.contents.prop.value.real_array.array[i] = atof(valPtr);
	    valPtr = strchr(valPtr, ',') + 1;
	}
    } else {
	if (valPtr[0] == '"' && valPtr[strlen(valPtr) - 1] == '"') {
	    valPtr[strlen(valPtr) - 1] = '\0';
	    valPtr++;
	}
	propPtr->obj.contents.prop.type = OCT_STRING;
	propPtr->obj.contents.prop.value.string = valPtr;
    }
}

collectBag(bagString, bagPtr)
char *bagString;	/* textual description of desired bag */
propDesc *bagPtr;	/* pointer to object to be filled in */
{
    char *handleContainers();

    bagPtr->obj.type = OCT_BAG;
    bagPtr->obj.contents.bag.name = handleContainers(bagString, bagPtr);
}

collectPropVal(propString, propPtr)
char *propString;	/* textual description of desired property */
propDesc *propPtr;	/* pointer to object to be filled in */
{
    char *handleContainers();

    propPtr->obj.type = OCT_UNDEFINED_OBJECT;
    propPtr->obj.contents.prop.name = handleContainers(propString, propPtr);
}

char *handleContainers(nameString, propPtr)
char *nameString;	/* textual description of desired property */
propDesc *propPtr;	/* pointer to object to be filled in */
{
    char *reStart, *reEnd;
    nameTree **placePtr;

    placePtr = &propPtr->container;

    while (reStart = strchr(nameString, '(')) {
	*placePtr = ALLOC(nameTree, 1);

	if (reStart == nameString) {
	    /* no object types specified, allow any */
	    (*placePtr)->objTypes = OCT_ALL_MASK;
	} else {
	    (*placePtr)->objTypes = 0;
	    while (reStart > nameString) {
		switch (*nameString++) {
		    case 't':	(*placePtr)->objTypes |= OCT_TERM_MASK; break;
		    case 'n':	(*placePtr)->objTypes |= OCT_NET_MASK; break;
		    case 'i':	(*placePtr)->objTypes |= OCT_INSTANCE_MASK; break;
		    case 'p':	(*placePtr)->objTypes |= OCT_PROP_MASK; break;
		    case 'b':	(*placePtr)->objTypes |= OCT_BAG_MASK; break;
		    case 'l':	(*placePtr)->objTypes |= OCT_LAYER_MASK; break;
		    default:
			optUsage();
		}
	    }
	}

	reEnd = ++reStart;
	do {
	    reEnd = strchr(reEnd+1, ')');
	    if (reEnd == NIL(char)) optUsage();
	} while (reEnd[-1] == '\\');

	*reEnd = '\0';
	(*placePtr)->name = ALLOC(char, strlen(reStart)+3);
	(void) sprintf((*placePtr)->name, "^%s$", reStart);

	placePtr = &(*placePtr)->next;
	nameString = reEnd + 1;
    }

    *placePtr = NIL(nameTree);
    return(nameString);
}

isInteger(string)
char *string;
{
    if (*string == '-' || *string == '+') string++;

    if (*string == '\0') return(0);
    while (*string != '\0') {
	if (!isdigit(*string++)) return(0);
    }
    return(1);
}

isReal(string)
char *string;
{
    int gotPeriod = 0;
    int digCount = 0;

    if (*string == '-' || *string == '+') string++;

    while (*string != '\0') {
	if (isdigit(*string)) {
	    digCount++;
	} else if (*string == '.' && !gotPeriod) {
	    gotPeriod = 1;
	} else if (*string == 'e' || *string == 'E') {
	    return( (digCount > 0) && isInteger(++string) );
	} else {
	    return(0);
	}
	string++;
    }
    return(digCount > 0);
}

isIntegerArray(string)
char *string;
{
    char *cPtr;
    int count, flag;

    cPtr = strchr(string, ',');
    if (cPtr == NIL(char)) return(0);

    for (count = 0; cPtr; count++) {
	*cPtr = '\0';
	flag = isInteger(string);
	*cPtr = ',';
	if ( ! flag) return(0);
	string = cPtr + 1;
	cPtr = strchr(string, ',');
    }

    if (isInteger(string)) {
	return(count + 1);
    } else {
	return(0);
    }
}

isRealArray(string)
char *string;
{
    char *cPtr;
    int count, flag;

    cPtr = strchr(string, ',');
    if (cPtr == NIL(char)) return(0);

    for (count = 0; cPtr; count++) {
	*cPtr = '\0';
	flag = isReal(string);
	*cPtr = ',';
	if ( ! flag) return(0);
	string = cPtr + 1;
	cPtr = strchr(string, ',');
    }

    if (isReal(string)) {
	return(count + 1);
    } else {
	return(0);
    }
}

insertObjects(facetPtr, objList, objCount)
octObject *facetPtr;		/* the facet */
propDesc *objList;
int objCount;
{
    int index;

    /* Open the facet */
    if (octOpenFacet(facetPtr) < OCT_OK) {
	char buffer[1000];
	(void) sprintf(buffer, "Can't open cell \"%s:%s:%s\" for reading",
		facetPtr->contents.facet.cell,
		facetPtr->contents.facet.view,
		facetPtr->contents.facet.facet);
	octError(buffer);
	exit(1);
    }

    /* add (or modify) objects to facet */
    for (index = 0; index < objCount; index++) {
	insertOne(facetPtr, objList[index].container, &objList[index].obj);
    }

    /* Write out the framed facet */
    if (octCloseFacet(facetPtr) != OCT_OK) {
	char buffer[1000];
	(void) sprintf(buffer, "Can't write out cell \"%s:%s:%s\"",
		facetPtr->contents.facet.cell,
		facetPtr->contents.facet.view,
		facetPtr->contents.facet.facet);
	octError(buffer);
	exit(1);
    }
}

insertOne(parentPtr, treePtr, objPtr)
octObject *parentPtr;
nameTree *treePtr;
octObject *objPtr;
{
    octGenerator gen;
    octStatus stat;
    octObject child;
    char *name;
    static char *currentRE = NIL(char);
#ifdef hpux
    int mesg;
    regex_t re;
#else
    char *mesg;
    char *re_comp();
#endif

    if (treePtr == NIL(nameTree)) {
	if (objPtr->type == OCT_UNDEFINED_OBJECT) {
	    getOne(parentPtr, objPtr);
	} else {
	    OH_ASSERT(octCreateOrModify(parentPtr, objPtr));
	}
	return;
    } else {
	OH_ASSERT(octInitGenContents(parentPtr, treePtr->objTypes, &gen));
	while ((stat = octGenerate(&gen, &child)) == OCT_OK) {
	    switch (child.type) {
		case OCT_TERM:
		    name = child.contents.term.name;
		    break;
		case OCT_NET:
		    name = child.contents.net.name;
		    break;
		case OCT_INSTANCE:
		    name = child.contents.instance.name;
		    break;
		case OCT_PROP:
		    name = child.contents.prop.name;
		    break;
		case OCT_BAG:
		    name = child.contents.bag.name;
		    break;
		case OCT_LAYER:
		    name = child.contents.layer.name;
		    break;
	    }

#ifdef hpux
	    if (currentRE != treePtr->name) {
		currentRE = treePtr->name;
		/* 0 is success; no subexpressions, don't treat nl specially */
		mesg = regcomp(&re, currentRE, REG_NOSUB|REG_NEWLINE);
		if (! mesg) {
		    (void) fprintf(stderr, "%s: bad RE (%s)\n",
					    optProgName, currentRE);
		    exit(1);
		}
	    }
	    /* 0 is success */
	    if (! regexec(&re, name, (size_t) 0, NULL, 0))
		insertOne(&child, treePtr->next, objPtr);
	    regfree(&re);
#else 
	    if (currentRE != treePtr->name) {
		currentRE = treePtr->name;
		mesg = re_comp(currentRE);
		if (mesg) {
		    (void) fprintf(stderr, "%s: bad RE (%s): %s\n",
					    optProgName, currentRE, mesg);
		    exit(1);
		}
	    }
	    if (re_exec(name) == 1) insertOne(&child, treePtr->next, objPtr);
#endif
	}
	OH_ASSERT(stat);
    }
}

getOne(parentPtr, objPtr)
octObject *parentPtr;
octObject *objPtr;
{
    octObject prop;

    prop.type = OCT_PROP;
    prop.contents.prop.name = objPtr->contents.prop.name;
    if (octGetByName(parentPtr, &prop) == OCT_OK) {
	switch (prop.contents.prop.type) {
	    case OCT_INTEGER:
		(void) printf("%d\n", prop.contents.prop.value.integer);
		break;
	    case OCT_REAL:
		(void) printf("%#g\n", prop.contents.prop.value.real);
		break;
	    case OCT_STRING:
		(void) printf("\"%s\"\n", prop.contents.prop.value.string);
		break;
	}
    }
}
