/*      DATA-STRUCTURE BUILDING ROUTINES 
 *       --------------------------------
 *
 *   File:       buildds.c
 *   Author:     Umakanta Choudhury 
 *   Date:       18th June 1989
 *   Last modified:  Enrico Malavasi,  Aug. 1992
 *
 *
 *   Routines to build the data structures. 
 */

#include <stdio.h>
#include <string.h>
#include "utility.h"
#include "circuit.h"
#include "layout.h"
#include "avl.h"
#define OUT_OF_MEMORY(code) {printf("Out of memory\n");  exit(code);}

/*
*   INITIALIZE THE CIRCUIT  
*
* Starts a new circuit structure. 
* Returns pointer to the structure via the
* circuit_ptr parameter.
*/


static void
  init_chan(chan_info,netA,netB,channel)
struct Chan_info *chan_info;
struct Net *netA;
struct Net *netB;
struct Channel *channel;
{
    chan_info->netA = netA;
    chan_info->netB = netB;
    chan_info->channel = channel;
    chan_info->param_num = 0;
    chan_info->C_bound = 0;
    chan_info->Cx_max = 0;
    chan_info->Cx_min = 0;
    chan_info->Cy_max = 0;
    chan_info->Cy_min = 0;
    chan_info->Ccross_max = 0;
    chan_info->Ccross_min = 0;
    chan_info->C_max = 1000000.0;
    chan_info->C_min = 0;
    chan_info->C_min_shield = 0;
    chan_info->Lx_max = 0;
    chan_info->Ly_max = 0;
    chan_info->shielded = NO;
}



#define  DEFAULT_SIMULATOR_NAME    "/users/cad/bin/spice3.old"
#define  DEFAULT_TECHNOLOGY_NAME   "scmos"

void
  ds_InitCircuit( circuit )
struct Circuit *circuit;
{
    circuit->name = NULL;
    circuit->num_nodes = 0;
    circuit->node_list = NULL;
    circuit->num_matched_pairs = 0;
    circuit->matched_pair_list = NULL;
    circuit->num_perf  = 0;
    circuit->perf_list = NULL;
    circuit->max_process_var_percent = 0.0;
    circuit->max_mismatch_percent    = 0;
    if( (circuit->node_tree = avl_init_table(strcmp)) == NULL) {
        printf("Error starting binary tree for nodes\n");
	exit( -1 );
    }
    if( (circuit->perf_tree = avl_init_table(strcmp)) == NULL) {
        printf("Error starting binary tree for performance functions\n");
	exit( -1 );
    }

    /*  Defaults  */
    circuit->sim_name       = DEFAULT_SIMULATOR_NAME;
    circuit->technologyName = DEFAULT_TECHNOLOGY_NAME;
    circuit->deckFile       = NULL;
    circuit->out_node1_name = NULL;
    circuit->out_node2_name = NULL;
}

/*
*   INITIALIZE THE LAYOUT  
*
* Starts a new layout structure. 
* Returns pointer to the structure via the
* layout parameter.
*
* Return OK or ERROR.
*/

void
  ds_InitLayout(layout)
struct Layout *layout;
{
    layout->name = NULL;
    layout->num_chan = 0;
    layout->perf_name = NULL;
    layout->chan_list = NULL;
    layout->chan_info_list_link = NULL;
    layout->num_critpairs = 0;
    layout->num_bound_constraints = 0;
    layout->num_match_constraints = 0;
    layout->critical_pairs = NULL;
    layout->cap_threshold = 0;
}

/*
*   ADD A PERFORMANCE STRUCTURE 
*
* Return OK or ERROR.
*/

REPORT
  ds_MakePerformance(perf_name, type, performance_ptr, layout, circuit)
char *perf_name;
int type; 
struct Performance **performance_ptr;
struct Layout  *layout;
struct Circuit *circuit;
{

    struct Performance *performance;

    if( (performance = 
      (struct Performance *)nwAddLink(sizeof(struct  Performance),
      &(circuit->perf_list),NULL,(circuit->perf_list)) ) == NULL) {
	OUT_OF_MEMORY(ERROR);
    }

    circuit->num_perf++;
    layout->perf_name = REALLOC(char *, layout->perf_name, circuit->num_perf);

    performance->name = util_strsav(perf_name);
    performance->type = type;
    performance->number = circuit->num_perf;
    performance->pos_eqn_num = 0;
    performance->neg_eqn_num = 0;
    performance->pos_constraint = NO;
    performance->neg_constraint = NO;
    performance->max_pos_change = 0;
    performance->max_neg_change = 0;
    performance->delta_pos_min = 0;
    performance->delta_neg_min = 0;
    performance->pos_sens_list = NULL;
    performance->neg_sens_list = NULL;
    performance->pos_sens_stored_list = NULL;
    performance->neg_sens_stored_list = NULL;
    performance->freqRange.min = 1.0;
    performance->freqRange.max = 10.0e6;
    avl_insert(circuit->perf_tree,performance->name,performance);
    *performance_ptr = performance;

    return OK;
}

/*
*   ADD A NODE STRUCTURE 
*
* Return OK or ERROR.
*/

REPORT
  ds_MakeNode(node_name,node_ptr,circuit)
char *node_name;
struct Node **node_ptr;
struct Circuit *circuit;
{

    struct Node *node;
    
    if( (node = (struct Node *)nwAddLink(sizeof(struct  Node),
					 &(circuit->node_list),NULL,(circuit->node_list)) ) == NULL) {
	OUT_OF_MEMORY(ERROR);
    }

    node->name = util_strsav(node_name);
    node->matched = NO;
    node->matched_node = NULL;
    avl_insert(circuit->node_tree,node->name,node);
    *node_ptr = node;
    return OK;
}

/*
*   ADD A MATCHED_PAIR STRUCTURE 
*
* Return OK or ERROR.
*/

REPORT
  ds_MakeMatchedPair(node_nameA, node_nameB, matched_pair_ptr, circuit)
char *node_nameA, *node_nameB;
struct Matched_pair **matched_pair_ptr;
struct Circuit *circuit;
{

    struct Matched_pair *matched_pair;
    struct Node *nodeA, *nodeB;

    if( (matched_pair = 
      (struct Matched_pair *)nwAddLink(sizeof(struct  Matched_pair),
      &(circuit->matched_pair_list),NULL,circuit->matched_pair_list) ) == NULL)
      {
	  OUT_OF_MEMORY(ERROR);
      }

    /* Get nodeA and nodeB */
    if(avl_lookup(circuit->node_tree,node_nameA,&nodeA) == NULL){
	printf("matched node %s not found in circuit\n",node_nameA);
	exit( -1 );
    }
    if(avl_lookup(circuit->node_tree,node_nameB,&nodeB) == NULL){
	printf("matched node %s not found in circuit\n",node_nameB);
	exit( -1 );
    }
    nodeA->matched = YES;
    nodeB->matched = YES;
    nodeA->matched_node = nodeB;
    nodeB->matched_node = nodeA;
    matched_pair->nodeA = nodeA;
    matched_pair->nodeB = nodeB;
    matched_pair->processed = NO;
    *matched_pair_ptr = matched_pair;
    return OK;
}

/*
*   ADD A CRITCALPAIR STRUCTURE 
*
* Return OK or ERROR.
*/

REPORT
  ds_MakeCriticalPair(nameA, nameB, crit_pair_ptr, num_perf, layout)
char *nameA, *nameB;
struct CriticalPairLink **crit_pair_ptr;
int num_perf;
struct Layout *layout;
{

    struct CriticalPairLink *crit_pair;
    struct Net *netA, *netB;
    struct Net *save_netA, *save_netB;
    int NETA_PRESENT;
    int NETB_PRESENT;
    int FOUND_NETA;
    int FOUND_NETB;
    struct ChannelLink *chan_link;
    struct Channel *channel;
    struct Chan_info *chan_info;
    

    if( (crit_pair = (struct CriticalPairLink *)
	 nwAddLink(sizeof(struct CriticalPairLink), &(layout->critical_pairs),NULL,
		   layout->critical_pairs) ) == NULL)      {
	OUT_OF_MEMORY(ERROR);
    }
	/* initialization */
    (void) init_crit(crit_pair,nameA,nameB,num_perf);

    /* Get/create the net(s) */
	FOUND_NETA = 0;
	FOUND_NETB = 0;
	for(chan_link = layout->chan_list ; chan_link!= NULL ; chan_link = chan_link->next){ 
	    NETA_PRESENT = 0;
	    NETB_PRESENT = 0;
	    channel = chan_link->channel;
	    if(avl_lookup(channel->net_tree,nameA,&netA) != 0){
		FOUND_NETA = 1;
		NETA_PRESENT = 1; /* in this channel */
		save_netA = netA;
	    }
	    if(avl_lookup(channel->net_tree,nameB,&netB) != 0){
		FOUND_NETB = 1;
		NETB_PRESENT = 1; /* in this channel */
		save_netB = netB;
	    }
	    if((NETA_PRESENT)&&(NETB_PRESENT)){
		/* to be modified : create chan_info even if only
		   one of the nets is present (to take into
		   account inter channel coupling )*/
		
		if(ds_MakeChanInfo(&chan_info,crit_pair) !=OK)   return ERROR;
		
		/* initialization */
		init_chan(chan_info,netA,netB,channel);
	    }
	}
    if(!FOUND_NETA){
	fprintf(stderr, "Critical net %s not found in layout\n",nameA);
	return ERROR;
    }
    if(!FOUND_NETB){
	fprintf(stderr, "Critical net %s not found in layout\n",nameB);
	return ERROR;
    }
    
    *crit_pair_ptr = crit_pair;
    return OK;
}

/*
*   ADD A CHAN_INFO STRUCTURE 
*
* Return OK or ERROR.
*/

REPORT
  ds_MakeChanInfo(chan_info_ptr,crit_pair)
struct Chan_info **chan_info_ptr;
struct CriticalPairLink *crit_pair;
{

    struct Chan_info *chan_info;

    if( (chan_info = 
      (struct Chan_info *)nwAddLink(sizeof(struct  Chan_info),
      &(crit_pair->chan_info_list),NULL,crit_pair->chan_info_list) ) == NULL)
	{
	OUT_OF_MEMORY(ERROR);
	}

	*chan_info_ptr = chan_info;
    return OK;
}
/*
*   START A CHANNEL LIST STRUCTURE
*
* Return OK or ERROR.
*/

REPORT
  ds_MakeChanList(num_chan, layout)
int num_chan;
struct Layout *layout;
{
    int i;
    
    layout->num_chan = num_chan;
    if((layout->chan_list = REALLOC( struct ChannelLink, layout->chan_list, num_chan )) == NULL) {
	OUT_OF_MEMORY( ERROR );
    }
    for(i=0 ; i<num_chan ; i++) {
	layout->chan_list[i].prev = (i == 0) ? NULL : &(layout->chan_list[i - 1]);
	layout->chan_list[i].next = (i == (num_chan-1)) ? NULL : &(layout->chan_list[i + 1]);
    }
    return OK;
}
/*
*   START A CHANNEL STRUCTURE
*
* Return OK or ERROR.
*/
#define   DEFAULT_RESOLUTION      0.05
#define   DEFAULT_INFLUENCE_DIST  15.0
#define   DEFAULT_NET_SEPARATION   3.0
#define   DEFAULT_NET_WIDTH        3.0
REPORT
  ds_MakeChannel(chan_link, channel_ptr)
struct ChannelLink *chan_link;
struct Channel **channel_ptr;
{
    struct Channel *channel;
    
    if( (chan_link->channel = ALLOC(struct Channel, 1)) == NULL) {
	OUT_OF_MEMORY(-1);
    }
    channel = chan_link->channel;
    
    if( (channel->net_tree = avl_init_table(strcmp)) == NULL) {
       	fprintf(stderr, "Error starting binary tree for nets\n");
	exit(-1);
    }
    rr_dsStartChannel(channel);

    rr_dsSetLayoutResolution(DEFAULT_RESOLUTION,         channel);
    rr_dsSetInfluenceDistance(DEFAULT_INFLUENCE_DIST,    channel);
    rr_dsSetDefaultNetSeparation(DEFAULT_NET_SEPARATION, channel);
    rr_dsSetChanNetSeparation(DEFAULT_NET_SEPARATION,    channel);
    rr_dsSetDefaultNetWidth(DEFAULT_NET_WIDTH,           channel);
    *channel_ptr = channel;
    
    return OK;
}
/*
*   ADD A PIN_LIST_LINK STRUCTURE 
*
* Return OK or ERROR.
*/

REPORT
  ds_MakePinListLink(channel,edge,pin_list_link_ptr)
struct Channel *channel;
int edge;
struct Pin_list_link **pin_list_link_ptr;
{

    struct Pin_list_link *pin_list_link;

	if(edge == TOP){
    	if( (pin_list_link = 
	      (struct Pin_list_link *)nwAddLink(sizeof(struct  
		  Pin_list_link),&(channel->top_pin_lists),NULL,
		  channel->top_pin_lists) ) == NULL)
		{
		OUT_OF_MEMORY(ERROR);
		}
	}
	else if(edge == BOTTOM){
    	if( (pin_list_link = 
	      (struct Pin_list_link *)nwAddLink(sizeof(struct  
		  Pin_list_link),&(channel->bottom_pin_lists),NULL,
		  channel->bottom_pin_lists) ) == NULL)
		{
		OUT_OF_MEMORY(ERROR);
		}
	}
	else{
		printf("ds_MakeChanLink:edge not set properly\n"); 
		return ERROR;
	}

	/* initialization */
	pin_list_link->pin_list = NULL;
	pin_list_link->last_pin_link = NULL;
	pin_list_link->channel = NULL;

	*pin_list_link_ptr = pin_list_link;
    return OK;
}

/*
*   ADD A PIN_LINK  STRUCTURE 
*
* Return OK or ERROR.
*/

REPORT
  ds_MakePinLink(pin_list_link,pin_link_ptr)
struct Pin_list_link *pin_list_link;
struct PinLink **pin_link_ptr;
{

    struct PinLink *pin_link;

    if( (pin_link = 
      (struct PinLink *)nwAddLink(sizeof(struct PinLink),
      &(pin_list_link->pin_list),NULL,pin_list_link->pin_list) ) 
	  == NULL) 
	{
	OUT_OF_MEMORY(ERROR);
	}

	*pin_link_ptr = pin_link;
    return OK;
}
/*
*   ADD A CHANNEL_LINK STRUCTURE TO CHAN_LIST
*
* Return OK or ERROR.
*/

REPORT
  ds_MakeChanLink(channel,edge,chan_link_ptr)
struct Channel *channel;
int edge;
struct ChannelLink **chan_link_ptr;
{

    struct ChannelLink *chan_link;

	if(edge == TOP){
    	if( (chan_link = 
	   	  (struct ChannelLink *)nwAddLink(sizeof(struct  ChannelLink),
   		  &(channel->top_chan_list),NULL,channel->top_chan_list) ) == NULL)
		{
		OUT_OF_MEMORY(ERROR);
		}
	}
	else if(edge == BOTTOM){
		if( (chan_link = 
	      (struct ChannelLink *)nwAddLink(sizeof(struct  ChannelLink),
   	      &(channel->bottom_chan_list),NULL,channel->bottom_chan_list) ) == NULL)
		{
		OUT_OF_MEMORY(ERROR);
		}
	}
	else{
		printf("ds_MakeChanLink:edge not set properly\n"); 
		return ERROR;
	}

	*chan_link_ptr = chan_link;
    return OK;
}
/*
*   ADD A CRIT_TEMP STRUCTURE TO CRIT_TEMP_LIST
*
* Return OK or ERROR.
*/

REPORT
  ds_MakeCritTemp(crit_temp_list_ptr,crit_temp_ptr)
struct Crit_temp **crit_temp_list_ptr;
struct Crit_temp **crit_temp_ptr;
{

    struct Crit_temp *crit_temp;

    if( (crit_temp = 
      (struct Crit_temp *)nwAddLink(sizeof(struct  Crit_temp),
      crit_temp_list_ptr,NULL,*crit_temp_list_ptr) ) == NULL)
	{
	OUT_OF_MEMORY(ERROR);
	}

	*crit_temp_ptr = crit_temp;
    return OK;
}

void
  ds_MakeSensLink(perf, value, nodeA,nodeB, sens_link_ptr, stored_link_ptr)
struct Performance *perf;
float value;
struct Node *nodeA, *nodeB;
struct Sens_link **sens_link_ptr;
struct Sens_link **stored_link_ptr;
/*  Add a sens_link structure to pos_sens_list */
{
    struct Sens_link *sens_link;
    struct Sens_link *stored_link;
    
    if(value >= 0) {
    	if( (sens_link = (struct Sens_link *)
	     nwAddLink(sizeof(struct  Sens_link),
		       &(perf->pos_sens_list),NULL,perf->pos_sens_list) ) == NULL) {
	    OUT_OF_MEMORY(ERROR);
	}
    	if( (stored_link = (struct Sens_link *)
	     nwAddLink(sizeof(struct  Sens_link),
		       &(perf->pos_sens_stored_list),NULL, perf->pos_sens_stored_list)) == NULL) {
	    OUT_OF_MEMORY(ERROR);
	}
    } else {
    	if( (sens_link = (struct Sens_link *)
	     nwAddLink(sizeof(struct  Sens_link),
		       &(perf->neg_sens_list),NULL,perf->neg_sens_list) ) == NULL) {
	    OUT_OF_MEMORY(ERROR);
	}
    	if( (stored_link = (struct Sens_link *)
	     nwAddLink(sizeof(struct  Sens_link),
		       &(perf->neg_sens_stored_list),NULL,
		       perf->neg_sens_stored_list)) == NULL) {
	    OUT_OF_MEMORY(ERROR);
	}
    }
    sens_link->sens = value;
    sens_link->nodeA = nodeA;
    sens_link->nodeB = nodeB;
    stored_link->sens = value;
    stored_link->nodeA = nodeA;
    stored_link->nodeB = nodeB;
    
    *sens_link_ptr = sens_link;
    *stored_link_ptr = stored_link;
}
/*
*   ADD A CHAN_INFO_LIST_LINK STRUCTURE TO LAYOUT
*
* Return OK or ERROR.
*/

void
  ds_MakeChanInfoListLink(layout,nameA,nameB,chan_info_list_link_ptr)
struct Layout *layout;
char  *nameA, *nameB;
struct Chan_info_list_link **chan_info_list_link_ptr;
{
    struct Net	*netA;
    struct Net	*netB;
    struct Net	*save_netA;
    struct Net	*save_netB;
    struct Channel	*channel;
    struct ChannelLink	*chan_link;
    struct Chan_info	*chan_info;
    struct Chan_info_list_link	*chan_info_list_link;
    int	NETA_PRESENT;
    int	NETB_PRESENT;
    int	FOUND_NETA = 0;
    int	FOUND_NETB = 0;
    int	LIST_PRESENT;
    
    if( (chan_info_list_link = (struct Chan_info_list_link *)
	 nwAddLink(sizeof(struct Chan_info_list_link),&(layout->chan_info_list_link),NULL,
		   layout->chan_info_list_link)) == NULL) {
	OUT_OF_MEMORY(ERROR);
    }
    /* initialization */
    chan_info_list_link->chan_info_list = NULL;
    
    /* Get/create the net(s) */
    for(chan_link = layout->chan_list; chan_link!= NULL; chan_link = chan_link->next) {
	NETA_PRESENT = 0;
	NETB_PRESENT = 0;
	channel = chan_link->channel;
	if(avl_lookup(channel->net_tree,nameA,&netA) != 0){
	    FOUND_NETA = 1;
	    NETA_PRESENT = 1; /* in this channel */
	    save_netA = netA;
	}
	if(avl_lookup(channel->net_tree,nameB,&netB) != 0){
	    FOUND_NETB = 1;
	    NETB_PRESENT = 1; /* in this channel */
	    save_netB = netB;
	}
	if((NETA_PRESENT)&&(NETB_PRESENT)){
	    if( (chan_info = (struct Chan_info *)
		 nwAddLink(sizeof(struct Chan_info), &(chan_info_list_link->chan_info_list),NULL,
			   chan_info_list_link->chan_info_list)) == NULL) {
		OUT_OF_MEMORY(ERROR);
	    }
	    /* initialization */
	    init_chan(chan_info,netA,netB,channel);
	}
    }
    if(!FOUND_NETA){
	DEBUGG( printf("net %s not found in layout\n",nameA) );
    }
    if(!FOUND_NETB) {
	DEBUGG( printf("net %s not found in layout\n",nameB) );
    }
    *chan_info_list_link_ptr = chan_info_list_link;
}
