/*		READ CAPACITANCE LIBRARY FILE
*		-----------------------------
*
*   File:	cap_lib_read.c
*   Author: Umakanta Choudhury
*   Date:   10th June
*
*/


#include <stdio.h>
#include "layout.h"
#include "cap_library.h"
#define MAX_LINE_LENGTH 200
#define PRINT_LINE_DATA {printf("%s: line %d\n",text_file_name,linenum);}
#define DECREMENT_NUMBER_OF_ITEMS \
  {if( (--number_of_items) == 0) readin_status = READING_NORMAL;}

/*  READIN STATUS DEFINITIONS */
#define READING_NORMAL 0
#define READING_X_LAYER_COEFFS 1
#define READING_Y_LAYER_COEFFS 2
#define READING_XY_LAYER_COEFFS 3
#define READING_CROSSOVER_COEFFS 4
#define READING_LIMITS 5

/*
*  READ TEXT FILE
*
*
*  "capacitance library" input syntax:
*  ----------------------
*
*    "x_layer cap coefficients" - capacitance coefficients in
*								second order polynomial for capacitance
*								between two parallel segments in x 
*								layer 
*    "y_layer cap coefficients" - capacitance coefficients in
*								second order polynomial for capacitance
*								between two parallel segments in y 
*								layer 
*    "xy_layer cap coefficients" - capacitance coefficients in
*								second order polynomial for capacitance
*								between two parallel segments one in x 
*								and one in y layer (happens sometimes) 
*    "crossover cap coefficients" - capacitance coefficients in
*								second order polynomial for capacitance
*								between two crossing segments one in x 
*								and one in y layer 
*    "coeff value"                - assignment of value to the 
*								coefficient
*
*    example
*    
*    "x_layer cap coefficients"
*    "coeffX0 1"				an example line which sets coeffX0 to 1
*    "coeffX_sw1inv 1"			an example line which sets coeffX_sw1inv
*								to 1
*
*/

void
 cap_lib_read(text_file_name,cap_library)
char *text_file_name;		/* The input text file			*/
struct Cap_Library *cap_library;	/* the capacitance library */

{
    FILE	*text_file;
    char	text_line[MAX_LINE_LENGTH];
    int		linenum = 1;
    int		readin_status = READING_NORMAL;
    int		number_of_items = 0;
	float  value;

    /* Open the text file */
    if( (text_file = fopen(text_file_name,"r")) == NULL) {
        printf("\nE: \"%s\" file not found.\n",text_file_name);
	exit( -1 );
    }

    /* Read file, line by line */
    for(linenum = 1; ;linenum++) {
        if(fgets(text_line,MAX_LINE_LENGTH,text_file) == NULL)
            break;
    
        /* Check for comment or empty line */
        if(text_line[0] == '!') continue;
		if(strlen(text_line) == 1) continue;
    
	
	/* Status determines whether aything specific is expected next */
	switch(readin_status) {
	case(READING_NORMAL): {
	    if(sscanf(text_line,"x_layer cap coefficients %d",
						&number_of_items) == 1){
			readin_status = READING_X_LAYER_COEFFS;
		}
	    else if(sscanf(text_line,"y_layer cap coefficients %d",
						&number_of_items) == 1){
			readin_status = READING_Y_LAYER_COEFFS;
		}
	    else if(sscanf(text_line,"xy_layer cap coefficients %d",
						&number_of_items) == 1){
			readin_status = READING_XY_LAYER_COEFFS;
		}
	    else if(sscanf(text_line,"crossover cap coefficients %d",
						&number_of_items) == 1){
			readin_status = READING_CROSSOVER_COEFFS;
		}
	    else if(sscanf(text_line,"limits %d",
						&number_of_items) == 1){
			readin_status = READING_LIMITS;
		}
		else{
			PRINT_LINE_DATA;
			printf("Unrecognized text.\n");
			exit( -1 );
		}
	    break;
    }
	case(READING_X_LAYER_COEFFS): {
	    if(sscanf(text_line,"coeffX0 %g",&value)==1) {
			cap_library->coeffX0 = value;
			DECREMENT_NUMBER_OF_ITEMS;
		}
		else if(sscanf(text_line,"coeffX_sw1inv %g",&value)==1) {
			cap_library->coeffX_sw1inv = value;
			DECREMENT_NUMBER_OF_ITEMS;
		}
		else if(sscanf(text_line,"coeffX_sw2inv %g",&value)==1) {
			cap_library->coeffX_sw2inv = value;
			DECREMENT_NUMBER_OF_ITEMS;
		}
		else if(sscanf(text_line,"coeffX_sinv %g",&value)==1) {
			cap_library->coeffX_sinv = value;
			DECREMENT_NUMBER_OF_ITEMS;
		}
		else if(sscanf(text_line,"coeffX_sinv_sq %g",&value)==1) {
			cap_library->coeffX_sinv_sq = value;
			DECREMENT_NUMBER_OF_ITEMS;
		}
	    else {
			PRINT_LINE_DATA;
			printf("Expecting more x_layer coeff data\n");
			exit( -1 );
		}
	   	break;
 	}
	case(READING_Y_LAYER_COEFFS): {
	    if(sscanf(text_line,"coeffY0 %g",&value)==1) {
			cap_library->coeffY0 = value;
			DECREMENT_NUMBER_OF_ITEMS;
		}
		else if(sscanf(text_line,"coeffY_sw1inv %g",&value)==1) {
			cap_library->coeffY_sw1inv = value;
			DECREMENT_NUMBER_OF_ITEMS;
		}
		else if(sscanf(text_line,"coeffY_sw2inv %g",&value)==1) {
			cap_library->coeffY_sw2inv = value;
			DECREMENT_NUMBER_OF_ITEMS;
		}
		else if(sscanf(text_line,"coeffY_sinv %g",&value)==1) {
			cap_library->coeffY_sinv = value;
			DECREMENT_NUMBER_OF_ITEMS;
		}
		else if(sscanf(text_line,"coeffY_sinv_sq %g",&value)==1) {
			cap_library->coeffY_sinv_sq = value;
			DECREMENT_NUMBER_OF_ITEMS;
		}
	    else {
			PRINT_LINE_DATA;
			printf("Expecting more y_layer coeff data\n");
			exit( -1 );
		}
	   	break;
 	}
	case(READING_CROSSOVER_COEFFS): {
	    if(sscanf(text_line,"coeffC0 %g",&value)==1) {
			cap_library->coeffC0 = value;
			DECREMENT_NUMBER_OF_ITEMS;
		}
		else if(sscanf(text_line,"coeffC_w1 %g",&value)==1) {
			cap_library->coeffC_w1 = value;
			DECREMENT_NUMBER_OF_ITEMS;
		}
		else if(sscanf(text_line,"coeffC_w2 %g",&value)==1) {
			cap_library->coeffC_w2 = value;
			DECREMENT_NUMBER_OF_ITEMS;
		}
		else if(sscanf(text_line,"coeffC_w1w2 %g",&value)==1) {
			cap_library->coeffC_w1w2 = value;
			DECREMENT_NUMBER_OF_ITEMS;
		}
	    else {
			PRINT_LINE_DATA;
			printf("Expecting more crossover coeff data\n");
			exit( -1 );
		}
	   	break;
 	}
	case(READING_LIMITS): {
	    if(sscanf(text_line,"s_min %g",&value)==1) {
			cap_library->s_min = value;
			DECREMENT_NUMBER_OF_ITEMS;
		}
		else if(sscanf(text_line,"s_max %g",&value)==1) {
			cap_library->s_max = value;
			DECREMENT_NUMBER_OF_ITEMS;
		}
		else if(sscanf(text_line,"w_min %g",&value)==1) {
			cap_library->w_min = value;
			DECREMENT_NUMBER_OF_ITEMS;
		}
		else if(sscanf(text_line,"w_max %g",&value)==1) {
			cap_library->w_max = value;
			DECREMENT_NUMBER_OF_ITEMS;
		}
	    else {
			PRINT_LINE_DATA;
			printf("Expecting more limit data\n");
			exit( -1 );
		}
	   	break;
 	}
	default: {
	    PRINT_LINE_DATA;
	    printf("Program error. Nothing expected - Why not?\n");
	    exit( -1 );
	}
    } /* End of switch */
    } /* End of text line processor */
}

