/*      layout_ds.c
 *
 *   File:       layout_ds.c
 *   Original Author: Nicholas Weiner
 *   Date:       18th March 1988
 *   Modified by Umakanta Choudhury
 *   Date:       9th June 1989
 *   Last modified:  Enrico Malavasi,  Aug. 1992
 *
 *
 *   Routines to build the channel data structure.
 */

#include <stdio.h>
#include <string.h>
#include "utility.h"
#include "layout.h"
#define OUT_OF_MEMORY(code) {printf("Out of memory\n"); exit( -1 );}


/*
*   START ROUTING CHANNEL
*
* Initializes a new channel structure. 
*
* Return OK or ERROR.
*/

REPORT
  rr_dsStartChannel(channel)
struct Channel *channel;
{
    channel->left = HUGE;
    channel->right = -HUGE;
    channel->top = -HUGE;
    channel->bottom = HUGE;
    channel->num_nets = 0;
    channel->net_list = NULL;
    channel->num_top_pins = 0;
    channel->top_pins = NULL;
    channel->num_bottom_pins = 0;
    channel->num_left_crosses = 0;
    channel->left_crosses = NULL;
    channel->num_right_crosses = 0;
    channel->right_crosses = NULL;
    channel->num_bottom_crosses = 0;
    channel->bottom_crosses = NULL;
    channel->num_top_crosses = 0;
    channel->top_crosses = NULL;
    channel->left_chan = NULL;
    channel->right_chan = NULL;
    channel->top_chan_list = NULL;
    channel->last_top_chan_link = NULL;
    channel->bottom_chan_list = NULL;
    channel->last_bottom_chan_link = NULL;
    channel->top_pin_lists = NULL;
    channel->bottom_pin_lists = NULL;
    return OK;
}

void
  rr_dsSetNumberOfNets(number_of_nets,channel)
int number_of_nets;
struct Channel *channel;
{
    channel->num_nets += number_of_nets;
}

void
  rr_dsSetNumberOfCorners(number_of_corners,channel)
int number_of_corners;
struct Channel *channel;
{
    if( (number_of_corners != 2) && (number_of_corners != 4) )   {
	sprintf(RRwarning,"Roadrunner deals only with rectangular channels");
	rrPrintWarning(W_NUMBER_OF_CORNERS);
    }
}


/*
*   SET DEFAULT NET SEPARATION 
*
*  Returns OK or ERROR.
*/
REPORT
  rr_dsSetDefaultNetSeparation(default_net_separation,channel)
float default_net_separation;
struct Channel *channel;
{
    channel->default_net_separation = default_net_separation;
    
    return OK;
}

/*
*   SET CHANNEL-NET SEPARATION 
*
*  Returns OK or ERROR.
*/
REPORT rr_dsSetChanNetSeparation(chan_net_separation,channel)
float chan_net_separation;
struct Channel *channel;
{
    channel->chan_net_separation = chan_net_separation;
    
    return OK;
}

/*
*   SET INFLUENCE DISTANCE 
*
*  Returns OK or ERROR.
*/
REPORT rr_dsSetInfluenceDistance(influence_distance,channel)
float influence_distance;
struct Channel *channel;
{
    channel->influence_dist= influence_distance;
    
    return OK;
}


/*
*   SET DEFAULT NET WIDTH 
*
*  Returns OK or ERROR.
*/
REPORT rr_dsSetDefaultNetWidth(default_net_width,channel)
float default_net_width;
struct Channel *channel;
{
    channel->default_net_width = default_net_width;
    
    return OK;
}

/*
*  SET LAYOUT RESOLUTION 
*
*  Returns OK;
*/
REPORT rr_dsSetLayoutResolution(resolution,channel)
float resolution;
struct Channel *channel;
{
    channel->layout_resolution = resolution;
    
    return OK;
}


/*
*    SET CORNER POSITION
*
*  Compares x and y positions of given corner with lower left
*  and upper right positions and changes these positions as 
*  necessary.
*
*  Returns OK.
*/
REPORT rr_dsSetCorner(xpos,ypos,channel)
float xpos, ypos;
struct Channel *channel;
{
    channel->left = MIN(xpos,channel->left);
    channel->right = MAX(xpos,channel->right);
    channel->bottom = MIN(ypos,channel->bottom);
    channel->top = MAX(ypos,channel->top);
    
    return OK;
}

/*
*  RECORD PIN
*
*  Puts new pin into the data structure. Name of net and pin position
*  are given. First we find out whether we already have the net.
*  If we do then just add new pin to net. Otherwise have to first create
*  the net.
*
*  Returns OK or ERROR.
*/
REPORT
  rr_dsRecordPin(net_name, xpos, ypos, channel)
char *net_name;
float xpos, ypos;
struct Channel *channel;
{
    struct Net *rr_dsFindNet();

    struct Net *net;
    struct Pin *pin;
    struct PinLink *pin_link;
    int edge;				/* Left, right, top or bottom	*/

    /* Get/create the net */
    if( (net = rr_dsFindNet(net_name,channel)) == NULL) {
	fprintf(stderr, "** Net %s not found\n", net_name);
	exit( -1 );
    }

    /* Decide which edge the pin is on */
    edge = rr_dsWhichEdge(xpos,ypos,channel);
    
    /* Make the pin */
    if( (pin=ALLOC(struct Pin,1 )) == NULL) OUT_OF_MEMORY(ERROR);
    pin->x_position = xpos;
    pin->y_position = ypos;
    pin->edge = edge;
    pin->net = net;
    pin->true_half_width = net->true_half_width;
	
    /* Add the pin to the net */
    if( (pin_link = (struct PinLink *)nwAddLink(sizeof(struct PinLink),
       &net->pins,NULL,net->pins)) == NULL) OUT_OF_MEMORY(ERROR);
    pin_link->pin = pin;

    /* Add the pin to the channel */
    if(edge == TOP)
	{
	channel->num_top_pins++;
        if( (pin_link = (struct PinLink *)nwAddLink(sizeof(struct PinLink),
	   &channel->top_pins,NULL,channel->top_pins)) == NULL) 
	   OUT_OF_MEMORY(ERROR);
        pin_link->pin = pin;
	}
    else if(edge == BOTTOM)
	{
	channel->num_bottom_pins++;
        if( (pin_link = (struct PinLink *)nwAddLink(sizeof(struct PinLink),
	   &channel->bottom_pins,NULL,channel->bottom_pins)) == NULL) 
	   OUT_OF_MEMORY(ERROR);
        pin_link->pin = pin;
	}
    else if(edge == RIGHT)
	{
	channel->num_right_pins++;
        if( (pin_link = (struct PinLink *)nwAddLink(sizeof(struct PinLink),
	   &channel->right_pins,NULL,channel->right_pins)) == NULL) 
	   OUT_OF_MEMORY(ERROR);
        pin_link->pin = pin;
	}
    else if(edge == LEFT) {
	channel->num_left_pins++;
        if( (pin_link = (struct PinLink *)
	     nwAddLink(sizeof(struct PinLink), &channel->left_pins,NULL,channel->left_pins)) == NULL) 
	  OUT_OF_MEMORY(ERROR);
        pin_link->pin = pin;
    } else {
	    printf("Error - Pin not on top,bottom,left or right edge of channel\n");
	    exit( -1 );
	}
}

/*
*  RECORD END CROSSING
*
*  Records the fact that the given net crosses left,right,top or bottom
*  end os channel.
*  First find out whether we already the net. If we do then just record 
*  the crossing. Otherwise have to first create the net.
*
*  Returns OK or ERROR.
*/

REPORT rr_dsRecordCrossing(net_name,edge,channel)
char *net_name;
int edge;
struct Channel *channel;
{
    struct Net *rr_dsFindNet();

    struct Net *net;
    struct NetLink *net_link;

    /* Get/create the net */
    if( (net = rr_dsFindNet(net_name,channel)) == NULL) return ERROR;

    /* Add the crossing to the channel */
    if(edge == LEFT)
	{
	channel->num_left_crosses++;
        if( (net_link = (struct NetLink *)nwAddLink(sizeof(struct NetLink),
	   &channel->left_crosses,NULL,channel->left_crosses)) == NULL) 
	   OUT_OF_MEMORY(ERROR);
        net_link->net = net;
        net->left_end_crossing = TRUE;
	}
    else if(edge == RIGHT)
	{
	channel->num_right_crosses++;
        if( (net_link = (struct NetLink *)nwAddLink(sizeof(struct NetLink),
	   &channel->right_crosses,NULL,channel->right_crosses)) == NULL) 
	   OUT_OF_MEMORY(ERROR);
        net_link->net = net;
        net->right_end_crossing = TRUE;
	}
    else if(edge == BOTTOM)
	{
	channel->num_bottom_crosses++;
        if( (net_link = (struct NetLink *)nwAddLink(sizeof(struct NetLink),
	   &channel->bottom_crosses,NULL,channel->bottom_crosses)) == NULL) 
	   OUT_OF_MEMORY(ERROR);
        net_link->net = net;
        net->bottom_end_crossing = TRUE;
	}
    else if(edge == TOP)
	{
	channel->num_top_crosses++;
        if( (net_link = (struct NetLink *)nwAddLink(sizeof(struct NetLink),
	   &channel->top_crosses,NULL,channel->top_crosses)) == NULL) 
	   OUT_OF_MEMORY(ERROR);
        net_link->net = net;
        net->top_end_crossing = TRUE;
	}
    else
	{
	printf("Error - Bad edge in rr_dsRecordCrossing.\n");
	return ERROR;
	}

    return OK;
}


/*
*  MAKE NEW NET
*
*  Allocates space for new net structure, initialises structurew members,
*  and puts net into the net tree. 
*
*  The WIDTH of the new net is set to the current value of the
*  channel->default_net_width.
*
*  Returns pointer to the new net, or NULL on error.
*/
struct Net *rr_dsMakeNet(net_name,channel)
char *net_name;
struct Channel *channel;

    {
    struct Net *net;			/* The new Net		*/
    struct NetLink *net_link;


    /* Allocate memory */
    if( (net = ALLOC(struct Net, 1)) == NULL) return NULL;

    /* Fill net structure */
    net->name = util_strsav(net_name);
    net->channel = channel;
    net->pins = NULL;
    net->left_end_crossing = FALSE;
    net->right_end_crossing = FALSE;
    net->bottom_end_crossing = FALSE;
    net->top_end_crossing = FALSE;

    net->true_half_width = channel->default_net_width/2.0;
    net->left = HUGE;
    net->right = -HUGE;

    /* Put net into binary tree (has no error check) */
    avl_insert(channel->net_tree,net->name,net);

    /* Put net into the linear linked list of nets */
    if( (net_link = (struct NetLink *)nwAddLink(sizeof(struct NetLink),
      &channel->net_list,NULL,channel->net_list)) == NULL) OUT_OF_MEMORY(NULL);
    net_link->net = net;

    return net;
    }


/*
*  WHICH EDGE
*
* Given coordinate pair of point and channel - decides which channel edge 
* the point is one. If the point is on one of the corners, then either
* top or bottom is assumed.
*
* Return LEFT, RIGHT, TOP, BOTTOM or NO_EDGE.
*/
int rr_dsWhichEdge(xpos,ypos,channel)
float xpos, ypos;
struct Channel *channel;

    {
    if( (ypos < channel->bottom) || (ypos > channel->top)
     || (xpos < channel->left) || (xpos > channel->right) )
	return NO_EDGE;
    
    if(ypos == channel->top) return TOP;
    if(ypos == channel->bottom) return BOTTOM;
    if(xpos == channel->left) return LEFT;
    if(xpos == channel->right) return RIGHT;

    return NO_EDGE;
    }

/*
*  SET TRUE NET WIDTH
*
*  Given a net and a width. Finds the net (creating it if necessary)
*  and sets its (true) width as specified.
*
*  Returns OK or ERROR.
*/
REPORT rr_dsSetNetWidth(net_name,true_width,channel)
char *net_name;
float true_width;
struct Channel *channel;
    {
    struct Net *rr_dsFindNet();

    struct Net *net;
    struct PinLink *pin_link;

    /* Get/create the net */
    if( (net = rr_dsFindNet(net_name,channel)) == NULL) return ERROR;

    /* Set the width */
    net->true_half_width = true_width/2;

    /* And all of its pins */
    for(pin_link=net->pins; pin_link!=NULL; pin_link=pin_link->next)
	{
	pin_link->pin->true_half_width = true_width/2;
	}

    return OK;
    }

/*
*  SET ORIENTATION OF CHANNEL
* top-bottom or left-right
*
*  Returns OK or ERROR.
*/
REPORT rr_dsSetChanOrient(orient,channel)
char *orient;
struct Channel *channel;
    {

	if(strcmp(orient,"top_bottom") == 0) 
		channel->orient = TOP_BOTTOM;
	else if(strcmp(orient,"left_right") == 0) 
		channel->orient = LEFT_RIGHT;
	else printf("error in orientation of channel %s\n",channel->name); 
    return OK;
    }
