/*		READ PERFORMANCE AND CIRCUIT 
*		RELATED INFO FILE
*		-----------------------------
*
*   File:	perf_read.c
*   Author: Umakanta Choudhury
*   Date:   14th June
*
*/
extern int strcmp();

#include <stdio.h>
#include "utility.h"
#include "layout.h"
#include "circuit.h"
#include "avl.h"
#define MAX_LINE_LENGTH 200
#define PRINT_LINE_DATA {printf("%s: line %d\n",text_file_name,linenum);}
#define CHECK_RETURN(X) \
  {int R; R=(X); if(R!=OK){PRINT_LINE_DATA;if(R==ERROR)return ERROR;}}
#define DECREMENT_NUMBER_OF_ITEMS \
  {if( (--number_of_items) == 0) readin_status = READING_NORMAL;}

/*  READIN STATUS DEFINITIONS */
#define READING_NORMAL 0
#define READING_PERFORMANCE 1
#define READING_NODES 2
#define READING_MATCHEDPAIRS 3
#define READING_CRITPAIRS 4

/*
*  READ TEXT FILE
*
*
*  "critical pair file" input syntax:
*  --------------------------
*
*    "number of performance functions [value]"  
*							- of the circuit of interest
*							- (should be the first line )
*    followed by:
*  	"performance function [name]"   - name of the function
*    	followed by one or both of: 
*   	 
* 		"maximum positive change allowed [value]" - (positive number) 
*								due to routing parasitics
*
* 		"maximum negative change allowed [value]" - (negative number) 
*									due to routing parasitics
*    	(repeat format for all the performance functions)
*  
*
*  	followed by:
*	"critpair ([net1],[net2]) [number] sensitivities"
*			- critpair net names followed by number of 
*				sensitivities going to be specified below
*   	followed by:
*		"performance [name] sensitivity [value]"
*            < required number of times >
*
*  		(repeat format for all the critical pairs)
*
*
*/

static struct Circuit *thisCircuit;
static struct Layout  *thisLayout;
static int	       readin_status = READING_NORMAL;

extern  struct Net *rr_dsMakeNet();

void
  setSimulator( name )
char *name;
{
    thisCircuit->sim_name = util_strsav( name );
    VERBOSE( printf(" Simulator is %s\n", name ) );
}

void
  setNumPerf( name )
char *name;
{
    sscanf( name, "%d", &(thisCircuit->num_perf) );
    thisLayout->perf_name =  ALLOC(char *, thisCircuit->num_perf + 1);
    VERBOSE( printf(" No. Performances =  %s\n", name ) );
} 

void
  setNodeNumber( name )
char *name;
{
    sscanf( name, "%d", &(thisCircuit->num_nodes) );
    VERBOSE( printf(" No. of nodes = %d\n", thisCircuit->num_nodes ) );
}

void
  setOutNodes( name1, name2 )
char *name1, *name2;
{
    thisCircuit->out_node1_name = util_strsav( name1 );
    thisCircuit->out_node2_name = util_strsav( name2 );
    VERBOSE( printf(" Output nodes: %s, %s\n", name1, name2 ) );
}

void
  setMatchedNum( name )
char *name;
{
    sscanf( name, "%d", &(thisCircuit->num_matched_pairs) );
    VERBOSE( printf(" No. of matched pairs = %d\n", thisCircuit->num_matched_pairs ) );
}

void
  setNode( name )
char *name;
{
    static int counter = 0;
    struct Node	*node;

    ds_MakeNode(name, &node, thisCircuit);

    if(thisLayout->mode == NORMAL){
	rr_dsMakeNet(name, thisLayout->chan_list->channel);
    }

    VERBOSE( printf(" %s ", name ) );

    counter++;
    if( counter == thisCircuit->num_nodes ) {
	readin_status = READING_NORMAL;
	VERBOSE( printf("\n" ) );
    }
    if( counter > thisCircuit->num_nodes ) {
	fprintf(stderr, "ERROR: more nodes than declared\n");
	exit( -1 );
    }
}

void
  setMatchedPair( name1, name2 )
char *name1, *name2;
{
    static int counter = 0;
    struct Matched_pair *matched_pair; 

    ds_MakeMatchedPair(name1, name2, &matched_pair, thisCircuit) ;

    VERBOSE( printf(" (%s, %s) ", name1, name2 ) );

    counter++;
    if( counter == thisCircuit->num_matched_pairs) {
	readin_status = READING_NORMAL;
	VERBOSE( printf("\n" ) );
    }
    if( counter > thisCircuit->num_matched_pairs ) {
	fprintf(stderr, "ERROR: more matched pairs than declared\n");
	exit( -1 );
    }
}

setProcVariation( name )
char *name;
{
    float x;
    sscanf( name, "%f", &x);
    thisCircuit->max_process_var_percent = x;
    VERBOSE( printf(" Max process variation = %f\n", thisCircuit->max_process_var_percent ) );
}
    
setProcMismatch( name )
char *name;
{
    float x;
    sscanf( name, "%f", &x );
    thisCircuit->max_mismatch_percent = x;
    VERBOSE( printf(" Max process mismatch = %f%c\n", thisCircuit->max_mismatch_percent, '\%' ) );
}

static struct Performance *thisPerf; 

void
  setPerfHeader( name, type, numConstr, simFile )
char *name;
char *type;
char *numConstr;
char *simFile;
{
    static int counter = 1;

    ds_MakePerformance(name, type, counter, &thisPerf, thisCircuit);
    
    thisLayout->perf_name[counter] = util_strsav( name );
    thisPerf->ckt_file_name = util_strsav( simFile );
    counter++;

    VERBOSE( printf(" Performance %s\n", name ));
}

void
  setPerfChange( name, type )
char *name;
int   type;
{
    float x;

    sscanf( name, "%f", &x );
    if( (x*type) < 0 ) {
	fprintf(stderr, "ERROR: Wrong sign for the maximum change \n");
	exit( -1 );
    }
    if( type > 0 ) {
	thisPerf->max_pos_change = x;
	thisPerf->delta_pos_min  = 0;
	thisPerf->pos_eqn_num    = 0;

	thisPerf->pos_constraint = YES;
    } else {
	thisPerf->max_neg_change = x;
	thisPerf->delta_neg_min  = 0;
	thisPerf->neg_eqn_num    = 0;

	thisPerf->neg_constraint = YES;
    }
}

REPORT
  perf_read(text_file_name, circuit,layout)
char           *text_file_name;		/* The input text file			*/
struct Circuit *circuit;	        /* The circuit structure */
struct Layout  *layout;	                /* The layout structure */

{
    FILE       *text_file;		/* The input text file			*/
    char	text_line[MAX_LINE_LENGTH];
    int		linenum = 1;
    int		prev_linenum = 1;
    int		number_of_items = 0;
    float  value;
    struct	CriticalPairLink *crit_pair; 
    int		num_perf;
    int		num_critical ;
    int		i ;
    char	name[30];
    char	*nameA[30];
    char	*nameB[30];
    struct Net	*net;
    struct Node	*node;
    int	number;
    int	perf_num;
    
    struct Net	*netA;
    struct Net	*netB;
    struct Net	*save_netA;
    struct Net	*save_netB;
    struct Node	*nodeA;
    struct Node	*nodeB;
    int NETA_PRESENT;
    int NETB_PRESENT;
    int FOUND_NETA;
    int FOUND_NETB;
    int type;
    struct ChannelLink *chan_link;
    struct Channel *channel;
    struct Chan_info *chan_info;


    thisCircuit  =  circuit;
    thisLayout   =  layout;

    yyopenfile(text_file_name);
    yyparse();

    /* Open the text file */
    if( (text_file = fopen(text_file_name,"r")) == NULL) {
        fprintf(stderr, "\nE: \"%s\" file not found.\n",text_file_name);
        return ERROR;
    }
    if(fgets(text_line,MAX_LINE_LENGTH,text_file) == NULL){
	printf("nothing in the file %s\n",text_file_name);
	exit(-1);
    }
    if(sscanf(text_line,"num_performance %d, simulator %s",&num_perf, name) == 2) {
/*	circuit->num_perf = num_perf; */
/*	circuit->sim_name = util_strsav(name);; */
/*	layout->perf_name = (char **)malloc((num_perf + 1)*sizeof(char *)); */ ;
    } else {
	printf("first line of file %s should specify number of performance functions\n",text_file_name);
	exit(-1);
    }

    /* Read file, line by line */
    perf_num = 1;
    for(linenum = 2; ;linenum++) {
        if(fgets(text_line,MAX_LINE_LENGTH,text_file) == NULL)
            break;
    
        /* Check for comment or empty line */
        if(text_line[0] == '!') continue;
		if(strlen(text_line) == 1) continue;
	/* first read the performance data */	

	/* Status determines whether aything specific is expected next */
	switch(readin_status) {
	    case(READING_NORMAL): {
		if(sscanf(text_line,"performance type num_constraints and sim_file: %s %d %d %s",nameA,&type,&number_of_items,nameB) == 4){
		    readin_status = READING_PERFORMANCE;
		    
/*		    if(ds_MakePerformance(nameA,type,perf_num,&performance, circuit) != OK) */
/*		      return ERROR; */
/*		     */
/*		    *(layout->perf_name + perf_num) = util_strsav(nameA); */
/*		    performance->ckt_file_name = util_strsav(nameB); */
/*		    perf_num++; */
		}
		else if(sscanf(text_line,"number of nodes %d out_node1 %s out_node2 %s",
			       &number_of_items,nameA,nameB) == 3) {
		    /*		circuit->num_nodes = number_of_items; */
		    /*		circuit->out_node1_name = util_strsav(nameA); */
		    /*		circuit->out_node2_name = util_strsav(nameB); */
		    readin_status = READING_NODES;
		}
		else if(sscanf(text_line,"number of matched pairs %d",
			       &number_of_items) == 1) {
		    /*		circuit->num_matched_pairs = number_of_items; */
		    readin_status = READING_MATCHEDPAIRS;
		}
		else if(sscanf(text_line,"maximum process variation in percent %g",
			       &value) == 1){
		    /*			circuit->max_process_var_percent = value; */
		}
	    else if(sscanf(text_line,"maximum mismatch in percent %g",
			&value) == 1){
/*			circuit->max_mismatch_percent = value; */
		}
	    else if(sscanf(text_line,"performance end %d", &number) == 1){
		prev_linenum = linenum;
		goto next;
	    }
	    else {
		PRINT_LINE_DATA;
		printf("text_line = %s\n",text_line);
		printf("performance function data expected\n");
		return ERROR;
		}
	    break;
	}
	case(READING_PERFORMANCE): {
	    if(sscanf(text_line,"maximum positive change %g",&value)==1) {
/*		if(value <=0){ */
/*		    PRINT_LINE_DATA; */
/*		    printf("maximum positive change should be a \ */
/*					positive number \n"); */
/*		    return ERROR; */
/*		} */
		DECREMENT_NUMBER_OF_ITEMS;
/*		if(performance->pos_constraint == NO){ */
/*		    performance->pos_constraint = YES; */
/*		} */
/*		performance->max_pos_change = value;  */
/*		performance->delta_pos_min = 0;  */
/*		performance->pos_eqn_num = 0;  */
	    }
	    else if(sscanf(text_line,"maximum negative change %g",
			   &value)==1) {
/*		if(value >=0){ */
/*		    PRINT_LINE_DATA; */
/*		    printf("maximum negative  change should be a \ */
/*					negative number \n"); */
/*		    return ERROR; */
/*		} */
		DECREMENT_NUMBER_OF_ITEMS;
/*		if(performance->neg_constraint == NO){ */
/*		    performance->neg_constraint = YES; */
/*		} */
/*		performance->max_neg_change = value;  */
/*		performance->delta_neg_min = 0;  */
/*		performance->neg_eqn_num = 0;  */
	    }
	    else{
		PRINT_LINE_DATA;
		printf("Unrecognised text.\n");
		return ERROR;
	    }
	    break;
	}
	case(READING_NODES): {
	    if(sscanf(text_line,"%s",name) == 1) {
		/*		if(ds_MakeNode(name,&node,circuit) != OK){ */
		/*		    printf("error in ds_MakeNode() \n"); */
		/*		    return ERROR; */ 
		/*	    } */
		/*		if(layout->mode == NORMAL){ */
		/*		    if(rr_dsMakeNet(name,layout->chan_list->channel) == NULL){ */
		/*			printf("error in rr_dsMakeNet() \n"); */
		/*			return ERROR; */
		/*		    } */
		/*		} */
		DECREMENT_NUMBER_OF_ITEMS;
	    }
	    break;
	}
	case(READING_MATCHEDPAIRS): {
	    if(sscanf(text_line,"%s %s",nameA,nameB)==2) {
		/*			if(ds_MakeMatchedPair(nameA,nameB,&matched_pair,circuit)  */
		/*				!= OK){ */
		/*				printf("error in ds_MakeMatchedPair() \n"); */
		/*				return ERROR; */
		/*			} */
		DECREMENT_NUMBER_OF_ITEMS;
	    }
	    break;
	}
	default: {
	    PRINT_LINE_DATA;
	    printf("Program error. Nothing expected - Why not?\n");
	    return ERROR;
	}
 	} /* End of switch */
	} 
next: ;
	for(linenum = prev_linenum+1; ;linenum++) {
        if(fgets(text_line,MAX_LINE_LENGTH,text_file) == NULL)
            break;
    
        /* Check for comment or empty line */
        if(text_line[0] == '!') continue;
		if(strlen(text_line) == 1) continue;

		/* now read the critical pair data */	
		i = 0;
		/* Status determines whether aything specific is expected next */
		switch(readin_status) {
		case(READING_NORMAL): {
		    if(sscanf(text_line,"critical pair %s %s %d \
				sensitivities",nameA,nameB,&number_of_items) == 3){
				readin_status = READING_CRITPAIRS;
				if(ds_MakeCriticalPair(nameA,nameB,&crit_pair,num_perf,
					layout) != OK){ 
					printf("error in ds_MakeCriticalPair() \n");
					return ERROR;
				}
			}
			else{
				PRINT_LINE_DATA;
				printf("text_line = %s\n",text_line);
				printf("critical pair data expected \n");
				return ERROR;
			}
		    break;
   	    }
		case(READING_CRITPAIRS): {
		    if(sscanf(text_line,"performance %s sensitivity %g",
						name,&value)==2) {
			for(i=1; i<= circuit->num_perf; i++){
				if(strcmp(*(layout->perf_name + i),name) == 0){
					if(value >= 0){
						*(crit_pair->pos_sens + i) = value; 	
					}
					else{
						*(crit_pair->neg_sens + i) = value; 	
					}
					break;
				}
				if(i == circuit->num_perf){
					printf("performance %s not found\n",name);
					return ERROR;
				}
			}
			DECREMENT_NUMBER_OF_ITEMS;
		}
		else{
			PRINT_LINE_DATA;
			printf("critical sensitivity data expected.\n");
			return ERROR;
		}
	    break;
 	}
	default: {
	    PRINT_LINE_DATA;
	    printf("Program error. Nothing expected - Why not?\n");
	    return ERROR;
	}
 	} /* End of switch */
	} /* End of text line processor */

    return OK;
}
