/*   cap_maxmin.c
 *
 *   Author:     Umakanta Choudhury
 *   Date:       9th June 1989
 *   Last modified:  Enrico Malavasi,  Aug. 1992
 *
 *   Purpose:     Estimate maximum and minimum capacitances between nets
 *                in a channel
 */
#include <stdio.h>
#include <string.h>
#include "layout.h"

extern double cap_parallel();

static REPORT
  cx_maxmin(chan_info)
struct Chan_info *chan_info;
{
    struct Channel *channel;
    struct ChannelLink *top_chan_link;
    struct ChannelLink *bottom_chan_link;
    struct Channel	*top_chan;
    struct Channel	*bottom_chan;
    float net_separation;
    float chan_net_sep;
    float channel_width;
    struct Net *net1;
    struct Net *net2;
    float W1; /* width of net1 */
    float W2; /* width of net2 */
    float left; 
    float right; 
    float d_min; /* minimum distance between two adjacent nets running
		    parallel used for computation of maximum capacitance */
    float cap; 
    float max_common_span; 
    int	first_pin;
    float	chan_intersect;
    float	Cx_max_top_12;
    float	Cx_max_top_21;
    float	Cx_max_bottom_12;
    float	Cx_max_bottom_21;
    float	Cx_max_top;
    float	Cx_max_bottom;

    DEBUGG(printf("cx_maxmin\n"));

    channel = chan_info->channel;
    net_separation = channel->default_net_separation;
    chan_net_sep= channel->chan_net_separation;
    channel_width = channel->top - channel->bottom;
    net1 = chan_info->netA;
    net2 = chan_info->netB;
    W1 = 2*(net1->true_half_width);
    W2 = 2*(net2->true_half_width);


    /* compute the maximum common span */
    left = MAX(net1->max_span_left,net2->max_span_left);
    right = MIN(net1->max_span_right,net2->max_span_right);
    if(right < left) max_common_span = 0;
    else max_common_span = right - left;
    
    /* first the coupling inside the channel */
    d_min = net_separation;
    cap = cap_parallel(W1,W2,d_min,X_LAYER);
    chan_info->Cx_max = max_common_span * cap;
    chan_info->Cx_min = 0; /* when the separation is very large */
    
    chan_info->Lx_max += max_common_span;
    
    
#ifdef DEBUG
    printf("net1 %s ; net2  %s\n",net1->name,net2->name);
    printf("net1_max_left = %f net1_max_right = %f\n",
	   net1->max_span_left,net1->max_span_right);
    printf("net2_max_left = %f net2_max_right = %f\n",
	   net2->max_span_left,net2->max_span_right);
    printf("max_common_span = %f W1 = %f W2 = %f\n",max_common_span,
	   W1,W2);
    printf("d_min = %f cap = %f \n",d_min,cap);
#endif 
									
    /* then consider the inter channel coupling w.r.t top and 
       bottom channels (coupling w.r.t left and right channels 
       will be considered while considering those channels to 
       avoid duplication) */
    
    /* compute maximum coupling w.r.t top channels */
    
    Cx_max_top_12 = 0; /* net1 inside and net2 outside */
    Cx_max_top_21 = 0; /* net2 inside and net1 outside */
    
    for(top_chan_link = channel->top_chan_list; top_chan_link != NULL;
	top_chan_link = top_chan_link->next){
	
	top_chan = top_chan_link->channel;
	
	/* look for coupling between 
	   net1 inside and net2 in top_chan */ 
	cx_top_maxmin(chan_info,top_chan,net1,net2);

	Cx_max_top_12 += chan_info->Cx_max_top;

	DEBUGG(printf("Cx_max_top_12 = %g \n",Cx_max_top_12));
	
	/* look for coupling between 
	   net2 inside and net1 in top_chan */ 
	cx_top_maxmin(chan_info,top_chan,net2,net1);

	Cx_max_top_21 += chan_info->Cx_max_top;

	DEBUGG(printf("Cx_max_top_21 = %g \n",Cx_max_top_21));
    }
    
    /* compute maximum coupling w.r.t bottom channels */
    
    Cx_max_bottom_12 = 0; /* net1 inside and net2 outside */
    Cx_max_bottom_21 = 0; /* net2 inside and net1 outside */
    
    for(bottom_chan_link = channel->bottom_chan_list; bottom_chan_link 
	!= NULL; bottom_chan_link = bottom_chan_link->next){
	
	bottom_chan = bottom_chan_link->channel;
	
	/* look for coupling between 
	   net1 inside and net2 in bottom_chan */ 
	cx_bot_maxmin(chan_info,bottom_chan,net1,net2);

	Cx_max_bottom_12 += chan_info->Cx_max_bottom;
	
	/* look for coupling between 
	   net2 inside and net1 in bottom_chan */ 
	cx_bot_maxmin(chan_info,bottom_chan,net2,net1);

	Cx_max_bottom_21 += chan_info->Cx_max_bottom;	
    }
    
    /*
      if(max_common_span == 0){
      Cx_max_top = Cx_max_top_12 + Cx_max_top_21;
      Cx_max_bottom = Cx_max_bottom_12 + Cx_max_bottom_21;
      }
      else{
      Cx_max_top = MAX(Cx_max_top_12, Cx_max_top_21);
      Cx_max_bottom = MAX(Cx_max_bottom_12, Cx_max_bottom_21);
      }
      */
    
    Cx_max_top = Cx_max_top_12 + Cx_max_top_21;
    Cx_max_bottom = Cx_max_bottom_12 + Cx_max_bottom_21;

    chan_info->Cx_max += Cx_max_top + Cx_max_bottom;
    return(OK);
}


static REPORT
  cy_maxmin(chan_info)
struct Chan_info *chan_info;
{
    /* fixed pins of net1 and pins of net2 */
    if(net1fix_net2(chan_info) != OK) 		return(ERROR);

    /* top floating pins of net1 and pins of net2 */
    if(net1fl_net2(chan_info,TOP) != OK) 	return(ERROR);
    
    /* bottom floating pins of net1 and pins of net2 */
    if(net1fl_net2(chan_info,BOTTOM) != OK) 	return(ERROR);

    return OK;
}

static REPORT
  ccross_maxmin(chan_info)
struct Chan_info *chan_info;
{
    struct Channel *channel;
    struct Channel *top_chan;
    struct Channel *bottom_chan;
    struct ChannelLink *top_chan_link;
    struct ChannelLink *bottom_chan_link;
    struct Net *net1;
    struct Net *net2;
    double left; 
    double right; 
    double W1; 
    double W2; 
    double minW; 
    double max_common_span; /* maximum common span in x direction of 
			       the two nets */
    double min_common_span; /* minimum common span in x direction of 
			       the two nets */
    float chan_net_sep;
    
    channel = chan_info->channel;
    chan_net_sep = channel->chan_net_separation;
    net1 = chan_info->netA;
    net2 = chan_info->netB;
    W1 = 2*(net1->true_half_width);
    W2 = 2*(net2->true_half_width);
    minW = MIN(W1,W2);
    

    /* compute the maximum common span */
    left = MAX(net1->max_span_left,net2->max_span_left);
    right = MIN(net1->max_span_right,net2->max_span_right);
    if(right < left) max_common_span = 0;
    else max_common_span = right - left;
    
    chan_info->Ccross_1above2 = 0;
    chan_info->Ccross_2above1 = 0;
    
    if(max_common_span != 0){
	
	/* compute the maximum crossover capacitance due to
	   fixed pins */
	
	if(fixpin_cross(chan_info,left,right) != OK){
	    return(ERROR);
	}
	
	/* compute the maximum crossover capacitance due to
	   floating pins */
	
	/* scan through the top channels */
	for(top_chan_link = channel->top_chan_list; 
	    top_chan_link != NULL;
	    top_chan_link = top_chan_link->next){
	    
	    top_chan = top_chan_link->channel;
	    
	    if((top_chan->top - chan_net_sep  > left) && 
	       (top_chan->bottom + chan_net_sep  < right)){
		/* intersects with the common span */
		if(flpin_cross(chan_info,top_chan,TOP) != OK){
		    return(ERROR);
		}
	    }
	}
	for(bottom_chan_link = channel->bottom_chan_list; 
	    bottom_chan_link != NULL;
	    bottom_chan_link = bottom_chan_link->next){
	    
	    bottom_chan = bottom_chan_link->channel;
	    
	    if((bottom_chan->top - chan_net_sep > left) && 
	       (bottom_chan->bottom + chan_net_sep < right)){
		/* intersects with the common span */
		if(flpin_cross(chan_info,bottom_chan,BOTTOM) != OK){
		    return(ERROR);
		}
	    }
	}
    }
    
    chan_info->Ccross_max = MAX(chan_info->Ccross_1above2,
				chan_info->Ccross_2above1);
    
    
    /* compute the minimum common span */
    left = MAX(net1->min_span_left,net2->min_span_left);
    right = MIN(net1->min_span_right,net2->min_span_right);
    if(right < left) min_common_span = 0;
    else min_common_span = right - left;
    
    chan_info->Ccross_1above2 = 0;
    chan_info->Ccross_2above1 = 0;
    
    if(min_common_span != 0){
	
	/* compute the minimum crossover capacitance due to
	   fixed pins */
	
	if(fixpin_cross(chan_info,left,right) != OK){
	    return(ERROR);
	}
	/* compute the minimum crossover capacitance due to
	   floating pins */
	
	/* scan through the top channels */
	for(top_chan_link = channel->top_chan_list; 
	    top_chan_link != NULL;
	    top_chan_link = top_chan_link->next){
	    
	    top_chan = top_chan_link->channel;
	    
	    
	    if((top_chan->top - chan_net_sep - minW < right) && 
	       (top_chan->bottom + chan_net_sep + minW > left)){
		/* must intersect the minimum common span */
		if(flpin_cross(chan_info,top_chan,TOP) != OK){
		    return(ERROR);
		}
	    }
	}
	/* scan through the bottom channels */
	for(bottom_chan_link = channel->bottom_chan_list; 
	    bottom_chan_link != NULL; 
	    bottom_chan_link = bottom_chan_link->next){
	    
	    bottom_chan = bottom_chan_link->channel;
	    
	    if((bottom_chan->top - chan_net_sep - minW < right) 
	       && (bottom_chan->bottom + chan_net_sep + minW > left)){
		/* must intersect the minimum common span */
		if(flpin_cross(chan_info,bottom_chan,BOTTOM) != OK){
		    return(ERROR);
		}
	    }
	}
    }
    
    chan_info->Ccross_min = MIN(chan_info->Ccross_1above2,
				chan_info->Ccross_2above1);
    
    return(OK);
}


REPORT
  cap_maxmin(chan_info)
struct Chan_info *chan_info;
{

    /* compute Cx_max,Cx_min (maximum and minimum cap. betw. segments in x dir.) */
    cx_maxmin( chan_info );

    /* compute Cy_max,Cy_min (maximum and minimum capacitances
       between segments in y direction) */
    if(cy_maxmin(chan_info) != OK)      return(ERROR);

    /* compute Ccross_max,Ccross_min (maximum and minimum crossover 
       capacitances between the two nets) */
    if(ccross_maxmin(chan_info) != OK) 	return(ERROR);

    chan_info->C_max = chan_info->Cx_max + chan_info->Cy_max + chan_info->Ccross_max;
    chan_info->C_min = chan_info->Cx_min + chan_info->Cy_min + chan_info->Ccross_min;
    chan_info->C_min_shield = chan_info->C_min;
    
    return(OK);
}
