/*   keep_crit.c
 *
 *   Author:     Umakanta Choudhury
 *   Date:       2nd July 1989
 *   Last modified:  Enrico Malavasi,  Aug. 1992
 *
 *   Purpose: Eliminate non-critical parasitics from further analysis
 */
#include <stdio.h>
#include "utility.h"
#include "macros.h"
#include "layout.h"
#include "circuit.h"

#define OUT_OF_MEMORY(code) {fprintf(stderr, "Out of memory\n"); exit(code);}


static void
  init_crit(critpair,nameA,nameB,num_perf)
struct CriticalPairLink *critpair;
char *nameA;
char *nameB;
int num_perf;
{
    int i;
    
    critpair->netA_name = util_strsav(nameA);
    critpair->netB_name = util_strsav(nameB);
    critpair->chan_info_list = NULL;
    
    if((critpair->pos_sens=ALLOC(float, num_perf+1)) == NULL) {
	OUT_OF_MEMORY( -1 );
    }
    if((critpair->neg_sens=ALLOC(float, num_perf+1)) == NULL) {
	OUT_OF_MEMORY( -1 );
    }
    for(i = 1; i<= num_perf; i++){
	*(critpair->pos_sens + i) = 0;
    }
    for(i = 1; i<= num_perf; i++){
	*(critpair->neg_sens + i) = 0;
    }
}

void
  keep_crit(circuit, layout, debug_file, channelMode)
struct Circuit *circuit;
struct Layout *layout;
FILE *debug_file;
int   channelMode;
{
    struct Performance *perf;
    struct Sens_link *sens_link;
    struct Chan_info *chan_info;
    struct CriticalPairLink *critpair;
    struct Chan_info_list_link *chan_info_list_link;
    int	perf_num;
    float	max_perf_change;
    float	C_max;
    float	tot_pos_sens;
    float	tot_neg_sens;
    float	cap_threshold;
    int	max_comp;
    int	max_zero;
    char	*nameA;
    char	*nameB;


    DEBUGG(printf("keep_crit\n") );

    /* sort the sensitivities for each performance function */
	
    for(perf = circuit->perf_list; perf != NULL; perf = perf->next){ 
	sens_sort(&(perf->pos_sens_list));
	for(sens_link = perf->neg_sens_list; sens_link != NULL;
	    sens_link = sens_link->next){
	    /* lowest sensitivity (in magnitude) should be the 
	       first in the sorted list */
	    /* temporary */
	    sens_link->sens *= -1;
	}
	sens_sort(&(perf->neg_sens_list));
	for(sens_link = perf->neg_sens_list; sens_link != NULL;
	    sens_link = sens_link->next){
	    /* restore the actual sensitivities */
	    sens_link->sens *= -1;
	}
    }
    
    /*  compute the maximum capacitances for 
     *  various pairs ; and eliminate the non-critical 
     *	parasitics
     */
    if( channelMode ){
	for(chan_info_list_link = layout->chan_info_list_link;
	    chan_info_list_link != NULL;
	    chan_info_list_link = chan_info_list_link->next){
	    for(chan_info = chan_info_list_link->chan_info_list; 
		chan_info != NULL; chan_info = chan_info->next){
		DEBUGG( printf("keep_crit 71: channel : %s\n",chan_info->channel->name) );
		if(cap_maxmin(chan_info) != OK) {
		    exit( -1 );
		}
	    }
	}
    }
	
    layout->cap_threshold = HUGE;
    for(perf = circuit->perf_list; perf != NULL; perf = perf->next){ 
	perf_num = perf->number;
	tot_pos_sens = 0;
	tot_neg_sens = 0;
	if(perf->pos_constraint == YES){
	    max_perf_change = 0;
	    max_comp = YES;
	    for(sens_link = perf->pos_sens_list; sens_link != NULL;
		sens_link = sens_link->next){
		if(max_comp == YES){
		    C_max = 0;
		    for(chan_info = sens_link->chan_info_list; 
			chan_info != NULL; chan_info = chan_info->next){
			C_max += chan_info->C_max;
		    }
		    max_perf_change += C_max * sens_link->sens;
		    if(max_perf_change >= 0.01 * perf->max_pos_change){
			max_comp = NO;
		    }
		}
		if(max_comp == NO){
		    /* form a critical pair if C_max is nonzero for 
		       at least one channel and
		       if has not already been formed */
		    max_zero = YES;
		    for(chan_info = sens_link->chan_info_list; 
			chan_info != NULL; 
			chan_info = chan_info->next){
			
			if(chan_info->C_max != 0) {
			    max_zero = NO;
			    break;
			}
		    }
		    if(max_zero == YES) continue; 
		    /* go to next sens_link */
		    
		    nameA = sens_link->nodeA->name;
		    nameB = sens_link->nodeB->name;
		    for(critpair = layout->critical_pairs; 
			critpair != NULL; critpair = critpair->next){
			if(((strcmp(critpair->netA_name,nameA) == 0)&&
			    (strcmp(critpair->netB_name,nameB) == 0)) ||
			   ((strcmp(critpair->netA_name,nameB) == 0)&&
			    (strcmp(critpair->netB_name,nameA) == 0))){
			    /* found the critpair */
			    goto next1;
			}
		    }
		    /* form a new critpair structure */
		    if( (critpair = (struct CriticalPairLink *)
			 nwAddLink(sizeof(struct CriticalPairLink), 
				   &(layout->critical_pairs),NULL,
				   layout->critical_pairs) ) == NULL)
		      {
			  OUT_OF_MEMORY(ERROR);
		      }
		    layout->num_critpairs ++;
		    
		    /* initialization */
		    init_crit(critpair,nameA,nameB, circuit->num_perf);
		    
		    critpair->chan_info_list = 
		      sens_link->chan_info_list;
		    
		  next1:				*(critpair->pos_sens + perf_num) 
		    = sens_link->sens;
		    tot_pos_sens += sens_link->sens; 
		}
	    }
	    if(tot_pos_sens != 0){
		cap_threshold 
		  = (0.01 * perf->max_pos_change)/tot_pos_sens;
		layout->cap_threshold = MIN(cap_threshold,
					    layout->cap_threshold);
	    }
	}
	if(perf->neg_constraint == YES){
	    max_perf_change = 0;
	    max_comp = YES;
	    for(sens_link = perf->neg_sens_list; sens_link != NULL;
		sens_link = sens_link->next){
		DEBUGG(	printf("keep_crit 127 (%s,%s)\n", sens_link->nodeA->name,sens_link->nodeB->name));

		if(max_comp == YES){
		    C_max = 0;
		    for(chan_info = sens_link->chan_info_list; 
			chan_info != NULL; chan_info = chan_info->next){
			C_max += chan_info->C_max;
			DEBUGG(printf("keep_crit 135 C_max(%s,%s)=%g\n",
				      sens_link->nodeA->name,sens_link->nodeB->name,
				      C_max));
		    }
		    
		    max_perf_change += C_max * sens_link->sens;
		    if(max_perf_change <= 0.01 * perf->max_neg_change){
			max_comp = NO;
		    }
		}
		if(max_comp == NO){
		    /* form a critical pair if C_max is nonzero for 
		       at least one channel and
		       if has not already been formed */
		    max_zero = YES;
		    for(chan_info = sens_link->chan_info_list; 
			chan_info != NULL; 
			chan_info = chan_info->next){
			
			if(chan_info->C_max != 0) {
			    max_zero = NO;
			    break;
			}
		    }
		    if(max_zero == YES) continue; 
		    /* go to next sens_link */
		    
		    nameA = sens_link->nodeA->name;
		    nameB = sens_link->nodeB->name;
		    for(critpair = layout->critical_pairs; 
			critpair != NULL; critpair = critpair->next){
			if(((strcmp(critpair->netA_name,nameA) == 0)&&
			    (strcmp(critpair->netB_name,nameB) == 0)) ||
			   ((strcmp(critpair->netA_name,nameB) == 0)&&
			    (strcmp(critpair->netB_name,nameA) == 0))){
			    /* found the critpair */
			    goto next2;
			}
		    }
		    /* form a new critpair structure */
		    if( (critpair = (struct CriticalPairLink *)
			 nwAddLink(sizeof(struct CriticalPairLink), 
				   &(layout->critical_pairs),NULL,
				   layout->critical_pairs) ) == NULL)
		      {
			  OUT_OF_MEMORY(ERROR);
		      }
		    
		    /* initialization */
		    init_crit(critpair,nameA,nameB, circuit->num_perf);
		    
		    critpair->chan_info_list = 
		      sens_link->chan_info_list;
		    
		  next2:				*(critpair->neg_sens + perf_num) 
		    = sens_link->sens;
		    tot_neg_sens += sens_link->sens; 
		    
		    
		}
	    }
	    if(tot_neg_sens != 0){
		cap_threshold 
		  = (0.01 * perf->max_neg_change)/tot_neg_sens;
		layout->cap_threshold = MIN(cap_threshold,
					    layout->cap_threshold);
	    }
	}
    }
}
