/*   main.c
 *
 *   Author: Umakanta Choudhury
 *   Date:   11th June
 *   Last modified:  Enrico Malavasi,  Aug. 1992
 *
 */
#include <stdio.h>
#include <math.h>
#include "oct.h"
/* #include "error.h" */
#include "utility.h"
#include "options.h"
#include "oh.h"
#include "th.h"
#include "pep.h"
#include "vov.h"

#include "layout.h"
#include "circuit.h"
#include "macros.h"

optionStruct optionList[] = {
    {"F", "format",            "Formatted output: string made of 's', 'b', 'w' [def.: 'sb']"},
    {"m", "maxmin_file",       "Max/min capacitance estimate file [def.: 'maxmin_file']"},
    {"S", "sens_file",         "Run only senscalc and write its output into the given file"},
    {"s", "sens_file",         "Sensitivity information file [def.: build it with senscalc]"},
    {"p", "perf_file",         "Performance constraints file [def.: 'perf_file']"},
    {"o", "out_file",          "Output information: constraints [def.: stdout]"},
    {"v", "",                  "Verbose option"},
    {"d", "",                  "Debug option"},
    {"x", "",                  "Skip deletion of temporary files"},
    { 0, 0, 0 }
};

unsigned int debug_flag    = 0;
unsigned int verbose_flag  = 0;
unsigned int senscalc_flag = 0;
unsigned int skipdel_flag  = 0;
unsigned int output_format = BOUND_FORMAT | SENS_FORMAT;

static void
  cleanTemporaryFiles()
{
    char command[128];
    sprintf( command, "rm -f /tmp/parcar.*_%d", getpid() );
    if(! skipdel_flag )  system( command );
}

main(argc,argv)
int argc;
char **argv;

{
    int option;
    struct Circuit circuit;
    struct Layout  layout;
    struct CriticalPairLink *critpair;
    struct Chan_info *chan_info;
    int    channelFlag;
    int error;
    FILE *debug_file;
    char filename[50];
    char *outFileName = NULL;
    char *perf_file_name;
    char *sensFileName = NULL;
    char *maxmin_file_name;
    char *mightyFileName = NULL;


    sprintf(filename,"%s_%d","/tmp/parcar.debug", getpid()); 
    FOPEN( debug_file, filename, "w");
    maxmin_file_name = util_strsav("maxmin_file");
    perf_file_name   = util_strsav("perf_file");

    while ( ( option = optGetOption (argc, argv)) != EOF) {
	switch (option) {
	  case 'F':
	    output_format = 0;
	    while((*optarg) != '\0') {
		if((*optarg) == 'b')        output_format |= BOUND_FORMAT;
		else if((*optarg) == 'w') {
		    output_format |= WEIGHT2_FORMAT;  /*  Default  */
		}
		else if((*optarg) == 's')   output_format |= SENS_FORMAT;
		optarg++;
	    }
	    break;
	  case 'x':
	    skipdel_flag = 1;
	    break;
	  case 'v':
	    verbose_flag = 1;
	    break;
	  case 'd':
	    debug_flag = skipdel_flag = 1;
	    break;
	  case 'm':
	    maxmin_file_name = util_strsav(optarg);
	    break;
	  case 's':
	    sensFileName = util_strsav( optarg );
	    break;
	  case 'S':
	    senscalc_flag = 1;
	    sensFileName = util_strsav( optarg );
	    break;
	  case 'p':
	    perf_file_name = util_strsav(optarg);
	    break;
	  case 'o':
	    outFileName = util_strsav( optarg );
	    break;
	  default:
	    optUsage();
	}
    }
    if( optind != argc ) optUsage();

    channelFlag = performanceRead(perf_file_name, &circuit, &layout,
				  (sensFileName==NULL)||senscalc_flag);
    DEBUGG( perf_pr(&circuit, &layout,debug_file) );

    if((sensFileName == NULL) || senscalc_flag) {
	char *inpFileName[128];
	char *nodesFileName[128];
	char  command[200];

	VERBOSE( printf( "Generating the sensitivity information ...\n"));
        sprintf(nodesFileName,  "%s_%d", "/tmp/parcar.senscalc_nodes", getpid()); 
        sprintf(inpFileName,    "%s_%d", "/tmp/parcar.senscalc_inp",   getpid()); 
        if(senscalc_flag == 0) {
	    sensFileName = ALLOC(char, 128);
	    sprintf(sensFileName, "%s_%d", "/tmp/parcar.senscalc_out",   getpid());
	}

        (void) create_node_file( &circuit, nodesFileName );
        (void) create_senscalc_inp( &circuit, inpFileName );

        sprintf(command,"senscalc -n %s -i %s -o %s", nodesFileName, inpFileName, sensFileName); 
	VERBOSE( printf( "Running\n %s\n", command));
        system( command );
	VERBOSE( printf( "Done\n"));

	if(senscalc_flag) {
	    cleanTemporaryFiles();
	    exit(0);
	}
    }

    (void) sens_read(sensFileName, &circuit, &layout);
    (void) sens_match(&circuit, &layout);
    DEBUGG( sens_match_pr(&circuit,&layout,debug_file) );

    if(! channelFlag ) {
	(void) maxmin_read(maxmin_file_name, &layout, &circuit);
    }

    (void) keep_crit(&circuit, &layout, debug_file, (channelFlag));
    DEBUGG( crit_pr(&circuit,&layout,debug_file) );
    DEBUGG( maxmin_pr(&layout,debug_file) );
    
    (void) delta_min(&layout,&circuit);
    DEBUGG( delta_min_pr(&layout,&circuit,debug_file) );

    if( channelFlag ) {
	shield_decide(&layout,&circuit);
	DEBUGG( shield_pr(&layout,&circuit,debug_file) );
    }
    
    (void) bound_gen(&layout, &circuit, debug_file);
    (void) bound_pr(&layout, &circuit, debug_file);

    if( channelFlag != NULL ) {
	infl_dist(&layout,&circuit,debug_file);
    }
    
    (void) constraint_pr(&layout, &circuit, outFileName);
    
    FCLOSE(debug_file);

    cleanTemporaryFiles();
    exit(0);
}
