/*   mightyChannel.c
*   Author:	Enrico Malavasi
*   last update:   July 1992
*
*   "Mighty" input syntax has been adopted. All of the "mighty" 
*   input is accepted - some is ignored.
*/
#include <stdio.h>
#include <string.h>
#include "utility.h"
#include "layout.h"


#define MAX_LINE_LENGTH 250

/*
*  "mighty" input syntax:
*  ----------------------
*
*    "number_of_nets integer"    - Self evident;
*    "rectagoncorners integer"   - For "roadrunner", the integer must be 2 or 4. Immediately after
*                                  this line there should be correponding corner positions:
*    "x-position y-position"	 - Corner coordinate (floating point).
*    "number_of_pins integer"	 - Number of pins on top and bottom of channel. Immediately after
*                                  this line there should be the correctnumber of line:
*    "net_name x-pos y-pos layer"- Net name may be any alpha-numeric string with no spaces.
*                                  x-pos and x-pos are floats.
*    "left_list integer"	 - Gives number of nets crossing left end of channel. This line should
*                                  be immediately followed by:
*    "net_name net_name .." 	 - Any number of net names per line until the correct numebr of nets
*                                  has been given.
*    "right_list integer"	 - Gives number of nets crossing right end of channel. This line should
*                                  be immediately followed by:
*    "net_name net_name .." 	 - Any number of net names per line until the correct numebr of nets
*                                  has been given.
*    "sensitive_nets integer"    - Gives number of nets with total length, secondary layer length, and
*                                  number of cross-overs to be minimised. This line should be immediately
*                                  followed by:
*    "net_name net_name .." 	 - Any number of net names per line until the correct numebr of nets has
*                                  been given.
*    "obstacles integer"	 - Ignored.
*
*/

static  char text_line[MAX_LINE_LENGTH];

static  struct _netList {
    char            *name;
    struct _netList *next;
} *netList = NULL;

static void
  addNewNet( name )
char *name;
{
    struct _netList *newNet;
    newNet = ALLOC( struct _netList, 1 );
    newNet->name = util_strsav( name );
    newNet->next = netList;
    netList = newNet;
}
static int
  checkNetList( name )
char *name;
{
    struct _netList *net;
    for(net=netList ; net!=NULL ; net=net->next) {
	if(! strcmp( name, net->name )) return 1;
    }
    return 0;
}

char *
  OLDnextWords( fin, s, n )
FILE *fin;
char *s;
int   n;
{
    int l = 0;

    s[0] = '\0';
    while( (n--) > 0 ) {
	if( fscanf( fin, "%s ", s+l ) != 1 )  return NULL;
	strcat(s, " ");
	l = strlen(s);
    }
    s[l-1] = '\0';
    return s;
}

static char last;
char *
  nextWords( fin, s, n )
FILE *fin;
char *s;
int   n;
{
    int l = 0;
    int k;

    for(k=0 ; k<n ; k++) {
	while(((s[l]=getc(fin)) == ' ') || (s[l] == '\t') || (s[l] == '\n')) ;
	l++;
	while(((s[l]=getc(fin)) != ' ') && (s[l] != '\t') && (s[l] != '\n')) {
	    if(s[l] == EOF) break;
	    l++;
	}
	if(s[l] == EOF) break;
	last = s[l];
	s[l++] = ' ';
    }
    if( k < n )  return  NULL;
    s[l-1] = '\0';
    return s;
}

int
  readMightyChannel( fileName, channel, layout )
char  *fileName;
struct Channel *channel;
struct Layout *layout;	        /* layout structure     */
{
    int    layerNo;
    int    i;
    char   c;
    FILE *fin;
    float	xpos, ypos;
    int    number_of_nets  = (-1);
    int    rectagoncorners = (-1);
    int    number_of_pins  = (-1);
    int    left_list       = (-1);
    int    right_list      = (-1);
    int    obstacles       = (-1);
    int    sensitive_nets  = (-1);
    char   netName[64];


    FOPEN( fin, fileName, "r");

    /* Read file, line by line */
    while( nextWords( fin, text_line, 1 ) != NULL)  {
	    
	if(text_line[0] == '!') {
	    if( last != '\n') {
		while(((c=getc( fin )) != '\n') && (c != EOF))  ;
	    }
	    continue;  /*  Comment  */
	}

	/*  Mighty commands */
	if(! strcmp(text_line, "number_of_nets")) {
	    fscanf(fin, "%d", &number_of_nets);
	    VERBOSE( printf( "Number of nets: %d\n", number_of_nets));
	    rr_dsSetNumberOfNets(number_of_nets, channel);
	    continue;
	}
	
	if(! strcmp(text_line,"rectagoncorners")) {
	    fscanf(fin, "%d ", &rectagoncorners);
	    VERBOSE( printf( "Reading %d rectagon-corners\n", rectagoncorners));
	    if( rectagoncorners != 4 )   {
		fprintf(stderr, "** SORRY, only rectangular channels (4 corners) are accepted\n");
		exit( -1 );
	    }
	    for(i=0 ; i<rectagoncorners ; i++) {
		if( nextWords( fin, text_line, 2 ) != NULL)  sscanf(text_line, "%g %g", &xpos, &ypos);
		else {
		    fprintf(stderr, "Short file\n" );
		    exit( -1 );
		}
		rr_dsSetCorner(xpos, ypos, channel);
	    }
	    continue;
	}
	
	if(! strcmp(text_line,"number_of_pins")) {
	    fscanf(fin, "%d ", &number_of_pins);
	    VERBOSE( printf( "Reading %d pins\n", number_of_pins));
	    for(i=0 ; i<number_of_pins ; i++) {
		if( nextWords( fin, text_line, 4 ) != NULL) {
		    sscanf(text_line, "%s %g %g %d", netName, &xpos, &ypos, &layerNo);
		    addNewNet( netName );
		} else {
		    fprintf(stderr, "Short file\n" );
		    exit( -1 );
		}
		rr_dsRecordPin(netName, xpos, ypos, channel);
	    }
	    continue;
	}
	
	if(! strcmp(text_line, "left_list")) {
	    fscanf(fin, "%d ", &left_list);
	    VERBOSE( printf( "Reading left list\n"));
	    for(i=0 ; i<left_list ; i++) {
		if( nextWords( fin, text_line, 1 ) != NULL)  sscanf(text_line, "%s ", netName);
		else {
		    fprintf(stderr, "Short file\n" );
		    exit( -1 );
		}
		if(! checkNetList( netName ) ) {
		    fprintf(stderr, "Nonexisting net: %s\n", netName);
		    exit( -1 );
		}		    
		rr_dsRecordCrossing(netName, LEFT, channel);
	    }
	    continue;
	}
	
	if(! strcmp(text_line, "right_list")) {
	    fscanf(fin, "%d ", &right_list);
	    VERBOSE( printf( "Reading right list\n"));
	    for(i=0 ; i<left_list ; i++) {
		if( nextWords( fin, text_line, 1 ) != NULL)  sscanf(text_line, "%s ", netName);
		else {
		    fprintf(stderr, "Short file\n" );
		    exit( -1 );
		}
		if(! checkNetList( netName ) ) {
		    fprintf(stderr, "Nonexisting net: %s\n", netName);
		    exit( -1 );
		}		    
		rr_dsRecordCrossing(netName, RIGHT, channel);
	    }
	    continue;
	}
	
	if(! strcmp(text_line,"obstacles")) {
	    fscanf(fin, "%d ",&obstacles);
	    /* Obstacles are ignored */
	    for(i=0 ; i<obstacles ; i++) {
		if( nextWords( fin, text_line, 5 ) == NULL) {
		    fprintf(stderr, "Short file\n" );
		    exit( -1 );
		}
	    }
	    continue;
	}
	
	if(! strcmp(text_line,"sensitive_nets")) {
	    fscanf(fin, "%d ", &sensitive_nets);
	    /* Sensitive nets are ignored */
	    for(i=0 ; i<sensitive_nets ; i++) {
		if( nextWords( fin, text_line, 1 ) == NULL) {
		    fprintf(stderr, "Short file\n" );
		    exit( -1 );
		}
	    }
	    continue;
	}
	
	fprintf(stderr, "*** %s : unknown command\n", text_line);
	exit( -1 );
    }

    {
	char *message = NULL;

	if( number_of_nets < 0 )  message = util_strsav("Number of nets in the channel is undefined");
	if( rectagoncorners < 0)  message = util_strsav("Rectagoncorners is undefined");
	if( number_of_pins < 0 )  message = util_strsav("Number of pins is undefined");
	if( message != NULL) {
	    fprintf(stderr, message );
	    exit( -1 );
	}
    }

    fclose( fin );
    return;
}
