/*     perf_read.c
 *
 *   Original author: Umakanta Choudhury
 *   Completely rewritten by Enrico Malavasi
 *   Ver. 1.00 - Aug. 1992
 *
 *   Purpose: read the input performance specification file
 *            The order in which statements and commands are given is
 *            uninfluent. See the doc file for the syntax of input specs.
 */
#include <stdio.h>
#include <string.h>
#include "utility.h"
#include "layout.h"
#include "circuit.h"

#define  MISSING_INFO(x)  { fprintf(stderr,"** ERROR: %s not specified\n",x); exit(-1); }

static struct Circuit *thisCircuit;
static struct Layout  *thisLayout;
static char           *cleanDeck;

/*  In what follows the static structures are used for storing the commands
 *  while they are read. In this way, they can be executed in the proper order
 *  after parsing the input file, independently of the order in which they appear
 *  there. */
/*  ============================================================
 *  SIMULATOR field
 *  ============================================================ */
typedef struct _simulOption {
    char  *simulator;   /*  Simulator to be used - Default: DEFAULT_SIMULATOR_NAME */
    char  *deckfile;    /*  Deck file - no default */
    char  *out1, *out2; /*  Output nodes  -  no default */
} _simulOption;

static _simulOption *simulOption = NULL;

#define  DEFAULT_SIMULATOR_NAME    "/users/cad/bin/spice3.old"
static void
  emInitSimulator()
/* Initialize the local static structure for the SIMULATION statements */
{
    simulOption = ALLOC( _simulOption, 1 );
    simulOption->simulator  = DEFAULT_SIMULATOR_NAME;
    simulOption->deckfile   = NULL;
    simulOption->out1       = NULL;
    simulOption->out2       = NULL;
}

/*  The following functions read from the file the simulator name, 
 *  the deck file and the output nodes respectively */
void
  setSimulator( name )
char *name;
{
    if(simulOption == NULL)  emInitSimulator();
    simulOption->simulator = util_strsav( name );
}
void
  setDeckfile( deckfile )
char *deckfile;
{
    if(simulOption == NULL)  emInitSimulator();
    simulOption->deckfile = util_strsav( deckfile );
}
void
  setOutNodes( name1, name2 )
char *name1, *name2;
{
    if(simulOption == NULL)  emInitSimulator();
    simulOption->out1 = util_strsav( name1 );
    simulOption->out2 = util_strsav( name2 );
}

/*  ============================================================
 *  TECHNOLOGY field
 *  ============================================================ */
typedef struct _techOption {
    char  *name;      /* Techology name - Default: DEFAULT_TECHNOLOGY_NAME */
    float variation;  /* Max % variation - Default: DEFAULT_TECH_VARIATION */
    float mismatch;   /* Max % mismatch - Default: DEFAULT_TECH_MISMATCH */
} _techOption;

static _techOption *techOption = NULL;

#define  DEFAULT_TECHNOLOGY_NAME   "scmos"
#define  DEFAULT_TECH_VARIATION    20.0
#define  DEFAULT_TECH_MISMATCH     10.0
static void
  emInitTechnology()
/* Initialize the local static structure for the TECHNOLOGY statements */
{
    techOption = ALLOC( _techOption, 1 );
    techOption->name      = DEFAULT_TECHNOLOGY_NAME;
    techOption->variation = DEFAULT_TECH_VARIATION;
    techOption->mismatch  = DEFAULT_TECH_MISMATCH;
}

/*  The following functions read from the file the technology name and
 *  the max percent variation and mismatch */
setTechnologyName( name )
char *name;
{
    if( techOption == NULL )  emInitTechnology();
    techOption->name = util_strsav( name );
}
void
  setProcessVariation( name )
char *name;
{
    if( techOption == NULL )  emInitTechnology();
    sscanf( name, "%f", &(techOption->variation));
}
void
  setProcessMismatch( name )
char *name;
{
    if( techOption == NULL )  emInitTechnology();
    sscanf( name, "%f", &(techOption->mismatch));
}

/*  ============================================================
 *  MATCHNODES field
 *  ============================================================ */
typedef struct _matchOption {
    char  *name1, *name2;        /* Names of the matched nodes */
    struct _matchOption *next;   /* It is a list structure */
} _matchOption;

static _matchOption *matchOption = NULL;

void
  setMatchedPair( name1, name2 )
char *name1, *name2;
/*  read from the input file the name of a pair of matched nodes  */
{
    _matchOption  *match;

    match = ALLOC( _matchOption, 1 );
    match->name1 = util_strsav( name1 );
    match->name2 = util_strsav( name2 );
    match->next  = matchOption;
    matchOption = match;
}

/*  ============================================================
 *  PERFORMANCE field  -  No defaults
 *  ============================================================ */
typedef struct _perfOption {
    char  *name;     /*  Performance name  */
    struct {
	float pos;   /*  Max pos variation */
	float neg;   /*  Max neg. variation */
    } maxvar;
    struct {
	float min;   /*  Min frequency (for the spice simulation) */
	float max;   /*  Max frequency "" "" */
    } range;
} _perfOption;

static _perfOption *perfOption = NULL;

typedef struct _perfList {
    _perfOption *perf;         /*  List pointing to all the performances read. */
    struct _perfList *next;
} _perfList;
static _perfList *perfList = NULL;

emInitPerf()
/* Initialize the local static structure for the PERFORMANCE statements */
{
    perfOption = ALLOC( _perfOption, 1 );
    perfOption->name = NULL;
    perfOption->maxvar.pos = (-1);
    perfOption->maxvar.neg = (-1);
    perfOption->range.min = (-1);
    perfOption->range.max = (-1);
}

/*  The following functions read from the input file the performance name,
 *  its range and max variations */
void
  setPerfName( name )
char *name;
{
    if(perfOption == NULL) emInitPerf();
    perfOption->name = util_strsav( name );
}
void
  setPerfRange( val1, val2 )
char *val1, *val2;
{
    float x;

    if(perfOption == NULL) emInitPerf();
    sscanf( val1, "%f", &x);
    perfOption->range.min = x;
    sscanf( val2, "%f", &x);
    perfOption->range.max = x;
}
void
  setPerfChange( negbound, posbound )
char *negbound, *posbound;
{
    float xneg, xpos;

    if(perfOption == NULL) emInitPerf();

    if(strcmp( negbound, "infinity" ) && strcmp( negbound, "-infinity" ) ) {
	sscanf( negbound, "%f", &xneg );
	
	if( xneg > 0 ) {
	    xneg = (-xneg);
	    fprintf(stderr, " Warning: max negative perf. variation set to %f \n", xneg);
	}
	perfOption->maxvar.neg = (-xneg);
    }

    if( strcmp( posbound, "infinity" ) ) {
	sscanf( posbound, "%f", &xpos );
	if( xpos < 0 ) {
	    fprintf(stderr, "ERROR: Wrong sign for the max positive perf. variation \n");
	    exit( -1 );
	}
	perfOption->maxvar.pos = xpos;
    }
}

static char *
  getSimFile( perfName, rangeMin, rangeMax )
char *perfName;
float rangeMin, rangeMax;
/*  After a performance has been defined, this function copies the input deck
 *  file into as many decks as performances. Each deck is completed with the 
 *  analysis commands necessary to compute its sensitivities. The names of the 
 *  decks contain the name of the performance they refer to. */
{
    char  *simFile = NULL;
    char   command[128];

    simFile = ALLOC(char, 128);
    sprintf( simFile, "/tmp/parcar.%s-deck_%d", perfName, getpid());
    sprintf(command, "cp %s %s", cleanDeck, simFile);
    system( command );
    
    if((! strcmp( perfName, "ugb" )) ||
       (! strcmp( perfName, "pm" ))) {
	sprintf(command, "echo \".ac LIN 100 %f %f\" >> %s",
		rangeMin, rangeMax, simFile);
	system( command );
	if(! strcmp(thisCircuit->out_node2_name, "0")) {
	    sprintf(command, "echo \".print ac v(%s)\" >> %s",
		    thisCircuit->out_node1_name, simFile);
	} else {
	    sprintf(command, "echo \".print ac v(%s,%s)\" >> %s",
		    thisCircuit->out_node1_name, thisCircuit->out_node2_name, simFile);
	}
	system( command );
    } else if(! strcmp( perfName, "lfg" )) {
	sprintf(command, "echo \".ac LIN 2 %f %f\" >> %s",
		rangeMin, rangeMax, simFile);
	system( command );
	if(! strcmp(thisCircuit->out_node2_name, "0")) {
	    sprintf(command, "echo \".print ac v(%s)\" >> %s",
		    thisCircuit->out_node1_name, simFile);
	} else {
	    sprintf(command, "echo \".print ac v(%s,%s)\" >> %s",
		    thisCircuit->out_node1_name, thisCircuit->out_node2_name, simFile);
	}
	system( command );
    } else {
	fprintf( stderr, "** %s: Unknown performance name\n", perfName );
	exit( -1 );
    }

    sprintf(command, "echo .end >> %s", simFile);
    system( command );

    return simFile;
}

void
  endPerformance()
/*  To be executed after a performance ha been read. It stores the static
 *  structure of this performance into a local list */
{
    _perfList  *pl;

    if(perfOption == NULL)  emInitPerf();
    if( perfOption->name == NULL ) MISSING_INFO( "performance name" );

    pl = ALLOC( _perfList, 1 );
    pl->perf = perfOption;
    pl->next = perfList;
    perfList = pl;
    perfOption = NULL;
}

/*  ============================================================
 *  LAYOUT field
 *  ============================================================ */
typedef struct _channelList {
    char *name;                    /*  channel name */
    struct _channelList *next;     /*  this is a list structure  */
} _channelList;
static _channelList *channelList = NULL;

#define   DEFAULT_RESOLUTION      0.05
#define   DEFAULT_NET_WIDTH        3.0
#define   DEFAULT_NET_SEPARATION   3.0
#define   DEFAULT_INFLUENCE_DIST  15.0
static int  channelFlag = 0;           /*  1 if at least 1 channel has been read */
static float  defaultNetSep   = DEFAULT_NET_SEPARATION; /*  Default net separation */
static float  defaultNetWidth = DEFAULT_NET_WIDTH;      /*  Default net width */
static float  defaultResolution = DEFAULT_RESOLUTION;   /*  Default layout resolution */
/* Both previous variables are local and, if positive, they define the values of the 
 * corresponding parameters for channels. If either of them is negative,
 * the corresponding parameter is set to its default (see rr_dsSetDefaultNet...). */

void
  readChannel( channelFileName )
char *channelFileName;
/*  Input the name of a file containing a channel */
{
    struct Channel *channel;
    _channelList   *newChannel;

    newChannel = ALLOC( _channelList, 1 );
    newChannel->name = util_strsav( channelFileName );
    newChannel->next = channelList;
    channelList = newChannel;
}

void
  lyParameter( param, value )
char *param;
char *value;
{
    if(! strcmp( param, "layout_resolution" )) {
	sscanf(value, "%g", defaultResolution);
    }
}

/*  The following functions introduce the data read from the input file
 *  into the permanent data structure of the program. They are all called, in
 *  the proper order, by  performanceRead. */
static void
  setTechnology()
{
    if( techOption == NULL )  emInitTechnology();
    if( techOption->name == NULL )  MISSING_INFO( "technology name" );
    thisCircuit->technologyName          = techOption->name;
    thisCircuit->max_process_var_percent = techOption->variation;
    thisCircuit->max_mismatch_percent    = techOption->mismatch;

    VERBOSE( fprintf(stderr, "Technology is %s\n", thisCircuit->technologyName ));
    VERBOSE( fprintf(stderr, " Max process variation = %4.1f%c\n", thisCircuit->max_process_var_percent, '\%' ) );
    VERBOSE( fprintf(stderr, " Max process mismatch = %4.1f%c\n", thisCircuit->max_mismatch_percent, '\%' ) );
    (void) techInit( thisCircuit->technologyName );
}

static void
  setMatchedNodes()
{
    struct Matched_pair *matched_pair; 
    _matchOption       *match;

    VERBOSE( fprintf(stderr, "Matched pairs:\n " ) );
    for(match=matchOption ; match!=NULL ; match=match->next ) {
	thisCircuit->num_matched_pairs++;
	ds_MakeMatchedPair(match->name1, match->name2, &matched_pair, thisCircuit) ;
	VERBOSE( fprintf(stderr, "(%s, %s) ", match->name1, match->name2 ) );
    }
    VERBOSE( fprintf(stderr, "\n" ) );
}

static void
  setPerformances( genDeckFiles )
int    genDeckFiles;        /* 1 if deck files need to be generated */
{
    int type = 0;
    _perfList *pl;
    struct Performance *thisPerf;

    for(pl=perfList ; pl!=NULL ; pl=pl->next ) {

	if((pl->perf->maxvar.neg < 0) && (pl->perf->maxvar.pos < 0)) {
	    fprintf(stderr, "WARNING: unconstrained performance %s will be ignored\n", pl->perf->name);
	    continue;
	}
	ds_MakePerformance(pl->perf->name, type, &thisPerf, thisLayout, thisCircuit);
	thisLayout->perf_name[thisCircuit->num_perf-1] = util_strsav( pl->perf->name );
	
	if(pl->perf->range.min >= 0) thisPerf->freqRange.min = pl->perf->range.min;
	if(pl->perf->range.max >= 0) thisPerf->freqRange.max = pl->perf->range.max;
	
	if(pl->perf->maxvar.neg >= 0 ) {
	    thisPerf->max_neg_change = (-pl->perf->maxvar.neg);
	    thisPerf->delta_neg_min  = 0;
	    thisPerf->neg_eqn_num    = 0;	
	    thisPerf->neg_constraint = YES;
	}
	if(pl->perf->maxvar.pos >= 0 ) {
	    thisPerf->max_pos_change = pl->perf->maxvar.pos;
	    thisPerf->delta_pos_min  = 0;
	    thisPerf->pos_eqn_num    = 0;
	    thisPerf->pos_constraint = YES;
	}

	if( genDeckFiles ) {
	    thisPerf->ckt_file_name = getSimFile( thisPerf->name,
						 thisPerf->freqRange.min, thisPerf->freqRange.max );
	}
	VERBOSE( fprintf(stderr, "Performance %s\n", thisPerf->name ));
	VERBOSE( fprintf(stderr, " Constraints: [%s, %s]\n",
			 (thisPerf->neg_constraint == YES) ?
			 util_pretty_print(thisPerf->max_neg_change) : "-inf.",
			 (thisPerf->pos_constraint == YES) ?
			 util_pretty_print(thisPerf->max_pos_change) : "inf.") );
    }
}

static void
  setSimulation()
{
    static int counter = 0;
    struct Node	*node;
    FILE        *stream;
    char        *name, *popNextNode();
    extern  struct Net *rr_dsMakeNet();

    if(simulOption == NULL)  emInitSimulator();
    if( simulOption->simulator == NULL )   MISSING_INFO( "simulator name" );
    thisCircuit->sim_name = simulOption->simulator;
    if( simulOption->deckfile == NULL )    MISSING_INFO( "deck file name" );
    thisCircuit->deckFile = simulOption->deckfile;
    if( simulOption->out1 == NULL )        MISSING_INFO( "output node names" );
    thisCircuit->out_node1_name = simulOption->out1;
    thisCircuit->out_node2_name = simulOption->out2;

    cleanDeck = ALLOC(char, strlen("/tmp/parcar.cleanDeck.")+10 );
    sprintf(cleanDeck, "/tmp/parcar.cleanDeck_%d", getpid() );

    VERBOSE( fprintf(stderr,  "Reading nodes from the deck file '%s':\n", thisCircuit->deckFile ) );
    spiceParse( thisCircuit->deckFile, cleanDeck );

    while( (name = popNextNode())  != NULL ) {

	thisCircuit->num_nodes++;
	ds_MakeNode(name, &node, thisCircuit);
	if(thisLayout->mode == NORMAL) {
	    rr_dsMakeNet(name, thisLayout->chan_list->channel);
	}
	VERBOSE( fprintf(stderr, " %s ", name ) );
    }

    VERBOSE( fprintf(stderr, "\n" ) );
    VERBOSE( fprintf(stderr, " Simulator   : %s\n", thisCircuit->sim_name ) );
    VERBOSE( fprintf(stderr, " Output nodes: %s, %s\n", thisCircuit->out_node1_name, thisCircuit->out_node2_name ) );
}

static void
  setLayout()
{
    struct Channel *ds_MakeChannel();
    int  channelNum = 0;
    _channelList   *chan;
    
    for( chan=channelList ; chan!=NULL ; chan=chan->next ) {
	ds_MakeChanList(channelNum+1, thisLayout);
	thisLayout->chan_list[channelNum].channel = 
	  ds_MakeChannel(chan->name, defaultNetSep, defaultNetSep,
			 defaultNetWidth, defaultResolution, DEFAULT_INFLUENCE_DIST);
	/* Open the text file */
	readMightyChannel( chan->name, thisLayout->chan_list[channelNum].channel, thisLayout );
	channelNum++;
    }

    if( channelNum == 0 ) {
	ds_MakeChanList(1, thisLayout);
	thisLayout->chan_list[0].channel =
	  ds_MakeChannel( " - ", defaultNetSep, defaultNetSep,
			 defaultNetWidth, defaultResolution, DEFAULT_INFLUENCE_DIST);
    } else {
	channelFlag = 1;
	(void) trans_chan( thisLayout );
	(void) link_chan( thisLayout );
	(void) sublist( thisLayout );
	(void) span_maxmin( thisLayout);
    }
}

int
  performanceRead(fileName, circuit, layout, genDeckFiles)
char           *fileName;		/* Input text file	*/
struct Circuit *circuit;	        /* The circuit structure */
struct Layout  *layout;	                /* The layout structure */
int    genDeckFiles;        /* 1 if deck files need to be generated */
/*  Read the input performance file (default name: perf_file).
 *  The input statements are read and stored in local static variables.
 *  The permanent storage happens in the proper order at the end of 
 *  the parsing */
{
    thisCircuit  =  circuit;
    thisLayout   =  layout;
    
    (void) ds_InitLayout( layout );
    (void) ds_InitCircuit( circuit );
    
    yyopenfile(fileName);
    yyparse();
    
    setLayout();
    setSimulation();
    setMatchedNodes();
    setTechnology();
    setPerformances( genDeckFiles );

    return  channelFlag;
}
