/*          "PIECE-WISE CONSTANT" UTILITY 
*           -----------------------------
*
*   Functions: 	pwcInit, pwcAddRectangle, pwcAddStep,  pwcGenerateFunction.
*               pwcGetSteps, pwcGetInterval, pwcMaxValue, 
*		pwcMaxValueAndInterval, pwcMinValue, pwcMinValueAndInterval,
*               pwcAddFunctions, pwcMultiplyFunctions, pwcNormalise, 
*               pwcReduceRange, pwcBinaryMax, pwcBinaryMin,
*		pwcDeleteFunction
*   File:	pwc.c
*   Author:	Nicholas Weiner 
*   Date:       29th May 1989.
*
*  Piece-wise constant functions are specified by giving: 1) rectangular
*  pulses - i.e. intervals with associated values; 2) steps - *  i.e. points
*  with associated step sizes. The complete function is given by the summing
*  these primitive functions.
*
*  After the function has been generated further rectangle and steps may
*  be added. The function will be iteratively updated. It is quicker,
*  though, to enter all of the primitive first, and then generate the 
*  function.
*  
*/

#include <stdio.h>
#include "utility.h"
#include "pwc.h"

#define PWCALLOC(x) (( x *)malloc(sizeof(x)))
#define OUT_OF_MEMORY(code) {printf("Out of memory\n"); return (code);}

/*
*   pwcInit
*
*  Starts a new piece-wise constant funtion.
* 
*  Returns pointer to the pwc function. Null on error.
*/
struct PWC *pwcInit(name,left,right)
char *name;
float left, right;
    {
    struct PWC *function;

    if( (function = PWCALLOC(struct PWC)) == NULL) OUT_OF_MEMORY(NULL);
    if( (function->name = util_strsav(name)) == NULL)  OUT_OF_MEMORY(NULL);
    function->left = left;
    function->right = right;
    function->step_list = NULL;
    function->interval_list = NULL;
    function->generated = FALSE;

    return function;
    }

/*
*   pwcAddStep
*
*   Given step position and value. Puts new entry into the step list
*   and, if the funtion has already been generated, updates it.
*
*   RETURNS PWC_OK or PWC_ERROR.
*/
PWC_REPORT pwcAddStep(position,size,function)
float position, size;
struct PWC *function;
    {
    extern struct StepLink *nwAddLink();

    struct StepLink *new_step;
    struct IntervalLink *interval, *new_interval;

    /* Make sure that the break-point is in the function range */
    if( (position < function->left) || (position > function->right))
	{
	printf("E: in pwcAddStep. %f not in range\[%f,%f\]\n",
	    position,function->left,function->right);
	return PWC_ERROR;
	}

    if(function->generated != TRUE) 
	{
        /* Create the new step */
	if( (new_step=(struct StepLink *)nwAddLink(sizeof(struct StepLink),
	    &(function->step_list),NULL,function->step_list)) == NULL) 
	    {
	    printf("E: in pwcAddStep\n");
	    return PWC_ERROR;
	    }

	new_step->size = size;

	return PWC_OK;
	}
    else
	{
        /* Update the function */
	/* Get interval constaining the position */
	if(pwcGetInterval(position,&interval,function)!=PWC_OK)
	    {
	    printf("E: in pwcAddStep\n");
	    return PWC_ERROR;
	    }

	/* Create a new interval */
	if( (new_interval=(struct IntervalLink *)
	    nwAddLink(sizeof(struct IntervalLink),&(function->interval_list),
	    interval, interval->next)) == NULL) 
	    {
	    printf("E: in pwcAddStep\n");
	    return PWC_ERROR;
	    }
	new_interval->left = position;
	new_interval->right = interval->right;
	interval->right = position;
	new_interval->value = interval->value;

	for(interval=new_interval;interval!=NULL;interval=interval->next)
            interval->value += size;

        return PWC_OK;
	}
    }


/*
*   pwcAddRectangle
*
*   Given the pulse interval and height. Puts two new entry into the step list
*   and, if the funtion has already been generated, updates it.
*
*   RETURNS PWC_OK or PWC_ERROR.
*/
PWC_REPORT pwcAddRectangle(left,right,height,function)
float left, right, height;
struct PWC *function;
    {
    struct StepLink *new_step;
    struct IntervalLink *interval, *new_interval;

    /* Assume an error if left > right */
    if(left > right)
	{
	printf("E: in pwcAddRectangle. left \(%f\) > right \(%f\)\n",
	    left,right);
	return PWC_ERROR;
	}

    /* Make sure that the rectangle is in the function range */
    if( (left < function->left) || (right > function->right))
	{
	printf("E: in pwcAddRectangle. \[%f,%f\] not in range\[%f,%f\]\n",
	    left,right,function->left,function->right);
	return PWC_ERROR;
	}

    if(function->generated != TRUE) 
	{
        /* Create the right step */
	if( (new_step=nwAddLink(sizeof(struct StepLink),&(function->step_list),
	NULL,function->step_list)) == NULL) 
	    {
	    printf("E: in pwcAddRectangle\n");
	    return PWC_ERROR;
	    }

	new_step->position = right;
	new_step->size = -height;

        /* Create the left step */
	if( (new_step=(struct StepLink *)nwAddLink(sizeof(struct StepLink),
	    &(function->step_list),NULL,function->step_list)) == NULL) 
	    {
	    printf("E: in pwcAddRectangle\n");
	    return PWC_ERROR;
	    }

	new_step->position = left;
	new_step->size = height;

	return PWC_OK;
	}
    else
	{
        /* Create the new interval to the left of the rectange */
	/* Get interval constaining the left position */
	if(pwcGetInterval(left,&interval,function)!=PWC_OK) 
	    {
	    printf("E: in pwcAddRectangle\n");
	    return PWC_ERROR;
	    }

	/* Create a new interval */
	if( (new_interval=(struct IntervalLink *)
	    nwAddLink(sizeof(struct IntervalLink),&(function->interval_list),
	    interval->prev,interval)) == NULL) 
	    {
	    printf("E: in pwcAddRectangle\n");
	    return PWC_ERROR;
	    }
	new_interval->right = left;
	new_interval->left = interval->left;
	interval->left = left;
	new_interval->value = interval->value;


        /* Create the new interval to the right of the rectangle */
	/* Get interval constaining the right position */
	if(pwcGetInterval(right,&interval,function)!=PWC_OK) 
	    {
	    printf("E: in pwcAddRectangle\n");
	    return PWC_ERROR;
	    }
	/* Create a new interval */
	if( (new_interval=(struct IntervalLink *)
	    nwAddLink(sizeof(struct IntervalLink),&(function->interval_list),
	    interval,interval->next)) == NULL) 
	    {
	    printf("E: in pwcAddRectangle\n");
	    return PWC_ERROR;
	    }
	new_interval->left = right;
	new_interval->right = interval->right;
	interval->right = right;
	new_interval->value = interval->value;

        /* Update the intervening intervals */
	for(interval=new_interval->prev;interval->left>=left;
	    interval=interval->prev) interval->value += height;

        return PWC_OK;
	}
    }

#define TYPE struct StepLink
#define DIRECT_COMPARE
#define SORT1 pwcSortSteps
#define FIELD position
#include "lsort.h"

/* pwcGenerateFunction
*
*  The linear list of steps is ordered according to their positions.
*  Uses Richard Rudell's list sorting package.
*
*  The piece-wise constant function is generated from the sorted list.
*
* Return PWC_OK or PWC_ERROR.
*/
PWC_REPORT pwcGenerateFunction(function)
struct PWC *function;
    {
    struct StepLink *step_link, *prev_link;
    struct IntervalLink *interval;

    /* Order the steps */
    function->step_list=pwcSortSteps(function->step_list,NULL);
    /* Rudell's routines only change the "next" pointers. Still have to do
       the "prev" pointers. */
    prev_link = NULL;
    for(step_link=function->step_list;step_link!=NULL;step_link=step_link->next)
        {
        step_link->prev = prev_link;
        prev_link = step_link;
	}

    /* Create the function */
    /* Interval to left of first break point */
    if( (interval=(struct IntervalLink *)nwAddLink(sizeof(struct IntervalLink),
	   &(function->interval_list),NULL,NULL)) == NULL) 
	{
	printf("E: in pwcGenerateFunction\n");
	return PWC_ERROR;
	}
    interval->left = function->left;
    interval->value = 0.0;

    /* One interval to the right of each break point */
    for(step_link=function->step_list;step_link!=NULL;step_link=step_link->next)
	{
	/* Create a new interval (put it at end of list, after last interval
           created) */
	if( (interval=(struct IntervalLink *)
	    nwAddLink(sizeof(struct IntervalLink),&(function->interval_list),
	    interval,NULL)) == NULL) 
	    {
	    printf("E: in pwcGenerateFunction\n");
	    return PWC_ERROR;
	    }
  	interval->left = interval->prev->right = step_link->position;
  	interval->value = interval->prev->value + step_link->size;
	}

    /* Interval to right of last break point stretches on forever */
    interval->right = function->right;

    /* Remove any single point intervals */
    for(interval=function->interval_list;interval!=NULL;interval=interval->next)
	{
	if(interval->left==interval->right)
	    {
	    nwDeleteLink(interval,&(function->interval_list));
	    }
	}

    function->generated = TRUE;

    return PWC_OK;
    }

/*
*   GET STEPS
*
* Given a position, find the two break points between which it lies.
*
* This routine should not be used before the routine pwdGenerateFunction.
*
* Points returned as left_step and right_step. 
* Returns PWC_OK or PWC_ERROR.
*/
PWC_REPORT pwcGetSteps(position,left_step,right_step,function)
float position;
struct StepLink  **left_step, **right_step;
struct PWC *function;
    {
    struct StepLink *step;

    /* Have steps been ordered yet? */
    if(function->generated == FALSE)
	{
	printf("E: Can not do pwcGetSteps before pwdGenerateFunction\n");
	return PWC_ERROR;
	}

    /* Any steps in list */
    if(function->step_list == NULL)
	{
	*left_step = *right_step = NULL;
	return PWC_OK;
	}

    /* Search through the step list */
    for(step=function->step_list; step->next!=NULL; step=step->next)
	{
	if(step->position > position)
	    {
	    *left_step = step->prev;
	    *right_step = step;
	    return PWC_OK;
	    }
    	}
    
    /* Got to here. Position is to right of last step */
    *left_step = step;
    *right_step = NULL;
    return PWC_OK;
    }

/*
*   GET INTERVAL
*
* Given a position, find the interval in which it lies.
*
* This routine should not be used before the routine pwdGenerateFunction.
*
* Interval returned as "interval". 
* Returns PWC_OK or PWC_ERROR.
*/
PWC_REPORT pwcGetInterval(position,interval,function)
float position;
struct IntervalLink **interval;
struct PWC *function;
    {
    struct IntervalLink *test_interval;

    /* Have intervals been generated yet? */
    if(function->generated == FALSE)
	{
	printf("E: Can not do pwcGetInterval before pwdGenerateFunction\n");
	return PWC_ERROR;
	}


    /* Search through the interval list */
    for(test_interval=function->interval_list; test_interval!=NULL;
	test_interval=test_interval->next)
	{
	if((test_interval->left<=position) && (test_interval->right>=position))
	    {
	    *interval = test_interval;
	    return PWC_OK;
	    }
    	}
    
    printf("E: In pwcGetInterval. %f not in any of the intervals\n",position);
    printf("   Range = \[%f,%f\]\n",function->left,function->right);
    return PWC_ERROR;
    }

/* 
*  MAX VALUE
* 
*  Given an interval, find the maximum value of the function within
*  the interval.
*
*  Return PWC_OK or PWC_ERROR.
*/
PWC_REPORT pwcMaxValue(left,right,value,function)
float left,right;
float *value;
struct PWC *function;
    {
    struct IntervalLink *interval_link;
    float max_value;

    /* Have intervals been generated yet? */
    if(function->generated == FALSE)
	{
	printf("E: Can not do pwcMaxValue before pwdGenerateFunction\n");
	return PWC_ERROR;
	}

    /* Get the interval constaining the left end */
    if(pwcGetInterval(left,&interval_link,function)!=PWC_OK) 
	{
	printf("E: in pwcMaxValue\n");
	return PWC_ERROR;
	}

    /* Initialise the maximum value */
    max_value = interval_link->value;

    /* Work through the interval list */
    for(; interval_link->left <= right; interval_link=interval_link->next)
	{
	if(INTERSECT(interval_link->left,interval_link->right,left,right))
	    max_value = MAX(max_value,interval_link->value);
	}

    *value = max_value;
    return PWC_OK;
    }


/* 
*  MAX VALUE AND INTERVAL
* 
*  Given an interval, find the maximum value of the function within
*  the interval and the (largest) sub-interval in which the maximum occurs.
*
*  Return PWC_OK or PWC_ERROR.
*/
PWC_REPORT pwcMaxValueAndInterval(left,right,value,max_left,max_right,function)
float left,right;
float *value;
float *max_left,*max_right;
struct PWC *function;
    {
    struct IntervalLink *interval_link;
    float max_value;
    float max_int_left;
    float max_int_right;

    /* Have intervals been generated yet? */
    if(function->generated == FALSE)
	{
	printf("E: pwcMaxValueAndINterval before pwdGenerateFunction\n");
	return PWC_ERROR;
	}

    /* Get the interval constaining the left end */
    if(pwcGetInterval(left,&interval_link,function)!=PWC_OK) 
	{
	printf("E: in pwcMaxValueAndInterval\n");
	return PWC_ERROR;
	}

    /* Initialise the maximum value */
    max_value = interval_link->value;
    max_int_left = interval_link->left;
    max_int_right = interval_link->right;

    /* Work through the interval list */
    for(; interval_link->left <= right; interval_link=interval_link->next)
	{
	if(INTERSECT(interval_link->left,interval_link->right,left,right))
 	    {
	    if(  (interval_link->value > max_value)
	       ||( (interval_link->value == max_value)
		 &&(   (interval_link->right-interval_link->left) 
		     > (max_int_right - max_int_left) ) ) )
		{
	        max_value = interval_link->value;
		max_int_left = interval_link->left;
		max_int_right = interval_link->right;
		}
	    }
	}

    *max_left = max_int_left;
    *max_right = max_int_right;
    *value = max_value;
    return PWC_OK;
    }


/* 
*  MIN VALUE
* 
*  Given an interval, find the minimum value of the function within
*  the interval.
*
*  Return PWC_OK or PWC_ERROR.
*/
PWC_REPORT pwcMinValue(left,right,value,function)
float left,right;
float *value;
struct PWC *function;
    {
    struct IntervalLink *interval_link;
    float min_value;

    /* Have intervals been generated yet? */
    if(function->generated == FALSE)
	{
	printf("E: Can not do pwcMinalue before pwdGenerateFunction\n");
	return PWC_ERROR;
	}

    /* Get the interval constaining the left end */
    if(pwcGetInterval(left,&interval_link,function)!=PWC_OK) 
	{
	printf("E: in pwcMinValue\n");
	return PWC_ERROR;
	}

    /* Initialise the minimum value */
    min_value = interval_link->value;

    /* Work through the interval list */
    for(; interval_link->left <= right; interval_link=interval_link->next)
	{
	if(INTERSECT(interval_link->left,interval_link->right,left,right))
	    min_value = MIN(min_value,interval_link->value);
	}

    *value = min_value;
    return PWC_OK;
    }


/* 
*  MIN VALUE AND INTERVAL
* 
*  Given an interval, find the minimum value of the function within
*  the interval and the (largest) sub-interval in which the minimum occurs.
*
*  Return PWC_OK or PWC_ERROR.
*/
PWC_REPORT pwcMinValueAndInterval(left,right,value,min_left,min_right,function)
float left,right;
float *value;
float *min_left,*min_right;
struct PWC *function;
    {
    struct IntervalLink *interval_link;
    float min_value;
    float min_int_left;
    float min_int_right;

    /* Have intervals been generated yet? */
    if(function->generated == FALSE)
	{
	printf("E: pwcMinValueAndInterval before pwdGenerateFunction\n");
	return PWC_ERROR;
	}

    /* Get the interval constaining the left end */
    if(pwcGetInterval(left,&interval_link,function)!=PWC_OK) 
	{
	printf("E: in pwcMinValueAndInterval\n");
	return PWC_ERROR;
	}

    /* Initialise the minumum value */
    min_value = interval_link->value;
    min_int_left = interval_link->left;
    min_int_right = interval_link->right;

    /* Work through the interval list */
    for(; interval_link->left <= right; interval_link=interval_link->next)
	{
	if(INTERSECT(interval_link->left,interval_link->right,left,right))
 	    {
	    if(  (interval_link->value < min_value)
	       ||( (interval_link->value == min_value)
		 &&(   (interval_link->right-interval_link->left) 
		     > (min_int_right - min_int_left) ) ) )
		{
	        min_value = interval_link->value;
		min_int_left = interval_link->left;
		min_int_right = interval_link->right;
		}
	    }
	}

    *min_left = min_int_left;
    *min_right = min_int_right;
    *value = min_value;
    return PWC_OK;
    }

/*
*  MULTIPTLY TWO FUNCTIONS
*
*  Given to piece-wise constant functions, functionA(x) and functionB(x), 
*  generates the product function, product(x) = functionA(x) * functionB(x).
*
*  Product is defined over the intersection of te ranges of functionA and
*  functionB
*
*  Functions A and B must both have been "generated" before 
*  pwcMultiplyFunctions can be used.
*
*  Creates intervals for the product function. Does not
*  create a step list.
*
*  Returns pointer to the product function, or NULL on error.
*/
struct PWC *pwcMultiplyFunctions(functionA,functionB,product_name)
struct PWC *functionA, *functionB;
char       *product_name;

    {
    struct PWC *product;
    struct IntervalLink *intervalA, *intervalB;
    struct IntervalLink *interval = NULL;


    /* Make sure that functions A and B have been "generated" */
    if( (functionA->generated != TRUE) || (functionB->generated != TRUE) )
	{
	printf("E: In pwcMultiplyFunctions, functions not \"generated\"\n");
	return NULL;
	}

    /* Start the new function */
    if((product=pwcInit(product_name,functionA->left,functionB->right)) == NULL)
	 return NULL;

    /* Get the range of the product */
    product->left = MAX(functionA->left,functionB->left);
    product->right = MIN(functionA->right,functionB->right);

    /* Get the left-most intervals to multiply */
    if(pwcGetInterval(product->left,&intervalA,functionA)!=PWC_OK) 
	return NULL;
    if(pwcGetInterval(product->left,&intervalB,functionB)!=PWC_OK) 
	return NULL;

    /* Work through the two functions */
    while( (intervalA!=NULL) && (intervalB != NULL) )
        {
	/* Create a new interval (put it at end of list, after last interval
           created) */
	if( (interval=(struct IntervalLink *)
	    nwAddLink(sizeof(struct IntervalLink),&(product->interval_list),
	    interval,NULL)) == NULL) 
	    {
	    printf("E: in pwcMultiplyFunctions\n");
	    return NULL;
	    }
        interval->left = MAX(intervalA->left,intervalB->left);
        interval->right = MIN(intervalA->right,intervalB->right);
        interval->value = intervalA->value * intervalB->value;

	/* Get the next interval */
	if(intervalA->right < intervalB->right)
	    {
	    intervalA = intervalA->next;
	    }
	else
	    {
	    intervalB = intervalB->next;
	    }
	}

    /* Remove any single point intervals */
    for(interval=product->interval_list;interval!=NULL;interval=interval->next)
	{
	if(interval->left==interval->right)
	    {
	    nwDeleteLink(interval,&(product->interval_list));
	    }
	}

    product->generated = TRUE;

    return product;
    }


/*
*  ADD TWO FUNCTIONS
*
*  Given to piece-wise constant functions, functionA(x) and functionB(x), 
*  generates the sum function, sum(x) = functionA(x) + functionB(x).
*
*  Sum is defined over the intersection of the ranges of the two function.
*
*  Functions A and B must both have been "generated" before pwcAdd
*  can be used.
*
*  Creates intervals for the sum function. Does not
*  create a step list.
*
*  Returns pointer to the sum function, or NULL on error.
*/
struct PWC *pwcAddFunctions(functionA,functionB,sum_name)
struct PWC *functionA, *functionB;
char       *sum_name;

    {
    struct PWC *sum;
    struct IntervalLink *intervalA, *intervalB;
    struct IntervalLink *interval = NULL;

    /* Make sure that functions A and B have been "generated" */
    if( (functionA->generated != TRUE) || (functionB->generated != TRUE) )
	{
	printf("E: In pwcSum, functions not \"generated\"\n");
	return NULL;
	}

    /* Start the new function */
    if( (sum=pwcInit(sum_name,functionA->left,functionB->right)) == NULL)
	 return NULL;

    /* Get the range of the sum */
    sum->left = MAX(functionA->left,functionB->left);
    sum->right = MIN(functionA->right,functionB->right);

    /* Get the left-most intervals to add */
    if(pwcGetInterval(sum->left,&intervalA,functionA)!=PWC_OK) 
	return NULL;
    if(pwcGetInterval(sum->left,&intervalB,functionB)!=PWC_OK) 
	return NULL;

    /* Work through the two functions */
    while( (intervalA!=NULL) && (intervalB != NULL) )
        {
	/* Create a new interval (put it at end of list, after last interval
           created) */
	if( (interval=(struct IntervalLink *)
	    nwAddLink(sizeof(struct IntervalLink),&(sum->interval_list),
	    interval,NULL)) == NULL) 
	    {
	    printf("E: in pwcSum\n");
	    return NULL;
	    }
        interval->left = MAX(intervalA->left,intervalB->left);
        interval->right = MIN(intervalA->right,intervalB->right);
        interval->value = intervalA->value + intervalB->value;

	/* Get the next interval */
	if(intervalA->right < intervalB->right)
	    {
	    intervalA = intervalA->next;
	    }
	else
	    {
	    intervalB = intervalB->next;
	    }
	}

    /* Remove any single point intervals */
    for(interval=sum->interval_list;interval!=NULL;interval=interval->next)
	{
	if(interval->left==interval->right)
	    {
	    nwDeleteLink(interval,&(sum->interval_list));
	    }
	}

    sum->generated = TRUE;

    return sum;
    }

/* 
*  NORMALISE 
* 
*  Normalises a function. Mutiplies through so that the maximum
*  value (NB NOT the max absolute value) is the given "normal". 
*
*  Return PWC_OK or PWC_ERROR.
*/
PWC_REPORT pwcNormalise(normal,function)
float normal;
struct PWC *function;
    {
    float max_value;
    float ratio;
    struct IntervalLink *interval_link;

    /* Have intervals been generated yet? */
    if(function->generated == FALSE)
	{
	printf("E: Can not do pwcNoramise before pwdGenerateFunction\n");
	return PWC_ERROR;
	}

    /* Get the maximum value */
    if(pwcMaxValue(function->left,function->right,&max_value,function)!=PWC_OK)
	return PWC_ERROR;
    
    /* Multiplication ratio */
    if(max_value == 0.0) return PWC_OK;
    ratio = normal/max_value;

    /* Work through the interval list */
    for(interval_link=function->interval_list; interval_link != NULL; 
	interval_link=interval_link->next)
	{
        interval_link->value = ratio * interval_link->value;
	}

    return PWC_OK;
    }




/* 
*  MAKE INTO BINARY - MAX
* 
*  Converts given function into a binary function, defined over
*  the same interval.
*
*  Intervals that are within delta of the maximum value evaluate
*  to 1, all other intervals eveluate to 0.
*
*  Return PWC_OK or PWC_ERROR.
*/
PWC_REPORT pwcBinaryMax(delta,function)
float delta;
struct PWC *function;
    {
    float max_value;
    struct IntervalLink *interval_link, *next_link;

    /* Have intervals been generated yet? */
    if(function->generated == FALSE)
	{
	printf("E: Can not do pwcNoramise before pwdGenerateFunction\n");
	return PWC_ERROR;
	}

    /* Get the maximum value */
    if(pwcMaxValue(function->left,function->right,&max_value,function)!=PWC_OK)
	return PWC_ERROR;
    
    /* Work through the interval list */
    for(interval_link=function->interval_list; interval_link != NULL; 
	interval_link=next_link)
	{
	next_link =interval_link->next;

	if(interval_link->value >= (max_value - delta))
	    {
	    interval_link->value = 1;
	    }
	else 
	    {
	    interval_link->value = 0;
	    }

	/* Clean up */
	if( (interval_link->prev!=NULL)
	 && (interval_link->prev->value == interval_link->value))
	    {
	    interval_link->prev->right = interval_link->right;
   	    nwDeleteLink(function->interval_list,&(function->interval_list));
	    }
	}

    return PWC_OK;
    }


/* 
*  MAKE INTO BINARY - MIN
* 
*  Converts given function into a binary function, defined over
*  the same interval.
*
*  Intervals that are within delta of the minimum value evaluate
*  to 1, all other intervals eveluate to 0.
*
*  Return PWC_OK or PWC_ERROR.
*/
PWC_REPORT pwcBinaryMin(delta,function)
float delta;
struct PWC *function;
    {
    float min_value;
    struct IntervalLink *interval_link, *next_link;

    /* Have intervals been generated yet? */
    if(function->generated == FALSE)
	{
	printf("E: Can not do pwcNoramise before pwdGenerateFunction\n");
	return PWC_ERROR;
	}

    /* Get the minimum value */
    if(pwcMinValue(function->left,function->right,&min_value,function)!=PWC_OK)
	return PWC_ERROR;
    
    /* Work through the interval list */
    for(interval_link=function->interval_list; interval_link != NULL; 
	interval_link=next_link)
	{
	next_link = interval_link->next;

	if(interval_link->value <= (min_value + delta))
	    {
	    interval_link->value = 1;
	    }
	else 
	    {
	    interval_link->value = 0;
	    }

	/* Clean up */
	if( (interval_link->prev!=NULL)
	 && (interval_link->prev->value == interval_link->value))
	    {
	    interval_link->prev->right = interval_link->right;
   	    nwDeleteLink(function->interval_list,&(function->interval_list));
	    }
	}

    return PWC_OK;
    }


/* 
*  REDUCE RANGE 
* 
*  Reduces the range over which the given function is defined.
* 
*  Remove all intervals to the left and right of the new range.
*  Adjusts the spans of the intervals crossing the end points of the
*  new range.
*
*  Return PWC_OK or PWC_ERROR.
*/
PWC_REPORT pwcReduceRange(new_left,new_right,function)
float new_left, new_right;
struct PWC *function;
    {
    struct IntervalLink *left_interval, *right_interval;

    /* Have intervals been generated yet? */
    if(function->generated == FALSE)
	{
	printf("E: Can not do pwcReduceRange before pwdGenerateFunction\n");
	return PWC_ERROR;
	}

    /* Make sure new interval is within the old one */
    if( (new_left < function->left) || (new_right > function->right))
	{
	printf("E: In pwcReduceRange - new interval not within old one.\n");
	return PWC_ERROR;
	}

    /* Get the interval crossing the new left end */
    if(pwcGetInterval(new_left,&left_interval,function)!=PWC_OK) 
	return PWC_ERROR;
    if(pwcGetInterval(new_right,&right_interval,function)!=PWC_OK) 
	return PWC_ERROR;

    /* Adjust the new end intervals */
    left_interval->left = new_left;
    right_interval->right = new_right;

    /* Remove intervals to left of left_interval */
    while(function->interval_list != left_interval)
	{
	nwDeleteLink(function->interval_list,&(function->interval_list));
	}

    /* Remove intervals to right of right_interval */
    nwDeleteList(&(right_interval->next));

    function->left = new_left;
    function->right = new_right;

    return PWC_OK;
    }


/*
*  DELETE FUNCTION
*
*  Deletes given function.
*
*  If NULL function is given then nothing is done and PWC_OK returned.
* 
*  Returns PWC_OK or PWC_ERROR.
*/
pwcDeleteFunction(function)
struct PWC *function;
    {
    if(function == NULL) return PWC_OK;

    nwDeleteList(&function->step_list);
    nwDeleteList(&function->interval_list);
    free(function->name);
    free(function);

    return PWC_OK;
    }
