c    la01cd - a routine to solve the linear programming problem.
c      the revised simplex method is used further modified so
c      that in the event of an infeasible initial solution
c      phase i is entered with only one artificial variable.
c
      subroutine la01cd(n,m,l,a,b,c,x,f,ia,iprint,ind,v,ier,
     *                 u,iu,w)
      implicit double precision(a-h,o-z)
      double precision a(ia,1),b(1),c(1),x(1),u(iu,1),w(1),v(1)
      integer ind(1)
c
c    COMMENT ADDED BY BILL NYE:
c
c    Solves an extended standard linear programming problem which involves
c    both equality as well as inequality constraints:
c
c      { min  C'*X   |   A(1~l)'*X <= B(1~l) ; A(l+1~m)'*X == B(l+1~m) }
c         X
c
c    where "A(j~k)" means rows j through k of matrix A().
c
c    END COMMENT ADDED BY BILL NYE.
c
c    m      integer scalar; (user:*); no. of constraints.
c    n      integer scalar; (user:*); no. of user variables.
c    l      integer scalar; (user:*); no. of inequalities, l<=m.
c    a      real array(m by n); (user:*); the constraint matrix,
c           the l inequalities must be in the first l rows.
c    b      real array(m); (user:*); constraint right hand sides.
c    c      real array(n); (user:*); objective function coeffs.
c    x      real array(n); (*:la01cd); the solution.
c    f      real scalar; (*:la01cd); value of objective function.
c    ia     integer scalar; (user:*); 1st dimension of a.
c    iprint integer scalar; (user:*); print option,
c             iprint=0  suppress printing
c             iprint=1  final result only
c             iprint=2  iteration function values
c             iprint=3,4 more detailed iteration print outs
c    ind    integer array(m+1); (user:la01cd); basis indicator.
c           the user may force the initial basis by setting nbasis
c           (in common la01dd) and the 1st nbasis elements of ind.
c           on return ind will hold the final basis.
c    v      real array(m+1); (*:la01cd); workspace, contains the
c           basic marginal costs on return.
c    ier    integer scalar; (*:la01cd); error flag,
c             ier=0  success
c             ier<0  success but required !ier! reinversions
c             ier=1  allowed no. of reinversions exeeded
c             ier=2  max basic marginal cost > tolerance
c             ier=3  phase i failed to find a feasible solution
c             ier=4  la01cd failed to find m independent cols
c             ier=5  solution unbounded
c             ier=6  invalid arguments (n,m,l)
c    u      real array(m+1 by m+1); (*:la01cd); transpose of basis
c           inverse.
c    iu     integer scalar; (user:*); 1st dimension of u
c    w      real array(m+1); (*:la01cd); workspace, lp solution
c
c    common: see description in block data
c
      common/la01dd/ eps,maxinv,nbasis,ip,isp,tol,name
      double precision name
c
c      constants:
c
c      rnderr: slight over-estimate off rounding error
c      big: largest floating point number
c      one: the constant 1.0
c      zero: the constant 0.0
cx
cx    double precision rnderr/1e-16/,big/z7fffffffffffffff/,one/1d0/,zero/0d0/
cx
cx: the previous line was refused by f77; it has been replaced by the next 2
cx
      double precision rnderr, big, one, zero
      data rnderr/1.0d-16/,   big/1.0d38/,   one/1.0d0/,   zero/0.0d0/
cx
cx begin change
cx data statement added by Bill Nye to set the common variables, per page
cx 4 of the Harwell LA01B writeup.
cx
      data eps,  maxinv,  nbasis,  ip,  isp,  tol,     name
     * / 0.0d0,    10  ,     0  ,   6,    0,  1.0d-4,  'la01bd' /
cx end change
c
c***************************************************
c
c      prologue
c
c*******************************************************
c
c      test argument values
      if(n.le.0) go to 540
      if(m.le.0) go to 540
      if(l.gt.m) go to 540
      if(m.gt.n+l) go to 540
c
c      set print options
      ipr=iprint
      if(ip.le.0) ipr=0
      if(ipr.ge.4) write(ip,10) name,n,m,l
   10 format('0',a8,': no. vars ',i3,'     no. constraints ',i4,
     *    '     no. inequalities ',i4)
c
c      no. of re-inversions of basis
      inv=0
c      error flag (0 = no errors)
      ier=0
      m1=m+1
c
c*****************************************************
c
c      select and invert initial basis
c
c******************************************************
c
c      la01ed builds the inverse a column at a time putting
c      in the slacks, if any, first.
c      note: the inverse is held in u in transposed form.
c
c      if nbasis > 0 the user is forcing cols into the basis
c      and these are in ind(j) j=1,nbasis
c
c
      call la01ed(m,n,l,a,ia,nbasis,ind,u,iu,ie)
c
c      test for success
      if(ie.ge.0) go to 500
c
c***************************************************
c
c      compute current solution, w( ),
c      and test to see if it is feasible
c
c******************************************************
c
   20 amax=zero
      do 30 i=1,m
      w(i)=fm01ad(m,u(1,i),b)
c      if we have to introduce an artificial variable we remove
c      basis col which has max(dabs(w(i))) to ensure stability
      if(dabs(w(i)).le.amax) go to 30
      amax=dabs(w(i))
      id=i
   30 continue
c
      if(amax.eq.zero) go to 90
      small=-amax*rnderr
      do 40 i=1,m
      if(w(i).lt.small) go to 50
   40 continue
      go to 90
c
c***************************************************
c
c      phase i
c
c********************************************************
c
c      the solution is not feasible, so introduce one artificial
c      variable by forcing the current solution to be w(i)=1 i=1,m
c
   50 iphase=1
c
c      update basis indicator and inverse
c
      ind(id)=n+l+1
      do 70 j=1,m
      delta=u(j,id)/w(id)
      do 60 k=1,m
   60 u(j,k)=u(j,k)-delta*(w(k)-one)
   70 u(j,id)=delta
c
c      set current solution
      do 80 i=1,m
   80 w(i)=one
c
c      set up for phase i
c
c      reduce is used to ensure the artificial variable gets
c      removed from the basis in the event of a tie.
      reduce=dabs(one-rnderr)
c
c      dimension of inverse is m
      ml=m
      npr=4
      go to 120
c
c*****************************************************
c
c      phase ii
c
c*********************************************************
c
c      introduce objective function and increase dimension
c      of basis to m + 1.
   90 iphase=2
c      compute (m+1)th row of augumented basis inverse
      do 110 i=1,m
      u(i,m1)=zero
      do 100 j=1,m
      if(ind(j).gt.n) go to 100
      u(i,m1)=u(i,m1)+u(i,j)*c(ind(j))
  100 continue
  110 continue
c      w(m1) = objective function value
      w(m1)=fm01ad(m,u(1,m1),b)
c
c      set up for phase ii
c
c      dimension of inverse is m+1
      ml=m1
      npr=3
      id=m1
c
c*********************************************************
c
c      main iteration of the simplex method
c
c      this code is shared by phase i and phase ii
c
c*********************************************************
c
c
  120 icnt=0
c
      if(ipr.ge.2) write(ip,130) name,iphase
  130 format('0',a8,': phase ',i1)
c
  140 icnt=icnt+1
c
      if(ipr.ge.2) write(ip,150) name,icnt,w(id)
  150 format('0',a8,': iteration no. ',i3,'  cost value ',1pe15.6)
      if(ipr.lt.npr) go to 180
      write(ip,160) name
  160 format(' ',a8,': lp basis variables')
      write(ip,170) (ind(i),w(i),i=1,m)
  170 format((4('   ',i3,' ',1pe23.16)))
c
c      compute marginal costs and search for a maximum
c      to decide which col is to enter the basis
c
  180 amax=zero
      jk=0
c
c      marginal costs for columns of a
      do 190 i=1,n
      sum=fm01ad(m,u(1,id),a(1,i))
      if(iphase.eq.2) sum=sum-c(i)
      if(amax.ge.sum) go to 190
      amax=sum
      jk=i
  190 continue
      if(l.le.0) go to 210
c
c      marginal costs for slacks
      do 200 i=1,l
      if(amax.ge.u(i,id)) go to 200
      amax=u(i,id)
      jk=i+n
  200 continue
c
c      test for end of lp iterations,
c      are all marginal costs <= 0 ?
  210 if(jk.le.0) go to 310
c      or, has a basic marginal cost been chosen ?
      do 220 j=1,m
      if(jk.eq.ind(j)) go to 310
  220 continue
c
c      continue with lp iterations, save chosen marginal cost
      v(m1)=amax
c
c      decide which column is to leave the basis
c
      theta=big
      il=0
      if(jk.le.n) go to 240
c
c      incoming column is a slack
      do 230 i=1,m
      v(i)=u(jk-n,i)
      if(v(i).le.zero) go to 230
      y=w(i)/v(i)
      if(theta.le.y) go to 230
      theta=y
      il=i
  230 continue
      go to 260
c
c      incoming column is from a
  240 do 250 i=1,m
      v(i)=fm01ad(m,u(1,i),a(1,jk))
      if(v(i).le.zero) go to 250
      y=w(i)/v(i)
      if(theta.le.y) go to 250
      theta=y
      il=i
  250 continue
c
c
  260 if(iphase.eq.2) go to 270
c      make sure artificial variable is removed in event of tie
      if(reduce*w(id)/v(id).gt.theta) go to 270
      il=id
  270 if(il.eq.0) go to 520
c      update basis indicator
      ind(il)=jk
c      update current lp solution
      do 280 i=1,ml
  280 w(i)=w(i)-theta*v(i)
      w(il)=theta
c
c      update basis inverse
c
      do 300 j=1,ml
      delta=u(j,il)/v(il)
      do 290 i=1,ml
  290 u(j,i)=u(j,i)-delta*v(i)
  300 u(j,il)=delta
c
c      did we just remove artificial variable ?
      if(il.ne.id) go to 140
c      yes: switch to phase ii
      iphase=2
      go to 90
c
c*****************************************************
c
c      a normal exit from the lp iterations
c
c*********************************************************
c
c      do we re-invert basis inverse ?
  310 if(icnt.le.1) go to 370
      if(inv.ge.maxinv) go to 350
c
c      yes: set up to re-invert basis
      ier=-inv
c      no. of re-inversions
      inv=inv+1
      nb=m
      if(iphase.eq.2) go to 320
c      discard artificial variable before trying again
      ind(id)=ind(m)
      nb=m-1
c
c      re-invert basis
  320 call la01ed(m,n,l,a,ia,nb,ind,u,iu,ie)
c
c      test for success
      if(ie.lt.0) go to 330
c      not successful, have we tried using just the cols that
c      are independent and scanning the tableau for the others ?
      if(nb.lt.ie) go to 500
c      no: try it then
      nb=ie
      go to 320
c
  330 if(ipr.ge.2) write(ip,340) name
  340 format('0',a8,': basis re-inverted')
c
c      go and see if solution is still feasible
      go to 20
c
c*****************************************************
c
c      warning: permitted no. of re-inversions exeeded
c
c******************************************************
c
  350 if(maxinv.le.0) go to 370
      if(isp.gt.0) write(isp,360) name,maxinv
  360 format('0',a8,' warning: max no. of ',i3,' basis re-inversions',
     *    ' exeeded')
      ier=1
c
c*****************************************************
c
c      compute marginal costs associated with basis
c
c*****************************************************
c
c      compute marginal costs and find maximum
  370 eps=zero
      do 400 i=1,m
c      ignore artificial variable cost (phase 1 only)
      if(i.eq.id) go to 400
      k=ind(i)
      if(k.gt.n) go to 380
c      column from a
      v(i)=fm01ad(m,u(1,id),a(1,k))
      if(iphase.eq.2) v(i)=v(i)-c(k)
      go to 390
c      column from slacks
  380 v(i)=u(k-n,m1)
  390 if(dabs(v(i)).le.eps) go to 400
      eps=dabs(v(i))
  400 continue
      v(id)=w(id)
c      are we still in phase 1 ?
      if(iphase.eq.1) go to 480
c
c      if largest marginal cost is > tol set error flag
      if(eps.gt.tol) ier=2
c
c      copy lp solution into caller's array
c
      do 410 k=1,n
  410 x(k)=zero
      do 430 j=1,m
      if(ind(j)-n)420,420,430
  420 x(ind(j))=w(j)
  430 continue
c
c      return objective function value
c
      f=w(m1)
c
      if(ipr.eq.0) go to 570
c
c******************************************************
c
c      output section for final results
c
c*******************************************************
c
      write(ip,440) name,f
  440 format('0',a8,': optimum cost value ',1pe25.16)
      write(ip,450) name,ier,eps
  450 format('0',a8,': error flags ',i4,' and ',1pe13.6)
      write(ip,460) name
  460 format('0',a8,': optimum solution')
      write(ip,470) (i,x(i),i=1,n)
  470 format((3('   ',i3,' ',1pe23.16)))
      go to 570
c
c******************************************************
c
c      error: phase i failed to find a feasible solution
c
c******************************************************
c
  480 if(isp.gt.0) write(isp,490) name
  490 format('0',a8,' error: phase i has failed to find a feasible',
     *' solution')
      ier=3
      go to 570
c
c*****************************************************
c
c      error: inverse building section failed to find
c      m independent columns from the tableau
c
c********************************************************
c
  500 if(isp.gt.0) write(isp,510) name
  510 format('0',a8,' error: rank of constraint matrix < m')
      ier=4
      go to 570
c
c*****************************************************
c
c      error: lp iteration predicts an unbounded solution
c
c******************************************************
c
  520 if(isp.gt.0) write(isp,530) name
  530 format('0',a8,' warning: solution is unbounded')
      ier=5
      go to 570
c
c******************************************************
c
c      error: invalid arguments
c
c******************************************************
c
  540 if(isp.le.0) go to 560
      write(isp,550) name
  550 format('0',a8,' error: invalid argument')
      write(isp,10) name,n,m,l
  560 ier=6
      go to 570
c
c******************************************************
c
c      return to caller (this is the only exit point)
c
c*****************************************************
  570 return
      end










