/*   sens_match.c
 *
 *   Author:     Umakanta Choudhury
 *   Date:       30th June 1989
 *   Last modified:  Enrico Malavasi,  Aug. 1992
 *
 *   Purpose: Compute the sensitivities of matched parasitics
 */
#include <stdio.h>
#include <string.h>
#include "utility.h"
#include "layout.h"
#include "circuit.h"

static struct Sens_link *
  add_link(sens_link_list_ptr)
struct	Sens_link **sens_link_list_ptr;
{
    struct Sens_link *sens_link;
    
    sens_link = ALLOC(struct Sens_link, 1);
    sens_link->prev = NULL;
    sens_link->next = (*sens_link_list_ptr);
    if(*sens_link_list_ptr != NULL)  (*sens_link_list_ptr)->prev = sens_link;
    *sens_link_list_ptr = sens_link;
    return sens_link;
}

static REPORT
  remove_link(sens_link_list_ptr,nameA,nameB)
struct	Sens_link **sens_link_list_ptr;
char	*nameA;
char	*nameB;
{
    struct Sens_link *link;
    struct Sens_link *prev_link;
    struct Sens_link *next_link;
    
    for(link = *sens_link_list_ptr; link != NULL; link = link->next){
	if(((strcmp(link->nodeA->name,nameA) == 0) &&
	    (strcmp(link->nodeB->name,nameB) == 0)) ||
	   ((strcmp(link->nodeA->name,nameB) == 0) &&
	    (strcmp(link->nodeB->name,nameA) == 0))) {
	    /* found the link */
	    next_link = link->next;
	    prev_link = link->prev;
	    if(next_link != NULL) next_link->prev = prev_link;
	    if(prev_link != NULL) prev_link->next = next_link;
	    else *sens_link_list_ptr = next_link;
	    free(link);
	    break;
	}
    }
    return OK; /* the link not found */
}

static struct Sens_link  *
  find_link(perf, nameA, nameB)
struct	Performance *perf;
char	*nameA;
char	*nameB;
{
    struct	Sens_link	*sens_link;
    
    for(sens_link=perf->pos_sens_list ; sens_link!=NULL ; sens_link=sens_link->next){
	if( (strcmp(sens_link->nodeA->name, nameA) == 0) &&
	   (strcmp(sens_link->nodeB->name, nameB) == 0)){
	    /*  EUREKA  */
	    return sens_link;
	}
	/* try the reverse order */
	if( (strcmp(sens_link->nodeA->name,nameB) == 0) &&
	   (strcmp(sens_link->nodeB->name,nameA) == 0)){
	    /*  EUREKA  */
	    return sens_link;
	}
    }

    for(sens_link = perf->neg_sens_list; sens_link != NULL; sens_link = sens_link->next){
	if( (strcmp(sens_link->nodeA->name,nameA) == 0) &&
	   (strcmp(sens_link->nodeB->name,nameB) == 0)){
	    /*  EUREKA  */
	    return sens_link;
	}
	/* try the reverse order */
	if( (strcmp(sens_link->nodeA->name,nameB) == 0) &&
	   (strcmp(sens_link->nodeB->name,nameA) == 0)){
	    /*  EUREKA  */
	    return sens_link;
	}
    }

    return NULL;
}


void
  sens_match(circuit,layout)
struct Circuit *circuit;
struct Layout *layout;
{
    struct Performance *perf;
    struct Matched_pair *matchpair1;
    struct Matched_pair *matchpair2;
    float sens_1A2A;
    float sens_1B2B;
    float sens_1A2B;
    float sens_1B2A;
    float sens_1A2;
    float sens_1B2;
    float sens_mismatch;
    float max_matched_sens;
    float min_matched_sens;
    float r_max;
    float r_mis;
    char	*name1A;
    char	*name1B;
    char	*name2A;
    char	*name2B;
    char	*name1;
    char	*name2;
    struct	Sens_link	*link_1A2A;
    struct	Sens_link	*link_1B2B;
    struct	Sens_link	*link_1A2B;
    struct	Sens_link	*link_1B2A;
    struct	Sens_link	*link_1A2;
    struct	Sens_link	*link_1B2;
    struct	Sens_link	*link_12;
    struct	Sens_link	*new_link;
    struct	Node	*node;
    struct	Node	*node1;
    struct	Node	*node2;

    /* compute the sensitivities w.r.t independent parasitics
       taking matching into account */

    DEBUGG(printf("sens_match\n"));

    r_max = circuit->max_process_var_percent * 0.01;
    r_mis = circuit->max_mismatch_percent * 0.01;

    for(perf = circuit->perf_list; perf != NULL; perf = perf->next){ 

	DEBUGG( printf("sens_match perf %s\n",perf->name));

	for(matchpair1=circuit->matched_pair_list ; matchpair1!=NULL ; matchpair1=matchpair1->next) {
	    matchpair1->processed = NO;
	}
	for(matchpair1=circuit->matched_pair_list ; matchpair1!=NULL; matchpair1=matchpair1->next) {
	    name1A = matchpair1->nodeA->name;
	    name1B = matchpair1->nodeB->name;
	    matchpair1->processed = YES;
	    /* consider first all the unprocessed matched pairs */
	    for(matchpair2=circuit->matched_pair_list ; matchpair2!=NULL ; matchpair2=matchpair2->next) {
		name2A = matchpair2->nodeA->name;
		name2B = matchpair2->nodeB->name;
		/* Check whether it has already been processed to avoid duplication */
		if(matchpair2->processed == YES)     continue;
		
		/* matching of C(1A,2A) and C(1B,2B) */
		/* find the sensitivity link for C(name1A,name2A) */
		if((link_1A2A = find_link(perf,name1A,name2A)) == NULL){
		    sens_1A2A = 0;
		} else {
		    sens_1A2A = link_1A2A->sens;
		}
		/* find the sensitivity link for C(name1B,name2B) */
		if((link_1B2B = find_link(perf,name1B,name2B)) == NULL){
		    sens_1B2B = 0;
		} else {
		    sens_1B2B = link_1B2B->sens;
		}

		if(sens_range(sens_1A2A,sens_1B2B,r_max,r_mis,&max_matched_sens,&min_matched_sens) != OK){
		    exit( -1 );
		}
		/* C(1A,2A) is treated as the independent parasitic */
		/* first remove the link associated with C(1B,2B) */
		if(sens_1B2B >= 0){
		    remove_link(&(perf->pos_sens_list),name1B,name2B);
		}
		else{
		    remove_link(&(perf->neg_sens_list),name1B,name2B);
		}
		
		if(max_matched_sens > 0){
		    /* (1A,2A) should be in pos_link */
		    if(sens_1A2A > 0){
			/* (1A,2A) already present in pos_link */
			/* modify the sensitivity */
			link_1A2A->sens = max_matched_sens;
		    }
		    else{
			/* form a positive link */
			new_link = add_link(&(perf->pos_sens_list));
			new_link->nodeA = link_1A2A->nodeA;
			new_link->nodeB = link_1A2A->nodeB;
			new_link->chan_info_list = link_1A2A->chan_info_list;
			new_link->sens = max_matched_sens;
		    }
		}
		else{
		    /* remove the positive link if it exists */
		    remove_link(&(perf->pos_sens_list),name1A,name2A);
		}
		if(min_matched_sens < 0){
		    /* (1A,2A) should be in neg_link */
		    if(sens_1A2A < 0){
			/* (1A,2A) already present in neg_link */
			/* modify the sensitivity */
			link_1A2A->sens = min_matched_sens;
		    }
		    else{
			/* form a negative link */
			new_link = add_link(&(perf->neg_sens_list));
			new_link->nodeA = link_1A2A->nodeA;
			new_link->nodeB = link_1A2A->nodeB;
			new_link->chan_info_list 
			  = link_1A2A->chan_info_list;
			new_link->sens = min_matched_sens;
		    }
		}
		else{
		    /* remove the negative link if it exists */
		    remove_link(&(perf->neg_sens_list),name1A,name2A);
		}
		
		/* matching of C(1A,2B) and C(1B,2A) */
		
		/* find the sensitivity link for C(name1A,name2B) */
		if((link_1A2B = find_link(perf,name1A,name2B)) == OK){
		    sens_1A2B = 0;
		}
		else{
		    sens_1A2B = link_1A2B->sens;
		}
		/* find the sensitivity link for C(name1B,name2A) */
		if((link_1B2A = find_link(perf,name1B,name2A)) == NULL){
		    sens_1B2A = 0;
		}
		else{
		    sens_1B2A = link_1B2A->sens;
		}
		if(sens_range(sens_1A2B,sens_1B2A,r_max,r_mis,
					&max_matched_sens,&min_matched_sens) != OK){
		    exit( -1 );
		}
		/* C(1A,2B) is treated as the independent parasitic */
		/* first remove the link associated with C(1B,2A) */
		if(sens_1B2A > 0){
		    remove_link(&(perf->pos_sens_list),name1B,name2A);
		}
		else{
		    remove_link(&(perf->neg_sens_list),name1B,name2A);
		}
		
		if(max_matched_sens > 0){
		    /* (1A,2B) should be in pos_link */
		    if(sens_1A2B > 0){
			/* (1A,2B) already present in pos_link */
			/* modify the sensitivity */
			link_1A2B->sens = max_matched_sens;
		    }
		    else{
			/* form a positive link */
			new_link = add_link(&(perf->pos_sens_list));
			new_link->nodeA = link_1A2B->nodeA;
			new_link->nodeB = link_1A2B->nodeB;
			new_link->chan_info_list 
			  = link_1A2B->chan_info_list;
			new_link->sens = max_matched_sens;
		    }
		}
		else{
		    /* remove the positive link if it exists */
		    remove_link(&(perf->pos_sens_list),name1A,name2B);
		}
		if(min_matched_sens < 0){
		    /* (1A,2B) should be in neg_link */
		    if(sens_1A2B < 0){
			/* (1A,2B) already present in neg_link */
			/* modify the sensitivity */
			link_1A2B->sens = min_matched_sens;
		    }
		    else{
			/* form a negative link */
			new_link = add_link(&(perf->neg_sens_list));
			new_link->nodeA = link_1A2B->nodeA;
			new_link->nodeB = link_1A2B->nodeB;
			new_link->chan_info_list 
			  = link_1A2B->chan_info_list;
			new_link->sens = min_matched_sens;
		    }
		}
		else{
		    /* remove the negative link if it exists */
		    remove_link(&(perf->neg_sens_list),name1A,name2B);
		}
	    }
	    
	    /* now consider all the unmatched nodes */
	    for(node = circuit->node_list; node != NULL; 
		node = node->next){
		if(node->matched == YES) continue;
		name2 = node->name;
		
		/* matching of C(1A,2) and C(1B,2) */
		/* find the sensitivity link for C(name1A,name2) */
		if((link_1A2 = find_link(perf,name1A,name2)) == NULL){
		    sens_1A2 = 0;
		} else {
		    sens_1A2 = link_1A2->sens;
		}
		/* find the sensitivity link for C(name1B,name2) */
		if((link_1B2 = find_link(perf,name1B,name2)) == NULL){
		    sens_1B2 = 0;
		} else {
		    sens_1B2 = link_1B2->sens;
		}
		
		if(sens_range(sens_1A2,sens_1B2,r_max,r_mis,
			      &max_matched_sens,&min_matched_sens) != OK){
		    exit( -1 );
		}


		/* C(1A,2) is treated as the independent parasitic */
		/* first remove the link associated with C(1B,2) */
		if(sens_1B2 > 0){
		    remove_link(&(perf->pos_sens_list),name1B,name2);
		} else {
		    remove_link(&(perf->neg_sens_list),name1B,name2);
		}
		
		if(max_matched_sens > 0){
		    /* (1A,2) should be in pos_link */
		    if(sens_1A2 > 0){
			/* (1A,2) already present in pos_link */
			/* modify the sensitivity */
			link_1A2->sens = max_matched_sens;
		    } else {
			/* form a positive link */
			new_link = add_link(&(perf->pos_sens_list));
			new_link->nodeA = link_1A2->nodeA;
			new_link->nodeB = link_1A2->nodeB;
			new_link->chan_info_list 
			  = link_1A2->chan_info_list;
			new_link->sens = max_matched_sens;
		    }
		} else {
		    /* remove the positive link if it exists */
		    remove_link(&(perf->pos_sens_list),name1A,name2);
		}
		if(min_matched_sens < 0){
		    /* (1A,2) should be in neg_link */
		    if(sens_1A2 < 0){
			/* (1A,2) already present in neg_link */
			/* modify the sensitivity */
			link_1A2->sens = min_matched_sens;
		    } else {
			/* form a negative link */
			new_link = add_link(&(perf->neg_sens_list));
			new_link->nodeA = link_1A2->nodeA;
			new_link->nodeB = link_1A2->nodeB;
			new_link->chan_info_list 
			  = link_1A2->chan_info_list;
			new_link->sens = min_matched_sens;
		    }
		} else {
		    /* remove the negative link if it exists */
		    remove_link(&(perf->neg_sens_list),name1A,name2);
		}
	    }
	}
	/* now consider coupling between the unmatched nodes */
	for(node1 = circuit->node_list; node1 != NULL; 
	    node1 = node1->next){
	    if(node1->matched == YES) continue;
	    for(node2 = circuit->node_list; node2 != NULL; 
		node2 = node2->next){
		if(node2->matched == YES) continue;
		name1 = node1->name;
		name2 = node2->name;
		/* find the sensitivity link for C(name1,name2) */
		if((link_12 = find_link(perf,name1,name2)) != NULL){
		    link_12->sens *= (1 + r_max);
		}
	    }
	}
    }
}
