/*    spiceParse.c
 *
 *    Author: Enrico Malavasi
 *    Ver. 1.00 - July 1992
 *
 *    Purpose: parse a spice deck to accomplish two targets:
 *             1.) Extract a list of all the electrical nodes
 *             2.) Filter off the command lines which could collide
 *                 with specific simulation commands, such as .ac, .dc, .print
 *             The second action is executed only if the "cleanDeck" file name
 *             is non null, in this case it specifies the name of the file where
 *             the filtered deck must be stored.
 */
#include <stdio.h>
#include "utility.h"

typedef struct _myNode  {
    char *name;
    struct _myNode *next;
} _myNode;

static int nodeNum  = 0;
static _myNode  *nodeList = NULL;

static FILE *fout = NULL;


static void
  checkNewNode( name )
char *name;
/*  Check whether the given node is new or it's been previously found  */
{
    _myNode *node;
    for(node=nodeList ; node!=NULL ; node=node->next) {
	if(! strcmp( node->name, name ))  return;
    }

    node = ALLOC( _myNode, 1);
    node->name = util_strsav( name );
    node->next = nodeList;
    nodeList   = node;
}

static void
  processCommand( s )
char *s;
/*  Filter off the undesired commands  */
{
    if(! strncmp( s, "ac ", 3))    return;
    if(! strncmp( s, "dc ", 3))    return;
    if(! strncmp( s, "end ", 4))   return;
    if(! strncmp( s, "end\n", 4))  return;
    if(! strncmp( s, "print ", 6)) return;
    if(! strncmp( s, "plot ", 5))  return;
    
    if(! strncmp( s, "AC ", 3))    return;
    if(! strncmp( s, "DC ", 3))    return;
    if(! strncmp( s, "END ", 4))   return;
    if(! strncmp( s, "END\n", 4))  return;
    if(! strncmp( s, "PRINT ", 6)) return;
    if(! strncmp( s, "PLOT ", 5))  return;

    if( fout != NULL ) 	fprintf( fout, ".%s", s );
}

static void
  processNodes( line, no )
char *line;
int   no;
/*  Extract from the line the 'no' electric node names  */
{
    char  name[128];
    int   i;

    while( (line[0] == ' ') || (line[0] == '\t') )  line++;
    while( (line[0] != ' ') && (line[0] != '\t') )  line++;
    
    while( no > 0 ) {
	while( (line[0] == ' ') || (line[0] == '\t') )  line++;
	for(i=0 ; (line[0] != ' ') && (line[0] != '\t') ; i++) {
	    name[i] = line[0];
	    line++;
	}
	name[i] = '\0';
	checkNewNode( name );
	no--;
    }
}


static void
  processLine( line )
char *line;
/*  Determine which kind of line we've got - extract the nodes and
 *  eventually filter it off if it is a command line */
{
    switch( line[0] ) {
      case '.':
	processCommand( line+1 );
	return;

      case 'm':
      case 'M':
      case 't':
      case 'T':
      case 'o':
      case 'O':
      case 'g':
      case 'G':
      case 'e':
      case 'E':
	processNodes( line, 4 );
	break;

      case 'u':
      case 'U':
      case 'q':
      case 'Q':
      case 'j':
      case 'J':
      case 'z':
      case 'Z':
	processNodes( line, 3 );
	break;

      case 'r':
      case 'R':
      case 'c':
      case 'C':
      case 'l':
      case 'L':
      case 'f':
      case 'F':
      case 'h':
      case 'H':
      case 'b':
      case 'B':
      case 'v':
      case 'V':
      case 'i':
      case 'I':
      case 'd':
      case 'D':
	processNodes( line, 2 );
	break;

      default:
	break;
    }

    if( fout != NULL ) 	fprintf( fout, "%s", line );
}

spiceParse( deckFile, cleanDeck)
char *deckFile;
char *cleanDeck;
/*  This function parses a spice deck, looking for all its electrical node names
 *  and filtering off the undesired commands */
{
    FILE  *fin;
    char  c, line[512];
    int   i;

    if( cleanDeck != NULL ) {
	if((fout=fopen(cleanDeck, "w")) == NULL ) {
	    fprintf(stderr, "ERROR: Couldn't open file %s\n", cleanDeck);
	    exit( -1 );
	}
    }

    if((fin=fopen(deckFile, "r")) == NULL) {
	fprintf(stderr, "ERROR: Couldn't read file %s\n", deckFile);
	exit( -1 );
    }

    for(i=0 ; (line[i]=getc( fin )) != EOF ; i++) {
	if( line[i] == '\n' ) {
	    line[i+1] = '\0';
	    processLine( line );
	    i = (-1);
	}
    }

    fclose( fin );
    fclose( fout );
}

char *
  popNextNode()
/*  Extract the next node from the list of the nodes extracted from
 *  the deck. After the last node has been extracted, return NULL. */
{
    _myNode *node;

    node = nodeList;
    if(node == NULL) return NULL;

    nodeList = nodeList->next;
    return node->name;
}
