#include "copyright.h"
#include "port.h"
#include "utility.h"
#include "oct.h"
#include "oh.h"
#include "vov.h"
#include "symlib.h"
#include "options.h"
#include "tap.h"
#include "th.h"

enum { SET_UNITS, CLEAN_CONSTRAINT, REGULAR_CONSTRAINT, SIDE_CONSTRAINT, 
	 ACTIVE_CONSTRAINT, SIDE_ACTIVE_CONSTRAINT };
/*
 *  Add constraints to oct facets
 *
 */



static void getInstOrTerm( facet, name, inst )
    octObject* facet;
    char* name;
    octObject* inst;
{
    octObject term;
    octObject aterm;

    inst->type = OCT_INSTANCE;
    inst->contents.instance.name = name;
    if (octGetByName(facet, inst) != OCT_OK)  {
	term.type = OCT_TERM;
	term.contents.term.name = name;
	if (octGetByName(facet, &term) != OCT_OK) {
	    errRaise( "putconst", 1, "can't get instance/term '%s'", name);
	}
	if (octGenFirstContent(&term, OCT_TERM_MASK, &aterm) != OCT_OK) {
	    errRaise( "putconst", 1, "can't get term from fterm '%s'", name);
	}
	if (octGenFirstContainer(&aterm, OCT_INSTANCE_MASK, inst) != OCT_OK) {
	    errRaise( "putconst", 1, "can't get inst from fterm '%s'", name);
	}
    }
}


optionStruct optionList[] = {
    { "p", 0,		"Print constraints" },
    { "c", 0,		"Clean all constraints" },
    { "h", 0,		"Help with constraints file format" },
    { "o", "cell:view", "Specify output facet" },
    { "f", "file",	"Constraints file" },
    { OPT_RARG, "cell:view", "Input facet" },
    { 0 , 0 , 0 }
};



static int decode(  s )
    char* s;
{
    if ( strcmp( s , "UNITS" ) == 0 ) {
	return SET_UNITS;
    }
    if ( strcmp( s , "CLEAN" ) == 0 ) {
	return CLEAN_CONSTRAINT;
    }
    if ( strcmp( s , "NORMAL" ) == 0 ) {
	return REGULAR_CONSTRAINT;
    }
    if ( strcmp( s , "SIDE" ) == 0 ) {
	return SIDE_CONSTRAINT;
    }
    if ( strcmp( s , "ACTIVE" ) == 0 ) {
	return ACTIVE_CONSTRAINT;
    }
    if ( strcmp( s , "SIDEACTIVE" ) == 0 ) {
	return SIDE_ACTIVE_CONSTRAINT;
    }
    errRaise( "putconst", 1 , "Cannot decode constraint type %s", s  );
    return 0;			/* Not reached */
}

static int getLambda( facet )
    octObject* facet;
{

    static int lambda = 0;
    
    if ( lambda == 0 ) {
	thTechnology( facet );
	lambda = thLambda();
    }
    return lambda;
}

static int getNumber( s, lambda )
    char* s;
{
    if ( strcmp( s, "---" ) == 0 ) {
	return SYM_PUT_CONST_UNDEFINED;
    }
    return atoi( s ) * lambda;
}

static int getDirection( s  )
  char* s;
{
    int direction;

    *s = tolower( *s );
    if (strcmp( s , "x") == 0)  {
	direction = SYM_HORIZONTAL;
    } else if (strcmp( s, "y") == 0)  {
	direction = SYM_VERTICAL;
    } else  {
	errRaise( "putconst", 1, "Bad format: must be either an x or a y constraint." );
    }
    return direction;
}

static int getSide( s )
    char* s;
{
    int l = strlen( s );
    int i = 0;
    for ( i = 0; i < strlen( s ); i++) s[i] = tolower(s[i]);
    if (strcmp(s, "left")==0)  {
	return  SYM_PUT_CONST_LEFT; 
    } else if  (strcmp(s, "right")==0)  {
	return  SYM_PUT_CONST_RIGHT; 
    } else if  (strcmp(s, "top")==0)   {
	return  SYM_PUT_CONST_TOP; 
    } else if  (strcmp(s, "bottom")==0 )  {
	return  SYM_PUT_CONST_BOTTOM; 
    } else {
	errRaise( "putconst", 1, "Unknown side %s", s );
    }
    return 0;			/* Make lint happy */
}

static char* getWord( fp )
    FILE* fp;
{
    static int index = 0;
    static char words[8][1024];
    
    if ( ++index > 7 ) index = 0;

    while ( 1 ) {
	if ( fscanf( fp, "%s", words[index] ) != 1 ) {
	    errRaise( "putconst", 1, "Premature end of file" );
	}
	if ( *words[index] == '#' ) {
	    while ( fgetc( fp ) != '\n' ); /* Skip over the comment */
	} else {
	    break;		/*  */
	}
    }

    return words[index];
}

static int units = -1;

getUnits()
{
    if ( units == - 1) {
	errRaise( "putconst", 1, "You MUST specify the units (e.g. UNITS lambda)" );
    }
    return units;
}

void setUnits( facet, fp )
    octObject* facet;
    FILE* fp;
{
    char *code = getWord(fp);
    int i = 0;

    while( code[i] != '\0' ) {
	code[i] = tolower( code[i] );
	i++;
    }

    if ( !strcmp(code,"lambda") ) {
	units = getLambda( facet );
    } else if ( !strcmp( code, "oct_units" ) ) {
	units = 1;
    } else {
	errRaise( "putconst", 1, 
		 "Illegal units (%s): Should be either \"lambda\" or \"oct_units\"" , code );
    }
}

void doRegular( facet, fp )
    octObject* facet;
    FILE* fp;
{
    octObject inst1, inst2;
    int direction = getDirection( getWord( fp ) );
    char* name1 = getWord( fp );
    char* name2 = getWord( fp ); 
    int units = getUnits(  facet );
    int min = getNumber( getWord( fp ), units );
    int max = getNumber( getWord( fp ), units );

    getInstOrTerm( facet, name1, &inst1 );
    getInstOrTerm( facet, name2, &inst2 );
    

    if ( min != SYM_PUT_CONST_UNDEFINED && max != SYM_PUT_CONST_UNDEFINED && min > max ) {
	int t = max; max = min; min = t;
    }

    symNormalConstraint( direction, &inst1, &inst2, min, max );
}    


void doSide( facet, fp )
    octObject* facet;
    FILE* fp;
{
    octObject inst;
    int side = getSide( getWord( fp ) );
    char* name = getWord( fp );
    int units = getUnits( facet );
    int min  = getNumber( getWord( fp ), units );
    int max  = getNumber( getWord( fp ), units );
    
    getInstOrTerm( facet, name, &inst );

    symSideConstraint( &inst, side, min, max );
}

void doActive( facet, fp )
    octObject* facet;
    FILE* fp;
{
    octObject inst[4];
    int i;
    int direction = getDirection( getWord( fp ) );
    
    for ( i = 0; i < 4 ; i++ ) {
	char* name = getWord( fp );
	getInstOrTerm( facet, name, &inst[i] );
    }

    symActiveConstraint( direction, &inst[0], &inst[1], &inst[2], &inst[3] );
}

void doSideActive( facet, fp )
    octObject* facet;
    FILE* fp;
{
    octObject inst1, inst2;
    int side = getSide( getWord( fp ) );
    char* name1 = getWord( fp );
    char* name2 = getWord( fp );
    
    getInstOrTerm( facet, name1, &inst1 );
    getInstOrTerm( facet, name2, &inst2 );

    symActiveSideConstraint( side, &inst1, &inst2 );
}


void addConst( facet, fp )
    octObject* facet;
    FILE* fp;
{
    char codeStr[1024];
    while( fscanf( fp, "%s", codeStr ) == 1 ) {
	if ( *codeStr == '#' ) {
	    while ( fgetc( fp ) != '\n' );
	    continue;
	}

	switch( decode( codeStr ) ) {
	case SET_UNITS:
	    setUnits( facet, fp );
	    break;
	case CLEAN_CONSTRAINT:
	    symCleanAllConstraints( facet );
	    break;
	case REGULAR_CONSTRAINT:
	    doRegular( facet, fp );
	    break;
	case SIDE_CONSTRAINT:
	    doSide( facet, fp );
	    break;
	case ACTIVE_CONSTRAINT:
	    doActive( facet, fp );
	    break;
	case SIDE_ACTIVE_CONSTRAINT:
	    doSideActive( facet, fp );
	    break;
	default:
	    errRaise( "putconst", 1, "cannot parse %s", codeStr );
	}

    }

}

void helpFormat()
{
    printf( "\n\
CLEAN\n\
UNITS [lambda|oct_units]\n\
NORMAL dir name name min max\n\
SIDE side name min max\n\
ACTIVE dir name name name name\n\
SIDEACTIVE side name name\n\
\n\
# dir  : x or y\n\
# name : name of a formal terminal or of an instance\n\
# min  : positive integer (or --- if undefined)\n\
# max  : positive integer (or --- if undefined)\n\
# side : left or right or top or bottom\n\
" );
    exit( 0 );
}


main(argc, argv)
    int argc;
    char *argv[];
{
    octObject facet;
    char* fileName = 0;
    char* outfacetName = 0;
    char* infacetName = 0;

    FILE* fp;
    int option;
    int printOnly = 0;
    int cleanOnly = 0;
    extern int optind;
    extern char *optarg;
    
    

    while (( option = optGetOption( argc, argv ) ) != EOF ) {
	switch ( option ) {
	case 'p':
	    printOnly = 1;
	    break;
	case 'h':
	    helpFormat();
	    break;
	case 'c':
	    cleanOnly = 1;
	    break;
	case 'f':
	    fileName = optarg;
	    break;
	case 'o':
	    outfacetName = optarg;
	    break;
	default:
	    optUsage();
	}
    }
    
    if ( optind != argc - 1) {
	optUsage();
    }
    infacetName = argv[optind];

    VOVbegin( argc, argv );
    octBegin();

    ohWorkFacet( &facet, infacetName, outfacetName );

    if ( printOnly ) {
	printf( "UNITS lambda\n" );
	symPrintAllConstraints( &facet );
	VOVend( 0 );
    }
    if ( cleanOnly ) {
	symCleanAllConstraints( &facet );
	OH_ASSERT( octFlushFacet( &facet ) );
	VOVend( 0 );
    }

    if ( fileName == 0  ) {
	printf( "You should specify a constraints file (option -f).\n" );
	VOVend( 1 );
    }

    VOVinput( VOV_UNIX_FILE, fileName );
    fp = fopen( util_file_expand(fileName), "r" );
    if ( fp == 0 ) {
	fprintf( stderr, "Cannot open file %s\n", fileName );
	VOVend( 1 ) ;
    }

    addConst( &facet, fp );

    if (octCloseFacet(&facet) != OCT_OK)  {
	octError( ohFormatName( &facet ) );
	VOVend(1);
    }

    VOVend(0);
}









