/*
 *  FILE: nleMacro.c
 *
 *  DESCRIPTION:
 *
 *
 *  CREATION DATE: xx/xx/xx		AUTHOR:
 *
 *  FUNCTIONS:
 * checkArgs(spot, cmdList)
 * rpcSoftMacro(chip, nName, InstBagP)
 * rpcHardMacro(chip, nName, InstBagP)
 * 
 */

 /*
 ** These routines take a number of instances from Chip and create a new macro cell
 ** containing those instances.
 ** This is accomplished with the following sequence of steps:
 **   1:  create the new cell and put all the instances in it
 **       At the same time mark all the terminals of all the instances to be deleted from Chip
 **   2:  Create all the formal terminals of the new cell. Do this by 
 **       checking all the nets, to see if there is any one that crosses the boundary of the new
 **       cell and also by checking the formal terminals of Chip
 **       In this step, for every terminal created, record all the connections that must be 
 **       made after the new macro has been instantiated in Chip
 **   3:  Instantiate the new cell in Chip and make all the connections previously recorded
 **   4:  Delete all old instances, useless nets, and temporary objects
 **   5:  Finish off, by adding all the properies and bags to the new cell
 */
#include "rpc.h"
#include "nle.h"

static int rpcGenericMacro();


/* >>> GLOBAL VARIABLES */

char 	    *RPCstrsave();

RPCArg    *gArg1, *gArg2;

octObject Chip;			/* The facet on which we work */
octObject MacroContents;	/* Contents facet of new cell */


/*
 *  FUNCTION: rpcSoftMerge
 *  takes a bag full of instances, creates a new master
 *  fills it up with all the instances, then updates the net list
 *  VEM ARGLIST:  Objects "name of new cell" : make-macro
 */
rpcSoftMacro(spot, cmdList , userOptionWord)
     RPCSpot *spot;
     lsList cmdList;
     long userOptionWord;
{
  rpcGenericMacro(spot, cmdList , userOptionWord, NLE_SOFT);
  return(RPC_OK);
}

rpcHardMacro(spot, cmdList , userOptionWord)
     RPCSpot *spot;
     lsList cmdList;
     long userOptionWord;
{
  rpcGenericMacro(spot, cmdList , userOptionWord, NLE_HARD);
  return(RPC_OK);
}


static rpcGenericMacro(spot, cmdList , userOptionWord,type)
     RPCSpot *spot;
     lsList cmdList;
     long userOptionWord;
    int type;
{
  char *name;
  octObject iBag;
  octObject NewInst;
  octObject NewMacro;
  char 		buffer[128];
  octStatus	status;
  struct octBox redrawBox;
  
  if (checkArgs(spot, cmdList) != NLE_OK) return(RPC_OK);

  Chip.objectId = spot->facet;
  octGetById(&Chip);
  iBag.objectId = gArg1->argData.objArg.theBag;
  name = RPCstrsave(gArg2->argData.string);
  octGetById(&iBag);
  
  sprintf(buffer,"Making %s Macro ...\n", type == NLE_SOFT ? "Soft" : "Hard");
  vemMessage(buffer, MSG_DISP|MSG_NOLOG);
 
  NewMacro.type = OCT_FACET;
  NewMacro.contents.facet.cell = name;
  NewMacro.contents.facet.view = "nlerpc";
  NewMacro.contents.facet.facet = "contents";
  NewMacro.contents.facet.version = OCT_CURRENT_VERSION;
  NewMacro.contents.facet.mode = "w";
  status = octOpenFacet(&NewMacro);
  if(status == OCT_OLD_FACET) {
    sprintf(buffer,"%s:%s:%s already exists",
	    NewMacro.contents.facet.cell,
	    NewMacro.contents.facet.view,
	    NewMacro.contents.facet.facet);
    if (dmConfirm("Warning", buffer, "overwrite", "abort")!=VEM_OK) {
      vemMessage("Aborting\n", MSG_DISP|MSG_NOLOG);
      return(RPC_OK);
    }
  } else if (status == OCT_NO_PERM){
    sprintf(buffer,"%s:%s:%s permission denied",
	    NewMacro.contents.facet.cell,
	    NewMacro.contents.facet.view,
	    NewMacro.contents.facet.facet);
    vemMessage(buffer, MSG_DISP|MSG_NOLOG);
    return(RPC_OK);
  }
      
  if (nleMakeMacro( &NewMacro, &iBag , &NewInst, type) != NLE_OK) {
    vemMessage("Macro failed\n", MSG_DISP|MSG_NOLOG);
    return(RPC_OK);
  }

  /*
   * Redisplay chip and open a window on the new cell
   */
  if (octBB(&NewInst,&redrawBox)!=OCT_OK) {
    Warning("No bounding box in the new instance\n");
  } else {
    vemWnQRegion(&Chip,&redrawBox);
    vemWnFlush();
  }

  vemMessage("Macro done. Remember to save everything\n", MSG_DISP|MSG_NOLOG);
  return(RPC_OK);
}


rpcSoftMergeUsage()
{
  vemMessage("format: [object set] \"macro-name\" : make-macro\n", MSG_DISP|MSG_NOLOG);
}

checkArgs(spot, cmdList)
     RPCSpot *spot;
     lsList cmdList;
{
  octObject Facet;
  
  if (lsLength(cmdList) != 2) {
    rpcSoftMergeUsage();
    return(NLE_NOT_OK);
  }
  Facet.objectId = spot->facet;
  if (octGetById(&Facet) != OCT_OK) {
    Warning("Could not find facet");
    return(NLE_NOT_OK);
  }
  lsDelBegin(cmdList, (Pointer *) &gArg1);
  if (gArg1->argType != VEM_OBJ_ARG) {
    rpcSoftMergeUsage();
    return(NLE_NOT_OK);
  }
  lsDelBegin(cmdList, (Pointer *) &gArg2);
  if (gArg2->argType != VEM_TEXT_ARG) {
    rpcSoftMergeUsage();
    return(NLE_NOT_OK);
  }
  return(NLE_OK);
}





