#include "copyright.h"
#include "port.h"
#include "oct.h"
#include "st.h"
#include "oh.h"
#include "vulcan.h"
#include "options.h"
#include "oh.h"
#include "region.h"
#include "utility.h"
#include "tap.h"

#define AVEX_L(bb) ((((bb).lowerLeft.x + (bb).upperRight.x) / 2 ) / get_lambda(0))
#define AVEY_L(bb) ((((bb).lowerLeft.y + (bb).upperRight.y) / 2) / get_lambda(0))

#define AVEX(bb) ((((bb).lowerLeft.x + (bb).upperRight.x) / 2 ))
#define AVEY(bb) ((((bb).lowerLeft.y + (bb).upperRight.y) / 2))




typedef struct path_info {
    octObject path;
    octObject layer;
    int isVertical;
    int numTerms;		/* Number of terminals attached */
    int numNets;		/* Number of nets attached to this path. */
    int32 numPoints;
    struct octPoint* points;
    struct octBox bb;
    struct octObject term[2];
} pathInfo;


static int pointInBox( point, bb )
    octPoint* point;
    octBox* bb;
{
    if ((point->x < bb->lowerLeft.x)  ||
	(point->x > bb->upperRight.x) ||
	(point->y < bb->lowerLeft.y)  ||
	(point->y > bb->upperRight.y)) {
	return 0;
    } 
    return 1;
}



char* pathPrint( info )
    pathInfo *info;
{
    static char buf[256];
    sprintf( buf, "segment on %s at (%d,%d)", info->layer.contents.layer.name,
	    AVEX_L(info->bb), AVEY_L(info->bb) );
    return buf;
}


void pathError( info, msg, action, errcount )
    pathInfo *info;
    char *msg;
    int action;
    int *errcount;
{
    char buffer[256];
    (*errcount)++;
    sprintf( buffer, "msg", pathPrint(&info) );
    if ( action ) {
	symcleanReportError( 0, buffer );
	OH_ASSERT(octDelete( &info->path ));
    } else {
	symcleanReportError( &info->path, buffer );
    }
}

int pathBasicChecks( info, fixp )
    pathInfo *info;
    int fixp;
{
    int errcount = 0;

    if ( info->numTerms < 2) {
	pathError(info,"%s has fewer than 2 terminals\n", fixp, &errcount );
    } 
    if ( info->numTerms > 2) {
	pathError(info,"%s has more than 2 terminals\n", fixp, &errcount );
    }
    if ( info->numNets == 0 ) {
	pathError(info,"%s has no net\n", 1, &errcount );
    } 
    if ( info->numNets > 1 ) {
	pathError(info,"%s is attached to multiple nets\n", fixp, &errcount );
    } 
    if ( info->numPoints > 2 ) {
	pathError(info,"%s has more than 2 points\n", fixp, &errcount );
    } 
    if ( info->numPoints < 2 ) {
	pathError(info,"%s has less than 2 points\n", fixp, &errcount );
    } 

    return errcount;
}



/*
 * check that the two terminals on the segment actually intersect
 * the end points of the segment on the proper layer
 */
static int checkPathGeometry( info, fixp ) 
    pathInfo *info;
    int fixp;
{
    struct octPoint* points = info->points;
    int errcount = 0;
    int sx = points[0].x - points[1].x;
    int sy = points[0].y - points[1].y;

    if ( sx * sy ) {
	pathError( info, "Non manhattan %s\n", fixp, &errcount );
	if ( fixp ) {
	    if ( ABS(sx) < ABS(sy) ) {
		if ( (ABS(sx) / (double)ABS(sy))  < 0.25 ) {
		    points[1].x += sx;
		    octPutPoints( &info->path, 2, points );
		}
	    } else {
		if ( (ABS(sy) / (double)ABS(sx)) < 0.25 ) {
		    points[1].y += sy;
		    octPutPoints( &info-> path, 2, points );
		}
	    }
	}
    }
    if ( (sx + sy) == 0 ) {
	pathError( info, "Zero length %s\n", fixp, &errcount );
    }

    errcount += snapToLambda( &info->path, fixp );

    return errcount;
}



static int
checkSegmentEndPoints(info, fixp)
    pathInfo *info;
    int fixp;
{
    octGenerator genTerm;
    octObject term;
    struct octBox termBB;
    int foundit, i, j, count;
    int countTerms = 0;
    int32 num;
    char **lyrAry;
    int errcount = 0;
    octCoord termWidth;
    octObject *path = &info->path;
    char buffer[256];
    int firstTermIntersectIndex = -1; /* Which terminal does first point intersect? */
    int noIntersectFlag = 0;
    int intersectIndex = 0;


    OH_ASSERT(octInitGenContentsSpecial(path, OCT_TERM_MASK, &genTerm)) ;	/*  */

    while ( octGenerate(&genTerm, &term) == OCT_OK ) {

	if (regFindLyrImpl(&term, &termBB, &count, &lyrAry) != REG_OK) {
	    sprintf(buffer,"No terminal implementation on one of the end points for %s\n",
		    pathPrint( info ) );
	    symcleanReportError( path, buffer );
	}

	foundit = 0;
	for (i = 0; i < count; i++) {
	    if (strcmp(lyrAry[i], info->layer.contents.layer.name) == 0) {
		foundit = 1;
		break;
	    }
	}

	if (!foundit) {
	    sprintf(buffer,"%s does not have the proper implementation geometries\n",
		    ohFormatName( &term ));
	    symcleanReportError( 0, buffer );
	    errcount++;
	}

	/* Check if path is wider than terminal it connects to */
	
	termWidth = (info->isVertical) ? ohBoxWidth( &termBB ) : ohBoxHeight( &termBB );
	if ( termWidth <= 0 ) {
	    sprintf(buffer,"Term %s has no width. This cannot be fixed.\n", ohFormatName( &term ) );
	    symcleanReportError( &term, buffer );
	    errcount++;
	} else if ( path->contents.path.width > termWidth ) {
	    sprintf(buffer,"%s is wider (%d lambdas) than terminal (%d lambdas)\n",
		    pathPrint(info), path->contents.path.width /get_lambda(0) , termWidth/get_lambda(0) );
	    symcleanReportError( path, buffer );
	    if ( fixp ) {
		path->contents.path.width = termWidth;
		OH_ASSERT(octModify( path ));
	    }
	    errcount++;
	}
	
	noIntersectFlag = 0;
	if ( firstTermIntersectIndex == -1 ) {
	    if ( pointInBox( &info->points[0], &termBB ) ) {
		firstTermIntersectIndex = 0;
	    } else if ( pointInBox( &info->points[1], &termBB ) ) {
		firstTermIntersectIndex = 1;
	    } else {
		noIntersectFlag = 1;
		intersectIndex = 0; /* Arbitrarily force intersection with one terminal */
		firstTermIntersectIndex = 0;
	    }
	} else {
	    if ( ! pointInBox( &info->points[1-firstTermIntersectIndex], &termBB ) ) {
		noIntersectFlag = 1;
		intersectIndex = 1 - firstTermIntersectIndex;
	    }
	}


	if ( noIntersectFlag ) {
	    sprintf(buffer,"%s does not intersect %s\n", ohFormatName( &term), pathPrint(info));
	    symcleanReportError( &info->path, buffer );
	    if ( fixp ) {
		/* Make sure the path does not become of zero length */
		int i = intersectIndex;
		int j = 1 - i;
		if ( info->isVertical ) {
		    info->points[i].y = (termBB.lowerLeft.y!=info->points[j].y) ? 
		      termBB.lowerLeft.y : termBB.upperRight.y;
		} else {
		    info->points[i].x = (termBB.lowerLeft.x!=info->points[j].x) ? 
		      termBB.lowerLeft.x : termBB.upperRight.x;
		}
		octPutPoints( &info->path, 2 , info->points );

	    }
	    errcount++;
	}
    }
    octFreeGenerator( &genTerm );
    return errcount;
}



/* 
 * Check that the two terminals attached to
 * the segment are not on the same instance. If they are,
 * there might be an overconstraint for sparcs
*/
static int
checkSegmentTerminals(info)
    pathInfo *info;
{
    octId id1;
    octGenerator genTerm;
    octObject term;
    octObject *path = &info->path;
    int rtn = 0;
    char buffer[256];

    OH_ASSERT(octInitGenContentsSpecial(path,OCT_TERM_MASK,&genTerm)) ;
    
    OH_ASSERT(octGenerate(&genTerm,&term));
    id1 = term.contents.term.instanceId;
    OH_ASSERT(octGenerate(&genTerm,&term));
    if (term.contents.term.instanceId == id1) {
	sprintf(buffer,"%s has both terminals on the same instance.\n\
\tThis may cause an overconstraint in sparcs\n",
		pathPrint( info ) );
	symcleanReportError( path, buffer );
	rtn = 1;
    }
    octFreeGenerator( &genTerm );
    return rtn;
}


pathFillInfo( path, info )
    octObject *path;
    pathInfo *info;
{
    OH_ASSERT( ohGetById( &info->path, path->objectId ));
    OH_ASSERT( octBB( &info->path, &info->bb ) );

    info->numNets = ohCountContainers( &info->path, OCT_NET_MASK );
    info->numTerms = ohCountContents( &info->path, OCT_TERM_MASK );
    
    ohGetPoints(&info->path, &info->numPoints, &info->points);

    OH_ASSERT(octGenFirstContainer(&info->path, OCT_LAYER_MASK, &info->layer)) ; /* Get Layer */


    /* Assuming a manhattan path this is OK */
    info->isVertical = ( info->points[0].x == info->points[1].x );

}



symcleanDoPath( path, fixp )
    octObject *path;
    int fixp;
{
   pathInfo info;
   int errcount = 0;
   

   pathFillInfo( path, &info );

   errcount = pathBasicChecks( &info, fixp );
   
   if ( errcount == 0 ) {
       errcount += checkPathGeometry( &info, fixp );
       errcount += checkSegmentEndPoints(&info, fixp);		/*  implementation */
       errcount += checkSegmentTerminals(&info); /*  terminals are on different instances */
   }

   return errcount;
}
