#include "port.h"
#include "rpc.h"
#include "oh.h"
#include "pep.h"
#include "utility.h"
#include "errtrap.h"

#define PPP(v) util_pretty_print( v )

static jmp_buf pepjmp;

static void handler( pkgname, code, message )
    char* pkgname;
    int code;
    char* message;
{
    char buf[1024];
    
    if ( !strcmp( pkgname, PEP_PKG_NAME ) ||
	!strcmp( pkgname, "RPH" ) ) {

	sprintf( buf, "Error detected by %s:%s\n", pkgname, message );
	VEMMSG( buf );
	longjmp( pepjmp, 1 );
    } else {
	errPass( pkgname, code, message );
    }
}

void RPHgetBag( list, bag )
    lsList list;
    octObject *bag;
{
    RPCArg *arg;

    if ( lsLength( list ) < 1 ) {
	errRaise( "RPH", 1, "Argument list is empty" );
    }    
    lsFirstItem( list, (lsGeneric*) &arg , LS_NH );
    if ( arg->argType != VEM_OBJ_ARG ) {
	errRaise( "RPH", 1, "Wrong arg type (should be objects)" );
    }
    OH_ASSERT( ohGetById( bag, arg->argData.objArg.theBag ));
}

static char* capFormat = "Usage:\n<object> : capacitance\n";
static char* crossCapFormat = "Usage:\n<net|path|term> <net|path|term> : cross-cap\n";
static char* resFormat = "Usage:\n\t<term> <term> : resistance\n\t<segment> <segment> : resistance\n";

int SRcapacitance( spot, cmdList, uow )
    RPCSpot* spot;
    lsList  cmdList;
    long  uow;
{
    octObject obj, bag;
    octGenerator gen;
    double cap = 0.0;
    char buf[256];

    errPushHandler( handler );
    if ( setjmp( pepjmp ) == 0 ) {

	RPHgetBag( cmdList, &bag );
    
	octInitGenContents( &bag, OCT_ALL_MASK, &gen );
	while ( octGenerate( &gen, &obj ) == OCT_OK ) {
	    cap = pepCapacitance( &obj );
	    sprintf( buf, "PEP: %s: cap %sF\n", ohFormatName( &obj ), PPP(cap) );
	    VEMMSG( buf );
	}
    }
    errPopHandler();
    vemPrompt();
    return RPC_OK;
}



int SRcrossCapacitance( spot, cmdList, uow )
    RPCSpot* spot;
    lsList  cmdList;
    long  uow;
{
    octObject obj, bag, n1, n2;
    octGenerator gen;
    double cap = 0.0;
    char buf[256];

    errPushHandler( handler );
    if ( setjmp( pepjmp ) == 0 ) {

	RPHgetBag( cmdList, &bag );
    
	octInitGenContents( &bag, OCT_NET_MASK|OCT_GEO_MASK|OCT_TERM_MASK,  &gen );
	if ( octGenerate( &gen, &n1 ) != OCT_OK ||
	     octGenerate( &gen, &n2 ) != OCT_OK ) {
	    VEMMSG( crossCapFormat );
	} else {
	    cap = pepCrossCapacitance( &n1, &n2  );

	    sprintf( buf, "PEP: cross cap = %sFarads\n", PPP(cap));
	    VEMMSG( buf );
	}
	octFreeGenerator( &gen );
    }
    errPopHandler();
    vemPrompt();
    return RPC_OK;
}

int SRresistance( spot, cmdList, uow )
    RPCSpot* spot;
    lsList  cmdList;
    long  uow;
{
    octObject obj, bag, t1, t2;
    octGenerator gen;
    double res = 0.0;
    char buf[256];

    errPushHandler( handler );
    if ( setjmp( pepjmp ) == 0 ) {

	RPHgetBag( cmdList, &bag );
    
	octInitGenContents( &bag, OCT_TERM_MASK | OCT_PATH_MASK, &gen );
	if ( octGenerate( &gen, &t1 ) != OCT_OK ||
	     octGenerate( &gen, &t2 ) != OCT_OK ) {
	    VEMMSG( resFormat );
	} else {
	    res = pepResistance( &t1, &t2  );

	    if ( res <  0.0 ) {
		VEMMSG( "PEP: no connection\n" );
	    } else {
		sprintf( buf, "PEP: resistance = %sohms\n", PPP(res));
		VEMMSG( buf );
	    }
	}
	octFreeGenerator( &gen );
    }
    errPopHandler();
    vemPrompt();
    return RPC_OK;
}


int SRpepInfo( spot, cmdList, uow )
    RPCSpot* spot;
    lsList  cmdList;
    long  uow;
{
    dmConfirm("PARASITICS EXTRACTION INFORMATION","\n\
The values computed for parasitics assume\n\
a 'reasonable' symbolic layout, i.e. without\n\
redundant overlaps of instances and segments.\n\
\n\
The parasitics extraction routines use\n\
the packages: pep tap th.\n\
Check the documentation for those packages\n\
to have more information on the accuracy of these\n\
estimates.\n\
\n\
","Continue", 0);
    return RPC_OK;
}



