

/*    Appendix A.

    The following comprises the source code for 'senscalc'.  This
    program can be found on janus in ~sensor/clark/senscalc1.
*/
/*
 * 'SENSCALC' written by Tu-Cuong(Clark) Nguyen and Haifang Guo
 *
 * 'senscalc' computes the sensitivities of specified performance
 * functions to capacitive interconnect parasitics at each 
 * node-to-node junction and presents the results in decreasing
 * order so as to facilitate identification of the nodes most
 * important for a particular performance function.
 * 
 * The program takes several input files depending upon how
 * many performance functions are to be investigated.  The
 * necessary files are:
 *
 * (1) One input file to be given as argument to 'senscalc'.
 *     This file should contain cards to two lines, the first
 *     indicating the type of performance function (i.e. Phase
 *     margin, the second giving the name of the associated spice 
 *     file.
 *
 * (2) Several spice input files, one for each performance function.
 *
 * (3) A 'node.in' file, which gives networking (node) information
 *     concerning the circuit under investigation.
 *
 * Examples of each input file are contained in this directory.
 *
 * The output of 'senscalc' is put in sorted form into the file
 * 'senscalc.out'
 *
 *
 */

#include <stdio.h>
#include "utility.h"
#include "options.h"
#include "vov.h"
#include "formulate.h"

optionStruct optionList[] = {
    {"n", "node_file",         "Node information file [Def.: \"node_file\"]"},
    {"o", "out_file",          "Output file with sensitivities [Def.: standard output]"},
    {"i", "inp_file",          "Performance information file [Def.: \"inp_file\"]"},
    { OPT_RARG,	"",            ""},
    { 0, 0, 0 }
};

#define NOT_NULL	1	/* just as long as it's not 0 */

main(argc,argv)
int argc;
char *argv[];
{
    int option;
    FILE *fopen();
    FILE *in_file;
    FILE *out_file;
    int num_incards, num_nodes;
    char *node_file_name;
    char *inp_file_name;


#ifdef TRIAL
    if (argc == 1)   /* no arguments */
	printf("senscalc:  need an input file\n");
    else if (argc>2)	/* too many args */
        printf("senscalc:  too many arguments\n");
    else {
	if ((ifile=fopen(*++argv, "r")) == NULL) {
	    printf("senscalc:  can't open %s\n",*argv);
	    fclose(ifile);
	}
	else {
#endif

    node_file_name = util_strsav("node_file"); 
    inp_file_name = util_strsav("inp_file"); 
    out_file = stdout;
    while ( ( option = optGetOption (argc, argv)) != EOF) {
	switch (option) {
	  case 'n':
	    node_file_name = util_strsav(optarg);
	    break;
	  case 'i':
	    inp_file_name = util_strsav(optarg);
	    break;
	  case 'o':
	    if((out_file=fopen(util_file_expand(optarg), "w")) == NULL) {
		fprintf(stderr, "Couldn't open file %s\n", optarg);
		exit(1);
	    }
	    break;
	  default:
	    optUsage();
	}
    }

    num_incards = ReadInputFile(inp_file_name);
    num_nodes = ComputeGradSENS(num_incards,node_file_name);
    WriteSENS(num_incards,out_file);
}

/*
 * 'fgetsnoret' is the same as fgets, except that it strips
 * the input line of the ending '\n'.
 *
 */
char *
fgetsnoret(s,n,iop)
char *s;
int n;
register FILE *iop;
{
    register char c;
    register char *cs;

    cs=s;
    while (--n > 0 && (c=getc(iop)) != EOF)
      if ((*cs++ = c) == '\n') {
	  cs--;
	  break;
      }
    *cs = '\0';
    return((c == EOF && cs == s) ? NULL : s);
}


/*
 *
 * ReadInputFile takes a file pointer to the input file, allocates
 * storage for the INCARD structures (defined externally) which 
 * hold the input, fills the INCARD structures, and returns the
 * number of INCARDS taken.
 *
 */
int
ReadInputFile(inp_file_name)
char *inp_file_name;
{
    INCARD *incard;
    char inline[MAX_INLINE];
    int i,j;		/* dummy variables */
    FILE *ifile;

    if( (ifile = fopen(inp_file_name,"r")) == NULL) {
        printf("\nE: could not open file %s",inp_file_name);
        exit(-1);
    }
    incard = Incards = (INCARD *) malloc(sizeof(INCARD)*MAX_INCARDS);

    j=0;
    i=1;
    while (fgetsnoret(inline,MAX_INLINE,ifile) != NULL) {
        if (*inline == '$') {  /* card separator */
            incard = Incards+i;
	    i++;
  	} else {
	    if (j == 0) {
		incard->label = (char *) malloc(sizeof(char)*MAX_INLINE);
		strcpy(incard->label,inline);
  	   	j++;
	    } else {
		incard->spiceinf = (char *) malloc(sizeof(char)*MAX_INLINE);
		strcpy(incard->spiceinf,inline);
		j=0;
	    }
	}
    }
    /*
printf("Incards->label = %s\n",Incards->label);
printf("(Incards+1)->label = %s\n",(Incards+1)->label);
printf("(Incards+2)->label = %s\n",(Incards+2)->label);
*/
    fclose(ifile);
    return(i);
}



WriteSENS(num_incards,out_file)
int num_incards;
FILE *out_file;
{
    GRAD_info *grad_info;
    FILE *fopen();
    int i,j,k;
    int num_entries;

/*
printf("inside WriteSENS()\n");
    if ((ofile=fopen("senscalc.out","w")) == NULL) {
        printf("senscalc:  can't open senscalc.out");
    } else {
printf("num_incards = %d\n",num_incards);
*/
        for (i=0; i < num_incards; i++) {
/*
printf("inside the i for loop\n");
*/
    	    grad_info = Gradients+i;
            num_entries = 0;
	        for (j=0; j <= grad_info->num_grads-1; j++) {
/*
		        if (grad_info->grad_card[j].grad == 0) continue;
*/
                num_entries++;
	        }
    	    fprintf(out_file,"number of %s sensitivities = %d \n\n",grad_info->label,num_entries);
    	    fprintf(out_file,"!Node1  Node2  Sensitivity\n");
    	    for (j=0; j <= grad_info->num_grads-1; j++) {
/*
    		    if (grad_info->grad_card[j].grad == 0) continue;
*/
    		    fprintf(out_file,"%s %s %e\n",
    			    (grad_info->grad_card)[j].node1,
    			    (grad_info->grad_card)[j].node2,
    		            (grad_info->grad_card)[j].grad);
    	    }
    	    fprintf(out_file,"\n\n");
    	}
    fclose(out_file);
}
	





