/*
 * interpretSim.c
 *
 * The parser used here comes from build.c of the Crystal simulator.
 * Copyright (C) 1982 John K. Ousterhout
 * sccsid "@(#)build.c	2.15 (Berkeley) 10/14/88"
 *
 * The semantic actions associated with the parse, however, are all new.
 *
 * After some setup of the facet being created, the .sim file is parsed.
 * As statements are read in, they are interpreted and stored in
 * an oct form.
 */

#include "sim2facet.h"

extern s_options Options;

/* The following variables are used to keep from printing too many
 * messages about negative fet lengths and widths.
 */

int negCount = 0;
int negLimit = 50;

#define LINELENGTH 500
#define NAMELENGTH 100

/*
 * simToOct
 *
 * convert a sim file to an equivalent Oct representation.
 */
int
simToOct()
{
    int result;

    if(( result = parseSim( Options.simFileName)) != TRUE){
	fprintf(stderr,"sim2oct: Warning: there were errors while parsing %s\n",
		Options.simFileName);
	fprintf(stderr,"Check error messages and results carefully\n");
    }
    return ( result );
}

/*
 * parseSim
 *
 * This routine parses the .sim file.  As its going, oct routines are
 * called to store the data found here.
 */
int
parseSim( name )
char *name;
{
    FILE *f;
    s_simFet fet;
    s_simCapacitor capacitor;
    s_simNodeCap nodeCap;
    s_simNodeRes nodeRes;
    char line[LINELENGTH], *node1, *node2, *node3, *a1, *a2, *a3;
    char *ptr, *tmp, *NextField();
    char *errMsg, *parseFETAtt();
    float length, width;
    float xloc, yloc;
    float units, cap;
    float ma, mp, da, dp, pa, pp;
    int lineNumber, result;

    result = TRUE;
    negCount = 0;
    units = 1.0;
    f = fopen(name, "r");
    if (f == NIL(FILE))
    {
	errRaise( optProgName, -1, "Couldn't open .sim file `%s'.\n", name);
    }

    /* Process the file one line at a time.  The first character
     * of each line determines the statement type of the line.
     */

    lineNumber = 0;
    while (fgets(line, LINELENGTH, f) != NULL)
    {
	lineNumber += 1;
	switch (line[0])
	{
	    /* Transistor: e/d gate drain source length width x y.
	     * There may be up to three additional attribute strings at the
	     * end of the line.  E means nMOS enhancement, d means
	     * nMOS depletion.
	     */

	    case 'e':
	    case 'd':
	    case 'n':
	    case 'p':

		fet.sourceDir = fet.drainDir = DIR_UNDEF;
		fet.modelName = NIL(char);

		if (line[0] == 'd') fet.type = NDEP;
		else if (line[0] == 'e') fet.type = NENH;
		else if (line[0] == 'n') fet.type = NCHAN;
		else if (line[0] == 'p') fet.type = PCHAN;

		/* This routine parses the line itself rather than
		 * calling sscanf, because it saves huge amounts of time.
		 */

		ptr = &line[1];
		if ((node1 = NextField(&ptr)) == NULL)
		{
		    err1: printf( "Not enough info for");
		    printf(" transistor on line %d.\n", lineNumber);
		    result = FALSE;
		    continue;
		}
		if ((node2 = NextField(&ptr)) == NULL) goto err1;
		if ((node3 = NextField(&ptr)) == NULL) goto err1;
		if ((tmp = NextField(&ptr)) == NULL) goto err1;
		length = atof(tmp);
		if ((tmp = NextField(&ptr)) == NULL) goto err1;
		width = atof(tmp);
		if ((tmp = NextField(&ptr)) == NULL) goto err1;
		xloc = atof(tmp);
		if ((tmp = NextField(&ptr)) == NULL) goto err1;
		yloc = atof(tmp);
		a1 = NextField(&ptr);
		a2 = NextField(&ptr);
		a3 = NextField(&ptr);

		fet.gate = node1;
		fet.source = node2;
		fet.drain = node3;

		/* Renormalize distances to micron units. */

		if (length < 0)
		{
		    if (negCount <= negLimit)
		    {
		        printf("Warning: transistor has negative ");
		        printf("length on line %d.\n", lineNumber);
			if (negCount == negLimit)
			{
			    printf("    No more messages of this kind ");
			    printf("will be printed...\n");
			}
			negCount += 1;
		    }
		    length = - length;
		}
		if (width < 0)
		{
		    if (negCount <= negLimit)
		    {
		        printf("Warning: transistor has negative ");
		        printf("width on line %d.\n", lineNumber);
			if (negCount == negLimit)
			{
			    printf("    No more messages of this kind ");
			    printf("will be printed...\n");
			}
			negCount += 1;
		    }
		    width = - width;
		}
		fet.length = length * units;
		fet.width = width * units;
		fet.xloc = xloc * units;
		fet.yloc = yloc * units;


		/* Handle attributes, if any.  It's important to do
		 * this here, because the attributes may change the
		 * type of the transistor.
		 */

		if (a1 != NULL)
		{
		    errMsg = parseFETAtt(&fet, a1);
		    if (errMsg != NULL)
		    {
			printf("Error in 1st attribute on line %d: %s.\n",
			    lineNumber, errMsg);
			result = FALSE;
		    }
		}
		if (a2 != NULL)
		{
		    errMsg = parseFETAtt(&fet, a2);
		    if (errMsg != NULL)
		    {
			printf("Error in 2nd attribute on line %d: %s.\n",
			    lineNumber, errMsg);
			result = FALSE;
		    }
		}
		if (a3 != NULL)
		{
		    errMsg = parseFETAtt(&fet, a3);
		    if (errMsg != NULL)
		    {
			printf("Error in 3rd attribute on line %d: %s.\n",
			    lineNumber, errMsg);
			result = FALSE;
		    }
		}

		result = octPutFet( &fet );

		break;

	    /* Node area information:  compute capacitance and resistance. */

	    case 'N':
		/* Once again, parse here rather than calling sscanf.
		 * This saves time.
		 */

		ptr = &line[1];
		if ((node1 = NextField(&ptr)) == NULL)
		{
		    err2: printf("Not enough node info");
		    printf(" on line %d.\n", lineNumber);
		    result = FALSE;
		    continue;
		}
		if ((tmp = NextField(&ptr)) == NULL) goto err2;
		da = atof(tmp);
		if ((tmp = NextField(&ptr)) == NULL) goto err2;
		dp = atof(tmp);
		if ((tmp = NextField(&ptr)) == NULL) goto err2;
		pa = atof(tmp);
		if ((tmp = NextField(&ptr)) == NULL) goto err2;
		pp = atof(tmp);
		if ((tmp = NextField(&ptr)) == NULL) goto err2;
		ma = atof(tmp);
		if ((tmp = NextField(&ptr)) == NULL) goto err2;
		mp = atof(tmp);

		nodeCap.node = node1;
		nodeCap.da = da * units * units;
		nodeCap.dp = dp * units;
		nodeCap.pa = pa * units * units;
		nodeCap.pp = pp * units;
		nodeCap.ma = ma * units * units;
		nodeCap.mp = mp * units;

		result = octPutNodeCap( &nodeCap );

		break;

	    /* Capacitance information:  "C node1 node2 femtofarads" */

	    case 'C':
		/* As usual, parse locally to save time. */

		ptr = &line[1];
		if ((node1 = NextField(&ptr)) == NULL)
		{
		    err3: printf("Not enough capacitance info on line %d.\n",
			lineNumber);
		    result = FALSE;
		    continue;
		}
		if ((node2 = NextField(&ptr)) == NULL) goto err3;
		if ((tmp = NextField(&ptr)) == NULL) goto err3;
		cap = atof(tmp);
		capacitor.node1 = node1;
		capacitor.node2 = node2;
		capacitor.cap = cap / 1000;

		result = octPutCapacitor( &capacitor );

		break;
		
	    /* Explicit resistance information from the extractor:
	     * "R node ohms".  The amount is added to whatever was
	     * already there for the node.
	     */

	    case 'R':
		/* As usual, parse locally to save time. */

		ptr = &line[1];
		if ((node1 = NextField(&ptr)) == NULL)
		{
		    err4: printf("Not enough resistance info on line %d.\n",
			lineNumber);
		    result = FALSE;
		    continue;
		}
		if ((tmp = NextField(&ptr)) == NULL) goto err4;
		nodeRes.node = node1;
		nodeRes.res = atof(tmp);

		result = octPutNodeRes( &nodeRes);

		break;
		
	    /* Header information with technology and other junk.  Get
	     * unit scale factor. */

	    case '|':
		if (sscanf(line, "| units: %f", &units) != 1)
		{
		    printf("Funny looking header line in .sim file.\n");
		    result = FALSE;
		    units = 1.0;
		}
		units = units/100.0;
	    	break;
	
	    /* Node attribute information: A node attribute */

	    case 'A':
		/* not supported in sim2oct				*/
		if (sscanf(&line[1], " %s %s", node1, a1) != 2)
		{
		    printf("Not enough info for attribute on line %d.\n",
			lineNumber);
		    result = FALSE;
		    continue;
		}
		fprintf(stderr,"sim2oct: line %d: Attribute statement not supported\n",
				lineNumber);
		break;

	    /* An unknown line.  Ignore it. */

	    default: break;
	}
    }

    fclose (f);
    return result;
}

/*
 * parseFETAtt
 *
 * AttParseFET parses an attribute string specified for
 * a transistor.  The attribute string must be of the
 * form "x=a,b,c,...", where x is "s", "d", or "g" to
 * indicate that the attribute pertains to the transistor
 * source, drain or gate, and each of the comma-separated
 * strings a, b, c, ... is of the form "Crystal:goo" or
 * "Cr:goo", where goo is the attribute information we store.
 *
 */
char *
parseFETAtt(fet, string)
s_simFet *fet;
char *string;
{
    char *next;
    fetTermType term; /* What the attribute refers to (s, d, g). */

    /* Make sure the attribute has the "x=" stuff at the front. */

    switch (*string)
    {
	case 's': term = SOURCE;  break;
	case 'd': term = DRAIN; break;
	case 'g': term = GATE; break;
	default: return "bad terminal";
    }
    string += 1;
    if (*string != '=') return "bad format";
    string += 1;

    /* Now skim off the comma-separated attributes one at a time. */

    if (*string == 0) next = 0;
    else next = string;
    while (next != 0)
    {
	string = next;

	/* Find the comma that ends the attribute, and replace it with
	 * a null.  Then save the location of the next attribute.
	 */
	
	next = strchr(string, (int)',');
	if (next != 0) *next++ = 0;

	/* Make sure that the first characters of the attribute are
	 * the letters "Crystal:" or "Cr:".  Then skip them.
	 */
	
	if (strncmp(string, "Cr:", 3) == 0) string += 3;
	else if (strncmp(string, "Crystal:", 8) == 0) string += 8;
	else continue;
	if (*string == 0) continue;

	/* If this attribute is attached to the transistor gate,
	 * it is the name of the transistor's type.  Look up the
	 * name and set the type.
	 */
	
	if (term == GATE)
	{
	    fet->modelName = string;
	    continue;
	}

	/* If the attribute is "In" or "Out", just set a flag in
	 * the transistor.  Otherwise, make a flow record for the
	 * name and add it to the list for the transistor.
	 */
	
	if (strcmp(string, "In") == 0)
	{
	    if (term == SOURCE) fet->sourceDir = DIR_IN;
	    else fet->drainDir = DIR_IN;
	}
	else if (strcmp(string, "Out") == 0)
	{
	    if (term == SOURCE) fet->sourceDir = DIR_OUT;
	    else fet->drainDir = DIR_OUT;
	}
	else
	{
	    /* named flow, currently unsupported			*/
	    fprintf(stderr,"sim2Oct: named flow encountered. Not currently supported\n");
	}
    }
    return NULL;
}

/*
 * NextField
 *
 * borrowed from nextfield.c of Crystal
 *
 * Copyright (C) 1982 John K. Ousterhout
 * sccsid "@(#)nextfield.c	2.1 (Berkeley) 4/9/83"
 *
 * This procedure picks arguments off of strings.
 * The string pointed to indirectly by p is assumed
 * to consist of zero or more fields separated by
 * whitespace.
 */
char *
NextField(p)
char **p;			/* A pointer to a string pointer.*/
{
    register char *p1;
    char *p2;

    p1 = *p;

    /* Strip off leading white space. */

    while (isspace(*p1)) p1 += 1;
    p2 = p1;

    while ((*p1 != 0) && !isspace(*p1)) p1 += 1;
    if (*p1 != 0) *p1++ = 0;
    *p = p1;
    if (*p2 == 0) return NULL;
    else return p2;
}
