/*
 * Symbol Generator Driver
 *
 * David Harrison
 * University of California, Berkeley
 * 1989
 *
 * This is a front end for the symbol generator contained in symgen.c.
 * It allows users to generate new schematic symbols from a textual
 * description or from an existing contents facet.
 */

#include "copyright.h"
#include "port.h"
#include "options.h"
#include "utility.h"
#include "oct.h"
#include "tr.h"
#include "oh.h"
#include "tap.h"
#include "vl.h"
#include "errtrap.h"
#include "da.h"
#include "avl.h"

/* Defaults for various parameters */
#define DEF_LAMBDA	20
#define DEF_STEM	1
#define DEF_SPACE	4

extern char *version;		/* Exported by version.c */

/* Exported by symgen.c */
extern octStatus sg_box();
extern octStatus sg_box_size();
extern int sg_count();

/* Forward declarations */
static int do_contents();
static void init_vals();
static void gen_symbol();
static void description();
static struct octBox *term_size();
static void prep();
static void do_desc();
static void parse_terms();
static char *get_name();
static void count_init();
static int get_count();
static void final_opts();
static char **extract_side();
static char *side_str();

#define MS 	1024
#define LAMBDA	20
#define MINSIZE	3

#ifndef MAX
#define MAX(a,b)	((a) > (b) ? (a) : (b))
#endif

#define Fprintf	(void) fprintf
#define Printf 	(void) printf

#define MA(type, num) \
(type *) malloc((unsigned) (sizeof(type) * num))
#define STC(str) \
strcpy(malloc((unsigned) (strlen(str)+1)), str)

#define SC(call, num) \
if ((call) != num) { (void) printf("Syntax error\n");  exit(1); }

optionStruct optionList[] = {
    { "c",	"",	"Force symbol to be same size as contents facet"  },
    { "s",	"",	"Silent mode (no warnings printed)"		  },
    { "k",	"",	"Write over existing contents facet"		  },
    { "f",	"description_file", "File containing textual description" },
    { "n",	"symbol_name",	"Name to be placed in symbol"		  },
    { "l",	"term_list",	"Left side terminal names"		  },
    { "r",	"term_list",	"Right side terminal names"		  },
    { "t",	"term_list",	"Top side terminal names"		  },
    { "b",	"term_list",	"Bottom side terminal names"		  },
    { "L",	"lambda",	"Oct units per lambda"			  },
    { "W",	"width",	"Symbol width (in lambda)"		  },
    { "H",	"height",	"Symbol height (in lambda)"		  },
    { "S",	"stem_length",	"Stem length (in lambda)"		  },
    { "I",	"stem_interval","Distance between stems (in lambda)"	  },
    { OPT_RARG,	"cell[:view[:facet]]", "Input/output cell"		  },
    { OPT_DESC, 0,
	"Generates a schematic symbol.  If the specified cell has no " },
    { OPT_CONT, 0,
	"contents facet, the symbol is generated from the textual " },
    { OPT_CONT, 0,
	"description specified using the -f option and is written to " },
    { OPT_CONT, 0,
	"both the contents and interface facets of the cell.  If the " },
    { OPT_CONT, 0,
	"specified cell has a contents facet, it is used as the template " },
    { OPT_CONT, 0,
	"for generating the symbol (i.e., the terminal locations will be " },
    { OPT_CONT, 0,
	"determined by examining the terminal locations in the contents " },
    { OPT_CONT, 0,
	"facet).  However, a textual description is still accepted and " },
    { OPT_CONT, 0,
	"will override the terminal locations in the contents facet.  In " },
    { OPT_CONT, 0,
	"fact,  a textual description is required for any terminals that " },
    { OPT_CONT, 0,
	"have no implementation in the contents facet.  The format for " },
    { OPT_CONT, 0,
	"of the textual description is given in the manual page.  All of " },
    { OPT_CONT, 0,
	"the information specified in the textual description may be " },
    { OPT_CONT, 0,
	"overridden by the -l, -r, -t, -b, -L, -W, and -H command line options." },
    { 0, 0, 0 }
};
/*	This struct declaration was made into a shared header file so that
	symgen gen can get at it.

	Marco Zelada
	Intel Corp.
*/
#include "symgen.h"


main(argc, argv)
int argc;
char *argv[];
/*
 * Parses argument list only.
 */
{
    int option;
    opt_vals vals;

    init_vals(&vals);
    while ((option = optGetOption(argc, argv)) != EOF) {
	switch (option) {
	case 'c':
	    vals.contents_size = 1;
	    break;
	case 's':
	    vals.silent = 1;
	    break;
	case 'k':
	    vals.kill_contents = 1;
	    break;
	case 'f':
	    vals.desc_file = optarg;
	    break;
	case 'n':
	    vals.name_s = 1;
	    vals.name = optarg;
	    break;
	case 'l':
	    vals.left_s = 1;
	    vals.left = optarg;
	    break;
	case 'r':
	    vals.right_s = 1;
	    vals.right = optarg;
	    break;
	case 't':
	    vals.top_s = 1;
	    vals.top = optarg;
	    break;
	case 'b':
	    vals.bottom_s = 1;
	    vals.bottom = optarg;
	    break;
	case 'L':
	    vals.lambda_s = 1;
	    vals.lambda = atoi(optarg);
	    break;
	case 'W':
	    vals.width_s = 1;
	    vals.width = atoi(optarg);
	    break;
	case 'H':
            vals.height_s = 1;
            vals.height= atoi(optarg);
            break;
	case 'S':
	    vals.stem_s = 1;
	    vals.stem = atoi(optarg);
	    break;
	case 'I':
	    vals.space_s = 1;
	    vals.space = atoi(optarg);
	    break;
	default:
	    optUsage();
	    /*NOTREACHED*/
	}
    }

    if (optind != argc-1) {
	Fprintf(stderr, "%s: one cell must be specified\n", optProgName);
	optUsage();
	/*NOTREACHED*/
    }
    vals.cell_spec = argv[optind];

    if (!vals.silent) {
	Printf("%s (compiled on %s)\n", optProgName, version);
    }
    gen_symbol(&vals);
    return 0;
}

static void init_vals(vals)
opt_vals *vals;
/*
 * Intializes all values in the option value structure to good initial values.
 */
{
    vals->contents_size = 0;
    vals->silent = 0;
    vals->left_s = 0;
    vals->right_s = 0;
    vals->top_s = 0;
    vals->bottom_s = 0;
    vals->name_s = 0;
    vals->lambda_s = 0;
    vals->width_s = 0;
    vals->width_oct = 0;
    vals->height_s = 0;
    vals->height_oct = 0;
    vals->stem_s = 0;
    vals->stem_oct = 0;
    vals->space_s = 0;
    vals->space_oct = 0;
    vals->name = (char *) 0;
    vals->desc_file = (char *) 0;
    vals->cell_spec = (char *) 0;
}

/* Symbol table for terminals */
typedef enum term_side_defn { NONE, LEFT, RIGHT, TOP, BOTTOM } term_side;
typedef struct term_info_defn {
    term_side side;		/* What side for terminal    */
    unsigned int from_contents;	/* Non-zero if from contents */
    int order;			/* Relative order on side    */
    char *term_name;		/* Name (stored redundantly) */
} term_info;
#define NONE_I		((int) NONE)
#define LEFT_I		((int) LEFT)
#define RIGHT_I		((int) RIGHT)
#define BOTTOM_I	((int) BOTTOM)
#define LAST_I		BOTTOM_I+1

static term_side what_side();

static void gen_symbol(opts)
opt_vals *opts;			/* Options and specifications */
/*
 * This routine attempts to open the contents facet of `cell_spec'
 * for reading.  If it exists, it is opened and examined to build
 * an original list of terminals for each side.  If it does not
 * exist,  it is opened for write and a null description is built.
 * Next,  the description file is read in (if any) and augments
 * the terminal description.  Finally,  the box size is determined
 * using the largest number of terminals horizontally and vertically
 * (or if the c_flag option is on,  from the size of the contents facet).
 * The symbol is then created using sg_box.
 */
{
    octObject fct, new;
    avl_tree *tree;
    char **left, **right, **top, **bottom;
    octObject sym, wire;
    struct octBox box;
    int con_desc;

    tree = avl_init_table(strcmp);
/*
	The last parameter &box was added to allow the do_contents routine to
	keep the bonding box encoutered in the contents facet.

	Marco Zelada
	Intel Corp.
*/
    con_desc = do_contents(&fct, opts, tree, &sym, &wire, &box);
    do_desc(opts, con_desc, tree);
    left = extract_side(tree, LEFT);
    right = extract_side(tree, RIGHT);
    top = extract_side(tree, TOP);
    bottom = extract_side(tree, BOTTOM);
    final_opts(opts, &fct, left, right, top, bottom);

/* The opts->stem is used to adjust the size of the symbol box. Since we do not
   want this to be done when -c is on, we set it to 0.

	Marco Zelada
	Intel Corp.
*/

    if ( opts->contents_size )
	opts->stem = 0;
    else
	{
    	box.lowerLeft.x = -(opts->width/2/opts->space*opts->space);
    	box.lowerLeft.y = -(opts->height/2/opts->space*opts->space);
    	box.upperRight.x = box.lowerLeft.x + opts->width;
    	box.upperRight.y = box.lowerLeft.y + opts->height;
	};
/*
	The last parameter opts was added to allow the sg_box routine 
	access to the -c option switch.

	Marco Zelada
	Intel Corp.
*/
    (void) sg_box(&sym, &wire, &box, opts->name, opts->stem, opts->space,
		  left, right, top, bottom, opts);
    OH_ASSERT(octFlushFacet(&fct));
    if (!con_desc) {
	new = fct;
	new.contents.facet.facet = "interface";
	OH_ASSERT(octCopyFacet(&new, &fct));
    }
}

static int do_contents(fct, opts, term_table, symbol, wiring, box)
octObject *fct;			/* Facet to open for writing interface */
opt_vals *opts;			/* Options and flags                   */
avl_tree *term_table;		/* Table of all terminals              */
octObject *symbol, *wiring;	/* Two layers for creating symbol      */
struct octBox *box; 		/* The contents facet bounding Box     */
				/*				       */
				/*	Marco Zelada		       */
				/*	Intel Corp. Zelada		       */
/*
 * This routine attempts to open the contents facet of the cell specfied
 * by `spec' for reading.  If successful, the interface is opened for
 * write and the contents is examined for plausible terminal locations
 * based on the locations of formal terminals in the master.  If a terminal
 * location cannot be determined,  the terminal name is placed in `unclaimed'.
 * If the contents facet does not exist,  it will be opened for write
 * and a null description will be passed back.  Returns a non-zero
 * value if a description was derived from the contents facet.
 */
{
    int ret_val = 0;
	octObject inputFacet;       /* facet to be framed */
	octObject outputFacet;       /* facet to be framed */

/*  Generate the formal terminals via vulcan if the -c option is on. This will
    ensure that the terminals are in the same place in both the contents and
    interface. It would be nicer if we used a special version of vulcan that
    had a flag to turn off the frame generation within vulcan, but this will
    do the job for now.

	Marco Zelada
	Intel Corp.
*/
if ( opts->contents_size )
{
/* Open input facet */

    ohUnpackDefaults(&inputFacet, "r", ":schematic:contents");
    if (ohUnpackFacetName(&inputFacet, opts->cell_spec) != OCT_OK) {
	Fprintf(stderr, "%s: Cannot read contents `%s'.\n",
		optProgName, opts->cell_spec);
	exit(1);
	/*NOTREACHED*/
    }
    inputFacet.contents.facet.mode = "r";
    if (octOpenFacet(&inputFacet) < OCT_OK) {
     (void) fprintf(stderr, "%s: can't open cell ``", optProgName);
     (void) fprintf(stderr, "%s:%s:%s:%s",
                            inputFacet.contents.facet.cell,
                            inputFacet.contents.facet.view,
                            inputFacet.contents.facet.facet,
			    inputFacet.contents.facet.version);
    }
                
    ohUnpackDefaults(&outputFacet, "w", ":schematic:interface");
    if (ohUnpackFacetName(&outputFacet, opts->cell_spec) != OCT_OK) {
	Fprintf(stderr, "%s: Cannot write contents `%s'.\n",
		optProgName, opts->cell_spec);
	exit(1);
	/*NOTREACHED*/
    }
    outputFacet.contents.facet.mode = "w";
    if (octOpenFacet(&outputFacet) < OCT_OK) {
     (void) fprintf(stderr, "%s: can't open cell ``", optProgName);
     (void) fprintf(stderr, "%s:%s:%s:%s",
                            outputFacet.contents.facet.cell,
                            outputFacet.contents.facet.view,
                            outputFacet.contents.facet.facet,
			    outputFacet.contents.facet.version);
    }

/*	Generate the I/O terminals in their proper place */

	vulcan(&inputFacet, &outputFacet );
	octCloseFacet(&inputFacet);
	octCloseFacet(&outputFacet);
};

    ohUnpackDefaults(fct, "r", ":schematic:contents");
    if (ohUnpackFacetName(fct, opts->cell_spec) != OCT_OK) {
	Fprintf(stderr, "%s: Cannot parse cell specification `%s'.\n",
		optProgName, opts->cell_spec);
	exit(1);
	/*NOTREACHED*/
    }
    fct->contents.facet.facet = "contents";
    switch (octOpenFacet(fct)) {
    case OCT_OLD_FACET:
	ret_val = 1;
	break;
    case OCT_NO_EXIST:
	ret_val = 0;
	break;
    default:
	Fprintf(stderr, "%s FATAL: Cannot open `%s':\n  %s\n",
		optProgName, opts->cell_spec, octErrorString());
	
	exit(1);
	/*NOTREACHED*/
    }
    if (ret_val && !(opts->kill_contents)) {
	/* Contents exists -- generate terminal description from it */
	description(fct, opts, term_table);

/* Keep info about its bounding box so that the symbol will be acurate.

   Marco Zelada
   Intel Corp.
*/
	octBB(fct, box );

	/* Open interface for append */
	fct->contents.facet.facet = "interface";
	fct->contents.facet.mode = "a";
	if (octOpenFacet(fct) < OCT_OK) {
	    Fprintf(stderr,
		    "%s FATAL: Cannot open interface of cell `%s' for append.\n",
		    optProgName, opts->cell_spec);
	    exit(1);
	    /*NOTREACHED*/
	}
    } else {
	/* Contents does not exist or should be killed */
	if (ret_val && opts->kill_contents) {
	    (void) octCloseFacet(fct);
	    ret_val = 0;
	}
	fct->contents.facet.mode = "w";
	if (octOpenFacet(fct) < OCT_OK) {
	    Fprintf(stderr,
		    "%s FATAL: Cannot open contents of cell `%s' for writing:\n  %s\n",
		    optProgName, opts->cell_spec, octErrorString());
	    exit(1);
	    /*NOTREACHED*/
	}
    }
    prep(fct, symbol, wiring);
    return ret_val;
}    

static void description(fct, opts, term_table)
octObject *fct;			/* Facet to examine     */
opt_vals *opts;			/* Options and flags    */
avl_tree *term_table;		/* All terminals        */
/*
 * This routine examines the formal terminals of `fct' and assigns
 * them to the right, left, top, or bottom based on the distance
 * of their implementation from the respective boundaries of the cell.
 * The terminals are all added to `term_table'.
 * These terminals are then sorted using the appropriate dimension
 * and returned in `right', `left', `top', or `bottom' respectively.
 * Those terminals without a useful implementation are returned
 * in `unimpl'.  The implementation is intended to be correct not
 * necessarily quick.
 */
{
    octGenerator gen;
    octObject term;
    struct octBox *size, bb;
    term_info *info;

    if ((octBB(fct, &bb) >= OCT_OK) && (opts->contents_size)) {
	opts->width = bb.upperRight.x - bb.lowerLeft.x;
	opts->width_s = 1;
	opts->width_oct = 1;
	opts->height = bb.upperRight.y - bb.lowerLeft.y;
	opts->height_s = 1;
	opts->height_oct = 1;
    }
    OH_ASSERT(octInitGenContents(fct, OCT_TERM_MASK, &gen));
    while (octGenerate(&gen, &term) == OCT_OK) {
	if (avl_lookup(term_table, term.contents.term.name, (char **) &info)) {
	    /* Its already there -- this is fatal */
	    Fprintf(stderr, "%s FATAL: More than one terminal with the same name\n",
		    optProgName);
	    Fprintf(stderr, "%s FATAL (cont): is not allowed (terminal `%s').\n",
		    optProgName, term.contents.term.name);
	    exit(1);
	    /*NOTREACHED*/
	}
	info = ALLOC(term_info, 1);
	if (size = term_size(&term)) {
	    info->side = what_side(&bb, size);
	    info->from_contents = 1;
	    switch (info->side) {
	    case LEFT:
	    case RIGHT:
		info->order = size->lowerLeft.y;
		break;
	    case TOP:
	    case BOTTOM:
		info->order = size->lowerLeft.x;
		break;
	    default:
		errRaise(optProgName, 1, "Bad side returned from what_side()");
		/*NOTREACHED*/
	    }
	    info->term_name = term.contents.term.name;
	} else {
	    /* It can't be placed */
	    info->side = NONE;
	    info->from_contents = 1;
	    info->order = 0;
	    info->term_name = term.contents.term.name;
	}
	(void) avl_insert(term_table, info->term_name, (char *) info);
    }
}

#define MAX_ASN(val, comp) \
if ((comp) > (val)) (val) = (comp)

#define MIN_ASN(val, comp) \
if ((comp) < (val)) (val) = (comp)

struct octBox *merge_box(box_ptr, box_rec, new_box)
struct octBox *box_ptr;		/* Current box (if any)              */
struct octBox *box_rec;		/* Place to put it (if box_ptr is 0) */
struct octBox *new_box;		/* Box to merge                      */
/*
 * Merges `new_box' into `box_ptr'.  If `box_ptr' is zero,  it is
 * made to point at `box_rec' and then the box is merged.  The
 * routine returns `box_ptr'.
 */
{
    if (box_ptr == (struct octBox *) 0) {
	box_ptr = box_rec;
	box_ptr->lowerLeft.x = INT_MAX;   box_ptr->lowerLeft.y = INT_MAX;
	box_ptr->upperRight.x = -INT_MAX; box_ptr->upperRight.y = -INT_MAX;
    }
    MIN_ASN(box_ptr->lowerLeft.x, new_box->lowerLeft.x);
    MIN_ASN(box_ptr->lowerLeft.y, new_box->lowerLeft.y);
    MAX_ASN(box_ptr->upperRight.x, new_box->upperRight.x);
    MAX_ASN(box_ptr->upperRight.y, new_box->upperRight.y);
    return box_ptr;
}

static struct octBox *term_size(term)
octObject *term;		/* Terminal to examine */
/* 
 * This routine uses some policy to determine the size of a terminal.
 * First,  it examines the objects connected to the terminal.  If
 * the object has a bounding box,  it is merged into the final result.
 * If the object is an actual terminal, the corresponding formal
 * in the interface of the master is examined recursively and
 * transformed appropriately.  The routine returns zero if
 * no suitable implementation is found.
 */
{
    static struct octBox final;
    struct octBox *result, result_data, *sub_size, bb;
    octGenerator gen;
    octObject obj, inst, master, sub_term;
    octCoord temp;

    result = (struct octBox *) 0;
    OH_ASSERT(octInitGenContents(term, OCT_GEO_MASK|OCT_TERM_MASK, &gen));
    while (octGenerate(&gen, &obj) == OCT_OK) {
	if (obj.type == OCT_TERM) {
	    /* Recursive portion */
	    inst.objectId = obj.contents.term.instanceId;
	    OH_ASSERT(octGetById(&inst));
	    master.type = OCT_FACET;
	    master.contents.facet.facet = "interface";
	    master.contents.facet.mode = "r";
	    if (octOpenMaster(&inst, &master) >= OCT_OK) {
		OH_ASSERT(octGetByExternalId(&master,
					     obj.contents.term.formalExternalId,
					     &sub_term));
		/* Recursively find size */
		sub_size = term_size(&sub_term);
		/* Map back to our coordinates */
		tr_oct_transform(&(inst.contents.instance.transform),
				 &(sub_size->lowerLeft.x),
				 &(sub_size->lowerLeft.y));
		tr_oct_transform(&(inst.contents.instance.transform),
				 &(sub_size->upperRight.x),
				 &(sub_size->upperRight.y));
		/* Normalize box */
		if (sub_size->lowerLeft.x > sub_size->upperRight.x) {
		    temp = sub_size->lowerLeft.x;
		    sub_size->lowerLeft.x = sub_size->upperRight.x;
		    sub_size->upperRight.x = temp;
		}
		if (sub_size->lowerLeft.y > sub_size->upperRight.y) {
		    temp = sub_size->lowerLeft.y;
		    sub_size->lowerLeft.y = sub_size->upperRight.y;
		    sub_size->upperRight.y = temp;
		}
		result = merge_box(result, &result_data, sub_size);
	    }
	} else {
	    if (octBB(&obj, &bb) >= OCT_OK) {
		result = merge_box(result, &result_data, &bb);
	    }
	}
    }
    if (result) {
	final = *result;
	return &final;
    } else {
	return (struct octBox *) 0;
    }
}

static term_side what_side(bounds, term)
struct octBox *bounds;		/* Overall size of contents facet */
struct octBox *term;		/* Size of terminal               */
/*
 * This routine determines which edge of `bounds' is closest to
 * the box `term'.
 */
{
    octCoord left, right, top, bottom;
    octCoord smallest;

    left = ABS(bounds->lowerLeft.x - term->lowerLeft.x);
    right = ABS(bounds->upperRight.x - term->upperRight.x);
    top = ABS(bounds->upperRight.y - term->upperRight.y);
    bottom = ABS(bounds->lowerLeft.y - term->lowerLeft.y);
    smallest = INT_MAX;
    MIN_ASN(smallest, left);
    MIN_ASN(smallest, right);
    MIN_ASN(smallest, top);
    MIN_ASN(smallest, bottom);
    if (smallest == left) return LEFT;
    else if (smallest == right) return RIGHT;
    else if (smallest == top) return TOP;
    else if (smallest == bottom) return BOTTOM;
    else return NONE;
}

static void prep(fct, symbol, wiring)
octObject *fct;			/* Facet to set properties in */
octObject *symbol;		/* Newly created symbol layer */
octObject *wiring;		/* Newly created wiring layer */
/*
 * This routine sets the values for the standard properties
 * for the symbol facet and creates two layers: SYMBOL and WIRING
 * which are returned.
 */
{
    octObject prop;

    OH_ASSERT(ohCreatePropStr(fct, &prop, "TECHNOLOGY", "scmos"));
    OH_ASSERT(ohCreatePropStr(fct, &prop, "VIEWTYPE", "SCHEMATIC"));
    OH_ASSERT(ohCreatePropStr(fct, &prop, "EDITSTYLE", "PHYSICAL"));
    OH_ASSERT(ohCreatePropStr(fct, &prop, "CELLTYPE", "MODULE"));
    OH_ASSERT(ohCreatePropStr(fct, &prop, "CELLCLASS", "MODULE"));
    OH_ASSERT(ohCreateLayer(fct, symbol, "SYMBOL"));
    OH_ASSERT(ohCreateLayer(fct, wiring, "WIRING"));
}

static int stricmp(a, b)
register char *a, *b;
/*
 * This routine compares two strings disregarding case.
 */
{
    register int value;

    if ((a == (char *) 0) || (b == (char *) 0)) {
	return a - b;
    }

    for ( /* nothing */;
	 ((*a | *b) &&
	  !(value = ((isupper(*a) ? *a - 'A' + 'a' : *a) -
		     (isupper(*b) ? *b - 'A' + 'a' : *b))));
	 a++, b++)
      /* Empty Body */;

    return value;
}

#define FIND_S(name, flag, value) \
if ((val = vl_query(db, name)) && !(flag)) { \
   (flag) = 1;  (value) = val; \
}

#define FIND_I(name, flag, value) \
if ((val = vl_query(db, name)) && !(flag)) { \
   (flag) = 1;  (value) = atoi(val); \
}

static void do_desc(opts, p_flag, term_table)
opt_vals *opts;			/* Options and flags        */
int p_flag;			/* Previous specification   */
avl_tree *term_table;		/* Terminal table           */
/*
 * This routine reads the description from `desc_file' in options
 * structure (if any) using the readvals
 * package.  This package reads a file of name/value pairs.  The
 * names recognized here are "name", "width", "height", "left",
 * "right", "top", "bottom", "lambda", and "stem".  The values of 
 * `name', `width', `height', `lambda', and `stem' are directly 
 * translated into the appropriate values.
 * The values for terminal directions are parsed into a list of
 * terminal names and added to the terminal symbol table.
 */
{
    val_db db;
    char *val;
    FILE *df;

    if (opts->desc_file) {
	if (!(df = fopen(opts->desc_file, "r"))) {
	    if (strcmp(opts->desc_file, "-") == 0) {
		df = stdin;
	    } else {
		Fprintf(stderr, "%s FATAL: Cannot open description file `%s'\n",
			optProgName, opts->desc_file);
		exit(1);
	    }
	}
	db = vl_new(df, stricmp);
	FIND_S("name", opts->name_s, opts->name);
	FIND_I("width", opts->width_s, opts->width);
	FIND_I("height", opts->height_s, opts->height);
	FIND_I("lambda", opts->lambda_s, opts->lambda);
	FIND_I("stem", opts->stem_s, opts->stem);
	FIND_I("space", opts->space_s, opts->space);
    } else {
	db = (val_db) 0;
    }
    /* Initializes count for relative positions */
    count_init(term_table);
    if (db && (val = vl_query(db, "left")) && !(opts->left_s)) {
	parse_terms(val, opts, term_table, LEFT, p_flag);
    } else if (opts->left_s) {
	parse_terms(opts->left, opts, term_table, LEFT, p_flag);
    }
    if (db && (val = vl_query(db, "right")) && !(opts->right_s)) {
	parse_terms(val, opts, term_table, RIGHT, p_flag);
    } else if (opts->right_s) {
	parse_terms(opts->right, opts, term_table, RIGHT, p_flag);
    }
    if (db && (val = vl_query(db, "top")) && !(opts->top_s)) {
	parse_terms(val, opts, term_table, TOP, p_flag);
    } else if (opts->top_s) {
	parse_terms(opts->top, opts, term_table, TOP, p_flag);
    }
    if (db && (val = vl_query(db, "bottom")) && !(opts->bottom_s)) {
	parse_terms(val, opts, term_table, BOTTOM, p_flag);
    } else if (opts->bottom_s) {
	parse_terms(opts->bottom, opts, term_table, BOTTOM, p_flag);
    }
}

static void parse_terms(val, opts, term_table, side, con_flag)
char *val;			/* List of terminals                */
opt_vals *opts;			/* Options and flags                */
avl_tree *term_table;		/* Terminal table                   */
term_side side;			/* What side should it be placed on */
int con_flag;			/* Non-zero if description read from contents */
/*
 * Parses `val' as a list of terminal names separated by spaces.  Each
 * terminal is added to `term_table'.  Its side becomes `side'.  If
 * the terminal is already in the table,  this description overrides
 * the previous one and a warning is produced under some conditions.
 * If it isn't in there,  it is added and a warning is produced
 * if `con_flag' is non-zero.
 */
{
    char *name;
    term_info *info;

    while (name = get_name(&val)) {
	if (avl_lookup(term_table, name, (char **) &info)) {
	    if ((info->from_contents) && (info->side != NONE) &&
			(info->side != side) && (!opts->silent)) {
		Printf("%s WARNING: Terminal `%s' defined in contents was\n",
		       optProgName, name);
		Printf("%s WARNING (cont): redefined in text definition.  It\n",
		       optProgName);
		Printf("%s WARNING (cont): was %s and will now be %s\n",
		       optProgName, side_str(info->side), side_str(side));
	    } else if ((!info->from_contents) && (!opts->silent)) {
		Printf("%s WARNING: Terminal `%s' defined multiply in text\n",
		       optProgName, name);
		Printf("%s WARNING (cont): description. It was on %s side and\n",
		       optProgName);
		Printf("%s WARNING (cont): will now be on %s side.\n",
		       optProgName, side_str(info->side), side_str(side));
	    }
	} else {
	    if (con_flag) {
		if (!opts->silent) {
		    Printf("%s WARNING: Terminal `%s' defined in textual description\n",
			   optProgName, name);
		    Printf("%s WARNING (cont): but not in contents facet.  Terminal\n",
			   optProgName);
		    Printf("%s WARNING (cont): will be ignored.\n",
			   optProgName);
		}
		continue;
		/*NOTREACHED*/
	    } else {
		/* Add it */
		info = ALLOC(term_info, 1);
		(void) avl_insert(term_table, name, (char *) info);
	    }
	}
	/* Update record */
	info->side = side;
	info->from_contents = 0;
	info->order = get_count(side);
	info->term_name = name;
    }
}

static char *side_str(side)
term_side side;
/*
 * Returns a string representation for `side'.
 */
{
    switch (side) {
    case NONE:
	return "NONE";
    case LEFT:
	return "LEFT";
    case RIGHT:
	return "RIGHT";
    case TOP:
	return "TOP";
    case BOTTOM:
	return "BOTTOM";
    default:
	return "UNKNOWN";
    }
    /*NOTREACHED*/
}

static char *get_name(ptr)
char **ptr;
/*
 * Gets the next space separated word from `ptr'.  The returned
 * word is allocated and should be freed when no longer needed.
 */
{
    char *start, *result;

    if (**ptr) {
	/* First eat all white space */
	while (**ptr && isspace(**ptr)) *ptr += 1;
	start = *ptr;
	/* Then move to the next space or to end */
	while (**ptr && !isspace(**ptr)) *ptr += 1;
	result = ALLOC(char, *ptr-start+1);
	(void) strncpy(result, start, *ptr-start);
	result[*ptr-start] = '\0';
	return result;
    } else {
	return (char *) 0;
    }
}

static int count_array[LAST_I];

static void count_init(term_table)
avl_tree *term_table;		/* Terminal table */
/*
 * Sets up the relative order count for terminals.
 */
{
    int i;
    avl_generator *gen;
    char *name;
    term_info *info;

    for (i = 0;  i < LAST_I;  i++) count_array[i] = INT_MAX;
    gen = avl_init_gen(term_table, AVL_FORWARD);
    while (avl_gen(gen, &name, (char **) &info)) {
	if (info->order < count_array[(int) info->side]) {
	    count_array[(int) info->side] = info->order-1;
	}
    }
    avl_free_gen(gen);
}

static int get_count(side)
term_side side;
/*
 * Returns a count for ordering terminals on a given side.
 */
{
    return (count_array[(int) side])--;
}

#define TI(ptr, field) (*((term_info **) ptr))->field
static int term_cmp(obj1, obj2)
char *obj1, *obj2;
/*
 * The routine is actually passed (term_info **).  The comparison
 * is based on the relative order field.
 */
{
    return TI(obj2, order) - TI(obj1, order);
}

#define INIT_SIZE	10

static char **extract_side(term_table, side)
avl_tree *term_table;		/* Terminal table  */
term_side side;			/* Side to extract */
/*
 * This routine looks through `term_table' looking for records
 * whose side is `side'.  These records are placed into a
 * dynamic array which is then sorted.  The resulting array
 * is then compressed down into an ordered, null terminated,
 *  list of terminal names.   This list is returned.
 */
{
    dary terms = daAlloc(term_info *, INIT_SIZE);
    avl_generator *gen;
    char *name, **result;
    term_info *info;
    int i;
    
    gen = avl_init_gen(term_table, AVL_FORWARD);
    while (avl_gen(gen, &name, (char **) &info)) {
	if (info->side == side) {
	    *daSetLast(term_info *, terms) = info;
	}
    }
    avl_free_gen(gen);
    daSort(terms, term_cmp);
    result = ALLOC(char *, daLen(terms)+1);
    for (i = 0;  i < daLen(terms);  i++) {
	result[i] = (*daGet(term_info *, terms, i))->term_name;
    }
    result[i] = (char *) 0;
    daFree(terms, (void (*)()) 0);
    return result;
}

static void final_opts(opts, fct, left, right, top, bottom)
opt_vals *opts;			/* Flags and options */
octObject *fct;			/* Facet to write to */
char **left;			/* Left terminals    */
char **right;			/* Right terminals   */
char **top;			/* Top terminals     */
char **bottom;			/* Bottom terminals  */
/*
 * This routine fills in reasonable final default values into
 * `opts' for those values which aren't set.
 */
{
    struct octBox box;
    octCoord min_width, min_height;

    if (!opts->name_s) {
	opts->name_s = 1;
	opts->name = fct->contents.facet.cell;
    }
    if (!opts->lambda_s) {
	opts->lambda_s = 1;
	if (tapGetProp(fct, TAP_LAMBDA)) {
	    opts->lambda = TAP_LAMBDA->value.integer;
	} else {
	    opts->lambda = DEF_LAMBDA;
	}
    }
    if (!opts->stem_s) {
	opts->stem_s = opts->stem_oct = 1;
	opts->stem = (opts->lambda * DEF_STEM);
    } else if (!opts->stem_oct) {
	opts->stem_oct = 1;
	opts->stem *= opts->lambda;
    }
    if (!opts->space_s) {
	opts->space_s = opts->space_oct = 1;
	opts->space = (opts->lambda * DEF_SPACE);
    } else if (!opts->space_oct) {
	opts->space_oct = 1;
	opts->space *= opts->lambda;
    }

    /* Handle width and height */
    OH_ASSERT(sg_box_size(opts->stem, opts->space,
			  sg_count(left), sg_count(right),
			  sg_count(top), sg_count(bottom),
			  &box));
    min_width = box.upperRight.x - box.lowerLeft.x;
    min_height = box.upperRight.y - box.lowerLeft.y;
    if (!opts->width_s) {
	opts->width_s = opts->width_oct = 1;
	opts->width = min_width;
    } else if (!opts->width_oct) {
	opts->width_oct = 1;
	opts->width *= opts->lambda;
    }
    if (!opts->silent && (opts->width < min_width)) {
	Printf("%s WARNING: Box width too small (must be at least %d lambda).\n",
	       optProgName, min_width/opts->lambda);
    }
    if (!opts->height_s) {
	opts->height_s = opts->height_oct = 1;
	opts->height = min_height;
    } else if (!opts->height_oct) {
	opts->height_oct = 1;
	opts->height *= opts->lambda;
    }
    if (!opts->silent && (opts->height < min_height)) {
	Printf("%s WARNING: Box height too small (must be at least %d lambda).\n",
	       optProgName, min_height/opts->lambda);
    }
}

