/*
 * Symbol Generator
 *
 * David Harrison
 * University of California, Berkeley
 * 1988
 *
 * This file contains a routine for creating simple box symbols.
 */

#include "copyright.h"
#include "port.h"
#include "oct.h"
#include "symgen.h"

#define STEM_LENGTH	1
#define ANNOT_HEIGHT	2

#ifndef MAX
#define MAX(a, b)	(((a) > (b)) ? (a) : (b))
#endif

#ifndef MIN
#define MIN(a, b)	(((a) < (b)) ? (a) : (b))
#endif

static octStatus oct_ret;
static void comp_param();

#define OC(call) \
if ((oct_ret = (call)) < OCT_OK) return oct_ret

static octStatus do_box();
static octStatus do_vert();
static octStatus do_hort();
static octStatus do_stem();
static octStatus do_vert_stem();
static octStatus do_hort_stem();

int sg_count(term_list)
char **term_list;		/* Terminal list */
/*
 * This routine returns the number of terminals in a terminal
 * list.  It is exported for convenience.
 */
{
    int count = 0;

    if (term_list) while (term_list[count]) count++;
    return count;
}

octStatus sg_box_size(stem, space, n_left, n_right, n_top, n_bottom, box)
octCoord stem;			/* Stem length (oct units)    */
octCoord space;			/* Inter-stem spacing (oct units) */
int n_left;			/* Number of left terminals   */
int n_right;			/* Number of right terminals  */
int n_top;			/* Number of top terminals    */
int n_bottom;			/* Number of bottom terminals */
struct octBox *box;		/* Minimum size (returned)    */
/*
 * This routine is used to compute the minimum size for a box symbol
 * constructed using sg_box().  The information required by this
 * routine is a simple derivation of the information required
 * by sg_box().  Note that sg_box() will refuse to construct
 * a symbol if the box provided is smaller than that indicated
 * by this routine.
 */
{
    int max_vert, max_hort;

    max_vert = MAX(n_left, n_right);
    max_hort = MAX(n_top, n_bottom);
    box->lowerLeft.x = 0;
    box->lowerLeft.y = 0;
    box->upperRight.x = (max_hort+1) * space;
    box->upperRight.y = (max_vert+1) * space;
    /* Now add stem lengths */
    if (n_left) box->upperRight.x += stem;
    if (n_right) box->upperRight.x += stem;
    if (n_top) box->upperRight.y += stem;
    if (n_bottom) box->upperRight.y += stem;
    return OCT_OK;
}

octStatus sg_box(sl, il, box, name, stem, space, left, right, top, bottom, opts)
octObject *sl;			/* Layer to create symbol on */
octObject *il;			/* Layer for implementation  */
struct octBox *box;		/* Size of symbol            */
char *name;			/* Name to place in symbol   */
octCoord stem;			/* Stem size (in oct units)  */
octCoord space;			/* Stem spacing (int oct units) */
char **left;			/* Terminals on left edge    */
char **right;			/* Terminals on right edge   */
char **top;			/* Terminals on top edge     */
char **bottom;			/* Terminals on bottom edge  */
opt_vals *opts;			/* Command level options     */
				/*			     */
				/* 	Marco Zelada	     */
				/* 	Intel Corp. 	     */
/*
 * This routine creates a new box symbol on `lyr'.  A label with
 * the text `name' will be placed in the box.  `left', `right',
 * `top', and `bottom' are null-terminated arrays of terminal
 * names.  Stems with small terminals on the end will be created
 * for each terminal.  Terminal locations will be snapped to
 * `lambda' boundaries.  Labels describing the terminal names
 * will be placed in the box.  Returns a bad status code if
 * something goes wrong.
 */
{
    octObject fct, new_lbl;
    octCoord width, height;
    struct octPoint pnt1, pnt2;
    int int1, int2;
    int n_left, n_right, n_top, n_bottom;
    int text_height;

    /* Count terminals */
    n_left = sg_count(left);
    n_right = sg_count(right);
    n_top = sg_count(top);
    n_bottom = sg_count(bottom);

    /* Derive facet from layer */
    octGetFacet(sl, &fct);

    /* Start by creating box outline */
    OC(do_box(sl, stem, n_left, n_right, n_top, n_bottom, box));

/*      ignore all of this if -c is on.

        Marco Zelada
        Intel Corp.
*/

    if ( ! opts->contents_size )
    {
    	width = box->upperRight.x - box->lowerLeft.x;
    	height = box->upperRight.y - box->lowerLeft.y;

    	/* Compute vertical parameters */
    	pnt1.x = box->lowerLeft.x;
    	pnt2.x = box->upperRight.x;
    	pnt1.y = pnt2.y = box->upperRight.y;
    	comp_param(height, space, -1, n_left, n_right, &pnt1.y, &int1, &pnt2.y, &int2);
    	text_height = MIN(int1, int2);
    	/* Layout vertical stems */
    	OC(do_vert(&fct, sl, il, stem, &pnt1, int1, -1, n_left, left, text_height, width));
    	OC(do_vert(&fct, sl, il, stem, &pnt2, int2, 1, n_right, right, text_height, width));

    	/* Compute horizontal parameters */
    	pnt1.x = pnt2.x = box->lowerLeft.x;
    	pnt1.y = box->upperRight.y;
    	pnt2.y = box->lowerLeft.y;
    	comp_param(width, space, 1, n_top, n_bottom, &pnt1.x, &int1, &pnt2.x, &int2);
    	/* Layout horizontal stems */
    	OC(do_hort(&fct, sl, il, stem, &pnt1, int1, 1, n_top, top, text_height, int1));
    	OC(do_hort(&fct, sl, il, stem, &pnt2, int2, -1, n_bottom, bottom, text_height, int2));
    };

    /* Place internal name - centered in symbol */
    new_lbl.type = OCT_LABEL;
    new_lbl.contents.label.label = name;
    new_lbl.contents.label.region = *box;
    new_lbl.contents.label.textHeight = box->upperRight.y - box->lowerLeft.y;
    new_lbl.contents.label.vertJust = OCT_JUST_CENTER;
    new_lbl.contents.label.horizJust = OCT_JUST_CENTER;
    new_lbl.contents.label.lineJust = OCT_JUST_CENTER;
    OC(octCreate(sl, &new_lbl));

    return OCT_OK;
}

static octStatus do_box(lyr, stem_len, nl, nr, nt, nb, box)
octObject *lyr;			/* Creation layer     */
int stem_len;			/* Stem length        */
int nl, nr, nt, nb;		/* Terminal counts    */
struct octBox *box;		/* Outline dimensions */
/*
 * Creates a path of zero width on `lyr' that outlines `box'.
 */
{
    octObject new_path;
    struct octPoint pnts[5];

    /* Reduce size of box by stem lengths */
    if (nl) box->lowerLeft.x += stem_len;
    if (nr) box->upperRight.x -= stem_len;
    if (nt) box->upperRight.y -= stem_len;
    if (nb) box->lowerLeft.y += stem_len;

    /* Create the box */
    pnts[0] = box->lowerLeft;
    pnts[1] = pnts[0];  pnts[1].y = box->upperRight.y;
    pnts[2] = box->upperRight;
    pnts[3] = pnts[2];  pnts[3].y = box->lowerLeft.y;
    pnts[4] = pnts[0];
    new_path.type = OCT_PATH;
    new_path.contents.path.width = 0;
    OC(octCreate(lyr, &new_path));
    return octPutPoints(&new_path, 5, pnts);
}

static void comp(span, sp, dir, n1, pnt1, int1, n2, pnt2, int2)
octCoord span;			/* Overall length                 */
int sp;				/* Minimum space between stems    */
int dir;			/* Which way the stems go (1,-1)  */
int n1;				/* Larger set of terminals        */
octCoord *pnt1;			/* Start point for first set      */
int *int1;			/* Interval for first set         */
int n2;				/* Smaller set of terminals       */
octCoord *pnt2;			/* Start point for second set     */
int *int2;			/* Interval for second set        */
/*
 * Computes vertical parameters based on the fact that the first
 * set of terminals is the critical set for placement.
 */
{
    int extra, orig;

    *int1 = ((span/(n1+1))/sp)*sp;
    if (*int1 <= 0) *int1 = sp;
    extra = (span - (*int1 * (n1+1)))/sp/2*sp;
    *pnt1 += (*int1 + extra)*dir;

    *int2 = (((span/(n2+1))/sp)*sp)/(*int1) * (*int1);
    if (*int2 <= 0) *int2 = sp;
    orig = *int2 + (span - (*int2 * (n2+1)))/sp/2*sp;
    *pnt2 += (((orig-extra)/(*int1) * (*int1)) + extra)*dir;
}

static void comp_param(hgt, sp, dir, nl, nr, l_pnt, l_int, r_pnt, r_int)
octCoord hgt;			/* Overall height 		  */
octCoord sp;			/* Minimum spacing                */
int dir;			/* Direction (1, -1)              */
int nl, nr;			/* Number terms on each side      */
octCoord *l_pnt;		/* Side one init (RETURN)         */
int *l_int;			/* Side one interval (RETURN) 	  */
octCoord *r_pnt;		/* Side two init (RETURN)         */
int *r_int;			/* Side two interval (RETURN) 	  */
/*
 * Computes the start and interval parameters for stems.
 * Note that the points are modified assuming that they have been
 * set to the appropriate corner of the symbol box.
 */
{
    /* First decide which one is critical */
    if (nl > nr) {
	comp(hgt, sp, dir, nl, l_pnt, l_int, nr, r_pnt, r_int);
    } else {
	comp(hgt, sp, dir, nr, r_pnt, r_int, nl, l_pnt, l_int);
    }
}

static octStatus do_vert(fct, sl, il, stem, init, intr, dir, num, names, th, tw)
octObject *fct;			/* Facet to create stems in      */
octObject *sl;			/* Tab creation layer            */
octObject *il;			/* Terminal implementation layer */
int stem;			/* Stem length                   */
struct octPoint *init;		/* Start point                   */
octCoord intr;			/* Interval between stems        */
int dir;			/* Direction for stem            */
int num;			/* Number of terminals           */
char **names;			/* Terminal names                */
octCoord th, tw;		/* Annotation size               */
/*
 * Creates the tabs, terminals, and annotations for the vertical
 * sides of the symbols.  All tabs begin at `pnt' and are placed
 * at `intr' intervals from top to bottom.  `dir' specifies
 * which direction horizontally to place the tab.  The starting
 * point is modified.
 */
{
    int i;
    struct octPoint pnts[2];
    octObject new_lbl;

    for (i = 0;  i < num;  i++) {
	/* Stem itself is easy */
	pnts[0].x = init->x;
	pnts[1].x = init->x + (stem * dir);
	pnts[0].y = pnts[1].y = init->y;
	OC(do_stem(fct, sl, il, pnts, names[i]));

	/* Annotation is a bit trickier */
	new_lbl.type = OCT_LABEL;
	new_lbl.contents.label.label = names[i];
	new_lbl.contents.label.region.lowerLeft.y = init->y - (th/2);
	new_lbl.contents.label.region.upperRight.y =
	  new_lbl.contents.label.region.lowerLeft.y + th;
	new_lbl.contents.label.textHeight = th/2;
	new_lbl.contents.label.vertJust = OCT_JUST_CENTER;
	new_lbl.contents.label.lineJust = OCT_JUST_CENTER;
	if (dir > 0) {
	    /* Left to right */
	    new_lbl.contents.label.region.lowerLeft.x = init->x - tw;
	    new_lbl.contents.label.region.upperRight.x = init->x;
	    new_lbl.contents.label.horizJust = OCT_JUST_RIGHT;
	} else {
	    /* Right to left */
	    new_lbl.contents.label.region.lowerLeft.x = init->x;
	    new_lbl.contents.label.region.upperRight.x = init->x + tw;
	    new_lbl.contents.label.horizJust = OCT_JUST_LEFT;
	}
	OC(octCreate(sl, &new_lbl));
	init->y -= intr;
    }
    return OCT_OK;
}


static octStatus do_hort(fct, sl, il, stem, init, intr, dir, num, names, th, tw)
octObject *fct;			/* Facet to create stems in      */
octObject *sl;			/* Tab creation layer            */
octObject *il;			/* Terminal implementation layer */
int stem;			/* Stem length                   */
struct octPoint *init;		/* Start point                   */
octCoord intr;			/* Interval between stems        */
int dir;			/* Direction for stem            */
int num;			/* Number of terminals           */
char **names;			/* Terminal names                */
octCoord th, tw;		/* Annotation size               */
/*
 * Creates the tabs, terminals, and annotations for the horizontal
 * sides of the symbols.  All tabs begin at `pnt' and are placed
 * at `intr' intervals from left to right.  `dir' specifies
 * which direction vertically to place the tab.  The starting
 * point is modified.
 */
{
    int i;
    struct octPoint pnts[2];
    octObject new_lbl;

    for (i = 0;  i < num;  i++) {
	/* Stem itself is easy */
	pnts[0].x = pnts[1].x = init->x;
	pnts[0].y = init->y;
	pnts[1].y = pnts[0].y + (stem * dir);
	OC(do_stem(fct, sl, il, pnts, names[i]));

	/* Annotation is a bit trickier */
	new_lbl.type = OCT_LABEL;
	new_lbl.contents.label.label = names[i];
	new_lbl.contents.label.region.lowerLeft.x = init->x - (tw/2);
	new_lbl.contents.label.region.upperRight.x =
	  new_lbl.contents.label.region.lowerLeft.x + tw;
	new_lbl.contents.label.textHeight = th/2;
	new_lbl.contents.label.horizJust = OCT_JUST_CENTER;
	new_lbl.contents.label.lineJust = OCT_JUST_CENTER;
	if (dir > 0) {
	    /* Bottom to top */
	    new_lbl.contents.label.region.lowerLeft.y = init->y - th;
	    new_lbl.contents.label.region.upperRight.y = init->y;
	    new_lbl.contents.label.vertJust = OCT_JUST_TOP;
	} else {
	    /* Top to bottom */
	    new_lbl.contents.label.region.lowerLeft.y = init->y;
	    new_lbl.contents.label.region.upperRight.y = init->y + th;
	    new_lbl.contents.label.vertJust = OCT_JUST_BOTTOM;
	}
	OC(octCreate(sl, &new_lbl));
	init->x += intr;
    }
    return OCT_OK;
}



static octStatus do_stem(fct, pathLayer, termLayer, pnts, termName)
octObject *fct;			/* Facet for terminal creation       */
octObject *pathLayer;		/* Layer for stem path               */
octObject *termLayer;		/* Layer for terminal implementation */
struct octPoint pnts[2];	/* Stem points                       */
char *termName;			/* Formal terminal name              */
/*
 * This routine carries out the common parts of stem construction:
 * builds the path, the terminal implementation, and the formal
 * terminal.
 */
{
    octObject new_path, new_box, f_term;

    /* Stem path */
    new_path.type = OCT_PATH;
    new_path.contents.path.width = 0;
    OC(octCreate(pathLayer, &new_path));
    OC(octPutPoints(&new_path, 2, pnts));

    /* Formal terminal promotion */
    f_term.type = OCT_TERM;
    f_term.contents.term.name = termName;
    f_term.contents.term.instanceId = oct_null_id;
    f_term.contents.term.width = 1;
    OC(octGetOrCreate(fct, &f_term));

    while ( octGenFirstContent( &f_term, OCT_BOX_MASK, &new_box ) == OCT_OK ) {
	octDelete( &new_box );	/* Delete old implementation of terminal. */
    }
    /* Terminal implementation box */
    new_box.type = OCT_BOX;
    new_box.contents.box.lowerLeft.x = pnts[1].x-1;
    new_box.contents.box.lowerLeft.y = pnts[1].y-1;
    new_box.contents.box.upperRight.x = pnts[1].x+1;
    new_box.contents.box.upperRight.y = pnts[1].y+1;
    OC(octCreate(termLayer, &new_box));
    
    return octAttach(&f_term, &new_box);
}
