/*LINTLIBRARY*/
/*
 * Readvals
 *
 * David Harrison
 * University of California, Berkeley
 * 1989
 *
 * This is a small package for reading name/value pairs from
 * a text file.  The format of the file is name, an equals symbol, 
 * followed by a value.  If either the name or the value includes
 * spaces,  it can be enclosed in double quotes.  For example:
 *
 * name = foobar 
 * "border color" = red
 * long = "this is a value that contains spaces"
 *
 * The values are retrieved by name.
 */

#include "port.h"
#include "avl.h"
#include "da.h"
#include "vl.h"

/*
 * The actual representation for val_db is an avl tree implemented
 * using the avl package.
 */

static int getpair();
static int read_ident();
static int read_equal();

val_db vl_new(data_file, cmp)
FILE *data_file;		/* File to read        */
int (*cmp)();			/* Comparison function */
/* 
 * This function reads the given file and builds a name/value
 * pair database from it.  The names are compared using the
 * function `cmp' which should be declared as follows:
 *   int cmp(a, b)
 *   char *a, *b;
 * This should return less than, equal to, or greater than
 * zero based on whether the first string is less than, equal to,
 * or greater than the second string.  If this is zero,
 * the standard function `strcmp' will be used.
 */
{
    avl_tree *tree;
    char *name, *value;
    
    tree = avl_init_table(cmp);
    while (getpair(data_file, &name, &value)) {
	(void) avl_insert(tree, name, value);
    }
    return (val_db) tree;
}

char *vl_query(db, name)
val_db db;			/* Database for examination */
char *name;			/* Name to look up          */
/*
 * Queries the given database for the supplied name using
 * the comparison function specified for vl_new.  If
 * the name is not in the database,  the routine returns zero.
 * The string is owned by this package and should not be overwritten.
 * The string ceases to exist when the database is destroyed
 * using vl_delete().
 */
{
    avl_tree *tree = (avl_tree *) db;
    char *result;

    if (avl_lookup(tree, name, &result)) {
	return result;
    } else {
	return (char *) 0;
    }
}

void vl_delete(db)
val_db db;			/* Database to destroy */
/*
 * This routine reclaims the resources consumed by `db'.  All
 * strings returned by vl_query are reclaimed.
 */
{
    avl_tree *tree = (avl_tree *) db;

    avl_free_table(tree, free, free);
}

static int getpair(file, name, value)
FILE *file;			/* File to read           */
char **name;			/* Name of item (RETURN)  */
char **value;			/* Value of item (RETURN) */
/*
 * Parses the file looking for the next name/value pair.  The
 * form of a name/value pair is `name = value'.  Normally,
 * these are space separated identifiers.  However,  these
 * specifications can include spaces if enclosed double quotes.
 * Double quotes can appear in the identifier by using
 * two quotes.  For example: "He said ""Hello"" to Betty".
 */
{
    if (read_ident(file, name) &&
	read_equal(file) &&
	read_ident(file, value)) {
	return 1;
    } else {
	return 0;
    }
}    


#define INIT_SIZE	256

static int read_ident(file, val)
FILE *file;			/* File to read */
char **val;			/* Identifier   */
/*
 * Reads an identifer from the specified file and returns it
 * in `val'.  If EOF is detected,  the routine returns zero
 * and `val' is not modifed.
 */
{
    int ch, next;
    static dary buf = (dary) 0;
    char quoted;

    /* Eat white space */
    while ((ch = getc(file)) != EOF) {
	if (!isspace(ch)) break;
    }
    if (ch == EOF) return 0;
    if (ch == '"') {
	quoted = 1;
    } else {
	quoted = 0;
	(void) ungetc(ch, file);
    }
    if (!buf) {
	buf = daAlloc(char, INIT_SIZE);
    }
    daReset(buf);
    while ((ch = getc(file)) != EOF) {
	if (quoted) {
	    if (ch == '"') {
		/* May be end */
		next = getc(file);
		if (next != '"') {
		    /* End */
		    (void) ungetc(next, file);
		    break;
		}
	    }
	} else if (isspace(ch)) {
	    /* End */
	    break;
	}
	*daSetLast(char, buf) = ch;
    }
    *daSetLast(char, buf) = 0;
    *val = daCopyData(char, buf);
    return 1;
}

static int read_equal(file)
FILE *file;			/* File to read from */
/*
 * Reads the separating equal from the input stream.  If
 * there is no equal,  it still returns ok.
 */
{
    int ch;

    /* Eat white space */
    while ((ch = getc(file)) != EOF) {
	if (!isspace(ch)) break;
    }
    if (ch == EOF) return 0;
    if (ch != '=') (void) ungetc(ch, file);
    return 1;
}
