/*
 * Symbolic policy checker
 *
 * David Harrison
 * University of California, Berkeley
 *
 * 1990
 */

#include "spc.h"

#include <varargs.h>
#include "copyright.h"
#include "port.h"
#include "utility.h"
#include "errtrap.h"
#include "uprintf.h"
#include "tr.h"
#include "st.h"
#include "oh.h"

static int err_count = 0;

static void def_err();
static int supress();
static scErrFunc err_func = { def_err, (char *) stderr };

char *sc_pkg_name = "sc";

#define OC(call) \
if ((call) < OCT_OK) errRaise(sc_pkg_name, 0, \
			      "Oct call failed in %s at %d:\n  %s\n", \
			      __FILE__, __LINE__, octErrorString())


/* Simple linked lists */
struct ll {
    char *data;
    struct ll *tail;
};
static struct ll *tl;
static struct ll *ll_remove();

#define LL	struct ll *
#define NILL	(LL) 0
#define CAR(l)	((l) ? ((l)->data) : (char *) 0)
#define CDR(l)	((l) ? ((l)->tail) : (LL) 0)
#define CONS(d, l)	(tl = ALLOC(struct ll, 1), tl->data = (char *) (d), tl->tail = (l), tl)
#define CUT(l, f)	ll_remove(l, f)

#define POINT_IN_BOX(p,b)	((p)->x >= (b)->lowerLeft.x && \
				 (p)->x <= (b)->upperRight.x && \
				 (p)->y >= (b)->lowerLeft.y && \
				 (p)->y <= (b)->upperRight.y)

#define SEQ(a, b)	((a) ? ((b) ? strcmp(a, b) : -1) : ((b) ? 1 : 0))

typedef struct term_impl_defn {
    char *lyr_name;
    struct octBox bb;
} term_impl;

typedef enum jt_type_defn { NOTJT, LOCALNET, GLOBALNET } jt_type;

typedef struct term_info_defn {
    octObject at;		/* Actual terminal */
    LL impl;			/* List of layer/bb pairs */
    jt_type jt_p;		/* Non-zero if jump terminal */
    char *jt_n;			/* Jump terminal net name */
} term_info;


/* Forward declarations */
static void check_cell();
static void check_facet();
static void ck_props();
static void ck_insts();
static void ck_ibag();
static int get_interface();
static LL tm_info();
static term_info *get_term_info();
static void add_to_joined_list();
static LL build_final();
static LL ck_terms();
static void ck_seg();
static void err();
static void zap_term_info();
static char *disp_obj();
static void ck_geometry();
static void ck_geo();
static void lyr_ck();
static void path_ck();
static void ck_nets();
static void ck_one_net();
static st_table *find_attached();
static int conn_term_p();
static void visit_and_mark();
static int jump_term_p();
static void visit_obj();
static void visit_term();
static void visit_path();
static void check_items();
static void ck_fterms();
static void ck_one_term();
static void ck_eprop();
static void supress_reset();



/*
 * Error class names
 */

char *scErrClass[EOE] = {
  "Missing facet property",
  "Bad facet property",
  "Object not attached to appropriate bag",
  "Missing cell type property",
  "Missing interface facet",
  "Mismatched actual/formal terminal",
  "Joined terminal policy is incorrect",
  "Terminal incorrectly attached to net",
  "Dangling actual terminal",
  "Jump terminal policy is incorrect",
  "Segment attached to too many or too few layers",
  "Segment can't connect to terminal",
  "Geometry directly attached to facet",
  "Unrecognized geometry in facet",
  "Segment does not have exactly two points",
  "Segment connected to too few or too many nets",
  "Illegal object type",
  "Object not reachable from net or not attached to appropriate net",
  "Net with less than two terminals",
  "Duplicate formal terminal names",
  "Unexpected object type",
  "Formal terminal implementation problem",
  "Formal terminal not attached to net",
  "Bad property value"
};



int scCheck(fct)
octObject *fct;			/* Facet to check  */
/*
 * Checks a symbolic cell for correctness.  Returns the number of
 * serious errors found.  The error reporting function can be
 * changed using sc_set_error().
 */
{
    err_count = 0;

    supress_reset();
    ck_props(fct);
    ck_insts(fct);
    ck_geometry(fct);
    ck_nets(fct);
    ck_fterms(fct);

    return err_count;
}


void scSetErrorFunc(new, old)
scErrFunc *new;			/* New error function */
scErrFunc *old;			/* Old error function */
/*
 * Sets the error reporting function to `new' and returns the old
 * error reporting function in `old'.  The error reporting function
 * has the following form:
 *   void err_func(severity, class, obj, msg, data)
 *   int severity, class;
 *   octObject *obj;
 *   char *msg, *data;
 * The severity will be one of WARN, ERR, or FATAL.  The error
 * class will be one of those given in the header file.  The
 * bad object (if any) will be passed as `obj'.  A
 * suitable error message will be passed in `msg'.  The `data'
 * parameter will be the same as that specified in `new'.
 */
{
    if (old) {
	*old = err_func;
    }
    err_func = *new;
}



static char *props[] = { "TECHNOLOGY", "VIEWTYPE", "EDITSTYLE" };

static void ck_props(fct)
octObject *fct;
{
    octObject prop;
    int i;

    prop.type = OCT_PROP;
    for (i = 0;  i < sizeof(props)/sizeof(char *);  i++) {
	prop.contents.prop.name = props[i];
	if (octGetByName(fct, &prop) != OCT_OK) {
	    err(WARN, (octObject *) 0, MFP, "Facet has no %s property\n", props[i]);
	} 
    }
}


static void ck_insts(fct)
octObject *fct;
{
    octObject inst;
    octGenerator gen;

    OC(octInitGenContents(fct, OCT_INSTANCE_MASK, &gen));
    while (octGenerate(&gen, &inst) == OCT_OK) {
	ck_ibag(&inst);
	CUT(ck_terms(tm_info(&inst)), zap_term_info);
    }
}



static void ck_ibag(inst)
octObject *inst;
/* Checks to see if it is on the connectors or instances bag */
{
    octGenerator gen;
    octObject interface;
    octObject prop, bag;
    char *bag_name;

    if (get_interface(inst, &interface)) {
	prop.type = OCT_PROP;
	prop.contents.prop.name = "CELLTYPE";
	if (octGetByName(&interface, &prop) == OCT_OK) {
	    if (prop.contents.prop.value.string &&
		(strcmp(prop.contents.prop.value.string, "CONTACT") == 0)) {
		/* Contact -- should be on connectors bag */
		bag_name = "CONNECTORS";
	    } else {
		/* Should be on instances bag */
		bag_name = "INSTANCES";
	    }
	    OC(octInitGenContainers(inst, OCT_BAG_MASK, &gen));
	    while (octGenerate(&gen, &bag) == OCT_OK) {
		if (bag.contents.bag.name &&
		    (strcmp(bag.contents.bag.name, bag_name) == 0)) {
		    octFreeGenerator(&gen);
		    return;
		}
	    }
	    err(ERR, inst, BCE, "Instance is not on the %s bag\n", bag_name);
	} else {
	    err(WARN, inst, MCT, "Interface master of instance has no CELLTYPE property\n");
	}
    }
}


static int get_interface(inst, ifct)
octObject *inst;
octObject *ifct;
{
    static st_table *bad_instances = (st_table *) 0;
    octId *stored;

    ifct->type = OCT_FACET;
    ifct->contents.facet.facet = "interface";
    ifct->contents.facet.mode = "r";
    ifct->contents.facet.version = OCT_CURRENT_VERSION;
    if (octOpenMaster(inst, ifct) < OCT_OK) {
	if (!bad_instances) bad_instances = st_init_table(octIdCmp, octIdHash);
	if (!st_is_member(bad_instances, (char *) &inst->objectId)) {
	    err(FATAL, inst, MIF, "Instance has no interface\n");
	    stored = ALLOC(octId, 1);
	    *stored = inst->objectId;
	    (void) st_insert(bad_instances, (char *) stored, (char *) 0);
	}
    }
    return 1;
}


static enum st_retval freeKey(key, value, arg)
char *key, *value, *arg;
{
    FREE(key);
    return ST_CONTINUE;
}

static LL tm_info(inst)
octObject *inst;
/*
 * Returns a list of joined terminal lists for the instance.
 * Gathers information about each terminal.
 */
{
    octGenerator gen;
    octObject interface, aterm;
    LL result = NILL;
    st_table *joined_lists;
    term_info *info;

    joined_lists = st_init_table(octIdCmp, octIdHash);
    if (get_interface(inst, &interface)) {
	OC(octInitGenContents(inst, OCT_TERM_MASK, &gen));
	while (octGenerate(&gen, &aterm) == OCT_OK) {
	    info = get_term_info(&aterm, inst, &interface);
	    if ( info ) {
		add_to_joined_list(joined_lists, info, &interface);
	    }
	}
	result = build_final(joined_lists);
	st_foreach(joined_lists, freeKey, (char *) 0);
	st_free_table(joined_lists);
    }
    return result;
}


static term_info *get_term_info(at, inst, ifct)
octObject *at;			/* Actual terminal */
octObject *inst;		/* Instance for actual terminal */
octObject *ifct;		/* Interface master */
/* There should be a cache above this... */
{
    term_info *result;
    term_impl *one;
    octObject fterm, impl, geo, layer, prop;
    octGenerator gen;
    int32 temp;

    if (ohGetTerminal(ifct, at->contents.term.formalExternalId, &fterm) != OCT_OK) {
	/* OK, then try by name */
	if ( ohGetByTermName( ifct, &fterm, at->contents.term.name ) != OCT_OK ) {
	    err(FATAL, at, MAF,
		"Actual terminal has no corresponding formal terminal\n");
	    return (term_info *) 0;
	}
    }

    result = ALLOC(term_info, 1);
    result->at = *at;
    result->impl = NILL;
    OC(octInitGenContents(&fterm, OCT_GEO_MASK, &gen));
    while (octGenerate(&gen, &geo) == OCT_OK) {
	one = ALLOC(term_impl, 1);
	OC(octBB(&geo, &(one->bb)));
	tr_oct_transform(&inst->contents.instance.transform,
			 &one->bb.lowerLeft.x, &one->bb.lowerLeft.y);
	tr_oct_transform(&inst->contents.instance.transform,
			 &one->bb.upperRight.x, &one->bb.upperRight.y);
	/* Normalize */
	if (one->bb.lowerLeft.x > one->bb.upperRight.x) {
	    temp = one->bb.lowerLeft.x;
	    one->bb.lowerLeft.x = one->bb.upperRight.x;
	    one->bb.upperRight.x = temp;
	}
	if (one->bb.lowerLeft.y > one->bb.upperRight.y) {
	    temp = one->bb.lowerLeft.y;
	    one->bb.lowerLeft.y = one->bb.upperRight.y;
	    one->bb.upperRight.y = temp;
	}
	OC(octGenFirstContainer(&geo, OCT_LAYER_MASK, &layer));
	one->lyr_name = layer.contents.layer.name;
	result->impl = CONS(one, result->impl);
    }
    /* Jump terminals */
    OC(octInitGenContents(&fterm, OCT_PROP_MASK, &gen));
    result->jt_p = NOTJT;
    while (octGenerate(&gen, &prop) == OCT_OK) {
	if (prop.contents.prop.name &&
	    (strcmp(prop.contents.prop.name, "GLOBALNET") == 0)) {
	    octFreeGenerator(&gen);
	    result->jt_p = GLOBALNET;
	    result->jt_n = prop.contents.prop.value.string;
	}
	if (prop.contents.prop.name &&
	    (strcmp(prop.contents.prop.name, "LOCALNET") == 0)) {
	    octFreeGenerator(&gen);
	    result->jt_p = LOCALNET;
	    result->jt_n = prop.contents.prop.value.string;
	}
    }
    return result;
}



static void add_to_joined_list(jl, info, ifct)
st_table *jl;			/* Hash table of all joined lists */
term_info *info;		/* Terminal information           */
octObject *ifct;			/* Interface master               */
/*
 * Looks up the corresponding formal terminal in the interface 
 * master and locates the joined bag it is a member of (assumes
 * only one joined bag per terminal).  It looks up this bag in `jl'
 * and adds `info' to the list stored there.  If it isn't there,
 * a new list is created and stored in `jl' under its bag id.  If
 * there is no joined bag associated with the terminal, it is placed
 * in the table under its own id.
 */
{
    octObject fterm, bag;
    octGenerator gen;
    octId jid = oct_null_id, *stored;
    LL aj;

    OC(ohGetTerminal(ifct, info->at.contents.term.formalExternalId, &fterm));
    OC(octInitGenContainers(&fterm, OCT_BAG_MASK, &gen));
    while (octGenerate(&gen, &bag) == OCT_OK) {
	if (bag.contents.bag.name &&
	    (strcmp(bag.contents.bag.name, "JOINED") == 0)) {
	    if (octIdIsNull(jid)) {
		jid = bag.objectId;
	    } else {
		err(ERR, &info->at, BJP,
		    "Terminal belongs to multiple joined bags\n");
	    }
	}
    }
    if (octIdIsNull(jid)) {
	/* No joined bag */
	stored = ALLOC(octId, 1);
	*stored = info->at.objectId;
	aj = CONS(info, NILL);
	(void) st_insert(jl, (char *) stored, (char *) aj);
    } else {
	/* Joined bag */
	aj = NILL;
	if (!st_lookup(jl, (char *) &jid, (char **) &aj)) {
	    aj = CONS(info, NILL);
	    stored = ALLOC(octId, 1);
	    *stored = jid;
	    (void) st_insert(jl, (char *) stored, (char *) aj);
	} else {
	    aj = CONS(info, aj);
	    (void) st_insert(jl, (char *) &jid, (char *) aj);
	}
    }
}


static LL build_final(jlt)
st_table *jlt;			/* Hash table of joined lists */
/*
 * Builds a linked list from the hash table of linked lists
 * in `jlt'.
 */
{
    LL result;
    st_generator *gen;
    char *key, *value;

    gen = st_init_gen(jlt);
    result = NILL;
    while (st_gen(gen, &key, &value)) {
	result = CONS(value, result);
    }
    st_free_gen(gen);
    return result;
}



static LL ck_terms(ti)
LL ti;				/* Terminal information */
/*
 * This routine makes sure all joined terminals are connected
 * to the same net and insures segments are attached to the
 * terminals appropriately.  Jump terminals are checked to
 * make sure the net associated with them is named appropriately.
 */
{
    LL i;
    LL j;
    LL jl;
    octId net_id;
    term_info *info;
    octGenerator gen, seg_gen;
    octObject seg_or_term, net, fterm, fct, prop;
    int net_p;
    
    for (i = ti;  i != NILL;  i = CDR(i)) {
	jl = (LL) CAR(i);
	net_id = oct_null_id;
	for (j = jl;  j != NILL;  j = CDR(j)) {
	    info = (term_info *) CAR(j);
	    net_p = 0;
	    OC(octInitGenContainers(&info->at, OCT_NET_MASK, &gen));
	    while (octGenerate(&gen, &net) == OCT_OK) {
		net_p = 1;
		if (octIdIsNull(net_id)) {
		    net_id = net.objectId;
		} else if (!octIdsEqual(net_id, net.objectId)) {
		    err(FATAL, &info->at, TNF,
			"Terminal attached to too many nets\n");
		}
	    }
	    OC(octInitGenContainers(&info->at, OCT_PATH_MASK|OCT_TERM_MASK, &gen));
	    while (octGenerate(&gen, &seg_or_term) == OCT_OK) {
	        OC(octInitGenContainers(&seg_or_term, OCT_NET_MASK, &seg_gen));
		while (octGenerate(&seg_gen, &net) == OCT_OK) {
		    net_p = 1;
		    if (octIdIsNull(net_id)) {
			net_id = net.objectId;
		    } else if (!octIdsEqual(net_id, net.objectId)) {
			err(FATAL, &seg_or_term, TNF, "Object attached to wrong net\n");
		    }
		}
		/* Commented out because all these checks are already done in the paths.c routines */
		/* if (seg_or_term.type == OCT_PATH) ck_seg(&seg_or_term, info); */
	    }
	    if (!net_p) {
		err(WARN, &info->at, DAT, "Dangling terminal\n");
	    }
	    /* Jump terminal test */
	    if (info->jt_p != NOTJT && !octIdIsNull(net_id)) {
		net.objectId = net_id;
		OC(octGetById(&net));
		if (SEQ(net.contents.net.name, info->jt_n) != 0) {
		    err(ERR, &info->at, JTF,
			"Net name does not match jump terminal specification\n");
		}
		if (info->jt_p == GLOBALNET) {
		    fterm.type = OCT_TERM;
		    fterm.contents.term.name = info->jt_n;
		    if (octGetByName(&net, &fterm) == OCT_OK) {
			prop.type = OCT_PROP;
			prop.contents.prop.name = "GLOBALNET";
			if (octGetByName(&fterm, &prop) != OCT_OK) {
			    err(ERR, &info->at, JTF,
				"Global net formal terminal doesn't have required GLOBALNET property\n");
			}
		    } else {
			err(ERR, &info->at, JTF,
			    "Global net does not have required formal terminal\n");
		    }
		    
		}
	    }
	}
    }
    return ti;
}



static void ck_seg(seg, info)
octObject *seg;
term_info *info;
/*
 * Checks to make sure `seg' intersects `info' on a appropriate
 * layer.
 */
{
    octGenerator gen;
    octObject layer;
    struct octPoint pnts[2];
    int32 np = 2;
    int lyr_cnt = 0, j, isect, len, hi;
    LL i;
    term_impl *impl;
    char buf[1024];

    if (octGetPoints(seg, &np, pnts) == OCT_OK) {
	lyr_cnt = 0;
	OC(octInitGenContainers(seg, OCT_LAYER_MASK, &gen));
	while (octGenerate(&gen, &layer) == OCT_OK) {
	    lyr_cnt++;
	}
	if (lyr_cnt == 0) {
	    err(FATAL, seg, SLF, "Segment not attached to a layer\n");
	    return;
	} else if (lyr_cnt > 1) {
	    err(FATAL, seg, SLF, "Segment connected to multiple layers\n");
	    return;
	}
	for (i = info->impl;  i != NILL;  i = CDR(i)) {
	    impl = (term_impl *) CAR(i);
	    if (impl->lyr_name && layer.contents.layer.name &&
		(strcmp(impl->lyr_name, layer.contents.layer.name) == 0)) {
		/* Appropriate layer */
		isect = 0;
		for (j = 0;  j < 2;  j++) {
		    if (POINT_IN_BOX(pnts+j, &(impl->bb))) {
			isect++;
		    }			
		}
		if (isect >= 1) {
		    /* Check size of terminal compared to segment */
		    len = ABS(pnts[0].x - pnts[1].x);
		    hi = ABS(pnts[0].y - pnts[1].y);
		    if (len > hi) {
			/* Horizontal */
			if (seg->contents.path.width >
			    (impl->bb.upperRight.y - impl->bb.lowerLeft.y)) {
			    err(WARN, seg, STF, "Segment too wide for terminal %s\n",
				disp_obj(buf, &info->at));
			}
		    } else {
			/* Vertical */
			if (seg->contents.path.width >
			    (impl->bb.upperRight.x - impl->bb.lowerLeft.x)) {
			    err(WARN, seg, STF,
				"Segment too wide for terminal %s\n",
				disp_obj(buf, &info->at));
			}
		    }
		} else {
		    err(FATAL, seg, STF,
			"Segment can't connect to terminal %s\n",
			disp_obj(buf, &info->at));
		}
	    }
	}
    }
}


static void ck_geometry(fct)
octObject *fct;
/*
 * Checks segments for correctness.  Insures all segments are attached
 * to layers and nets, and that all segments are connected to exactly
 * two actual terminals.  It also checks to see if random geometry is
 * in the facet.  Connection validity is checked by ck_insts().
 */
{
    octGenerator gen, lgen;
    octObject geo, lyr;

    /* Catch segments attached to facet */
    OC(octInitGenContents(fct, OCT_GEO_MASK, &gen));
    while (octGenerate(&gen, &geo) == OCT_OK) {
	/* This is a warning itself */
	err(WARN, &geo, SFE, "Geometry attached directly to facet\n");
	ck_geo(&geo);
    }

    /* Now get regulars attached to layers */
    OC(octInitGenContents(fct, OCT_LAYER_MASK, &lgen));
    while (octGenerate(&lgen, &lyr) == OCT_OK) {
	OC(octInitGenContents(&lyr, OCT_GEO_MASK, &gen));
	while (octGenerate(&gen, &geo) == OCT_OK) {
	    ck_geo(&geo);
	}
    }
}


static void ck_geo(geo)
octObject *geo;
/*
 * First, only two-point paths and labels are valid geometry in a
 * symbolic cell.  They must be attached to exactly one layer.  Segments
 * (two-point paths) must be attached to exactly one net and must
 * contain exactly two actual terminals.
 */
{
    switch (geo->type) {
    case OCT_PATH:
	lyr_ck(geo);
	path_ck(geo);
	break;
    case OCT_LABEL:
	lyr_ck(geo);
	break;
    default:
	err(ERR, geo, BGE, "Only paths and labels are allowed in symbolic facets\n");
	break;
    }
}


static void lyr_ck(geo)
octObject *geo;
/*
 * Insures that the geometry is attached once to exactly one layer.
 */
{
    octGenerator gen;
    octObject lyr;
    int attach_count = 0;

    OC(octInitGenContainers(geo, OCT_LAYER_MASK, &gen));
    while (octGenerate(&gen, &lyr) == OCT_OK) {
	attach_count++;
    }
    if (attach_count < 1) {
	err(FATAL, geo, SLF, "Geometry not attached to layer\n");
    } else if (attach_count > 1) {
	err(FATAL, geo, SLF, "Geometry attached to too many layers\n");
    }
}


static void path_ck(path)
octObject *path;
/*
 * This performs path specific error checking.  It makes sure the segment
 * has exactly two points, is attached to exactly one net, and has exactly
 * two actual terminals attached to it.
 */
{
    octGenerator gen;
    octObject net, aterm;
    struct octPoint pnts[2];
    int32 np = 2;
    int net_count, at_count;

    if ((octGetPoints(path, &np, pnts) != OCT_OK) || (np != 2)) {
	err(FATAL, path, SPE, "Segment does not have exactly two points\n");
    }
    net_count = 0;
    OC(octInitGenContainers(path, OCT_NET_MASK, &gen));
    while (octGenerate(&gen, &net) == OCT_OK) {
	net_count++;
    }
    if (net_count < 1) {
	err(FATAL, path, SNF, "Segment not attached to a net\n");
    } else if (net_count > 1) {
	err(FATAL, path, SNF, "Segment attached to too many nets\n");
    }
    at_count = 0;
    OC(octInitGenContents(path, OCT_TERM_MASK, &gen));
    while (octGenerate(&gen, &aterm) == OCT_OK) {
	if (!octIdIsNull(aterm.contents.term.instanceId)) {
	    at_count++;
	}
    }
    if (at_count < 2) {
	err(FATAL, path, STF, "Segment attached to less than two actual terminals\n");
    } else if (at_count > 2) {
	err(FATAL, path, STF, "Segment attached to more than two actual terminals\n");
    }
}


static void ck_nets(fct)
octObject *fct;
/*
 * Checks the net structure of `fct' for problems.  Makes sure only
 * allowed object types are attached to the net.  Insures all segments
 * and terminals attached to the net are reachable by simple visitation.
 * Makes sure nets are non-trival (more than one terminal attached to
 * them).
 */
{
    octGenerator gen;
    octObject net;

    OC(octInitGenContents(fct, OCT_NET_MASK, &gen));
    while (octGenerate(&gen, &net) == OCT_OK) {
	ck_one_net(&net);
    }
}



static void ck_one_net(net)
octObject *net;
/*
 * Checks one net for validity.
 */
{    
    st_table *items;

    items = find_attached(net);
    visit_and_mark(net, items);
    check_items(net, items);
}


static st_table *find_attached(net)
octObject *net;
/*
 * Places all items directly attached to `net' into a hash table
 * that is returned.  The value field is an integer flag representing
 * whether the item has been touched.  As it builds this list, it
 * checks to make sure only the right types of items are attached
 * and that the net is non-trivial.
 */
{
    octGenerator gen;
    octId *key;
    int *value;
    octObject obj;
    st_table *result;
    char buf[1024];
    int term_count;

    result = st_init_table(octIdCmp, octIdHash);
    OC(octInitGenContents(net, OCT_ALL_MASK, &gen));
    term_count = 0;
    while (octGenerate(&gen, &obj) == OCT_OK) {
	switch (obj.type) {
	case OCT_TERM:
	    if (conn_term_p(&obj)) {
		err(ERR, &obj, TNF, "Connector terminal attached to net %s\n",
		    disp_obj(buf, net));
		break;
	    }
	    term_count++;
	    /* Intentional fall-through */
	case OCT_PATH:
	    key = ALLOC(octId, 1);
	    *key = obj.objectId;
	    value = ALLOC(int, 1);
	    *value = 0;
	    (void) st_insert(result, (char *) key, (char *) value);
	    break;
	case OCT_PROP:
	case OCT_BAG:
	    /* Intentionally empty */
	    break;
	default:
	    err(ERR, &obj, "Illegal object type attached to net %s\n",
		disp_obj(buf, net));
	    break;
	}
    }
    if (term_count < 2) {
	err(WARN, net, TNE, "Net with less than two terminals\n");
    }
    return result;
}


static int conn_term_p(term)
octObject *term;
/*
 * Returns a non-zero status if `term' is an actual terminal of a
 * connector instance.  This is checked using the simple bag
 * form of testing.
 */
{
    octObject inst, bag;
    octGenerator gen;

    if ((term->type == OCT_TERM) &&
	(!octIdIsNull(term->contents.term.instanceId))) {
	inst.objectId = term->contents.term.instanceId;
	OC(octGetById(&inst));
	OC(octInitGenContainers(&inst, OCT_BAG_MASK, &gen));
	while (octGenerate(&gen, &bag) == OCT_OK) {
	    if (SEQ(bag.contents.bag.name, "CONNECTORS") == 0) {
		octFreeGenerator(&gen);
		return 1;
	    }
	}
    }
    return 0;
}



static void visit_and_mark(net, items)
octObject *net;
st_table *items;
/*
 * Generates the first item off of `net' and begins searching through
 * the items by connectivity attachment.  Items that are found are
 * marked in `items'.  Objects found that are not in `items' are
 * flagged (unless they are connector terminals).
 *
 * Visitation is tricky.  It first looks through all of the terminals
 * connected to `net' and visits each jump terminal.  Then it
 * generates the first item on the net and visits it.  That should
 * visit all of the reachable items on the net.
 */
{
    octGenerator gen;
    octObject aterm, obj;

    OC(octInitGenContents(net, OCT_TERM_MASK, &gen));
    while (octGenerate(&gen, &aterm) == OCT_OK) {
	if (jump_term_p(&aterm)) {
	    visit_obj(&aterm, items, net);
	}
    }
    if (octGenFirstContent(net, OCT_TERM_MASK|OCT_PATH_MASK, &obj) == OCT_OK) {
	visit_obj(&obj, items, net);
    }
}


static int jump_term_p(term)
octObject *term;
/*
 * Returns a non-zero status if `term' is a jump terminal.  This
 * is checked exhaustively using master lookup.  This is a reasonably
 * costly function.
 */
{
    octGenerator gen;
    octObject inst, itf, ft, prop;

    if (!octIdIsNull(term->contents.term.instanceId)) {
	inst.objectId = term->contents.term.instanceId;
	OC(octGetById(&inst));
	if (get_interface(&inst, &itf) &&
	    (ohGetTerminal(&itf,
			   term->contents.term.formalExternalId,
			   &ft) == OCT_OK)) {
	    OC(octInitGenContents(&ft, OCT_PROP_MASK, &gen));
	    while (octGenerate(&gen, &prop) == OCT_OK) {
		if ((SEQ(prop.contents.prop.name, "GLOBALNET") == 0) ||
		    (SEQ(prop.contents.prop.name, "LOCALNET") == 0)) {
		    octFreeGenerator(&gen);
		    return 1;
		}
	    }
	}
    }
    return 0;
}


static void visit_obj(obj, vo, net)
octObject *obj;			/* Object itself   */
st_table *vo;			/* Visited Objects */
octObject *net;			/* Net (for errors) */
/*
 * This routine looks up `obj' in `vo'.  If it is not in the table and
 * it is not a connector terminal, it is flagged as an error.  If
 * it is in the table but not marked, it is marked and all objects
 * connected to it are visited.  If it is in the table and already
 * marked, the routine returns.
 */
{
    int *value;
    char buf[1024];

    if (st_lookup(vo, (char *) &(obj->objectId), (char **) &value)) {
	/* Found */
	if (*value == 0) {
	    *value = 1;
	    switch (obj->type) {
	    case OCT_TERM:
		visit_term(obj, vo, net);
		break;
	    case OCT_PATH:
		visit_path(obj, vo, net);
		break;
	    default:
		break;
	    }
	}
    } else {
	/* Not found */
	if (conn_term_p(obj)) {
	    visit_term(obj, vo, net);
	} else {
	    err(FATAL, obj, CER, "Object reachable but not attached to %s\n",
		disp_obj(buf, net));
	}
    }
}


static void visit_term(term, vo, net)
octObject *term;
st_table *vo;
octObject *net;
/*
 * Visits all segments attached to `term'.
 */
{
    octGenerator gen;
    octObject seg;

    OC(octInitGenContainers(term, OCT_PATH_MASK, &gen));
    while (octGenerate(&gen, &seg) == OCT_OK) {
	visit_obj(&seg, vo, net);
    }
}

static void visit_path(path, vo, net)
octObject *path;
st_table *vo;
octObject *net;
/*
 * Visits all terminals connected to `path'.
 */
{
    octGenerator gen;
    octObject term;

    OC(octInitGenContents(path, OCT_TERM_MASK, &gen));
    while (octGenerate(&gen, &term) == OCT_OK) {
	visit_obj(&term, vo, net);
    }
}


static enum st_retval check_one(key, value, arg)
char *key, *value, *arg;
/*
 * `value' is actually (int *).  If zero, an error message
 * is generated.
 */
{
    int *marked = (int *) value;
    octId *id = (octId *) key;
    octObject *net = (octObject *) arg;
    octObject obj;
    char buf[1024];

    if (!marked) {
	obj.objectId = *id;
	OC(octGetById(&obj));
	err(FATAL, obj, CER, "Unreachable object attached to net %s\n",
	    disp_obj(buf, net));
    }
    return ST_CONTINUE;
}

static enum st_retval free_net_item(key, value, arg)
char *key, *value, *arg;
/*
 * Reclaims resources.
 */
{
    FREE(key);
    FREE(value);
    return ST_CONTINUE;
}

static void check_items(net, items)
octObject *net;
st_table *items;
/*
 * This routine marches through `items' and looks for objects that
 * weren't touched in the visiting phase.  This objects are flagged
 * as errors.  The table is then reclaimed.
 */
{
    st_foreach(items, check_one, (char *) net);
    st_foreach(items, free_net_item, (char *) 0);
    st_free_table(items);
}


static void ck_fterms(fct)
octObject *fct;
/*
 * Checks for problems with formal terminals.  It makes sure formal
 * terminals are uniquely named, have a non-null implementation (unless
 * they are jump terminals), are connected to exactly one net, and
 * have termtype and direction annotation.
 */
{
    octGenerator gen;
    octObject fterm;
    st_table *term_names;
    char *value;

    term_names = st_init_table(strcmp, st_strhash);
    OC(octInitGenContents(fct, OCT_TERM_MASK, &gen));
    while (octGenerate(&gen, &fterm) == OCT_OK) {
	if (octIdIsNull(fterm.contents.term.instanceId)) {
	    if (st_lookup(term_names, fterm.contents.term.name, &value)) {
		err(ERR, &fterm, DTN, "Duplicate formal terminal name\n");
	    } else {
		(void) st_insert(term_names, fterm.contents.term.name,
				 (char *) 0);
	    }
	} else {
	    err(FATAL, &fterm, UOT, "Actual terminal attached to facet\n");
	}
	ck_one_term(&fterm);
    }
    st_free_table(term_names);
}


char *tt_vals[] = { "SIGNAL", "GROUND", "CLOCK", "SUPPLY", (char *) 0 };
char *dir_vals[] = { "INPUT", "OUTPUT", "INOUT", (char *) 0 };

static void ck_one_term(ft)
octObject *ft;
/*
 * Checks a formal terminal for correctness.  Makes sure the terminal
 * has a non-null implementation and the implementation is of the right
 * type, makes sure it is connected to exactly one terminal, and
 * produces warnings if there is no termtype or direction annotations.
 */
{
    octGenerator gen;
    octObject obj, prop;
    int impl_count, net_count;
    char buf[1024];

    impl_count = 0;
    OC(octInitGenContents(ft, OCT_ALL_MASK, &gen));
    while (octGenerate(&gen, &obj) == OCT_OK) {
	switch (obj.type) {
	case OCT_TERM:
	case OCT_PATH:
	    impl_count++;
	    break;
	case OCT_PROP:
	case OCT_BAG:
	    /* Empty intentionally */
	    break;
	default:
	    err(ERR, &obj, FIE, "Unexpected object attached to formal terminal %s\n",
		disp_obj(buf, ft));
	    break;
	}
    }
    if (impl_count < 1) {
	/* May be a warning if it doesn't have a global net property */
	prop.type = OCT_PROP;
	prop.contents.prop.name = "GLOBALNET";
	if (octGetByName(ft, &prop) != OCT_OK) {
	    err(WARN, ft, FIE, "Terminal has no implementation\n");
	}
    }
    net_count = 0;
    OC(octInitGenContainers(ft, OCT_NET_MASK, &gen));
    while (octGenerate(&gen, &obj) == OCT_OK) {
	net_count++;
    }
    if (net_count < 1) {
	err(ERR, ft, FNN, "No net associated with terminal\n");
    } else if (net_count > 1) {
	err(ERR, ft, FNN, "More than one net associated with terminal\n");
    }

    ck_eprop(ft, "TERMTYPE", tt_vals, 1);
    ck_eprop(ft, "DIRECTION", dir_vals, 1);

}


static void ck_eprop(parent, prop_name, prop_vals, present_p)
octObject *parent;		/* Object to search */
char *prop_name;		/* Property name    */
char *prop_vals[];		/* Property values  */
int present_p;			/* Presence required */
/*
 * Checks `parent' for the property named `prop_name'.  This
 * property must be string valued and have one of the values
 * given in `prop_vals'.  If `present_p' is non-zero, the
 * property must exist.  All errors are warnings.
 */
{
    octObject prop;
    char **i;

    prop.type = OCT_PROP;
    prop.contents.prop.name = prop_name;
    if (octGetByName(parent, &prop) == OCT_OK) {
	if (prop.contents.prop.type == OCT_STRING) {
	    for (i = prop_vals;  *i;  i++) {
		if (SEQ(prop.contents.prop.value.string, *i) == 0) {
		    return;
		}
	    }
	    err(WARN, parent, BPV, "%s property has unknown value %s\n",
		prop_name, prop.contents.prop.value.string);
	} else {
	    err(WARN, parent, BPV, "%s property is not a string\n",
		prop_name);
	}
    } else if (present_p) {
	err(WARN, parent, BPV, "Missing %s property\n", prop_name);
    }
}



static struct ll *ll_remove(list, func)
struct ll *list;
void (*func)();
{
    if (list) {
	ll_remove(list->tail, func );
	if (func) (*func)(list->data);
	FREE(list);
    }
    return NILL;
}


static void zap_impl(data)
char *data;
/* Called on layer/implemenation structures */
{
    term_impl *impl = (term_impl *) data;

    FREE(impl);
}

static void zap_info(data)
char *data;
/* This is called on terminal information structures */
{
    term_info *info = (term_info *) data;

    CUT(info->impl, zap_impl);
    FREE(info);
}

static void zap_term_info(data)
char *data;
/* This is called on a list of terminal information structures */
{
    CUT((LL) data, zap_info);
}



/*VARARGS4*/
static void err(va_alist)
va_dcl
/*
 * err(sev, obj, errtype, format, ... )
 */
{
    va_list ap;
    char buf[1024];
    int sev, errtype;
    octObject *obj;
    char *format;

    va_start(ap);
    sev = va_arg(ap, int);
    obj = va_arg(ap, octObject *);
    errtype = va_arg(ap, int);
    format = va_arg(ap, char *);

    /* Duplicate error message suppression */
    if (supress(obj, errtype)) return;
    if (sev > WARN) err_count++;
    (*err_func.err_func)(sev, errtype, obj, uprintf(buf, format, &ap),
			 err_func.user_data);
    va_end(ap);
}


static st_table *slf_errors = (st_table *) 0;

static int supress(obj, errtype)
octObject *obj;
int errtype;
/*
 * Some identical errors can be generated in many places.  This
 * routine recognizes duplicates and supresses all but the first.
 */
{
    octId *id;
    int result = 0;

    switch (errtype) {
    case SLF:
	if (!st_is_member(slf_errors, (char *) &(obj->objectId))) {
	    id = ALLOC(octId, 1);
	    *id = obj->objectId;
	    (void) st_insert(slf_errors, (char *) id, (char *) 0);
	} else {
	    result = 1;
	}
	break;
    default:
	/* No suppression */
	break;
    }
    return result;
}

static enum st_retval remSlf(key, value, arg)
char *key, *value, *arg;
{
    FREE(key);
    return ST_CONTINUE;
}

static void supress_reset()
/*
 * Resets any state associated with duplicate error
 * suppression
 */
{
   if (slf_errors) {
	(void) st_foreach(slf_errors, remSlf, (char *) 0);
	(void) st_free_table(slf_errors);
   }
   slf_errors = st_init_table(octIdCmp, octIdHash);
}


static void def_err(sev, type, obj, msg, data)
int sev;
int type;
octObject *obj;
char *msg;
char *data;
/*
 * The default error reporting function prints a message to the
 * file descriptor given in the `data' field.
 */
{
    char buf[1024];
    char* head;

    switch (sev) {
    case WARN:
	head = "warning: ";
	break;
    case ERR:
	head = "Error: ";
	break;
    case FATAL:
	head =  "FATAL: ";
	break;
    }
    if (obj) {
	(void) sprintf( buf, "%s %s: %s", head, ohFormatName( obj ), msg);
    } else {
	(void) sprintf( buf, "%s: %s", head, msg);
    }
    symcleanReportError( obj, buf );
}



static char *disp_obj(buf, obj)
char *buf;
octObject *obj;
/*
 * Produces a human readable form of `obj' and writes it into `buf'.
 */
{
    char xid_buf[128];
    int32 xid;
    
    octExternalId(obj, &xid);
    (void) sprintf(xid_buf, " [%d]", xid);
    (void) strcpy(buf, ohFormatName(obj));
    (void) strcat(buf, xid_buf);
    return buf;
}
