#include "copyright.h"
/*
 * program to clean up various problems in oct symbolic facets
 *
 *   Rick L Spickelmier, 4/2/88
 *
 *
 * does the following:
 *
 * delete floating nets
 * delete nets with only one terminal and no paths
 * report multiple nets with the same name (excluding NULL and "")
 * delete formal terminals with no implementation
 * report formal terminals contained by more than one net
 * attach formal terminals to nets
 * report multiple formal terminals with the same name
 * delete bags left by VEM
 * report segments (paths) with more than 2 terminals
 * delete segments with fewer than 2 terminals
 * report segments with more than 1 net
 * delete segments with no containing nets
 * report connectors that are of the wrong type
 * report connectors that are not physically on the ends of paths
 * detach connector terminals from nets
 * detach connectors from INSTANCES bag
 * attach connectors to CONNECTORS bag
 * delete floating connector instances
 * report missing interface/contents facets
 * report actual/formal terminal mismatches
 * report actual terminals connected to more than 1 net and merge the nets
 * report actual terminals promoted to multiple formal terminals
 * merge through JOINED bags
 * verifies interface/contents terminal consistency
 *
 * report segments with both terminals on the same instance (might create
 * a sparcs overconstraint)
 *
 * optional:
 * checks for actual terminals of real instances not connected to nets
 *
 * runs vulcan (-k if necessary)
 *
 */

#include "copyright.h"
#include "port.h"
#include "oct.h"
#include "st.h"
#include "oh.h"
#include "vulcan.h"
#include "options.h"
#include "oh.h"
#include "region.h"
#include "utility.h"
#include "tap.h"

#define AVEX_L(bb) ((((bb).lowerLeft.x + (bb).upperRight.x) / 2 ) / get_lambda(0))
#define AVEY_L(bb) ((((bb).lowerLeft.y + (bb).upperRight.y) / 2) / get_lambda(0))

#define AVEX(bb) ((((bb).lowerLeft.x + (bb).upperRight.x) / 2 ))
#define AVEY(bb) ((((bb).lowerLeft.y + (bb).upperRight.y) / 2))


#define SYM_OCT_ASSERT(fnc) \
  OH_ASSERT_DESCR(fnc, "SYMCLEAN: serious Oct error, aborting without updating the facet")


#define PROPEQ(p,s) (!strcmp((p)->contents.prop.value.string,(s)))


static char buffer[1024];


int get_lambda( obj )
    octObject *obj;
{
    static int l = 0;
    if ( l == 0 ) {
	tapGetProp( obj, TAP_LAMBDA );
	l = TAP_LAMBDA->value.integer;
    }
    if ( l == 0 ) {
	printf( "Lambda not set: return default 20\n");
	return 20;
    }
    return l;
}

static int
countAttachments(object, type_mask, direction)
    octObject *object;
    octObjectMask type_mask;
    int direction;
{
    octGenerator gen;
    octObject dummy;
    int count = 0;
    
    if (direction == 1) {
	OH_ASSERT(octInitGenContentsSpecial(object, type_mask, &gen));
    } else {
	OH_ASSERT(octInitGenContainersSpecial(object, type_mask, &gen));
    }

    while (octGenerate(&gen, &dummy) == OCT_OK) {
	count++;
    }
    
    return count;
}


static int
betterName(new, old)
octObject *new, *old;
{
    static char *goodNames[] = {"Vdd", "VDD", "Vdd!", "GND!", "GND", "Gnd"};
    int i;

    if (new->contents.net.name == NIL(char)) {
	return 0;
    }
    if (new->contents.net.name[0] == '\0') {
	return 0;
    }
    if (old->contents.net.name == NIL(char)) {
	return 1;
    }
    if (old->contents.net.name[0] == '\0') {
	return 1;
    }
    for (i = 0; i < sizeof(goodNames) / sizeof(char *); i++) {
	if (strcmp(new->contents.net.name, goodNames[i]) == 0) {
	    return 1;
	}
    }
    return 0;
}


static void (*errReportFunc)() = 0;

void symcleanSetErrorReport( f )
    void (*f)();
{
    errReportFunc = f;
}

void symcleanReportError( obj, msg )
    octObject* obj;
    char* msg;
{
    if ( errReportFunc ) {
	(*errReportFunc)( obj, msg );
    }
}



static void
mergeNets(net, oldnet)
octObject *net, *oldnet;
{
    octObject object;
    octGenerator gen;
    
    if (octIdsEqual(net->objectId, oldnet->objectId)) {
	return;
    }
    if (betterName(oldnet, net)) {
	char *temp;
	
	temp = net->contents.net.name;
	net->contents.net.name = oldnet->contents.net.name;
	oldnet->contents.net.name = temp;
	OH_ASSERT(octModify(net));
    }
    
    /* PRINTF("\tmerging net `%s' into net `%s'\n",
	           oldnet->contents.net.name, net->contents.net.name); */
    OH_ASSERT(octInitGenContents(oldnet, OCT_ALL_MASK, &gen));
    while (octGenerate(&gen, &object) == OCT_OK) {
	OH_ASSERT(octDetach(oldnet, &object));
	if (octIsAttached(net, &object) != OCT_OK) {
	    OH_ASSERT(octAttach(net, &object));
	}
    }
    return;
}


static void
mergeNetsOnTerminal( term )
    octObject *term;
{
    octObject net, oldnet;
    octGenerator gen;
    
    OH_ASSERT(octInitGenContainers(term, OCT_NET_MASK, &gen));
    OH_ASSERT(octGenerate(&gen, &net));
    while (octGenerate(&gen, &oldnet) == OCT_OK) {
	if (octIdsEqual(oldnet.objectId, net.objectId)) {
	    symcleanReportError( &oldnet,"same net connected multiple times\n");
	    OH_ASSERT(octDetach(&oldnet, term));
	    continue;
	}
	mergeNets(&net, &oldnet);
    }
    return;
}

/*
 * check that the two terminals on the segment actually intersect
 * the end points of the segment on the proper layer
 */
static int checkPath( path, fixp ) 
    octObject* path;
    int fixp;
{
    struct octPoint* points;
    int32 num;
    int errcount = 0;
    int lambda  = get_lambda( path );

    ohGetPoints(path, &num, &points);

    if ( num != 2 ) {
	sprintf( buffer, "%d points in path: deleting",num );
	errcount++;
	if ( fixp ) {
	    symcleanReportError( 0, buffer );
	    octDelete( path );
	} else {
	    symcleanReportError( path, buffer );
	}
    } else {
	int sx = points[0].x - points[1].x;
	int sy = points[0].y - points[1].y;
	if ( sx * sy ) {
	    errcount++;
	    sprintf( buffer, "Non manhattan path (%d,%d)->(%d,%d)\n",
		    points[0].x/lambda,points[0].y/lambda,
		    points[1].x/lambda,points[1].y/lambda );
	    symcleanReportError( path, buffer );
	    if ( fixp ) {
		if ( ABS(sx) < ABS(sy) ) {
		    if ( (ABS(sx) / (double)ABS(sy))  < 0.25 ) {
			points[1].x += sx;
			octPutPoints( path, 2, points );
		    }
		} else {
		    if ( (ABS(sy) / (double)ABS(sx)) < 0.25 ) {
			points[1].y += sy;
			octPutPoints( path, 2, points );
		    }
		}
	    }
	}
	if ( (sx + sy) == 0 ) {
	    errcount++;
	    sprintf( buffer, "Zero length path (%d,%d)\n",
		    points[0].x/lambda,points[0].y/lambda );
	    if ( fixp ) {
		symcleanReportError( 0, buffer );
		octDelete( path );
	    } else {
		symcleanReportError( path, buffer );
	    }
	}
    }
    return errcount;
}



symcleanFormals( facet, fixp )
    octObject* facet;
    int fixp;
{
    octObject interface, net, bag, term, prop;
    octObject aterm, layer, path, geo, master, instance;
    octObject fterm, joined, joinedterm, ajoinedterm, oldnet;
    struct octBox bb, origBB, finalBB;
    octGenerator gen, gen2, gen3;
    st_table *TermTable, *NetTable;
    int cnt, oncethrough = 0, schematicp = 0;
    int modifyp = 0, keepformals = 0, fullvulcanp = 0;
    int actualp = 0;
    int errcount = 0;

    TermTable = st_init_table(strcmp, st_strhash);

    OH_ASSERT(octInitGenContents(facet, OCT_TERM_MASK, &gen));
    while (octGenerate(&gen, &term) == OCT_OK) {
	/* Check for formal terminals with no implementation */
	if (countAttachments(&term, OCT_TERM_MASK, 1) == 0) {
	    errcount++;
	    sprintf(buffer,"Unimplemented formal terminal: %s\n", term.contents.term.name);
	    symcleanReportError( 0, buffer );
	    if ( fixp ) OH_ASSERT(octDelete(&term));
	    continue;
	}

	/* Check for more than one net containing the formal terminal */
	   
	if (countAttachments(&term, OCT_NET_MASK, 0) > 1) {
	    errcount++;
	    sprintf(buffer,"%s connected to more than 1 net\n", ohFormatName( &term ));
	    symcleanReportError( &term, buffer );
	    if ( fixp ) mergeNetsOnTerminal(&term);
	    continue;
	}


	/*
	 * connect up a formal terminal if not connected to a net
	 * (not necessary if the formal is implemented by the actual
	 * terminal of a cell instance and has no interconnect attached
	 * to it)
	 */
	if (countAttachments(&term, OCT_NET_MASK, 0) == 0) {
	    /* The term must have an implementation */
	    OH_ASSERT(octGenFirstContent(&term, OCT_TERM_MASK, &aterm));
	    /* 
	     * shouldn't really be done, but does not hurt and
	     * will help mosaico
	     */
	    if (octGenFirstContainer(&aterm, OCT_NET_MASK, &net) == OCT_OK) {
		sprintf(buffer,"Attaching formal terminal `%s' to net `%s'\n",
		       term.contents.term.name,
		       net.contents.net.name);
		symcleanReportError( &term, buffer );
		if ( fixp ) OH_ASSERT(octAttachOnce(&net, &term));
		continue;
	    }
	    
	    if ( octGenFirstContainer( &aterm, OCT_PATH_MASK, &path ) == OCT_OK ) {
		if ( octGenFirstContainer( &path, OCT_NET_MASK, &net ) == OCT_OK ) {
		    sprintf(buffer,"Attaching formal terminal `%s' to net `%s'\n",
			    term.contents.term.name,
			    net.contents.net.name);
		    symcleanReportError( &term, buffer );
		    if ( fixp ) OH_ASSERT(octAttachOnce(&net, &term));
		    continue;
		} else {
		    /* Path not attached to net. */
		    sprintf( buffer, "Path not attached to any net." );
		    if ( fixp ) {
			symcleanReportError( 0, buffer );
			octDelete( &path );
		    } else {
			symcleanReportError( &path, buffer );
		    }
		}
	    }
	}
	
	/* Check for multiple formal terminals with the same name */

	if (st_lookup(TermTable, term.contents.term.name, NIL(char *))) {
	    sprintf(buffer,"Warning: formal terminal name `%s' exists multiple times\n",
		   term.contents.term.name);
	    symcleanReportError( &term, buffer );
	} else {
	    (void) st_insert(TermTable, term.contents.term.name, NIL(char));
	}
    }
    st_free_table( TermTable );
    return errcount;
}

int symcleanNets( facet, fixp )
    octObject* facet;
    int fixp;
{

    octObject interface, net, bag, term, prop;
    octObject aterm, layer, path, geo, master, instance;
    octObject fterm, joined, joinedterm, ajoinedterm, oldnet;
    struct octBox bb, origBB, finalBB;
    octGenerator gen, gen2, gen3;
    st_table  *NetTable;
    int cnt, again = 0, oncethrough = 0, schematicp = 0;
    int option;
    extern int optind;
    extern char *optarg;
    int modifyp = 0, keepformals = 0, vulcanp = 1, fullvulcanp = 0;
    int actualp = 0, strictp = 1;
    int errcount = 0;


    OH_ASSERT(octInitGenContents( facet, OCT_NET_MASK, &gen));
    NetTable = st_init_table(strcmp, st_strhash);
    while (octGenerate(&gen, &net) == OCT_OK) {
	    
	/* Check for floating nets */
	    
	if (countAttachments(&net, OCT_ALL_MASK, 1) == 0) {
	    sprintf(buffer,"deleting a floating net (%s)\n", net.contents.net.name);
	    symcleanReportError( 0, buffer );
	    errcount++;
	    if ( fixp ) OH_ASSERT(octDelete(&net));
	    continue;
	}
	

	/* Check for a net with only 1 terminal */
	
	if ((countAttachments(&net, OCT_TERM_MASK, 1) < 2) &&
	    (countAttachments(&net, OCT_PATH_MASK, 1) == 0)) {
	    errcount++;
	    sprintf(buffer,"deleting a net without geometry and fewer than two terminals (%s)\n", net.contents.net.name);
	    symcleanReportError( 0, buffer );
	    if (fixp) OH_ASSERT(octDelete(&net));
	    continue;
	}

	/* Check for multiple nets with the same name */
	
	if ((net.contents.net.name != NIL(char)) &&
	    ( net.contents.net.name[0] != '\0')) {
	    if (st_lookup(NetTable, net.contents.net.name, NIL(char *))) {
		errcount++;
		sprintf(buffer,"Warning: net name `%s' exists multiple times\n",
		       net.contents.net.name);
		symcleanReportError( 0, buffer );
	    } else {
		(void) st_insert(NetTable, net.contents.net.name, NIL(char));
	    }
	}
	
    }
    st_free_table( NetTable );
    return errcount;
}


int symcleanInstances( facet, fixp )
    octObject *facet;
    int fixp;
{
    octObject interface, net, bag, term, prop;
    octObject aterm, layer, path, geo, master, instance;
    octObject fterm, joined, joinedterm, ajoinedterm, oldnet;
    octObject instancesbag, connectorsbag;
    struct octBox bb, origBB, finalBB;
    octGenerator gen, gen2, gen3;
    st_table *TermTable, *NetTable;
    int cnt, again = 0, oncethrough = 0, schematicp = 0;
    int option;
    extern int optind;
    extern char *optarg;
    int modifyp = 0, keepformals = 0, vulcanp = 1, fullvulcanp = 0;
    int actualp = 0, strictp = 1;
    int errcount = 0;
    int lambda = get_lambda( facet );


    OH_ASSERT(octInitGenContentsSpecial( facet, OCT_INSTANCE_MASK, &gen));
    while (octGenerate(&gen, &instance) == OCT_OK) {

	/* look for contents and interface facets */
	master.type = OCT_FACET;
	master.contents.facet.version = instance.contents.instance.version;
	master.contents.facet.facet = "contents";
	master.contents.facet.mode = "r";
	if (octOpenMaster(&instance, &master) < OCT_OK) {
	    sprintf(buffer,"The contents facet of %s:%s does not exist\n",
		   instance.contents.instance.master,
		   instance.contents.instance.view);
	    symcleanReportError( &instance, buffer );
	}
	master.contents.facet.facet = "interface";
	master.contents.facet.mode = "r"; /* XXX Oct 2.0 trashes the mode */
	if (octOpenMaster(&instance, &master) < OCT_OK) {
	    sprintf(buffer,"The interface facet of %s:%s does not exist\n",
		   instance.contents.instance.master,
		   instance.contents.instance.view);
	    symcleanReportError( &instance, buffer );
	}

	/* look for connectors incorrectly attached
	 *   detach from INSTANCES bag, attach to CONNECTORS bag,
	 *   again = 1 to handle detaching actual terminal
	 */
	ohGetOrCreateBag( facet, &instancesbag, "INSTANCES" );
	ohGetOrCreateBag( facet, &connectorsbag, "CONNECTORS" );
	if (ohGetByPropName(&master, &prop, "CELLTYPE") == OCT_OK) {
	    if (strcmp(prop.contents.prop.value.string, "CONTACT") == 0) {
		if (octIsAttached(&instancesbag, &instance) == OCT_OK) {
		    OH_ASSERT(octDetach(&instancesbag, &instance));
		    sprintf(buffer,"Detaching connector from INSTANCES bag\n");
		    symcleanReportError( &instance, buffer );
		}
		if (octIsAttached(&connectorsbag, &instance) != OCT_OK) {
		    OH_ASSERT(octAttach(&connectorsbag, &instance));
		    sprintf(buffer,"Attaching connector to CONNECTORS bag\n");
		    symcleanReportError( &instance, buffer );
		}
	    }
	}
 	    
	OH_ASSERT(octInitGenContentsSpecial(&instance, OCT_TERM_MASK, &gen2));
	while (octGenerate(&gen2, &aterm) == OCT_OK) {
	    /* look for actual terminals promoted to multiple formals */
	    if (countAttachments(&aterm, OCT_TERM_MASK, 0) > 1) {
		sprintf(buffer,"More than one formal terminal attached to %s of %s (%s:%s) at (%d,%d)\n",
		       aterm.contents.term.name,
		       instance.contents.instance.name,
		       instance.contents.instance.master,
		       instance.contents.instance.view,
		       instance.contents.instance.transform.translation.x/lambda,
		       instance.contents.instance.transform.translation.y/lambda);
		symcleanReportError( &aterm, buffer );

	    }
	    /* look for actual terminals connected to more than one net */
	    if (countAttachments(&aterm, OCT_NET_MASK, 0) > 1) {
		sprintf(buffer,"More than 1 net attached to %s of %s (%s:%s) at (%d,%d)\n",
		       aterm.contents.term.name,
		       instance.contents.instance.name,
		       instance.contents.instance.master,
		       instance.contents.instance.view,
		       instance.contents.instance.transform.translation.x/lambda,
		       instance.contents.instance.transform.translation.y/lambda);
		symcleanReportError( &aterm, buffer );
		mergeNetsOnTerminal(&aterm);
		again = 1;
		continue;
	    }		    

	    /* look for actual terminals not on any net */
	    if (!oncethrough 
		&& actualp
		&& (octIsAttached(&instancesbag, &instance) == OCT_OK)
		&& (countAttachments(&aterm, OCT_NET_MASK, 0) == 0)) {
		sprintf( buffer, "Warning: actual terminal %s of %s (%s:%s) at (%d,%d) is not connected to a net\n",
		       aterm.contents.term.name,
		       instance.contents.instance.name,
		       instance.contents.instance.master,
		       instance.contents.instance.view,
		       instance.contents.instance.transform.translation.x/lambda,
		       instance.contents.instance.transform.translation.y/lambda);
		symcleanReportError( &aterm, buffer );
	    }
		
	    /* look for corresponding formal */
	    if (ohGetByTermName(&master, &fterm, aterm.contents.term.name) < OCT_OK) {
		sprintf(buffer,"Master terminal %s of %s (%s:%s) has disappeared\n",
		       aterm.contents.term.name,
		       instance.contents.instance.name,
		       instance.contents.instance.master,
		       instance.contents.instance.view);
		symcleanReportError( &aterm, buffer );
		continue;
	    }

	    /* handle joined terminals */
		
	    /* get the JOINED bag */
	    joined.type = OCT_BAG;
	    joined.contents.bag.name = "JOINED";
	    if (octGetContainerByName(&fterm, &joined) != OCT_OK) {
		continue;
	    }
	    /* get the net attached to the original aterm */
	    if (octGenFirstContainer(&aterm, OCT_NET_MASK, &net) != OCT_OK) {
		continue;
	    }
	    /* generate all fterms in the joined bag */
	    OH_ASSERT(octInitGenContentsSpecial(&joined, OCT_TERM_MASK, &gen3));
	    while (octGenerate(&gen3, &joinedterm) == OCT_OK) {
		if (octIdsEqual(joinedterm.objectId, fterm.objectId)) {
		    continue;
		}
		/* get the corresponding aterm for the fterm */
		if (ohGetByTermName(&instance, &ajoinedterm, joinedterm.contents.term.name) < OCT_OK) {
		    sprintf(buffer,"Warning: Can't find the actual terminal `%s' of `%s' (%s:%s)\n",
			   ajoinedterm.contents.term.name,
			   instance.contents.instance.name,
			   instance.contents.instance.master,
			   instance.contents.instance.view);
		    symcleanReportError( &ajoinedterm, buffer );
		    continue;
		}
		/* get the net attached to the aterm */
		if (octGenFirstContainer(&ajoinedterm, OCT_NET_MASK, &oldnet) != OCT_OK) {
		    sprintf(buffer,"Joined terminal `%s' of `%s' (%s:%s) is not attached to a net, attaching to net `%s'\n",
			   ajoinedterm.contents.term.name,
			   instance.contents.instance.name,
			   instance.contents.instance.master,
			   instance.contents.instance.view,
			   net.contents.net.name);
		    symcleanReportError( &ajoinedterm, buffer );

		    OH_ASSERT(octAttachOnce(&net, &ajoinedterm));
		    continue;
		} else {
		    if (octIdsEqual(net.objectId, oldnet.objectId)) {
			continue;
		    }
		    sprintf(buffer,"Joining net `%s' and net `%s'\n",
			   net.contents.net.name, oldnet.contents.net.name);
		    symcleanReportError( 0, buffer );
		    mergeNets(&net, &oldnet);
		    again = 1;
		    continue;
		}
	    }
	}
    }

    OH_ASSERT(octInitGenContentsSpecial(&connectorsbag, OCT_INSTANCE_MASK, &gen));
    while (octGenerate(&gen, &instance) == OCT_OK) {
	/* just to be safe, somebody managed to get instances
	 * in the policy bags, but not on the facet ... */
	OH_ASSERT(octAttachOnce( facet, &instance));
    }

    OH_ASSERT(octInitGenContentsSpecial(&instancesbag, OCT_INSTANCE_MASK, &gen));
    while (octGenerate(&gen, &instance) == OCT_OK) {
	/* just to be safe, somebody managed to get instances
	 * in the policy bags, but not on the facet ... */
	OH_ASSERT(octAttachOnce( facet, &instance));
    }


    OH_ASSERT(octInitGenContentsSpecial(&connectorsbag, OCT_INSTANCE_MASK, &gen));
    while (octGenerate(&gen, &instance) == OCT_OK) {
	OH_ASSERT(octBB(&instance, &bb));

	OH_ASSERT(octGenFirstContent(&instance, OCT_TERM_MASK, &term));
	    
	/* Check for connector terminal on net */
	if (octGenFirstContainer(&term, OCT_NET_MASK, &net) == OCT_OK) {
	    sprintf(buffer,"Detaching connector from net\n");
	    symcleanReportError( &term, buffer );
	    OH_ASSERT(octDetach(&net, &term));
	    again = 1;
	}

	/* Check for lonely connectors */
	if (countAttachments(&term, OCT_PATH_MASK|OCT_TERM_MASK, 0) == 0) {
	    sprintf(buffer,"Deleting the connector instance at (%d,%d)\n\tno paths or formal terminals attached to its terminal\n",  AVEX_L(bb), AVEY_L(bb));
	    symcleanReportError( 0, buffer );
	    if ( fixp ) OH_ASSERT(octDelete(&instance));
	}
    }


    return errcount;
}	


int symcleanSegments( facet, fixp)
    octObject* facet;
    int fixp;
{
    octObject interface, net, bag, term, prop;
    octObject aterm, layer, path, geo, master, instance;
    octObject fterm, joined, joinedterm, ajoinedterm, oldnet;
    struct octBox bb, origBB, finalBB;
    octGenerator gen, gen2, gen3;
    int cnt;
    int errcount = 0;
    int lambda = get_lambda( facet );

    OH_ASSERT(octInitGenContentsSpecial( facet, OCT_LAYER_MASK, &gen));
    while (octGenerate(&gen, &layer) == OCT_OK) {
	/* check for garbage geos on in symbolic views */
	OH_ASSERT(octInitGenContentsSpecial(&layer, OCT_GEO_MASK&~OCT_PATH_MASK&~OCT_LABEL_MASK, &gen2));
	while (octGenerate(&gen2, &geo) == OCT_OK) {
	    sprintf(buffer,"Bad geometry: %s\n", ohFormatName(&geo));
	    symcleanReportError( &geo, buffer );
	}
	    
	OH_ASSERT(octInitGenContents(&layer, OCT_PATH_MASK, &gen3));
	while (octGenerate(&gen3, &path) == OCT_OK) {
	    symcleanDoPath( &path, fixp );
	}
    }
    return errcount;
}


int symcleanLayers( facet, fixp)
    octObject* facet;
    int fixp;
{
    octObject layer,  geo;
    octGenerator gen, gen2;
    int errcount = 0;

    OH_ASSERT(octInitGenContentsSpecial( facet, OCT_LAYER_MASK, &gen));
    while (octGenerate(&gen, &layer) == OCT_OK) {
	/* check for garbage geos on in symbolic views */
	OH_ASSERT(octInitGenContentsSpecial(&layer, OCT_GEO_MASK&~OCT_PATH_MASK&~OCT_LABEL_MASK, &gen2));
	while (octGenerate(&gen2, &geo) == OCT_OK) {
	    errcount++;
	    sprintf(buffer,"Bad geometry on layer %s %s\n", ohFormatName(&geo));
	    symcleanReportError( &geo, buffer );
	}
    }
    return errcount;
}


int symcleanGarbage( facet, fixp )
    octObject* facet;
    int fixp;
{
    octObject interface, net, bag, term, prop;
    octObject aterm, layer, path, geo, master, instance;
    octObject fterm, joined, joinedterm, ajoinedterm, oldnet;
    struct octBox bb, origBB, finalBB;
    octGenerator gen, gen2, gen3;
    int cnt, again = 0, oncethrough = 0, schematicp = 0;
    int modifyp = 0, keepformals = 0, vulcanp = 1, fullvulcanp = 0;
    int actualp = 0, strictp = 1;
    int errcount = 0;

    /* Check for the VEM garbage bags */
    OH_ASSERT(octInitGenContents( facet, OCT_BAG_MASK, &gen));
    while (octGenerate(&gen, &bag) == OCT_OK) {
	if (strncmp(bag.contents.bag.name, ";;; ", 4) == 0) {
	    if ( octIsTemporary( &bag ) ) continue;
	    errcount++;
	    sprintf(buffer,"Garbage bag: %s\n", bag.contents.bag.name);
	    symcleanReportError( 0, buffer );
	    /* check for dummy geos to delete - really screws up bb */
	    OH_ASSERT(octInitGenContents(&bag, OCT_GEO_MASK, &gen2));
	    while (octGenerate(&gen2, &geo) == OCT_OK) {
		errcount++;
		if (octGenFirstContainer(&geo, OCT_LAYER_MASK, &layer) == OCT_GEN_DONE) {
		    sprintf(buffer,"\tdeleting dummy geo attached to the bag\n");
		    symcleanReportError( 0, buffer );
		    if ( fixp ) OH_ASSERT(octDelete(&geo));
		}
	    }
	    if ( fixp ) OH_ASSERT(octDelete(&bag));
	}
    }
    return errcount;
}

int symcleanAll( facet, fixp )
    octObject* facet;
    int fixp;
{
    octObject interface, net, bag, term, prop;
    octObject aterm, layer, path, geo, master, instance;
    octObject fterm, joined, joinedterm, ajoinedterm, oldnet;
    struct octBox bb, origBB, finalBB;
    octGenerator gen, gen2, gen3;
    st_table *TermTable, *NetTable;
    int cnt, again = 0, oncethrough = 0, schematicp = 0;
    int option;
    extern int optind;
    extern char *optarg;
    int modifyp = 0, keepformals = 0, vulcanp = 1, fullvulcanp = 0;
    int actualp = 0, strictp = 1;
    int errcount = 0;

    do {

	if (again) {
	    oncethrough = 1;
	}
	again = 0;
	errcount = 0;
	
	errcount += symcleanNets( facet, fixp );

	errcount += symcleanFormals( facet, fixp );

	errcount += symcleanSegments( facet, fixp);

	errcount += symcleanInstances( facet, fixp );

	errcount += symcleanGarbage( facet, fixp );

	errcount += scCheck( facet );




    } while (again);

    return errcount;
}


int symcleanInterface( facet, fixp )
    octObject* facet;
    int fixp;
{
    octObject interface, net, bag, term, prop;
    octObject aterm, layer, path, geo, master, instance;
    octObject fterm, joined, joinedterm, ajoinedterm, oldnet;
    struct octBox bb, origBB, finalBB;
    octGenerator gen, gen2, gen3;
    st_table *TermTable, *NetTable;
    int cnt, again = 0, oncethrough = 0, schematicp = 0;
    int option;
    extern int optind;
    extern char *optarg;
    int modifyp = 0, keepformals = 0, vulcanp = 1, fullvulcanp = 0;
    int actualp = 0, strictp = 1;




    /* final paranoid check */
    interface = *facet;
    interface.contents.facet.facet = "interface";
    interface.contents.facet.mode = "r";

    if (octOpenFacet(&interface) == OCT_OLD_FACET) {

	OH_ASSERT(octInitGenContentsSpecial( facet, OCT_TERM_MASK, &gen));
	while (octGenerate(&gen, &term) == OCT_OK) {
	    if (octGetByName(&interface, &term) != OCT_OK) {
		fullvulcanp = 1;
		sprintf(buffer,"Terminal `%s' is in the contents but not in the interface\n",
		       term.contents.term.name);
		symcleanReportError( 0, buffer );
	    }
	}

	OH_ASSERT(octInitGenContentsSpecial(&interface, OCT_TERM_MASK, &gen));
	while (octGenerate(&gen, &term) == OCT_OK) {
	    if (octGetByName( facet, &term) != OCT_OK) {
		fullvulcanp = 1;
	    sprintf(buffer,"Terminal `%s' is in the interface but not in the contents\n",
		   term.contents.term.name);
		symcleanReportError( 0, buffer );
	    }
	}
    
	OH_ASSERT(octCloseFacet(&interface));
    }

    if (strictp) {
	OH_ASSERT(octBB(facet, &finalBB));

	if ((origBB.lowerLeft.x != finalBB.lowerLeft.x) ||
	    (origBB.lowerLeft.y != finalBB.lowerLeft.y) ||
	    (origBB.upperRight.x != finalBB.upperRight.x) ||
	    (origBB.upperRight.y != finalBB.upperRight.y)) {
	    sprintf(buffer,"Warning: bounding box of the facet changed\n");
	    symcleanReportError( 0, buffer );
	}
    }

    if (modifyp && vulcanp) {
	interface = *facet;
	interface.contents.facet.facet = "interface";

	if (fullvulcanp) {
	    interface.contents.facet.mode = "w";
	} else {
	    interface.contents.facet.mode = "a";
	}
	
	OH_ASSERT(octOpenFacet(&interface));

	/* PRINTF("\nRunning vulcan %s\n\n", (fullvulcanp ? "(-k)" : "")); */
	
	(void) vulcan(facet, &interface);
    }
    
}

int symcleanAddNames( facet )
    octObject* facet;
{
    octGenerator genNets;
    octGenerator genInst;
    octObject net, inst;
    octObject ibag,cbag;

    int count = 0;
    int countNewNames = 0;

    octInitGenContentsSpecial( facet, OCT_NET_MASK, &genNets );
    while ( octGenerate( &genNets, &net ) == OCT_OK ) {
	if ( net.contents.net.name == NIL(char) ||  net.contents.net.name[0] == '\0' ) {
	    octObject tmpNet;
	    char buf[128];
	    tmpNet.type = OCT_NET;
	    do {
		sprintf( buf, "n%d", count++ );
		tmpNet.contents.net.name = buf;
	    } while ( octGetByName( facet, &tmpNet ) == OCT_OK );
	    net.contents.net.name = tmpNet.contents.net.name;
	    octModify( &net );
	    countNewNames++;
	}
    }

    count = 0;
    
    if (ohGetByBagName( facet, &ibag, "INSTANCES" ) == OCT_OK) {
	octInitGenContentsSpecial( facet, OCT_INSTANCE_MASK, &genInst );
	while ( octGenerate( &genInst, &inst ) == OCT_OK ) {
	    char* i_or_c = (octIsAttached( &ibag,&inst) == OCT_OK) ? "i":"c";
	    if ( inst.contents.instance.name[0] == '\0' ) {
		octObject tmpInst;
		char buf[128];
		tmpInst.type = OCT_INSTANCE;
		do {
		    sprintf( buf, "%s%d", i_or_c, count++ );
		    tmpInst.contents.instance.name = buf;
		} while ( octGetByName( facet, &tmpInst ) == OCT_OK );
		inst.contents.instance.name = tmpInst.contents.instance.name;
		octModify( &inst );
		countNewNames++;
	    }
	}
    }

    return  countNewNames;
}



int snapToLambda( obj , fixp )
    octObject* obj;
    int fixp;
{
    octBox bb;
    int w,h;
    int errcount = 0;
    int correctionX = 0;		/* Correction in transformation (used for paths) */
    int correctionY = 0;		/* Correction in transformation (used for paths) */

    int lambda = get_lambda( obj );

    OH_ASSERT( octBB( obj, &bb ) );
    w = bb.upperRight.x - bb.lowerLeft.x;
    h = bb.upperRight.y - bb.lowerLeft.y;

	    
    if ( obj->type == OCT_PATH ) {
	if ( checkPath( obj,fixp ) ) {
	    return 1;		/* Error detected there. */
	}

	if ( obj->contents.path.width % lambda ) {
	    sprintf(buffer, "Path with non integer width %.2f\n", obj->contents.path.width/(double)lambda);
	    symcleanReportError( obj, buffer );
	    return 1;
	}
	/* Move even width paths to the right/top */
	if ( !( (obj->contents.path.width / lambda ) & 0x1 ))  {
	    correctionY = lambda;
	    correctionX = lambda;
	}

	{
	    /* We only want to adjust paths in direction orthogonal to their direction */
	    int32 num;
	    struct octPoint *points;

	    ohGetPoints( obj, &num, &points );
	    if ( points[0].x == points[1].x ) { /* Vertical path? */
		bb.lowerLeft.y = bb.upperRight.y = 0; /* Ignore vertical dimention of bb */
	    }
	    if ( points[0].y == points[1].y ) { /* Horizontal path? */
		bb.lowerLeft.x = bb.upperRight.x  = 0; /* Ignore horizontal dimention of bb */
	    }
	}
    } else {
	if ( w % lambda || h % lambda ) {
	    errcount++;
	    sprintf(buffer, "%s has non integer bounding box (%.2fx%.2f)\n",
		   ohFormatName(obj),w/(double)lambda, h/(double)lambda);
	    symcleanReportError( obj, buffer );
	    return 1;
	}
	correctionX = (( w / lambda ) & 0x1) ?  0 : lambda;
	correctionY = (( h / lambda ) & 0x1) ?  0 : lambda;
    }
    if ( bb.lowerLeft.x % lambda || bb.lowerLeft.y % lambda ) {
	errcount++;
	/* sprintf(buffer, "%s not aligned to lambda grid\n", ohFormatName( obj )); */
	symcleanReportError( obj, buffer );
	if (fixp) {
	    octTransform transform;
	    /* Translate only dimension that needs it. */
	    int dx =  bb.lowerLeft.x % lambda;
	    int dy =  bb.lowerLeft.y % lambda;
	    if ( dx < 0 ) dx += lambda;
	    if ( dy < 0 ) dy += lambda;
	    transform.transformType = OCT_NO_TRANSFORM;
	    transform.translation.x = dx ? correctionX - dx : 0;
	    transform.translation.y = dy ? correctionY - dy : 0;
	    octTransformAndModifyGeo( obj, &transform );
	}
    }
    return errcount;
}



int symcleanLambdas( facet, fixp  )
    octObject* facet;
    int fixp;
{
    octGenerator genGeo;
    octGenerator genLayer;
    octGenerator genInst;
    octObject geo,layer,inst,prop;
    int errcount = 0;

    OH_ASSERT( ohGetByPropName( facet, &prop, "VIEWTYPE" ));
    if ( !PROPEQ( &prop, "SYMBOLIC") && !PROPEQ(&prop,"PHYSICAL") ) {
	/* This makes sense only for symbolic facets */
	return 0;
    }
    OH_ASSERT( octInitGenContentsSpecial( facet, OCT_INSTANCE_MASK, &genInst ));
    while ( octGenerate( &genInst, &inst ) == OCT_OK ) {
	errcount += snapToLambda( &inst, fixp );
    }
    OH_ASSERT( octInitGenContentsSpecial( facet, OCT_LAYER_MASK, &genLayer ) );
    while ( octGenerate( &genLayer, &layer ) == OCT_OK ) {
	OH_ASSERT( octInitGenContentsSpecial( &layer, OCT_GEO_MASK, &genGeo ) );
	while ( octGenerate( &genGeo, &geo ) == OCT_OK ) {
	    errcount += snapToLambda( &geo, fixp );
	}
    }
    return errcount;
}




#define MAX_REASONABLE_NAME_LENGTH 128

#define ERR_NAME_TOO_LONG -1


static int checkName( old_name, new_name, fix )
    char* old_name;
    char* new_name;
    int fix;
/*
* RETURN: 0 no change
* 	  >0 name has been changed.
*        ERR_NAME_TOO_LONG 
*/
{
    register char *c;			/* Pointer to character in name. */
    register char *ic;			/* Pointer to illegal character. */
    char *illegalChars = " \"\'\n\t";

    register int   has_illegal_char = 0;
    int   l = strlen( old_name );
    
    if ( l == 0 ) return 0;
    if (strlen(old_name) > MAX_REASONABLE_NAME_LENGTH ) {
	return ERR_NAME_TOO_LONG;
    }
    
    strcpy( new_name, old_name );
    c = new_name;

    do {
	ic = illegalChars;
	do {
	    if ( *ic == *c ) {
		has_illegal_char++;
		*c = '_';	/* Replace with a legal char. */
	    }
	} while ( *(++ic) != '\0' );
    } while ( *(++c) != '\0' );

    return has_illegal_char;
}


int symCheckName( obj, fixp )
    octObject *obj;
    int fixp;
{
    int fixed = 0;
    char **oldNamePtr;
    char newName[MAX_REASONABLE_NAME_LENGTH];


    switch ( obj->type ) {
    case OCT_INSTANCE: 
	oldNamePtr = &obj->contents.instance.name;
	break;
    case OCT_NET:
	oldNamePtr =  &obj->contents.net.name;
	break;
    case OCT_TERM:
	oldNamePtr =  &obj->contents.term.name;
	break;
    }

    fixed = checkName( *oldNamePtr, newName );
    if ( fixed == ERR_NAME_TOO_LONG ) {
	symcleanReportError( obj, "name too long" );
    } else if ( fixed > 0 ) {
	char buffer[1024];
	if ( fixp ) {
	    sprintf( buffer, "Name %s becomes %s\n", *oldNamePtr, newName );
	    (*oldNamePtr) = newName;
	    OH_ASSERT( octModify( obj ) );
	    symcleanReportError( obj, buffer );
	} else {
	    sprintf( buffer, "Name '%s' has strange chars.\n", *oldNamePtr );
	    symcleanReportError( obj, buffer );
	}
    }
    
    return fixed;
}




int symcleanCheckNames( facet, fixp )
    octObject *facet;
    int fixp;
{
    octGenerator gen;
    octObject obj;

    int errcount = 0;

    octInitGenContents( facet, OCT_INSTANCE_MASK | OCT_TERM_MASK | OCT_NET_MASK, &gen );
    while (octGenerate( &gen, &obj ) == OCT_OK ) {
	errcount += symCheckName( &obj, fixp );
    }
    return errcount;
}
