#include "copyright.h"
#include "port.h"

/*
 * RPC server (non oct and vem specific parts)
 *
 * Rick L Spickelmier, 3/20/86, last modified 3/2/87
 *
 */

#include "X11/Xlib.h"
#include "xvals.h"
#include "general.h"
#include "message.h"
#include "list.h"
#include "attributes.h"
#include "rpcServer.h"
#include "oh.h"

struct RPCApplication *RPCApplication;

rpc_fd_set *RPCSelectAccept;	/* mask for selecting on client connects */
rpc_fd_set *RPCSelectRequest;	/* mask for selecting on client requests */
rpc_fd_set *RPCFileRequest;     /* mask for selecting on file requests */

int *RPCFileDescriptors;        /* file descriptors for RPCFileRequest */

int Xfd;			/* X file descriptor */

lsList RPCDemonList;

void
rpcInit()
{
    /*
     * allow selection on the X file descriptor
     */
    Xfd = XConnectionNumber(xv_disp());  

    /* set up the bit arrays */
    RPCSelectRequest = RPCALLOC(rpc_fd_set);
    RPCSelectAccept = RPCALLOC(rpc_fd_set);
    RPCFileRequest = RPCALLOC(rpc_fd_set);
    RPC_FD_ZERO(RPCSelectRequest);
    RPC_FD_ZERO(RPCSelectAccept);
    RPC_FD_ZERO(RPCFileRequest);
    RPCFileDescriptors = RPCARRAYALLOC(int, getdtablesize());
    RPC_FD_SET(Xfd, RPCSelectRequest);
    RPCLargestFD = MAX(RPCLargestFD, Xfd);

    RPCDemonList = lsCreate();
}

/*
 * wait for and service requests
 *
 * Returns:
 *  RPC_OK - clean exit
 *  RPC_ERROR - error
 */
rpcStatus
vemServer()
{
    int eventType;		/* type the select event */
    int application;		/* application requesting service */
    int socket;                 /* socket that was selected on */
    long functionNumber;
    STREAM receiveStream;

    /*
     * wait for remote events and process them
     *
     * RPCGetEvent handles 'accept'
     *
     * returns:
     *	    server dies cleanly (RPC_EXIT)
     *	    something bad happened, abort (RPC_ABORT)
     *	    X event (RPC_X_EVENT)
     *      file event (RPC_FILE_EVENT)
     *	    rpc event (RPC_EVENT)
     */

    /* do initial X updates */
    vemProcessXEvent();

    while ((eventType = RPCGetEvent(RPCSelectAccept, RPCSelectRequest, RPCFileRequest,
				    &application, &socket)) != RPC_EXIT) {

	/* error */
	if (eventType == RPC_ABORT) {
	    vemMessage("RPC Error: server: panic!!!  server closing\n",
			      MSG_DISP);
	    RPC_FD_ZERO(RPCSelectRequest);
	    RPC_FD_SET(Xfd, RPCSelectRequest);
	    RPC_FD_ZERO(RPCSelectAccept);
	    RPC_FD_ZERO(RPCFileRequest);
	    continue;
	}

        /* process change lists demons */
        if (lsLength(RPCDemonList) > 0) {
            Pointer ptr;
            lsGen gen = lsStart(RPCDemonList);
            while (lsNext(gen, &ptr, 0) == LS_OK) {
		struct rpcDemon *demon;
                octObject changeList, cr;

		demon = (struct rpcDemon *) ptr;
                if (ohGetById(&changeList, demon->clId) == OCT_OK) {
		    if (octGenFirstContent(&changeList, OCT_ALL_MASK, &cr) == OCT_OK) {
			if (RPCDemonFunction(demon->app, &changeList) == RPC_OK) {
			    lsDelBefore(gen, 0);
			    VEMFREE(ptr);
			}
		    }
                }
            }
	}

	/* no event */
	if (eventType == RPC_NO_EVENT) {
	    continue;
	}

	/* X event */
	if (eventType == RPC_X_EVENT) {
	    /* process X events till there are no more */
	    while (vemPending() > 0) {
		if (vemProcessXEvent() != 0) {
		    return(RPC_OK);
		}
	    }
	    continue;
	}

	/* remote file request */

	if (eventType == RPC_FILE_EVENT) {
	    /* handle a request to write to a file */
	    RPCProcessFileWriteRequest(socket, RPCFileDescriptors[socket]);
	    continue;
	}

	/* remote request received, process it */

	if (eventType == RPC_EVENT) {

	    /* RPC remote function request */
	    receiveStream = RPCApplication[application].ReceiveStream;
	    if (!RPCReceiveLong(&functionNumber, receiveStream)) {
		
		/* error usually due to the application closing connection */
		vemMessage("RPC Error: server: application exited without calling RPCExit\n", MSG_DISP);
		
		/* tell the USER that the application has closed */
		(void) sprintf(RPCErrorBuffer, "Closing Application %s on host %s\n",
				        RPCApplication[application].name,
				        RPCApplication[application].host);

		vemMessage(RPCErrorBuffer, MSG_DISP);
		
		/* tell RPC that the application has closed */
		RPCServerEnd(application);
		
		/*
		 * tell VEM that the application has closed
		 */
		(void) vemCloseApplication(RPCApplication[application].host,
					   RPCApplication[application].name,
					   application);
		continue;
	    }

	    if (functionNumber > VEM_MAX_FUNCTION) {
		if (RPCOCTRequest(application, functionNumber) != RPC_OK) {
		    vemMessage("RPC Error: server: bad OCT request\n", MSG_DISP);
		    
		    /* tell the USER that the application has closed */
		    (void) sprintf(RPCErrorBuffer, "Closing Application %s on host %s\n",
				            RPCApplication[application].name,
				            RPCApplication[application].host);
		    vemMessage(RPCErrorBuffer, MSG_DISP);
		    
		    /* tell RPC that the application has closed */
		    RPCServerEnd(application);
		    
		    /*
		     * tell VEM that the application has closed
		     */
		    (void) vemCloseApplication(RPCApplication[application].host,
					       RPCApplication[application].name,
					       application);
		    continue;
		}
	    } else {
		if (RPCVEMRequest(application, functionNumber) != RPC_OK) {
		    vemMessage("RPC Error: server: bad VEM request\n", MSG_DISP);

		    /* tell the USER that the application has closed */
		    (void) sprintf(RPCErrorBuffer, "Closing Application %s on host %s\n",
				            RPCApplication[application].name,
				            RPCApplication[application].host);
		    vemMessage(RPCErrorBuffer, MSG_DISP);

		    /* tell RPC that the application has closed */
		    RPCServerEnd(application);

		    /*
		     * tell VEM that the application has closed
		     */
		    (void) vemCloseApplication(RPCApplication[application].host,
					       RPCApplication[application].name,
					       application);
		    continue;
		}
	    }
	    
	    while (vemPending() > 0) {
		if (vemProcessXEvent() != 0) {
		    return(RPC_OK);
		}
	    }
	    continue;
	}

	/* error */
	(void) sprintf(RPCErrorBuffer, "RPC Error: server: bad event type (%d)\n",
				eventType);
	vemMessage(RPCErrorBuffer, MSG_DISP);
	/* panic!!!  Go to just X events */
	RPC_FD_ZERO(RPCSelectRequest);
	RPC_FD_SET(Xfd, RPCSelectRequest);
	RPC_FD_ZERO(RPCSelectAccept);
	RPC_FD_ZERO(RPCFileRequest);
    }

    return(RPC_OK);
}
