/*
 * Symbolic Editing Initialization
 *
 * David Harrison
 * University of California, Berkeley
 * 1990
 *
 * This file contains the intitialization code for the symbolic
 * package.  Before any symbolic function is called in the library,
 * initialization is done to set up certain callback functions.
 */

#include "general.h"
#include "symbolic.h"
#include "vemDM.h"
#include "message.h"

#define NO_JUMP_NAME	";;; Unspecified Local/Global Net ;;;"

static int se_initialized = 0;



static void seJumpNameQuery(term, name, data)
octObject *term;		/* What terminal */
char name[SYM_MAX_MSG];		/* Net name      */
char *data;			/* User data     */
/*
 * This routine is called when the symbolic policy library encounters
 * a local or global net terminal whose property is empty.  This means
 * the name should come from the user.  We implement this by posting
 * a dialog.
 */
{
    dmTextItem item;

    item.itemPrompt = "Global net name";
    item.rows = 1;  item.cols = 40;
    item.value = (STR) 0;
    (void) sprintf(errMsgArea, "Terminal %s", term->contents.term.name);
    if (dmMultiText(errMsgArea, 1, &item) != VEM_OK) {
	(void) strcpy(name, NO_JUMP_NAME);
    } else {
	(void) strcpy(name, item.value);
    }
}


#define NULLNAME(actual) (STRLEN(actual) ? actual : "<No Name>" )

static int seMergeQuery(n1, n2, data)
octObject *n1, *n2;		/* Nets to merge */
char *data;			/* User data     */
/*
 * This routine is called when the symbolic policy package cannot
 * determine which of two nets to preserve when the nets must
 * be merged.  This is solved in VEM by asking the user using
 * a dialog.
 */
{
    char *name_one, *name_two;

    name_one = NULLNAME(n1->contents.net.name);
    name_two = NULLNAME(n2->contents.net.name);

    (void) sprintf(errMsgArea,
		   "Nets `%s' and `%s' must be merged.\nWhich one do you wish\nto preserve?",
		   name_one, name_two);
    if (dmConfirm("Net Maintenence", errMsgArea, name_one, name_two) == VEM_OK) {
	/* Preserve n1 */
	return 1;
    } else {
	/* Preserve n2 */
	return 0;
    }
}


static void seWarnFunc(sev, msg, data)
symSeverity sev;		/* Severity of message */
char *msg;			/* Message itself      */
char *data;			/* User data           */
/*
 * This routine is called when the symbolic package produces
 * diagnostics.  Based on the severity level, it either
 * prints it out in the console or posts a dialog.
 */
{
    switch (sev) {
    case SYM_INFO:
	vemMsg(MSG_C, "Symbolic message: %s\n", msg);
	break;
    case SYM_WARN:
	vemMsg(MSG_A, "Symbolic WARNING: %s\n", msg);
	break;
    case SYM_ERR:
	(void) dmConfirm("Symbolic Error", msg, "Dismiss", (STR) 0);
	break;
    }
}



void seInit()
/*
 * Initializes symbolic editing package.  Most of the callback functions
 * for symbolic are defined here.
 */
{
    static symJumpNameFunc jn = { seJumpNameQuery, (char *) 0 };
    static symMergeFunc merge = { seMergeQuery, (char *) 0 };
    static symWarnFunc warn = { seWarnFunc, (char *) 0 };

    if (!se_initialized) {
	symSetJumpNameFunc(&jn, (symJumpNameFunc *) 0);
	symSetMergeFunc(&merge, (symMergeFunc *) 0);
	symSetWarnFunc(&warn, (symWarnFunc *) 0);
	
	se_initialized = 1;
    }
}


